<?xml version="1.0" encoding="UTF-8"?>
<!--
 : HSLuv or HPLuv points
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:this="http://mathling.com/colour/hsl"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/hsl"
             exclude-result-prefixes="this xs math map array errors point util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <!--
    : this:hsl()
    : Make an HSL colour point
    :
    : $h: Hue [0.0, 360.0]
    : $s: Saturation [0.0, 100.0]
    : $l: Lightness [0.0, 100.0]
    :-->
   <xsl:function name="this:hsl" as="map(xs:string, item()*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="s" as="xs:double"/>
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      point:point($h, $s, $l)
    '/>
   </xsl:function>

   <!--
    : this:hsl()
    : Convert HSL string to HSL point
    :
    : $hsl The hsl string
    :-->
   <xsl:function name="this:hsl" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="xs:string"/>
      <xsl:sequence select='
       if (starts-with($hsl, "hsl(")) then (
         let $core := substring-before(substring-after($hsl, "hsl("),")") return
         let $parts := tokenize($core, "[,\s]+")[. ne ""] 
         return (
           if (count($parts) ne 3)
           then errors:error("COL-BADHSL", $hsl)
           else (),
           this:hsl(
             xs:double($parts[1]),
             xs:double($parts[2]),
             xs:double($parts[3])
           )
         )
       ) else (
         errors:error("COL-BADHSL", $hsl)
       )
    '/>
   </xsl:function>

   <!--
    : this:h()
    : Accessor for hue value of the point
    :
    : $hsl The point
    :-->
   <xsl:function name="this:h" as="xs:double">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($hsl)
    '/>
   </xsl:function>

   <!--
    : this:s()
    : Accessor for saturation value of the point
    :
    : $hsl The point
    :-->
   <xsl:function name="this:s" as="xs:double">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($hsl)
    '/>
   </xsl:function>

   <!--
    : this:l()
    : Accessor for luminance value of the point
    :
    : $hsl The point
    :-->
   <xsl:function name="this:l" as="xs:double">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($hsl)
    '/>
   </xsl:function>

   <!--
    : this:saturate()
    : When we play games with interpolations and multiplications, we can 
    : overrun the legal bounds. This caps things back in bounds.
    :-->
   <xsl:function name="this:saturate" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:hsl(
        util:remap-degrees(this:h($hsl)),
        util:clamp(this:s($hsl), 0.0, 100.0),
        util:clamp(this:l($hsl), 0.0, 100.0)
      )
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert HSL point value to string representation suitable for debugging
    :
    : $hsl The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "hsl("||
        this:h($hsl)||","||
        this:s($hsl)||","||
        this:l($hsl)||
      ")"
    '/>
   </xsl:function>

</xsl:package>
<!--hsl.xsl-->
