<?xml version="1.0" encoding="UTF-8"?>
<!--
 : RGB points
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/rgb"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"             
             version="3.0"
             name="http://mathling.com/colour/rgb"
             exclude-result-prefixes="this xs math map array errors point v util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>


   <!--
    : this:rgb()
    : Make an RGB colour point
    :
    : $r: red [0.0, 1.0]
    : $g: green [0.0, 1.0]
    : $b: blue [0.0, 1.0]
    :-->
   <xsl:function name="this:rgb" as="map(xs:string, item()*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="g" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      if (not(util:twixt($r, 0, 1) and util:twixt($g, 0, 1) and util:twixt($b, 0, 1))) then (
        errors:error("COL-BADRGB", ($r||","||$g||","||$b))
      ) else (
        point:point($r, $g, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgba()
    : Make an RGB colour point with alpha channel
    :
    : $r: red [0.0, 1.0]
    : $g: green [0.0, 1.0]
    : $b: blue [0.0, 1.0]
    : $alpha: alpha value
    :-->
   <xsl:function name="this:rgba" as="map(xs:string, item()*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="g" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:sequence select='
      if (not(util:twixt($r, 0, 1) and util:twixt($g, 0, 1) and util:twixt($b, 0, 1) and util:twixt($alpha, 0, 1))) then (
        errors:error("COL-BADRGB", ($r||","||$g||","||$b||","||$alpha))
      ) else (
        point:point($r, $g, $b, $alpha)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgba()
    :-->
   <xsl:function name="this:rgba" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:sequence select='
      if (not(util:twixt($alpha, 0, 1))) then (
        errors:error("COL-BADRGB", $alpha)
      ) else (
        point:point(this:r($rgb), this:g($rgb), this:b($rgb), $alpha)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgba()
    :-->
   <xsl:function name="this:rgba" as="map(xs:string, item()*)">
      <xsl:param name="rgba" as="xs:string"/>
      <xsl:sequence select='
      let $alpha := number(substring-after($rgba,")"))
      return
        if (exists($alpha) and $alpha=$alpha)
        then this:rgba(this:rgb($rgba), $alpha)
        else this:rgb($rgba)
    '/>
   </xsl:function>

   <!--
    : this:is-rgba()
    : Is there an alpha channel here?
    :-->
   <xsl:function name="this:is-rgba" as="xs:boolean">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:dimension($rgb)=4
    '/>
   </xsl:function>

   <!--
    : this:rgb()
    : Convert RGB string to RGB point with values [0.0, 1.0]
    : Strings come in two formats: rbg functional form and hex form
    : Also converts some named colours.
    : Example: "rgb(128,51,0)" => (0.5, 0.2, 0)
    : Example: "#FF00AA" => (1, 0, 0.666667)
    :
    : $rgb The rgb string
    :-->
   <xsl:function name="this:rgb" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="xs:string"/>
      <xsl:sequence select='
       if (starts-with($rgb, "rgb(")) then (
         let $core := substring-before(substring-after($rgb, "rgb("),")") return
         let $parts := tokenize($core, "[,\s]+")[. ne ""] 
         return (
           if (count($parts) ne 3)
           then errors:error("COL-BADRGB", $rgb)
           else (),
           this:rgb(
             xs:integer($parts[1]) div 255,
             xs:integer($parts[2]) div 255,
             xs:integer($parts[3]) div 255
           )
         )
       ) else if (starts-with($rgb, "#")) then (
         let $r := util:hex-to-integer(substring($rgb, 2, 2)) return
         let $g := util:hex-to-integer(substring($rgb, 4, 2)) return
         let $b := util:hex-to-integer(substring($rgb, 6, 2))
         return
           this:rgb(
             $r div 255,
             $g div 255,
             $b div 255
           )
       ) else (
         this:named-colour($rgb)
       )
    '/>
   </xsl:function>

   <!--
    : this:to-rgb()
    : Construct an RGB point from a vector.
    :-->
   <xsl:function name="this:to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      if (v:dimension($v)=4)
      then this:rgba(v:px($v), v:py($v), v:pz($v), v:pw($v))
      else this:rgb(v:px($v), v:py($v), v:pz($v))
    '/>
   </xsl:function>

   <!--
    : this:to-int()
    :-->
   <xsl:function name="this:to-int" as="xs:integer">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: ( r &lt;&lt; 16 ) | ( g &lt;&lt; 8 ) | ( b &lt;&lt; 0 ) :)
      let $coords := this:coordinates($rgb)
      return (
        util:or(
          xs:unsignedLong($coords[3]),
          util:or(
            util:lshift(xs:unsignedLong($coords[2]), 8),
            util:lshift(xs:unsignedLong($coords[1]), 16)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:from-int()
    :-->
   <xsl:function name="this:from-int" as="map(xs:string, item()*)">
      <xsl:param name="val" as="xs:integer"/>
      <xsl:sequence select='
      if ($val &lt; 0) then errors:error("COL-BADRGB", ($val)) else (),
      (:
       : float r = ((color >> 16) &amp; 0xff) / 255.0f;
       : float g = ((color >>  8) &amp; 0xff) / 255.0f;
       : float b = ((color      ) &amp; 0xff) / 255.0f;
       : float a = ((color >> 24) &amp; 0xff) / 255.0f;
       : &amp;0xff == &amp;255 = mod 256
       :)
      let $val := xs:unsignedLong($val) return
      let $ff := xs:unsignedLong(255) return
      let $a := util:rshift($val, 24) mod 256 return
      let $r := util:rshift($val, 16) mod 256 return
      let $g := util:rshift($val, 8) mod 256 return
      let $b := $val mod 256
      return
        if ($a = 0)
        then this:rgb($r div 255, $g div 255, $b div 255)
        else this:rgba($r div 255, $g div 255, $b div 255, $a div 255)
    '/>
   </xsl:function>

   <!--
    : this:r()
    : Accessor for red value of the point
    :
    : $rgb The point
    :-->
   <xsl:function name="this:r" as="xs:double">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($rgb)
    '/>
   </xsl:function>

   <!--
    : this:g()
    : Accessor for green value of the point
    :
    : $rgb The point
    :-->
   <xsl:function name="this:g" as="xs:double">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($rgb)
    '/>
   </xsl:function>

   <!--
    : this:b()
    : Accessor for blue value of the point
    :
    : $rgb The point
    :-->
   <xsl:function name="this:b" as="xs:double">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($rgb)
    '/>
   </xsl:function>

   <!--
    : this:alpha()
    : Accessor for alpha value of the point
    :
    : $rgb The point
    :-->
   <xsl:function name="this:alpha" as="xs:double">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:dimension($rgb)=4) then point:pw($rgb) else 1.0
    '/>
   </xsl:function>

   <!--
    : this:saturate()
    : When we play games with interpolations and multiplications, we can 
    : overrun the legal bounds. This caps things back in bounds.
    :-->
   <xsl:function name="this:saturate" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:dimension($rgb)=4) then ( 
        this:rgba(
          util:clamp(this:r($rgb), 0.0, 1.0),
          util:clamp(this:g($rgb), 0.0, 1.0),
          util:clamp(this:b($rgb), 0.0, 1.0),
          util:clamp(this:alpha($rgb), 0.0, 1.0)
        )
      ) else (
        this:rgb(
          util:clamp(this:r($rgb), 0.0, 1.0),
          util:clamp(this:g($rgb), 0.0, 1.0),
          util:clamp(this:b($rgb), 0.0, 1.0)
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert RGB point value to string representation suitable for use in SVG
    : Example: (0.5,0.2,0) => "rgb(128,51,0)"
    : Skip the alpha value, if any. SVG puts that in opacity property separately.
    :
    : $rgb The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: abs because sometimes conversions leave us w/ small negative
       : numbers which become -0 
       :)
      "rgb("||
        abs(round(this:r($rgb)*255))||","||
        abs(round(this:g($rgb)*255))||","||
        abs(round(this:b($rgb)*255))||
      ")"
    '/>
   </xsl:function>

   <!--
    : this:coordinates()
    : Get the RGB(A) coordinate values in the external range (0-255)
    :-->
   <xsl:function name="this:coordinates" as="xs:integer*">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pcoordinates($rgb)!round(. * 255)!abs(.)!(. cast as xs:integer)
    '/>
   </xsl:function>

   <!--
    : this:coordinates()
    : Get the RGB(A) coordinate values in the external range (0-255)
    : Pick whether to get RGB or RGBA via dimension parameter; e.g. 3 or 4
    :-->
   <xsl:function name="this:coordinates" as="xs:integer*">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:sequence select='
      if ($dim &lt; 4) then (
        point:pcoordinates($rgb, $dim)!round(. * 255)!abs(.)!(. cast as xs:integer)
      ) else (
        point:pcoordinates($rgb, 3)!round(. * 255)!abs(.)!(. cast as xs:integer), this:alpha($rgb)!round(. * 255)!abs(.)!(. cast as xs:integer)
      )
    '/>
   </xsl:function>

   <!--
    : this:raw-coordinates()
    : Get the raw coordinates, suitable for tonemapping etc.
    :-->
   <xsl:function name="this:raw-coordinates" as="xs:double*">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pcoordinates($rgb)
    '/>
   </xsl:function>

   <!--
    : this:named-colour()
    :-->
   <xsl:function name="this:named-colour" as="map(xs:string, item()*)">
      <xsl:param name="colour" as="xs:string"/>
      <xsl:sequence select='
      let $normalized := $this:KNOWN-COLOURS($colour)
      return
      (: Known SVG colour names :)
      if (empty($normalized)) then errors:error("COL-BADRGB", $colour)
      else if ($normalized = "aliceblue") then this:rgb("rgb(240, 248, 255)")
      else if ($normalized = "antiquewhite") then this:rgb("rgb(250, 235, 215)")
      else if ($normalized = "aqua") then this:rgb("rgb(0, 255, 255)")
      else if ($normalized = "aquamarine") then this:rgb("rgb(127, 255, 212)")
      else if ($normalized = "azure") then this:rgb("rgb(240, 255, 255)")
      else if ($normalized = "beige") then this:rgb("rgb(245, 245, 220)")
      else if ($normalized = "bisque") then this:rgb("rgb(255, 228, 196)")
      else if ($normalized = "black") then this:rgb("rgb(0, 0, 0)")
      else if ($normalized = "blanchedalmond") then this:rgb("rgb(255, 235, 205)")
      else if ($normalized = "blue") then this:rgb("rgb(0, 0, 255)")
      else if ($normalized = "blueviolet") then this:rgb("rgb(138, 43, 226)")
      else if ($normalized = "brown") then this:rgb("rgb(165, 42, 42)")
      else if ($normalized = "burlywood") then this:rgb("rgb(222, 184, 135)")
      else if ($normalized = "cadetblue") then this:rgb("rgb( 95, 158, 160)")
      else if ($normalized = "chartreuse") then this:rgb("rgb(127, 255, 0)")
      else if ($normalized = "chocolate") then this:rgb("rgb(210, 105, 30)")
      else if ($normalized = "coral") then this:rgb("rgb(255, 127, 80)")
      else if ($normalized = "cornflowerblue") then this:rgb("rgb(100, 149, 237)")
      else if ($normalized = "cornsilk") then this:rgb("rgb(255, 248, 220)")
      else if ($normalized = "crimson") then this:rgb("rgb(220, 20, 60)")
      else if ($normalized = "cyan") then this:rgb("rgb(0, 255, 255)")
      else if ($normalized = "darkblue") then this:rgb("rgb(0, 0, 139)")
      else if ($normalized = "darkcyan") then this:rgb("rgb(0, 139, 139)")
      else if ($normalized = "darkgoldenrod") then this:rgb("rgb(184, 134, 11)")
      else if ($normalized = "darkgray") then this:rgb("rgb(169, 169, 169)")
      else if ($normalized = "darkgreen") then this:rgb("rgb(0, 100, 0)")
      else if ($normalized = "darkgrey") then this:rgb("rgb(169, 169, 169)")
      else if ($normalized = "darkkhaki") then this:rgb("rgb(189, 183, 107)")
      else if ($normalized = "darkmagenta") then this:rgb("rgb(139, 0, 139)")
      else if ($normalized = "darkolivegreen") then this:rgb("rgb( 85, 107, 47)")
      else if ($normalized = "darkorange") then this:rgb("rgb(255, 140, 0)")
      else if ($normalized = "darkorchid") then this:rgb("rgb(153, 50, 204)")
      else if ($normalized = "darkred") then this:rgb("rgb(139, 0, 0)")
      else if ($normalized = "darksalmon") then this:rgb("rgb(233, 150, 122)")
      else if ($normalized = "darkseagreen") then this:rgb("rgb(143, 188, 143)")
      else if ($normalized = "darkslateblue") then this:rgb("rgb( 72, 61, 139)")
      else if ($normalized = "darkslategray") then this:rgb("rgb( 47, 79, 79)")
      else if ($normalized = "darkslategrey") then this:rgb("rgb( 47, 79, 79)")
      else if ($normalized = "darkturquoise") then this:rgb("rgb(0, 206, 209)")
      else if ($normalized = "darkviolet") then this:rgb("rgb(148, 0, 211)")
      else if ($normalized = "deeppink") then this:rgb("rgb(255, 20, 147)")
      else if ($normalized = "deepskyblue") then this:rgb("rgb(0, 191, 255)")
      else if ($normalized = "dimgray") then this:rgb("rgb(105, 105, 105)")
      else if ($normalized = "dimgrey") then this:rgb("rgb(105, 105, 105)")
      else if ($normalized = "dodgerblue") then this:rgb("rgb( 30, 144, 255)")
      else if ($normalized = "firebrick") then this:rgb("rgb(178, 34, 34)")
      else if ($normalized = "floralwhite") then this:rgb("rgb(255, 250, 240)")
      else if ($normalized = "forestgreen") then this:rgb("rgb( 34, 139, 34)")
      else if ($normalized = "fuchsia") then this:rgb("rgb(255, 0, 255)")
      else if ($normalized = "gainsboro") then this:rgb("rgb(220, 220, 220)")
      else if ($normalized = "ghostwhite") then this:rgb("rgb(248, 248, 255)")
      else if ($normalized = "gold") then this:rgb("rgb(255, 215, 0)")
      else if ($normalized = "goldenrod") then this:rgb("rgb(218, 165, 32)")
      else if ($normalized = "gray") then this:rgb("rgb(128, 128, 128)")
      else if ($normalized = "grey") then this:rgb("rgb(128, 128, 128)")
      else if ($normalized = "green") then this:rgb("rgb(0, 128, 0)")
      else if ($normalized = "greenyellow") then this:rgb("rgb(173, 255, 47)")
      else if ($normalized = "honeydew") then this:rgb("rgb(240, 255, 240)")
      else if ($normalized = "hotpink") then this:rgb("rgb(255, 105, 180)")
      else if ($normalized = "indianred") then this:rgb("rgb(205, 92, 92)")
      else if ($normalized = "indigo") then this:rgb("rgb( 75, 0, 130)")
      else if ($normalized = "ivory") then this:rgb("rgb(255, 255, 240)")
      else if ($normalized = "khaki") then this:rgb("rgb(240, 230, 140)")
      else if ($normalized = "lavender") then this:rgb("rgb(230, 230, 250)")
      else if ($normalized = "lavenderblush") then this:rgb("rgb(255, 240, 245)")
      else if ($normalized = "lawngreen") then this:rgb("rgb(124, 252, 0)")
      else if ($normalized = "lemonchiffon") then this:rgb("rgb(255, 250, 205)")
      else if ($normalized = "lightblue") then this:rgb("rgb(173, 216, 230)")
      else if ($normalized = "lightcoral") then this:rgb("rgb(240, 128, 128)")
      else if ($normalized = "lightcyan") then this:rgb("rgb(224, 255, 255)")
      else if ($normalized = "lightgoldenrodyellow") then this:rgb("rgb(250, 250, 210)")
      else if ($normalized = "lightgray") then this:rgb("rgb(211, 211, 211)")
      else if ($normalized = "lightgreen") then this:rgb("rgb(144, 238, 144)")
      else if ($normalized = "lightgrey") then this:rgb("rgb(211, 211, 211)")
      else if ($normalized = "lightpink") then this:rgb("rgb(255, 182, 193)")
      else if ($normalized = "lightsalmon") then this:rgb("rgb(255, 160, 122)")
      else if ($normalized = "lightseagreen") then this:rgb("rgb( 32, 178, 170)")
      else if ($normalized = "lightskyblue") then this:rgb("rgb(135, 206, 250)")
      else if ($normalized = "lightslategray") then this:rgb("rgb(119, 136, 153)")
      else if ($normalized = "lightslategrey") then this:rgb("rgb(119, 136, 153)")
      else if ($normalized = "lightsteelblue") then this:rgb("rgb(176, 196, 222)")
      else if ($normalized = "lightyellow") then this:rgb("rgb(255, 255, 224)")
      else if ($normalized = "lime") then this:rgb("rgb(0, 255, 0)")
      else if ($normalized = "limegreen") then this:rgb("rgb( 50, 205, 50)")
      else if ($normalized = "linen") then this:rgb("rgb(250, 240, 230)")
      else if ($normalized = "magenta") then this:rgb("rgb(255, 0, 255)")
      else if ($normalized = "maroon") then this:rgb("rgb(128, 0, 0)")
      else if ($normalized = "mediumaquamarine") then this:rgb("rgb(102, 205, 170)")
      else if ($normalized = "mediumblue") then this:rgb("rgb(0, 0, 205)")
      else if ($normalized = "mediumorchid") then this:rgb("rgb(186, 85, 211)")
      else if ($normalized = "mediumpurple") then this:rgb("rgb(147, 112, 219)")
      else if ($normalized = "mediumseagreen") then this:rgb("rgb( 60, 179, 113)")
      else if ($normalized = "mediumslateblue") then this:rgb("rgb(123, 104, 238)")
      else if ($normalized = "mediumspringgreen") then this:rgb("rgb(0, 250, 154)")
      else if ($normalized = "mediumturquoise") then this:rgb("rgb( 72, 209, 204)")
      else if ($normalized = "mediumvioletred") then this:rgb("rgb(199, 21, 133)")
      else if ($normalized = "midnightblue") then this:rgb("rgb( 25, 25, 112)")
      else if ($normalized = "mintcream") then this:rgb("rgb(245, 255, 250)")
      else if ($normalized = "mistyrose") then this:rgb("rgb(255, 228, 225)")
      else if ($normalized = "moccasin") then this:rgb("rgb(255, 228, 181)")
      else if ($normalized = "navajowhite") then this:rgb("rgb(255, 222, 173)")
      else if ($normalized = "navy") then this:rgb("rgb(0, 0, 128)")
      else if ($normalized = "oldlace") then this:rgb("rgb(253, 245, 230)")
      else if ($normalized = "olive") then this:rgb("rgb(128, 128, 0)")
      else if ($normalized = "olivedrab") then this:rgb("rgb(107, 142, 35)")
      else if ($normalized = "orange") then this:rgb("rgb(255, 165, 0)")
      else if ($normalized = "orangered") then this:rgb("rgb(255, 69, 0)")
      else if ($normalized = "orchid") then this:rgb("rgb(218, 112, 214)")
      else if ($normalized = "palegoldenrod") then this:rgb("rgb(238, 232, 170)")
      else if ($normalized = "palegreen") then this:rgb("rgb(152, 251, 152)")
      else if ($normalized = "paleturquoise") then this:rgb("rgb(175, 238, 238)")
      else if ($normalized = "palevioletred") then this:rgb("rgb(219, 112, 147)")
      else if ($normalized = "papayawhip") then this:rgb("rgb(255, 239, 213)")
      else if ($normalized = "peachpuff") then this:rgb("rgb(255, 218, 185)")
      else if ($normalized = "peru") then this:rgb("rgb(205, 133, 63)")
      else if ($normalized = "pink") then this:rgb("rgb(255, 192, 203)")
      else if ($normalized = "plum") then this:rgb("rgb(221, 160, 221)")
      else if ($normalized = "powderblue") then this:rgb("rgb(176, 224, 230)")
      else if ($normalized = "purple") then this:rgb("rgb(128, 0, 128)")
      else if ($normalized = "red") then this:rgb("rgb(255, 0, 0)")
      else if ($normalized = "rosybrown") then this:rgb("rgb(188, 143, 143)")
      else if ($normalized = "royalblue") then this:rgb("rgb( 65, 105, 225)")
      else if ($normalized = "saddlebrown") then this:rgb("rgb(139, 69, 19)")
      else if ($normalized = "salmon") then this:rgb("rgb(250, 128, 114)")
      else if ($normalized = "sandybrown") then this:rgb("rgb(244, 164, 96)")
      else if ($normalized = "seagreen") then this:rgb("rgb( 46, 139, 87)")
      else if ($normalized = "seashell") then this:rgb("rgb(255, 245, 238)")
      else if ($normalized = "sienna") then this:rgb("rgb(160, 82, 45)")
      else if ($normalized = "silver") then this:rgb("rgb(192, 192, 192)")
      else if ($normalized = "skyblue") then this:rgb("rgb(135, 206, 235)")
      else if ($normalized = "slateblue") then this:rgb("rgb(106, 90, 205)")
      else if ($normalized = "slategray") then this:rgb("rgb(112, 128, 144)")
      else if ($normalized = "slategrey") then this:rgb("rgb(112, 128, 144)")
      else if ($normalized = "snow") then this:rgb("rgb(255, 250, 250)")
      else if ($normalized = "springgreen") then this:rgb("rgb(0, 255, 127)")
      else if ($normalized = "steelblue") then this:rgb("rgb( 70, 130, 180)")
      else if ($normalized = "tan") then this:rgb("rgb(210, 180, 140)")
      else if ($normalized = "teal") then this:rgb("rgb(0, 128, 128)")
      else if ($normalized = "thistle") then this:rgb("rgb(216, 191, 216)")
      else if ($normalized = "tomato") then this:rgb("rgb(255, 99, 71)")
      else if ($normalized = "turquoise") then this:rgb("rgb( 64, 224, 208)")
      else if ($normalized = "violet") then this:rgb("rgb(238, 130, 238)")
      else if ($normalized = "wheat") then this:rgb("rgb(245, 222, 179)")
      else if ($normalized = "white") then this:rgb("rgb(255, 255, 255)")
      else if ($normalized = "whitesmoke") then this:rgb("rgb(245, 245, 245)")
      else if ($normalized = "yellow") then this:rgb("rgb(255, 255, 0)")
      else if ($normalized = "yellowgreen") then this:rgb("rgb(154, 205, 50)")
      else (
        (: Aliased named :)
        if (starts-with($normalized,"#"))
        then this:rgb($normalized)
        else this:named-colour($normalized)
      )
    '/>
   </xsl:function>


   <!--
    : $this:KNOWN-COLOURS
    :-->
   <xsl:variable name="this:KNOWN-COLOURS" as="map(xs:string,item()*)">
      <xsl:sequence select='
      map {
        "alice_blue":"aliceblue",
        "aliceblue":"aliceblue",
        "AliceBlue":"aliceblue",
        "antique_white":"antiquewhite",
        "antiquewhite":"antiquewhite",
        "AntiqueWhite":"antiquewhite",
        "AntiqueWhite1":"#FFEFDB",
        "AntiqueWhite2":"#EEDFCC",
        "AntiqueWhite3":"#CDC0B0",
        "AntiqueWhite4":"#8B8378",
        "aqua":"aqua",
        "aquamarine":"aquamarine",
        "aquamarine1":"aquamarine",
        "aquamarine2":"#76EEC6",
        "aquamarine3":"mediumaquamarine",
        "aquamarine4":"#458B74",
        "azure":"azure",
        "azure1":"azure",
        "azure2":"#E0EEEE",
        "azure3":"#C1CDCD",
        "azure4":"#838B8B",
        "beige":"beige",
        "bisque":"bisque",
        "bisque1":"bisque",
        "bisque2":"#EED5B7",
        "bisque3":"#CDB79E",
        "bisque4":"#8B7D6B",
        "black":"black",
        "blanched_almond":"blanchedalmond",
        "blanchedalmond":"blanchedalmond",
        "BlanchedAlmond":"blanchedalmond",
        "blue":"blue",
        "blue_violet":"blueviolet",
        "blue1":"blue",
        "blue2":"#0000EE",
        "blue3":"mediumblue",
        "blue4":"darkblue",
        "blueviolet":"blueviolet",
        "BlueViolet":"blueviolet",
        "brown":"brown",
        "brown1":"#FF4040",
        "brown2":"#EE3B3B",
        "brown3":"#CD3333",
        "brown4":"#8B2323",
        "burlywood":"burlywood",
        "burlywood1":"#FFD39B",
        "burlywood2":"#EEC591",
        "burlywood3":"#CDAA7D",
        "burlywood4":"#8B7355",
        "cadet_blue":"cadetblue",
        "cadetblue":"cadetblue",
        "CadetBlue":"cadetblue",
        "CadetBlue1":"#98F5FF",
        "CadetBlue2":"#8EE5EE",
        "CadetBlue3":"#7AC5CD",
        "CadetBlue4":"#53868B",
        "chartreuse":"chartreuse",
        "chartreuse1":"chartreuse",
        "chartreuse2":"#76EE00",
        "chartreuse3":"#66CD00",
        "chartreuse4":"#458B00",
        "chocolate":"chocolate",
        "chocolate1":"#FF7F24",
        "chocolate2":"#EE7621",
        "chocolate3":"#CD661D",
        "chocolate4":"saddlebrown",
        "coral":"coral",
        "coral1":"#FF7256",
        "coral2":"#EE6A50",
        "coral3":"#CD5B45",
        "coral4":"#8B3E2F",
        "cornflower_blue":"cornflowerblue",
        "cornflowerblue":"cornflowerblue",
        "CornflowerBlue":"cornflowerblue",
        "cornsilk":"cornsilk",
        "cornsilk1":"cornsilk",
        "cornsilk2":"#EEE8CD",
        "cornsilk3":"#CDC8B1",
        "cornsilk4":"#8B8878",
        "crimson":"crimson",
        "cyan":"aqua",
        "cyan1":"aqua",
        "cyan2":"#00EEEE",
        "cyan3":"#00CDCD",
        "cyan4":"darkcyan",
        "dark_goldenrod":"darkgoldenrod",
        "dark_green":"darkgreen",
        "dark_khaki":"darkkhaki",
        "dark_olive_green":"darkolivegreen",
        "dark_orange":"darkorange",
        "dark_orchid":"darkorchid",
        "dark_salmon":"darksalmon",
        "dark_sea_green":"darkseagreen",
        "dark_slate_blue":"darkslateblue",
        "dark_slate_gray":"darkslategray",
        "dark_slate_grey":"darkslategray",
        "dark_turquoise":"darkturquoise",
        "dark_violet":"darkviolet",
        "darkblue":"darkblue",
        "darkcyan":"darkcyan",
        "darkgoldenrod":"darkgoldenrod",
        "DarkGoldenrod":"darkgoldenrod",
        "DarkGoldenrod1":"#FFB90F",
        "DarkGoldenrod2":"#EEAD0E",
        "DarkGoldenrod3":"#CD950C",
        "DarkGoldenrod4":"#8B6508",
        "darkgray":"darkgray",
        "darkgreen":"darkgreen",
        "DarkGreen":"darkgreen",
        "darkgrey":"darkgray",
        "darkkhaki":"darkkhaki",
        "DarkKhaki":"darkkhaki",
        "darkmagenta":"darkmagenta",
        "darkolivegreen":"darkolivegreen",
        "DarkOliveGreen":"darkolivegreen",
        "DarkOliveGreen1":"#CAFF70",
        "DarkOliveGreen2":"#BCEE68",
        "DarkOliveGreen3":"#A2CD5A",
        "DarkOliveGreen4":"#6E8B3D",
        "darkorange":"darkorange",
        "DarkOrange":"darkorange",
        "DarkOrange1":"#FF7F00",
        "DarkOrange2":"#EE7600",
        "DarkOrange3":"#CD6600",
        "DarkOrange4":"#8B4500",
        "darkorchid":"darkorchid",
        "DarkOrchid":"darkorchid",
        "DarkOrchid1":"#BF3EFF",
        "DarkOrchid2":"#B23AEE",
        "DarkOrchid3":"#9A32CD",
        "DarkOrchid4":"#68228B",
        "darkred":"darkred",
        "darksalmon":"darksalmon",
        "DarkSalmon":"darksalmon",
        "darkseagreen":"darkseagreen",
        "DarkSeaGreen":"darkseagreen",
        "DarkSeaGreen1":"#C1FFC1",
        "DarkSeaGreen2":"#B4EEB4",
        "DarkSeaGreen3":"#9BCD9B",
        "DarkSeaGreen4":"#698B69",
        "darkslateblue":"darkslateblue",
        "DarkSlateBlue":"darkslateblue",
        "darkslategray":"darkslategray",
        "DarkSlateGray":"darkslategray",
        "DarkSlateGray1":"#97FFFF",
        "DarkSlateGray2":"#8DEEEE",
        "DarkSlateGray3":"#79CDCD",
        "DarkSlateGray4":"#528B8B",
        "darkslategrey":"darkslategray",
        "DarkSlateGrey":"darkslategray",
        "darkturquoise":"darkturquoise",
        "DarkTurquoise":"darkturquoise",
        "darkviolet":"darkviolet",
        "DarkViolet":"darkviolet",
        "deep_pink":"deeppink",
        "deep_sky_blue":"deepskyblue",
        "deeppink":"deeppink",
        "DeepPink":"deeppink",
        "DeepPink1":"deeppink",
        "DeepPink2":"#EE1289",
        "DeepPink3":"#CD1076",
        "DeepPink4":"#8B0A50",
        "deepskyblue":"deepskyblue",
        "DeepSkyBlue":"deepskyblue",
        "DeepSkyBlue1":"deepskyblue",
        "DeepSkyBlue2":"#00B2EE",
        "DeepSkyBlue3":"#009ACD",
        "DeepSkyBlue4":"#00688B",
        "dim_gray":"dimgray",
        "dim_grey":"dimgray",
        "dimgray":"dimgray",
        "DimGray":"dimgray",
        "dimgrey":"dimgray",
        "DimGrey":"dimgray",
        "dodger_blue":"dodgerblue",
        "dodgerblue":"dodgerblue",
        "DodgerBlue":"dodgerblue",
        "DodgerBlue1":"dodgerblue",
        "DodgerBlue2":"#1C86EE",
        "DodgerBlue3":"#1874CD",
        "DodgerBlue4":"#104E8B",
        "firebrick":"firebrick",
        "firebrick1":"#FF3030",
        "firebrick2":"#EE2C2C",
        "firebrick3":"#CD2626",
        "firebrick4":"#8B1A1A",
        "floral_white":"floralwhite",
        "floralwhite":"floralwhite",
        "FloralWhite":"floralwhite",
        "forest_green":"forestgreen",
        "forestgreen":"forestgreen",
        "ForestGreen":"forestgreen",
        "fuchsia":"fuchsia",
        "gainsboro":"gainsboro",
        "ghost_white":"ghostwhite",
        "ghostwhite":"ghostwhite",
        "GhostWhite":"ghostwhite",
        "gold":"gold",
        "gold1":"gold",
        "gold2":"#EEC900",
        "gold3":"#CDAD00",
        "gold4":"#8B7500",
        "goldenrod":"goldenrod",
        "goldenrod1":"#FFC125",
        "goldenrod2":"#EEB422",
        "goldenrod3":"#CD9B1D",
        "goldenrod4":"#8B6914",
        "gray":"silver",
        "gray0":"black",
        "gray1":"#030303",
        "gray10":"#1A1A1A",
        "gray100":"white",
        "gray11":"#1C1C1C",
        "gray12":"#1F1F1F",
        "gray13":"#212121",
        "gray14":"#242424",
        "gray15":"#262626",
        "gray16":"#292929",
        "gray17":"#2B2B2B",
        "gray18":"#2E2E2E",
        "gray19":"#303030",
        "gray2":"#050505",
        "gray20":"#333333",
        "gray21":"#363636",
        "gray22":"#383838",
        "gray23":"#3B3B3B",
        "gray24":"#3D3D3D",
        "gray25":"#404040",
        "gray26":"#424242",
        "gray27":"#454545",
        "gray28":"#474747",
        "gray29":"#4A4A4A",
        "gray3":"#080808",
        "gray30":"#4D4D4D",
        "gray31":"#4F4F4F",
        "gray32":"#525252",
        "gray33":"#545454",
        "gray34":"#575757",
        "gray35":"#595959",
        "gray36":"#5C5C5C",
        "gray37":"#5E5E5E",
        "gray38":"#616161",
        "gray39":"#636363",
        "gray4":"#0A0A0A",
        "gray40":"#666666",
        "gray41":"dimgray",
        "gray42":"#6B6B6B",
        "gray43":"#6E6E6E",
        "gray44":"#707070",
        "gray45":"#737373",
        "gray46":"#757575",
        "gray47":"#787878",
        "gray48":"#7A7A7A",
        "gray49":"#7D7D7D",
        "gray5":"#0D0D0D",
        "gray50":"#7F7F7F",
        "gray51":"#828282",
        "gray52":"#858585",
        "gray53":"#878787",
        "gray54":"#8A8A8A",
        "gray55":"#8C8C8C",
        "gray56":"#8F8F8F",
        "gray57":"#919191",
        "gray58":"#949494",
        "gray59":"#969696",
        "gray6":"#0F0F0F",
        "gray60":"#999999",
        "gray61":"#9C9C9C",
        "gray62":"#9E9E9E",
        "gray63":"#A1A1A1",
        "gray64":"#A3A3A3",
        "gray65":"#A6A6A6",
        "gray66":"#A8A8A8",
        "gray67":"#ABABAB",
        "gray68":"#ADADAD",
        "gray69":"#B0B0B0",
        "gray7":"#121212",
        "gray70":"#B3B3B3",
        "gray71":"#B5B5B5",
        "gray72":"#B8B8B8",
        "gray73":"#BABABA",
        "gray74":"#BDBDBD",
        "gray75":"#BFBFBF",
        "gray76":"#C2C2C2",
        "gray77":"#C4C4C4",
        "gray78":"#C7C7C7",
        "gray79":"#C9C9C9",
        "gray8":"#141414",
        "gray80":"#CCCCCC",
        "gray81":"#CFCFCF",
        "gray82":"#D1D1D1",
        "gray83":"#D4D4D4",
        "gray84":"#D6D6D6",
        "gray85":"#D9D9D9",
        "gray86":"#DBDBDB",
        "gray87":"#DEDEDE",
        "gray88":"#E0E0E0",
        "gray89":"#E3E3E3",
        "gray9":"#171717",
        "gray90":"#E5E5E5",
        "gray91":"#E8E8E8",
        "gray92":"#EBEBEB",
        "gray93":"#EDEDED",
        "gray94":"#F0F0F0",
        "gray95":"#F2F2F2",
        "gray96":"whitesmoke",
        "gray97":"#F7F7F7",
        "gray98":"#FAFAFA",
        "gray99":"#FCFCFC",
        "green":"green",
        "green_yellow":"greenyellow",
        "green1":"lime",
        "green2":"#00EE00",
        "green3":"#00CD00",
        "green4":"#008B00",
        "greenyellow":"greenyellow",
        "GreenYellow":"greenyellow",
        "grey":"grey",
        "grey0":"black",
        "grey1":"#030303",
        "grey10":"#1A1A1A",
        "grey100":"white",
        "grey11":"#1C1C1C",
        "grey12":"#1F1F1F",
        "grey13":"#212121",
        "grey14":"#242424",
        "grey15":"#262626",
        "grey16":"#292929",
        "grey17":"#2B2B2B",
        "grey18":"#2E2E2E",
        "grey19":"#303030",
        "grey2":"#050505",
        "grey20":"#333333",
        "grey21":"#363636",
        "grey22":"#383838",
        "grey23":"#3B3B3B",
        "grey24":"#3D3D3D",
        "grey25":"#404040",
        "grey26":"#424242",
        "grey27":"#454545",
        "grey28":"#474747",
        "grey29":"#4A4A4A",
        "grey3":"#080808",
        "grey30":"#4D4D4D",
        "grey31":"#4F4F4F",
        "grey32":"#525252",
        "grey33":"#545454",
        "grey34":"#575757",
        "grey35":"#595959",
        "grey36":"#5C5C5C",
        "grey37":"#5E5E5E",
        "grey38":"#616161",
        "grey39":"#636363",
        "grey4":"#0A0A0A",
        "grey40":"#666666",
        "grey41":"dimgray",
        "grey42":"#6B6B6B",
        "grey43":"#6E6E6E",
        "grey44":"#707070",
        "grey45":"#737373",
        "grey46":"#757575",
        "grey47":"#787878",
        "grey48":"#7A7A7A",
        "grey49":"#7D7D7D",
        "grey5":"#0D0D0D",
        "grey50":"#7F7F7F",
        "grey51":"#828282",
        "grey52":"#858585",
        "grey53":"#878787",
        "grey54":"#8A8A8A",
        "grey55":"#8C8C8C",
        "grey56":"#8F8F8F",
        "grey57":"#919191",
        "grey58":"#949494",
        "grey59":"#969696",
        "grey6":"#0F0F0F",
        "grey60":"#999999",
        "grey61":"#9C9C9C",
        "grey62":"#9E9E9E",
        "grey63":"#A1A1A1",
        "grey64":"#A3A3A3",
        "grey65":"#A6A6A6",
        "grey66":"#A8A8A8",
        "grey67":"#ABABAB",
        "grey68":"#ADADAD",
        "grey69":"#B0B0B0",
        "grey7":"#121212",
        "grey70":"#B3B3B3",
        "grey71":"#B5B5B5",
        "grey72":"#B8B8B8",
        "grey73":"#BABABA",
        "grey74":"#BDBDBD",
        "grey75":"#BFBFBF",
        "grey76":"#C2C2C2",
        "grey77":"#C4C4C4",
        "grey78":"#C7C7C7",
        "grey79":"#C9C9C9",
        "grey8":"#141414",
        "grey80":"#CCCCCC",
        "grey81":"#CFCFCF",
        "grey82":"#D1D1D1",
        "grey83":"#D4D4D4",
        "grey84":"#D6D6D6",
        "grey85":"#D9D9D9",
        "grey86":"#DBDBDB",
        "grey87":"#DEDEDE",
        "grey88":"#E0E0E0",
        "grey89":"#E3E3E3",
        "grey9":"#171717",
        "grey90":"#E5E5E5",
        "grey91":"#E8E8E8",
        "grey92":"#EBEBEB",
        "grey93":"#EDEDED",
        "grey94":"#F0F0F0",
        "grey95":"#F2F2F2",
        "grey96":"whitesmoke",
        "grey97":"#F7F7F7",
        "grey98":"#FAFAFA",
        "grey99":"#FCFCFC",
        "honeydew":"honeydew",
        "honeydew1":"honeydew",
        "honeydew2":"#E0EEE0",
        "honeydew3":"#C1CDC1",
        "honeydew4":"#838B83",
        "hot_pink":"hotpink",
        "hotpink":"hotpink",
        "HotPink":"hotpink",
        "HotPink1":"#FF6EB4",
        "HotPink2":"#EE6AA7",
        "HotPink3":"#CD6090",
        "HotPink4":"#8B3A62",
        "indian_red":"indianred",
        "indianred":"indianred",
        "IndianRed":"indianred",
        "IndianRed1":"#FF6A6A",
        "IndianRed2":"#EE6363",
        "IndianRed3":"#CD5555",
        "IndianRed4":"#8B3A3A",
        "indigo":"indigo",
        "ivory":"ivory",
        "ivory1":"ivory",
        "ivory2":"#EEEEE0",
        "ivory3":"#CDCDC1",
        "ivory4":"#8B8B83",
        "khaki":"khaki",
        "khaki1":"#FFF68F",
        "khaki2":"#EEE685",
        "khaki3":"#CDC673",
        "khaki4":"#8B864E",
        "lavender":"lavender",
        "lavender_blush":"lavenderblush",
        "lavenderblush":"lavenderblush",
        "LavenderBlush":"lavenderblush",
        "LavenderBlush1":"lavenderblush",
        "LavenderBlush2":"#EEE0E5",
        "LavenderBlush3":"#CDC1C5",
        "LavenderBlush4":"#8B8386",
        "lawn_green":"lawngreen",
        "lawngreen":"lawngreen",
        "LawnGreen":"lawngreen",
        "lemon_chiffon":"lemonchiffon",
        "lemonchiffon":"lemonchiffon",
        "LemonChiffon":"lemonchiffon",
        "LemonChiffon1":"lemonchiffon",
        "LemonChiffon2":"#EEE9BF",
        "LemonChiffon3":"#CDC9A5",
        "LemonChiffon4":"#8B8970",
        "light_blue":"lightblue",
        "light_coral":"lightcoral",
        "light_cyan":"lightcyan",
        "light_goldenrod":"#EEDD82",
        "light_goldenrod_yellow":"lightgoldenrodyellow",
        "light_gray":"lightgray",
        "light_grey":"lightgray",
        "light_pink":"lightpink",
        "light_salmon":"lightsalmon",
        "light_sea_green":"lightseagreen",
        "light_sky_blue":"lightskyblue",
        "light_slate_blue":"#8470FF",
        "light_slate_gray":"lightslategray",
        "light_slate_grey":"lightslategray",
        "light_steel_blue":"lightsteelblue",
        "light_yellow":"lightyellow",
        "lightblue":"lightblue",
        "LightBlue":"lightblue",
        "LightBlue1":"#BFEFFF",
        "LightBlue2":"#B2DFEE",
        "LightBlue3":"#9AC0CD",
        "LightBlue4":"#68838B",
        "lightcoral":"lightcoral",
        "LightCoral":"lightcoral",
        "lightcyan":"lightcyan",
        "LightCyan":"lightcyan",
        "LightCyan1":"lightcyan",
        "LightCyan2":"#D1EEEE",
        "LightCyan3":"#B4CDCD",
        "LightCyan4":"#7A8B8B",
        "LightGoldenrod":"#EEDD82",
        "LightGoldenrod1":"#FFEC8B",
        "LightGoldenrod2":"#EEDC82",
        "LightGoldenrod3":"#CDBE70",
        "LightGoldenrod4":"#8B814C",
        "lightgoldenrodyellow":"lightgoldenrodyellow",
        "LightGoldenrodYellow":"lightgoldenrodyellow",
        "lightgray":"lightgray",
        "LightGray":"lightgray",
        "lightgreen":"lightgreen",
        "lightgrey":"lightgray",
        "LightGrey":"lightgray",
        "lightpink":"lightpink",
        "LightPink":"lightpink",
        "LightPink1":"#FFAEB9",
        "LightPink2":"#EEA2AD",
        "LightPink3":"#CD8C95",
        "LightPink4":"#8B5F65",
        "lightsalmon":"lightsalmon",
        "LightSalmon":"lightsalmon",
        "LightSalmon1":"lightsalmon",
        "LightSalmon2":"#EE9572",
        "LightSalmon3":"#CD8162",
        "LightSalmon4":"#8B5742",
        "lightseagreen":"lightseagreen",
        "LightSeaGreen":"lightseagreen",
        "lightskyblue":"lightskyblue",
        "LightSkyBlue":"lightskyblue",
        "LightSkyBlue1":"#B0E2FF",
        "LightSkyBlue2":"#A4D3EE",
        "LightSkyBlue3":"#8DB6CD",
        "LightSkyBlue4":"#607B8B",
        "LightSlateBlue":"#8470FF",
        "lightslategray":"lightslategray",
        "LightSlateGray":"lightslategray",
        "lightslategrey":"lightslategray",
        "LightSlateGrey":"lightslategray",
        "lightsteelblue":"lightsteelblue",
        "LightSteelBlue":"lightsteelblue",
        "LightSteelBlue1":"#CAE1FF",
        "LightSteelBlue2":"#BCD2EE",
        "LightSteelBlue3":"#A2B5CD",
        "LightSteelBlue4":"#6E7B8B",
        "lightyellow":"lightyellow",
        "LightYellow":"lightyellow",
        "LightYellow1":"lightyellow",
        "LightYellow2":"#EEEED1",
        "LightYellow3":"#CDCDB4",
        "LightYellow4":"#8B8B7A",
        "lime":"lime",
        "lime_green":"limegreen",
        "limegreen":"limegreen",
        "LimeGreen":"limegreen",
        "linen":"linen",
        "magenta":"magenta",
        "magenta1":"fuchsia",
        "magenta2":"#EE00EE",
        "magenta3":"#CD00CD",
        "magenta4":"darkmagenta",
        "maroon":"maroon",
        "maroon1":"#FF34B3",
        "maroon2":"#EE30A7",
        "maroon3":"#CD2990",
        "maroon4":"#8B1C62",
        "maroonAlt":"#B03060",
        "medium_aquamarine":"mediumaquamarine",
        "medium_blue":"mediumblue",
        "medium_orchid":"mediumorchid",
        "medium_purple":"mediumpurple",
        "medium_sea_green":"mediumseagreen",
        "medium_slate_blue":"mediumslateblue",
        "medium_spring_green":"mediumspringgreen",
        "medium_turquoise":"mediumturquoise",
        "medium_violet_red":"mediumvioletred",
        "mediumaquamarine":"mediumaquamarine",
        "MediumAquamarine":"mediumaquamarine",
        "mediumblue":"mediumblue",
        "MediumBlue":"mediumblue",
        "mediumorchid":"mediumorchid",
        "MediumOrchid":"mediumorchid",
        "MediumOrchid1":"#E066FF",
        "MediumOrchid2":"#D15FEE",
        "MediumOrchid3":"#B452CD",
        "MediumOrchid4":"#7A378B",
        "mediumpurple":"mediumpurple",
        "MediumPurple":"mediumpurple",
        "MediumPurple1":"#AB82FF",
        "MediumPurple2":"#9F79EE",
        "MediumPurple3":"#8968CD",
        "MediumPurple4":"#5D478B",
        "mediumseagreen":"mediumseagreen",
        "MediumSeaGreen":"mediumseagreen",
        "mediumslateblue":"mediumslateblue",
        "MediumSlateBlue":"mediumslateblue",
        "mediumspringgreen":"mediumspringgreen",
        "MediumSpringGreen":"mediumspringgreen",
        "mediumturquoise":"mediumturquoise",
        "MediumTurquoise":"mediumturquoise",
        "mediumvioletred":"mediumvioletred",
        "MediumVioletRed":"mediumvioletred",
        "midnight_blue":"midnightblue",
        "midnightblue":"midnightblue",
        "MidnightBlue":"midnightblue",
        "mint_cream":"mintcream",
        "mintcream":"mintcream",
        "MintCream":"mintcream",
        "misty_rose":"mistyrose",
        "mistyrose":"mistyrose",
        "MistyRose":"mistyrose",
        "MistyRose1":"mistyrose",
        "MistyRose2":"#EED5D2",
        "MistyRose3":"#CDB7B5",
        "MistyRose4":"#8B7D7B",
        "moccasin":"moccasin",
        "navajo_white":"navajowhite",
        "navajowhite":"navajowhite",
        "NavajoWhite":"navajowhite",
        "NavajoWhite1":"navajowhite",
        "NavajoWhite2":"#EECFA1",
        "NavajoWhite3":"#CDB38B",
        "NavajoWhite4":"#8B795E",
        "navy":"navy",
        "navy_blue":"navy",
        "NavyBlue":"navy",
        "old_lace":"oldlace",
        "oldlace":"oldlace",
        "OldLace":"oldlace",
        "olive":"olive",
        "olive_drab":"olivedrab",
        "olivedrab":"olivedrab",
        "OliveDrab":"olivedrab",
        "OliveDrab1":"#C0FF3E",
        "OliveDrab2":"#B3EE3A",
        "OliveDrab3":"yellowgreen",
        "OliveDrab4":"#698B22",
        "orange":"orange",
        "orange_red":"orangered",
        "orange1":"orange",
        "orange2":"#EE9A00",
        "orange3":"#CD8500",
        "orange4":"#8B5A00",
        "orangered":"orangered",
        "OrangeRed":"orangered",
        "OrangeRed1":"orangered",
        "OrangeRed2":"#EE4000",
        "OrangeRed3":"#CD3700",
        "OrangeRed4":"#8B2500",
        "orchid":"orchid",
        "orchid1":"#FF83FA",
        "orchid2":"#EE7AE9",
        "orchid3":"#CD69C9",
        "orchid4":"#8B4789",
        "pale_goldenrod":"palegoldenrod",
        "pale_green":"palegreen",
        "pale_turquoise":"paleturquoise",
        "pale_violet_red":"palevioletred",
        "palegoldenrod":"palegoldenrod",
        "PaleGoldenrod":"palegoldenrod",
        "palegreen":"palegreen",
        "PaleGreen":"palegreen",
        "PaleGreen1":"#9AFF9A",
        "PaleGreen2":"lightgreen",
        "PaleGreen3":"#7CCD7C",
        "PaleGreen4":"#548B54",
        "paleturquoise":"paleturquoise",
        "PaleTurquoise":"paleturquoise",
        "PaleTurquoise1":"#BBFFFF",
        "PaleTurquoise2":"#AEEEEE",
        "PaleTurquoise3":"#96CDCD",
        "PaleTurquoise4":"#668B8B",
        "palevioletred":"palevioletred",
        "PaleVioletRed":"palevioletred",
        "PaleVioletRed1":"#FF82AB",
        "PaleVioletRed2":"#EE799F",
        "PaleVioletRed3":"#CD6889",
        "PaleVioletRed4":"#8B475D",
        "papaya_whip":"papayawhip",
        "papayawhip":"papayawhip",
        "PapayaWhip":"papayawhip",
        "peach_puff":"peachpuff",
        "peachpuff":"peachpuff",
        "PeachPuff":"peachpuff",
        "PeachPuff1":"peachpuff",
        "PeachPuff2":"#EECBAD",
        "PeachPuff3":"#CDAF95",
        "PeachPuff4":"#8B7765",
        "peru":"peru",
        "pink":"pink",
        "pink1":"#FFB5C5",
        "pink2":"#EEA9B8",
        "pink3":"#CD919E",
        "pink4":"#8B636C",
        "plum":"plum",
        "plum1":"#FFBBFF",
        "plum2":"#EEAEEE",
        "plum3":"#CD96CD",
        "plum4":"#8B668B",
        "powder_blue":"powderblue",
        "powderblue":"powderblue",
        "PowderBlue":"powderblue",
        "purple":"purple",
        "purple1":"#9B30FF",
        "purple2":"#912CEE",
        "purple3":"#7D26CD",
        "purple4":"#551A8B",
        "purpleAlt":"#A020F0",
        "red":"red",
        "red1":"red",
        "red2":"#EE0000",
        "red3":"#CD0000",
        "red4":"darkred",
        "rosy_brown":"rosybrown",
        "rosybrown":"rosybrown",
        "RosyBrown":"rosybrown",
        "RosyBrown1":"#FFC1C1",
        "RosyBrown2":"#EEB4B4",
        "RosyBrown3":"#CD9B9B",
        "RosyBrown4":"#8B6969",
        "royal_blue":"royalblue",
        "royalblue":"royalblue",
        "RoyalBlue":"royalblue",
        "RoyalBlue1":"#4876FF",
        "RoyalBlue2":"#436EEE",
        "RoyalBlue3":"#3A5FCD",
        "RoyalBlue4":"#27408B",
        "saddle_brown":"saddlebrown",
        "saddlebrown":"saddlebrown",
        "SaddleBrown":"saddlebrown",
        "salmon":"salmon",
        "salmon1":"#FF8C69",
        "salmon2":"#EE8262",
        "salmon3":"#CD7054",
        "salmon4":"#8B4C39",
        "sandy_brown":"sandybrown",
        "sandybrown":"sandybrown",
        "SandyBrown":"sandybrown",
        "sea_green":"seagreen",
        "seagreen":"seagreen",
        "SeaGreen":"seagreen",
        "SeaGreen1":"#54FF9F",
        "SeaGreen2":"#4EEE94",
        "SeaGreen3":"#43CD80",
        "SeaGreen4":"seagreen",
        "seashell":"seashell",
        "seashell1":"seashell",
        "seashell2":"#EEE5DE",
        "seashell3":"#CDC5BF",
        "seashell4":"#8B8682",
        "sienna":"sienna",
        "sienna1":"#FF8247",
        "sienna2":"#EE7942",
        "sienna3":"#CD6839",
        "sienna4":"#8B4726",
        "silver":"silver",
        "sky_blue":"skyblue",
        "skyblue":"skyblue",
        "SkyBlue":"skyblue",
        "SkyBlue1":"#87CEFF",
        "SkyBlue2":"#7EC0EE",
        "SkyBlue3":"#6CA6CD",
        "SkyBlue4":"#4A708B",
        "slate_blue":"slateblue",
        "slate_gray":"slategray",
        "slate_grey":"slategray",
        "slateblue":"slateblue",
        "SlateBlue":"slateblue",
        "SlateBlue1":"#836FFF",
        "SlateBlue2":"#7A67EE",
        "SlateBlue3":"#6959CD",
        "SlateBlue4":"#473C8B",
        "slategray":"slategray",
        "SlateGray":"slategray",
        "SlateGray1":"#C6E2FF",
        "SlateGray2":"#B9D3EE",
        "SlateGray3":"#9FB6CD",
        "SlateGray4":"#6C7B8B",
        "slategrey":"slategray",
        "SlateGrey":"slategray",
        "snow":"snow",
        "snow1":"snow",
        "snow2":"#EEE9E9",
        "snow3":"#CDC9C9",
        "snow4":"#8B8989",
        "spring_green":"springgreen",
        "springgreen":"springgreen",
        "SpringGreen":"springgreen",
        "SpringGreen1":"springgreen",
        "SpringGreen2":"#00EE76",
        "SpringGreen3":"#00CD66",
        "SpringGreen4":"#008B45",
        "steel_blue":"steelblue",
        "steelblue":"steelblue",
        "SteelBlue":"steelblue",
        "SteelBlue1":"#63B8FF",
        "SteelBlue2":"#5CACEE",
        "SteelBlue3":"#4F94CD",
        "SteelBlue4":"#36648B",
        "tan":"tan",
        "tan1":"#FFA54F",
        "tan2":"#EE9A49",
        "tan3":"peru",
        "tan4":"#8B5A2B",
        "teal":"teal",
        "thistle":"thistle",
        "thistle1":"#FFE1FF",
        "thistle2":"#EED2EE",
        "thistle3":"#CDB5CD",
        "thistle4":"#8B7B8B",
        "tomato":"tomato",
        "tomato1":"tomato",
        "tomato2":"#EE5C42",
        "tomato3":"#CD4F39",
        "tomato4":"#8B3626",
        "turquoise":"turquoise",
        "turquoise1":"#00F5FF",
        "turquoise2":"#00E5EE",
        "turquoise3":"#00C5CD",
        "turquoise4":"#00868B",
        "violet":"violet",
        "violet_red":"#D02090",
        "VioletRed":"#D02090",
        "VioletRed1":"#FF3E96",
        "VioletRed2":"#EE3A8C",
        "VioletRed3":"#CD3278",
        "VioletRed4":"#8B2252",
        "wheat":"wheat",
        "wheat1":"#FFE7BA",
        "wheat2":"#EED8AE",
        "wheat3":"#CDBA96",
        "wheat4":"#8B7E66",
        "white":"white",
        "white_smoke":"whitesmoke",
        "whitesmoke":"whitesmoke",
        "WhiteSmoke":"whitesmoke",
        "yellow":"yellow",
        "yellow_green":"yellowgreen",
        "yellow1":"yellow",
        "yellow2":"#EEEE00",
        "yellow3":"#CDCD00",
        "yellow4":"#8B8B00",
        "yellowgreen":"yellowgreen",
        "YellowGreen":"yellowgreen",
    
        (: Werner colours from Werner&#x27;s Nomenclature of Colours :)
        (: Whites :)
        "SnowWhite": "#F1E9CD",
        "snow_white": "#F1E9CD",
        "snow-white": "#F1E9CD",
        "werner1": "#F1E9CD",
        "ReddishWhite": "#F2E7CF",
        "reddish_white": "#F2E7CF",
        "reddish-white": "#F2E7CF",
        "werner2": "#F2E7CF",
        "PurplishWhite": "#ECE6D0",
        "purplish_white": "#ECE6D0",
        "purplish-white": "#ECE6D0",
        "werner3": "#ECE6D0",
        "YellowishWhite": "#F2EACC",
        "yellowish_white": "#F2EACC",
        "yellowish-white": "#F2EACC",
        "werner4": "#F2EACC",
        "OrangeColouredWhite": "#F3E9CA",
        "orange_coloured_white": "#F3E9CA",
        "orange-coloured-white": "#F3E9CA",
        "werner5": "#F3E9CA",
        "GreenishWhite": "#F2EBCD",
        "greenish_white": "#F2EBCD",
        "greenish-white": "#F2EBCD",
        "werner6": "#F2EBCD",
        "SkimmedMilkWhite": "#E6E1C9",
        "skimmed_milk_white": "#E6E1C9",
        "skimmed-milk-white": "#E6E1C9",
        "werner7": "#E6E1C9",
        "GreyishWhite": "#E2DDC6",
        "greyish_white": "#E2DDC6",
        "greyish-white": "#E2DDC6",
        "werner8": "#E2DDC6",
    
        (: Greys :)
        "AshGrey": "#CBC8B7",
        "ash_grey": "#CBC8B7",
        "ash-grey": "#CBC8B7",
        "werner9": "#CBC8B7",
        "SmokeGrey": "#BFBBB0",
        "smoke_grey": "#BFBBB0",
        "smoke-grey": "#BFBBB0",
        "werner10": "#BFBBB0",
        "FrenchGrey": "#BEBEB3",
        "french_grey": "#BEBEB3",
        "french-grey": "#BEBEB3",
        "werner11": "#BEBEB3",
        "PearlGrey": "#B7B5AC",
        "pearl_grey": "#B7B5AC",
        "pearl-grey": "#B7B5AC",
        "werner12": "#B7B5AC",
        "YellowishGrey": "#BAB191",
        "yellowish_grey": "#BAB191",
        "yellowish-grey": "#BAB191",
        "werner13": "#BAB191",
        "BluishGrey": "#9C9D9A",
        "bluish_grey": "#9C9D9A",
        "bluish-grey": "#9C9D9A",
        "werner14": "#9C9D9A",
        "GreenishGrey": "#8A8D84",
        "greenish_grey": "#8A8D84",
        "greenish-grey": "#8A8D84",
        "werner15": "#8A8D84",
        "BlackishGrey": "#5B5C61",
        "blackish_grey": "#5B5C61",
        "blackish-grey": "#5B5C61",
        "werner16": "#5B5C61",
    
        (: Blacks :)
        "GreyishBlack": "#555152",
        "greyish_black": "#555152",
        "greyish-black": "#555152",
        "werner17": "#555152",
        "BluishBlack": "#413F44",
        "bluish_black": "#413F44",
        "bluish-black": "#413F44",
        "werner18": "#413F44",
        "GreenishBlack": "#454445",
        "greenish_black": "#454445",
        "greenish-black": "#454445",
        "werner19": "#454445",
        "PitchOrBrownishBlack": "#423937",
        "pitch_or_brownish_black": "#423937",
        "pitch-or-brownish-black": "#423937",
        "werner20": "#423937",
        "ReddishBlack": "#433635",
        "reddish_black": "#433635",
        "reddish-black": "#433635",
        "werner21": "#433635",
        "InkBlack": "#252024",
        "ink_black": "#252024",
        "ink-black": "#252024",
        "werner22": "#252024",
        "VelvetBlack": "#241F20",
        "velvet_black": "#241F20",
        "velvet-black": "#241F20",
        "werner23": "#241F20",
    
        (: Blues :) 
        "ScotchBlue": "#281F3F",
        "scotch_blue": "#281F3F",
        "scotch-blue": "#281F3F",
        "werner24": "#281F3F",
        "PrussianBlue": "#1C1949",
        "prussian_blue": "#1C1949",
        "prussian-blue": "#1C1949",
        "werner25": "#1C1949",
        "IndigoBlue": "#4F638D",
        "indigo_blue": "#4F638D",
        "indigo-blue": "#4F638D",
        "werner26": "#4F638D",
        "ChinaBlue": "#383867",
        "china_blue": "#383867",
        "china-blue": "#383867",
        "werner27": "#383867",
        "AzureBlue": "#5C6B8F",
        "azure_blue": "#5C6B8F",
        "azure-blue": "#5C6B8F",
        "werner28": "#5C6B8F",
        "UltramarineBlue": "#657ABB",
        "ultramarine_blue": "#657ABB",
        "ultramarine-blue": "#657ABB",
        "werner29": "#657ABB",
        "FlaxFlowerBlue": "#6F88AF",
        "flax_flower_blue": "#6F88AF",
        "flax-flower-blue": "#6F88AF",
        "werner30": "#6F88AF",
        "BerlinBlue": "#7994B5",
        "berlin_blue": "#7994B5",
        "berlin-blue": "#7994B5",
        "werner31": "#7994B5",
        "VerditterBlue": "#6FB5A8",
        "verditter_blue": "#6FB5A8",
        "verditter-blue": "#6FB5A8",
        "werner32": "#6FB5A8",
        "GreenishBlue": "#719BA2",
        "greenish_blue": "#719BA2",
        "greenish-blue": "#719BA2",
        "werner33": "#719BA2",
        "GreyishBlue": "#8AA1A6",
        "greyish_blue": "#8AA1A6",
        "greyish-blue": "#8AA1A6",
        "werner34": "#8AA1A6",
    
        (: Purples:) 
        "BluishLilacPurple": "#D0D5D3",
        "bluish_lilac_purple": "#D0D5D3",
        "bluish-lilac-purple": "#D0D5D3",
        "werner35": "#D0D5D3",
        "BluishPurple": "#8590AE",
        "bluish_purple": "#8590AE",
        "bluish-purple": "#8590AE",
        "werner36": "#8590AE",
        "VioletPurple": "#3A2F52",
        "violet_purple": "#3A2F52",
        "violet-purple": "#3A2F52",
        "werner37": "#3A2F52",
        "PansyPurple": "#39334A",
        "pansy_purple": "#39334A",
        "pansy-purple": "#39334A",
        "werner38": "#39334A",
        "CampanulaPurple": "#6C6D94",
        "campanula_purple": "#6C6D94",
        "campanula-purple": "#6C6D94",
        "werner39": "#6C6D94",
        "ImperialPurple": "#584C77",
        "imperial_purple": "#584C77",
        "imperial-purple": "#584C77",
        "werner40": "#584C77",
        "AuriculaPurple": "#533552",
        "auricula_purple": "#533552",
        "auricula-purple": "#533552",
        "werner41": "#533552",
        "PlumPurple": "#463759",
        "plum_purple": "#463759",
        "plum-purple": "#463759",
        "werner42": "#463759",
        "RedLilacPurple": "#BFBAC0",
        "red_lilac_purple": "#BFBAC0",
        "red-lilac-purple": "#BFBAC0",
        "werner43": "#BFBAC0",
        "LavenderPurple": "#77747F",
        "lavender_purple": "#77747F",
        "lavender-purple": "#77747F",
        "werner44": "#77747F",
        "PaleBlackishPurple": "#4A475C",
        "pale_blackish_purple": "#4A475C",
        "pale-blackish-purple": "#4A475C",
        "werner45": "#4A475C",
    
        (: Greens :) 
        "CeladineGreen": "#B8BFAF",
        "celadine_green": "#B8BFAF",
        "celadine-green": "#B8BFAF",
        "werner46": "#B8BFAF",
        "MountainGreen": "#B2B599",
        "mountain_green": "#B2B599",
        "mountain-green": "#B2B599",
        "werner47": "#B2B599",
        "LeekGreen": "#979C84",
        "leek_green": "#979C84",
        "leek-green": "#979C84",
        "werner48": "#979C84",
        "BlackishGreen": "#5D6161",
        "blackish_green": "#5D6161",
        "blackish-green": "#5D6161",
        "werner49": "#5D6161",
        "VerdigrisGreen": "#61AC86",
        "verdigris_green": "#61AC86",
        "verdigris-green": "#61AC86",
        "werner50": "#61AC86",
        "BluishGreen": "#A4B6A7",
        "bluish_green": "#A4B6A7",
        "bluish-green": "#A4B6A7",
        "werner51": "#A4B6A7",
        "AppleGreen": "#ADBA98",
        "apple_green": "#ADBA98",
        "apple-green": "#ADBA98",
        "werner52": "#ADBA98",
        "EmeraldGreen": "#93B778",
        "emerald_green": "#93B778",
        "emerald-green": "#93B778",
        "werner53": "#93B778",
        "GrassGreen": "#7D8C55",
        "grass_green": "#7D8C55",
        "grass-green": "#7D8C55",
        "werner54": "#7D8C55",
        "DuckGreen": "#33431E",
        "duck_green": "#33431E",
        "duck-green": "#33431E",
        "werner55": "#33431E",
        "SapGreen": "#7C8635",
        "sap_green": "#7C8635",
        "sap-green": "#7C8635",
        "werner56": "#7C8635",
        "PistachioGreen": "#8E9849",
        "pistachio_green": "#8E9849",
        "pistachio-green": "#8E9849",
        "werner57": "#8E9849",
        "AsparagusGreen": "#C2C190",
        "asparagus_green": "#C2C190",
        "asparagus-green": "#C2C190",
        "werner58": "#C2C190",
        "OliveGreen": "#67765B",
        "olive_green": "#67765B",
        "olive-green": "#67765B",
        "werner59": "#67765B",
        "OilGreen": "#AB924B",
        "oil_green": "#AB924B",
        "oil-green": "#AB924B",
        "werner60": "#AB924B",
        "SiskinGreen": "#C8C76F",
        "siskin_green": "#C8C76F",
        "siskin-green": "#C8C76F",
        "werner61": "#C8C76F",
    
        (: Yellows :) 
        "SulphurYellow": "#CCC050",
        "sulphur_yellow": "#CCC050",
        "sulphur-yellow": "#CCC050",
        "werner62": "#CCC050",
        "PrimroseYellow": "#EBDD99",
        "primrose_yellow": "#EBDD99",
        "primrose-yellow": "#EBDD99",
        "werner63": "#EBDD99",
        "WaxYellow": "#AB9649",
        "wax_yellow": "#AB9649",
        "wax-yellow": "#AB9649",
        "werner64": "#AB9649",
        "LemonYellow": "#DBC364",
        "lemon_yellow": "#DBC364",
        "lemon-yellow": "#DBC364",
        "werner65": "#DBC364",
        "GambogeYellow": "#E6D058",
        "gamboge_yellow": "#E6D058",
        "gamboge-yellow": "#E6D058",
        "werner66": "#E6D058",
        "KingsYellow": "#EAD665",
        "kings_yellow": "#EAD665",
        "kings-yellow": "#EAD665",
        "werner67": "#EAD665",
        "SaffronYellow": "#D09B2C",
        "saffron_yellow": "#D09B2C",
        "saffron-yellow": "#D09B2C",
        "werner68": "#D09B2C",
        "GallstoneYellow": "#A36629",
        "gallstone_yellow": "#A36629",
        "gallstone-yellow": "#A36629",
        "werner69": "#A36629",
        "HoneyYellow": "#A77D35",
        "honey_yellow": "#A77D35",
        "honey-yellow": "#A77D35",
        "werner70": "#A77D35",
        "StrawYellow": "#F0D696",
        "straw_yellow": "#F0D696",
        "straw-yellow": "#F0D696",
        "werner71": "#F0D696",
        "WineYellow": "#D7C485",
        "wine_yellow": "#D7C485",
        "wine-yellow": "#D7C485",
        "werner72": "#D7C485",
        "SiennaYellow": "#F1D28C",
        "sienna_yellow": "#F1D28C",
        "sienna-yellow": "#F1D28C",
        "werner73": "#F1D28C",
        "OchreYellow": "#EFCC83",
        "ochre_yellow": "#EFCC83",
        "ochre-yellow": "#EFCC83",
        "werner74": "#EFCC83",
        "CreamYellow": "#F3DAA7",
        "cream_yellow": "#F3DAA7",
        "cream-yellow": "#F3DAA7",
        "werner75": "#F3DAA7",
    
        (: Oranges :) 
        "DutchOrange": "#DFA837",
        "dutch_orange": "#DFA837",
        "dutch-orange": "#DFA837",
        "werner76": "#DFA837",
        "BuffOrange": "#EBBC71",
        "buff_orange": "#EBBC71",
        "buff-orange": "#EBBC71",
        "werner77": "#EBBC71",
        "OrpimentOrange": "#D17C3F",
        "orpiment_orange": "#D17C3F",
        "orpiment-orange": "#D17C3F",
        "werner78": "#D17C3F",
        "BrownishOrange": "#92462F",
        "brownish_orange": "#92462F",
        "brownish-orange": "#92462F",
        "werner79": "#92462F",
        "ReddishOrange": "#BE7249",
        "reddish_orange": "#BE7249",
        "reddish-orange": "#BE7249",
        "werner80": "#BE7249",
        "DeepReddishOrange": "#BB603C",
        "deep_reddish_orange": "#BB603C",
        "deep-reddish-orange": "#BB603C",
        "werner81": "#BB603C",
    
        (: Reds :) 
        "TileRed": "#C76B4A",
        "tile_red": "#C76B4A",
        "tile-red": "#C76B4A",
        "werner82": "#C76B4A",
        "HyacinthRed": "#A75536",
        "hyacinth_red": "#A75536",
        "hyacinth-red": "#A75536",
        "werner83": "#A75536",
        "ScarletRed": "#B63E36",
        "scarlet_red": "#B63E36",
        "scarlet-red": "#B63E36",
        "werner84": "#B63E36",
        "VermilionRed": "#B5493A",
        "vermilion_red": "#B5493A",
        "vermilion-red": "#B5493A",
        "werner85": "#B5493A",
        "AuroraRed": "#CD6D57",
        "aurora_red": "#CD6D57",
        "aurora-red": "#CD6D57",
        "werner86": "#CD6D57",
        "ArterialBloodRed": "#711518",
        "arterial_blood_red": "#711518",
        "arterial-blood-red": "#711518",
        "werner87": "#711518",
        "FleshRed": "#E9C49D",
        "flesh_red": "#E9C49D",
        "flesh-red": "#E9C49D",
        "werner88": "#E9C49D",
        "RoseRed": "#EEDAC3",
        "rose_red": "#EEDAC3",
        "rose-red": "#EEDAC3",
        "werner89": "#EEDAC3",
        "PeachBlossomRed": "#EECFBF",
        "peach_blossom_red": "#EECFBF",
        "peach-blossom-red": "#EECFBF",
        "werner90": "#EECFBF",
        "CarmineRed": "#CE536B",
        "carmine_red": "#CE536B",
        "carmine-red": "#CE536B",
        "werner91": "#CE536B",
        "LakeRed": "#B74A70",
        "lake_red": "#B74A70",
        "lake-red": "#B74A70",
        "werner92": "#B74A70",
        "CrimsonRed": "#B7757C",
        "crimson_red": "#B7757C",
        "crimson-red": "#B7757C",
        "werner93": "#B7757C",
        "PurplishRed": "#612741",
        "purplish_red": "#612741",
        "purplish-red": "#612741",
        "werner94": "#612741",
        "CochinealRed": "#7A4848",
        "cochineal_red": "#7A4848",
        "cochineal-red": "#7A4848",
        "werner95": "#7A4848",
        "VeinousBloodRed": "#3F3033",
        "veinous_blood_red": "#3F3033",
        "veinous-blood-red": "#3F3033",
        "werner96": "#3F3033",
        "BrownishPurpleRed": "#8D746F",
        "brownish_purple_red": "#8D746F",
        "brownish-purple-red": "#8D746F",
        "werner97": "#8D746F",
        "ChocolateRed": "#4D3635",
        "chocolate_red": "#4D3635",
        "chocolate-red": "#4D3635",
        "werner98": "#4D3635",
        "BrownishRed": "#6E3B31",
        "brownish_red": "#6E3B31",
        "brownish-red": "#6E3B31",
        "werner99": "#6E3B31",
    
        (: Browns :) 
        "DeepOrangeColouredBrown": "#864735",
        "deep_orange_coloured_brown": "#864735",
        "deep-orange-coloured-brown": "#864735",
        "werner100": "#864735",
        "DeepReddishBrown": "#553D3A",
        "deep_reddish_brown": "#553D3A",
        "deep-reddish-brown": "#553D3A",
        "werner101": "#553D3A",
        "UmberBrown": "#613936",
        "umber_brown": "#613936",
        "umber-brown": "#613936",
        "werner102": "#613936",
        "ChestnutBrown": "#7A4B3A",
        "chestnut_brown": "#7A4B3A",
        "chestnut-brown": "#7A4B3A",
        "werner103": "#7A4B3A",
        "YellowishBrown": "#946943",
        "yellowish_brown": "#946943",
        "yellowish-brown": "#946943",
        "werner104": "#946943",
        "WoodBrown": "#C39E6D",
        "wood_brown": "#C39E6D",
        "wood-brown": "#C39E6D",
        "werner105": "#C39E6D",
        "LiverBrown": "#513E32",
        "liver_brown": "#513E32",
        "liver-brown": "#513E32",
        "werner106": "#513E32",
        "HairBrown": "#8B7859",
        "hair_brown": "#8B7859",
        "hair-brown": "#8B7859",
        "werner107": "#8B7859",
        "BroccoliBrown": "#9B856B",
        "broccoli_brown": "#9B856B",
        "broccoli-brown": "#9B856B",
        "werner108": "#9B856B",
        "CloveBrown": "#766051",
        "clove_brown": "#766051",
        "clove-brown": "#766051",
        "werner109": "#766051",
        "BlackishBrown": "#453B32",
        "blackish_brown": "#453B32",
        "blackish-brown": "#453B32",
        "werner110": "#453B32",
    
        (: Additional colours :)
        "middle_blue_green": "#9AD2CB",
        "middle-blue-green": "#9AD2CB",
        "MiddleBlueGreen": "#9AD2CB",
        "tea_green": "#D7EBBA",
        "tea-green": "#D7EBBA",
        "TeaGreen": "#D7EBBA",
        "cream": "#FEFFBE",
        "Cream": "#FEFFBE",
        "medium_champagne": "#EBD494",
        "medium-champagne": "#EBD494",
        "MediumChampagne": "#EBD494",
        "dark_byzantium": "#472835",
        "dark-byzantium": "#472835",
        "DarkByzantium": "#472835",
        
        "RedHotPoker": "#B17A5B",
        "red_hot_poker": "#B17A5B",
        "red-hot-poker": "#B17A5B",
        "HawthornLeaf": "#D97036",
        "hawthorn_leaf": "#D97036",
        "hawthorn-leaf": "#D97036",
        "BogMyrtle": "#FFC24B",
        "bog_myrtle": "#FFC24B",
        "bog-myrtle": "#FFC24B",
        "Silverweed": "#FECA75",
        "silverweed": "#FECA75",
        "GorseFlower": "#FAC567",
        "gorse_flower": "#FAC567",
        "gorse-flower": "#FAC567",
        "ReedFlower": "#C6AA26",
        "reed_flower": "#C6AA26",
        "reed-flower": "#C6AA26",
        "NettleLeaf": "#BBAC8A",
        "nettle_leaf": "#BBAC8A",
        "nettle-leaf": "#BBAC8A",
        "ColtsFoot": "#C8B995",
        "colts_foot": "#C8B995",
        "colts-foot": "#C8B995",
        "Acorn": "#6B6A69",
        "acorn": "#6B6A69",
        "IrisRoot": "#736B68",
        "iris_root": "#736B68",
        "iris-root": "#736B68"
    }'/>
   </xsl:variable>
</xsl:package>

<!--rgb.xsl-->
