<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Tonemapping
 : References:
 : https://64.github.io/tonemapping/ 
 : https://bartwronski.com/2016/09/01/dynamic-range-and-evs/
 : https://knarkowicz.wordpress.com/2016/01/06/aces-filmic-tone-mapping-curve/
 : 
 : There are two parallel APIs here: the higher level one takes XYZ points
 : and produces XYZ points, and the lower level one takes and produces
 : raw XYZ coordinates. You could use raw RGB coordinates instead, but
 : but the results aren't as nice (same as when you try to do interpolations
 : over raw RGB). If you are going to perform a chain of operations, the
 : lower-level API will be more efficient (less conversion/construction
 : between steps). 
 : 
 : You will need to call saturate() or saturatev() before you convert back
 : to RGB or risk getting illegal RGB values.
 :
 : Copyright© Mary Holstege 2021-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/tonemap"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:cs="http://mathling.com/colour/space"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xyz="http://mathling.com/colour/xyz"
             version="3.0"
             name="http://mathling.com/colour/tonemap"
             exclude-result-prefixes="this xs math map array util v cs xyz">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/space" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/xyz" package-version="*"/>


   <!--
    : this:ACES()
    : Perform ACES (Academy Color Encoding System) filmic tone mapping
    :
    : $colour: XYZ colour
    :-->
   <xsl:function name="this:ACES" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:ACESv($colour=>xyz:raw-coordinates())=>xyz:to-xyz()
    '/>
   </xsl:function>

   <!--
    : this:ACESv()
    : Perform ACES (Academy Color Encoding System) filmic tone mapping
    :
    : $colour: XYZ input colour as raw coordinates
    :-->
   <xsl:function name="this:ACESv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:sequence select='
      (: Multiply input matrix by input colour :)
      let $v := (
        v:dot((0.59719E0, 0.35458E0, 0.04823E0), $colour),
        v:dot((0.07600E0, 0.90834E0, 0.01566E0), $colour),
        v:dot((0.02840E0, 0.13383E0, 0.84777E0), $colour)
      ) return
      (: Curve fitting :)
      let $a := (: v * (v + 0.0245786) - 0.000090537 :)
        $v=>
          v:multiply($v=>v:plus(0.0245786E0))=>
          v:plus(-0.000090537E0)
      return
      let $b := (: v * (0.983729 * v + 0.4329510) + 0.238081 :)
        $v=>
          v:multiply(($v=>v:times(0.983729E0))=>v:plus(0.4329510E0))=>
          v:plus(0.238081E0)
      return
      let $ab := $a=>v:divide($b) return
      (: Multiply by output matrix to get output colour :)
      let $m2ab := (
        (1.60475E0, -0.53108E0, -0.07367E0)=>v:dot($ab),
        (-0.10208E0, 1.10813E0, -0.00605E0)=>v:dot($ab),
        (-0.00327E0, -0.07276E0, 1.07602E0)=>v:dot($ab)
      )
      return (
        $m2ab
      )
    '/>
   </xsl:function>

   <!--
    : this:Hablic-partial()
    : Basic Hablic operator with default parameters
    :
    : $colour: XYZ input colour as raw coordinates
    : $curve: Hablic constants in order (A, B, C, D, E, F)
    :-->
   <xsl:function name="this:Hablic-partial" as="xs:double*" visibility="private" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="curve" as="xs:double*"/>
      <xsl:sequence select='
      let $A := $curve[1] return
      let $B := $curve[2] return
      let $C := $curve[3] return
      let $D := $curve[4] return
      let $E := $curve[5] return
      let $F := $curve[6]
      return (
        v:map(
          function($x) {
            (($x*($A*$x+$C*$B)+$D*$E) div ($x*($A*$x+$B)+$D*$F))-$E div $F
          },
          $colour
        )
      )
    '/>
   </xsl:function>

   <!--
    : $this:HABLIC-DEFAULTS
    : Default Hablic constants
    : A=shoulder strength, B=linear strength, C=linear angle, 
    : D=toe strength, E=toe numerator, F=toe denominator, E/F=toe angle
    :-->
   <xsl:variable name="this:HABLIC-DEFAULTS" as="xs:double*">
      <xsl:sequence select='(
      (: A :) 0.15E0, (: 0.22 :)
      (: B :) 0.50E0, (: 0.30 :)
      (: C :) 0.10E0,
      (: D :) 0.20E0,
      (: E :) 0.02E0, (: 0.01 :)
      (: F :) 0.30E0
    )'/>
   </xsl:variable>

   <!--
    : this:Hablic()
    : Hablic tone mapping with default values.
    :
    : $colour: XYZ point
    :-->
   <xsl:function name="this:Hablic" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:Hablic($colour, 2.0E0, 11.2E0, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablicv()
    : Hablic tone mapping with default values.
    :
    : $colour: XYZ input colour as raw coordinates
    :-->
   <xsl:function name="this:Hablicv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:sequence select='
      this:Hablicv($colour, 2.0E0, 11.2E0, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablic()
    : Hablic tone mapping with default values except for exposure.
    :
    : $colour: XYZ input colour as raw coordinates
    : $exposure: amount of exposure
    :-->
   <xsl:function name="this:Hablic" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:sequence select='
      this:Hablic($colour, $exposure, 11.2E0, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablicv()
    : Hablic tone mapping with default values except for exposure.
    :
    : $colour: XYZ input colour as raw coordinates
    : $exposure: amount of exposure
    :-->
   <xsl:function name="this:Hablicv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:sequence select='
      this:Hablicv($colour, $exposure, 11.2E0, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablic()
    : Hablic tone mapping with default curve constants.
    :
    : $colour: XYZ colour point
    : $exposure: amount of exposure
    : $white: white point
    :-->
   <xsl:function name="this:Hablic" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:sequence select='
      this:Hablic($colour, $exposure, $white, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablicv()
    : Hablic tone mapping with default curve constants.
    :
    : $colour: XYZ input colour as raw coordinates
    : $exposure: amount of exposure
    : $white: white point
    :-->
   <xsl:function name="this:Hablicv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:sequence select='
      this:Hablicv($colour, $exposure, $white, $this:HABLIC-DEFAULTS)
    '/>
   </xsl:function>

   <!--
    : this:Hablic()
    : Hablic tone mapping with specific parameters.
    :
    : $colour: XYZ colour point
    : $exposure: amount of exposure
    : $white: white point
    : $curve: Hablic constants in order (A, B, C, D, E, F)
    :-->
   <xsl:function name="this:Hablic" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:param name="curve" as="xs:double*"/>
      <xsl:sequence select='
      xyz:to-xyz(
        this:Hablic-partial(xyz:raw-coordinates($colour), $curve)=>
          v:divide(this:Hablic-partial($white, $curve))
      )
    '/>
   </xsl:function>

   <!--
    : this:Hablicv()
    : Hablic tone mapping with specific parameters.
    :
    : $colour: XYZ input colour as raw coordinates
    : $exposure: amount of exposure
    : $white: white point
    : $curve: Hablic constants in order (A, B, C, D, E, F)
    :-->
   <xsl:function name="this:Hablicv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="exposure" as="xs:double"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:param name="curve" as="xs:double*"/>
      <xsl:sequence select='
      this:Hablic-partial($colour, $curve)=>
        v:divide(this:Hablic-partial($white, $curve))
    '/>
   </xsl:function>

   <!--
    : this:simple-Reinhard()
    : Basic Reinhard tonemapping: tends to grey out colours
    :
    : $colour: XYZ colour point
    :-->
   <xsl:function name="this:simple-Reinhard" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      xyz:to-xyz(this:simple-Reinhardv(xyz:raw-coordinates($colour)))
    '/>
   </xsl:function>

   <!--
    : this:simple-Reinhardv()
    : Basic Reinhard tonemapping: tends to grey out colours
    :
    : $colour: XYZ input colour as raw coordinates
    :-->
   <xsl:function name="this:simple-Reinhardv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double"/>
      <xsl:sequence select='
      $colour=>v:divide($colour=>v:plus(1E0))
    '/>
   </xsl:function>

   <!--
    : this:Reinhard()
    : (Extended) Reinhard tonemapping with luminance
    :
    : $colour: XYZ colour point
    : $white: white point
    :-->
   <xsl:function name="this:Reinhard" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:sequence select='
      xyz:to-xyz(this:Reinhardv(xyz:raw-coordinates($colour), $white))
    '/>
   </xsl:function>

   <!--
    : this:Reinhardv()
    : (Extended) Reinhard tonemapping with luminance
    :
    : $colour: XYZ input colour as raw coordinates
    : $white: white point
    :-->
   <xsl:function name="this:Reinhardv" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="white" as="xs:double"/>
      <xsl:sequence select='
      let $l-old := this:luminancev($colour) return
      let $numerator := $l-old * (1E0 + ($l-old div ($white*$white))) return
      let $l-new := $numerator div (1E0 + $l-old)
      return this:luminatev($colour, $l-new)
    '/>
   </xsl:function>

   <!--
    : this:luminance()
    : Get luminance of an XYZ value.
    :
    : $colour: XYZ colour point
    :-->
   <xsl:function name="this:luminance" as="xs:double" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      xyz:raw-coordinates($colour)=>v:dot((0.2125E0, 0.7152E0, 0.0722E0))
    '/>
   </xsl:function>

   <!--
    : this:luminancev()
    : Get luminance of an XYZ value.
    :
    : $colour: XYZ input colour as raw coordinates
    :-->
   <xsl:function name="this:luminancev" as="xs:double" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:sequence select='
      let $res := $colour=>v:dot((0.2125E0, 0.7152E0, 0.0722E0))
      return (
        if (some $c in $res satisfies $c != $c) then (
          trace((), "luninance XYZ=>NaN: "||util:quote($colour))
        ) else (),
        $res
      )
    '/>
   </xsl:function>

   <!--
    : this:luminate()
    : Change the luminance to the new luminance.
    :
    : $colour: XYZ input colour as raw coordinates
    : $luminance: Output luminance
    :-->
   <xsl:function name="this:luminate" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:param name="luminance" as="xs:double"/>
      <xsl:sequence select='
      this:luminatev(xyz:raw-coordinates($colour), $luminance)=>xyz:to-xyz()
    '/>
   </xsl:function>

   <!--
    : this:luminatev()
    : Change the luminance to the new luminance.
    :
    : $colour: XYZ input colour as raw coordinates
    : $luminance: Output luminance
    :-->
   <xsl:function name="this:luminatev" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:param name="luminance" as="xs:double"/>
      <xsl:sequence select='
      let $l := this:luminancev($colour)
      return
        if ($l = 0)
        then (0.0, 0.0, 0.0) (: black :)
        else $colour=>v:times($luminance div $l)
    '/>
   </xsl:function>

   <!--
    : this:saturate()
    : Generally the final step in tone-mapping: clamping colour values
    : If you use just this you'll get washed out whites
    :
    : $colour: XYZ colour point
    :-->
   <xsl:function name="this:saturate" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      xyz:saturate($colour)
    '/>
   </xsl:function>

   <!--
    : this:saturatev()
    : Generally the final step in tone-mapping: clamping colour values
    : If you use just this you'll get washed out whites
    :
    : $colour: XYZ input colour as raw coordinates
    :-->
   <xsl:function name="this:saturatev" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:sequence select='
      v:clamp($colour, 0.0E0, 1.0E0)
    '/>
   </xsl:function>

   <!--
    : this:gamma()
    : Inverse OETF gamma correction assuming 2.2 gamma display
    :-->
   <xsl:function name="this:gamma" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:gammav(xyz:raw-coordinates($colour))=>xyz:to-xyz()
    '/>
   </xsl:function>

   <!--
    : this:gammav()
    : Inverse OETF gamma correction assuming 2.2 gamma display
    :-->
   <xsl:function name="this:gammav" as="xs:double*" new-each-time="no">
      <xsl:param name="colour" as="xs:double*"/>
      <xsl:sequence select='
      v:map(
        function($c as xs:double) { math:pow($c, 1E0 div 2.2E0) }, 
        $colour
      )
    '/>
   </xsl:function>

   <!--
    : this:cmyk-snap()
    : Roundtrip to CMYK and back to get colours as printed.
    :
    : $colour: RGB colour
    :-->
   <xsl:function name="this:cmyk-snap" as="map(xs:string, item()*)" new-each-time="no">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $colour=>cs:rgb-to-cmyk()=>cs:cmyk-to-rgb()
    '/>
   </xsl:function>

</xsl:package>
<!--tonemap.xsl-->
