<?xml version="1.0" encoding="UTF-8"?>
<!--
 : XYZ points; linear space (used for internal colour-space conversions)
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/xyz"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/xyz"
             exclude-result-prefixes="this xs math map array point v util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <!--
    : this:xyz()
    : Make an XYZ colour point
    :
    : $x: x value
    : $y: y value
    : $z: z value
    :-->
   <xsl:function name="this:xyz" as="map(xs:string, item()*)">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:sequence select='
      point:point($x, $y, $z)
    '/>
   </xsl:function>

   <!--
    : this:xyza()
    : Make an XYZ colour point with alpha channel
    :
    : $x: x value
    : $y: y value
    : $z: z value
    : $alpha: alpha value
    :-->
   <xsl:function name="this:xyza" as="map(xs:string, item()*)">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:sequence select='
      point:point($x, $y, $z, $alpha)
    '/>
   </xsl:function>

   <!--
    : this:x()
    : Accessor for x value of the point
    :
    : $xyz The point
    :-->
   <xsl:function name="this:x" as="xs:double">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($xyz)
    '/>
   </xsl:function>

   <!--
    : this:y()
    : Accessor for y value of the point
    :
    : $xyz The point
    :-->
   <xsl:function name="this:y" as="xs:double">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($xyz)
    '/>
   </xsl:function>

   <!--
    : this:z()
    : Accessor for z value of the point
    :
    : $xyz The point
    :-->
   <xsl:function name="this:z" as="xs:double">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($xyz)
    '/>
   </xsl:function>

   <!--
    : this:alpha()
    : Accessor for alpha value of the point
    :
    : $xyz The point
    :-->
   <xsl:function name="this:alpha" as="xs:double">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:dimension($xyz)=4) then point:pw($xyz) else 1.0
    '/>
   </xsl:function>

   <!--
    : this:to-xyz()
    : Construct an XYZ point from a vector.
    :-->
   <xsl:function name="this:to-xyz" as="map(xs:string, item()*)">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      if (v:dimension($v)=4)
      then this:xyza(v:px($v), v:py($v), v:pz($v), v:pw($v))
      else this:xyz(v:px($v), v:py($v), v:pz($v))
    '/>
   </xsl:function>

   <!--
    : this:raw-coordinates()
    : Get the raw coordinates, suitable for tonemapping etc.
    :-->
   <xsl:function name="this:raw-coordinates" as="xs:double*">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pcoordinates($xyz)
    '/>
   </xsl:function>

   <!--
    : this:saturate()
    : When we play games with interpolations and multiplications, we can 
    : overrun the legal bounds. This caps things back in bounds.
    :-->
   <xsl:function name="this:saturate" as="map(xs:string, item()*)">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:dimension($xyz)=4) then ( 
        this:xyza(
          util:clamp(this:x($xyz), 0.0, 1.0),
          util:clamp(this:y($xyz), 0.0, 1.0),
          util:clamp(this:z($xyz), 0.0, 1.0),
          util:clamp(this:alpha($xyz), 0.0, 1.0)
        )
      ) else (
        this:xyz(
          util:clamp(this:x($xyz), 0.0, 1.0),
          util:clamp(this:y($xyz), 0.0, 1.0),
          util:clamp(this:z($xyz), 0.0, 1.0)
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert XYZ point value to string representation suitable for debugging
    :
    : $xyz The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:dimension($xyz)=4) then (
        "xyza("||
          this:x($xyz)||","||
          this:y($xyz)||","||
          this:z($xyz)||","||
          this:alpha($xyz)||
        ")"
      ) else (
        "xyz("||
          this:x($xyz)||","||
          this:y($xyz)||","||
          this:z($xyz)||
        ")"
      )
    '/>
   </xsl:function>

   <!--
    : this:alpha-blend()
    : Blend colours channel by channel
    :-->
   <xsl:function name="this:alpha-blend" as="map(xs:string, item()*)">
      <xsl:param name="xyza1" as="map(xs:string, item()*)"/>
      <xsl:param name="xyza2" as="map(xs:string, item()*)"/>
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:sequence select='
      this:xyza(
        this:x($xyza2) * $alpha + (this:x($xyza1) * (1 - $alpha)),
        this:y($xyza2) * $alpha + (this:y($xyza1) * (1 - $alpha)),
        this:z($xyza2) * $alpha + (this:z($xyza1) * (1 - $alpha)),
        this:alpha($xyza2) * $alpha + (this:alpha($xyza1) * (1 - $alpha))
      )
    '/>
   </xsl:function>

</xsl:package>
<!--xyz.xsl-->
