<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Callable function with annotations. An alternative approach to the problem (vs
 : special eval'd functions).
 :
 : Example:
 : let $example := 
 :    callable:named("ex:hello", function ($x as xs:string) {"Hello "||$x})
 : return ("Calling "||callable:name($example), callable:f($example)("world"))
 :
 : Copyright© Mary Holstege 2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/core/callable"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:art="http://mathling.com/art"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/core/callable"
             exclude-result-prefixes="this xs math map array errors">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>

   <!--
    : this:callable()
    : Create a callable annotated function.
    :
    : $f: the function
    : $annotations: map of annotations
    :-->
   <xsl:function name="this:callable" as="map(*)">
      <xsl:param name="f" as="function(*)"/>
      <xsl:param name="annotations" as="map(xs:QName, item()*)"/>
      <xsl:sequence select='
      map:merge((
        $annotations,
        map {
          "kind": "callable",
          "f": $f,
          "describe": this:describe#1
        }
      ))
    '/>
   </xsl:function>

   <!--
    : this:named()
    : Create a callable annotated function with a given name.
    :
    : $f: the function
    : $name: its name annotation
    :-->
   <xsl:function name="this:named" as="map(*)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="f" as="function(*)"/>
      <xsl:sequence select='
      map {
        "kind": "callable",
        "f": $f,
        "describe": this:describe#1,
        xs:QName("art:name"): $name
      }
    '/>
   </xsl:function>

   <!--
    : this:describe()
    : Description function for callables.
    : 
    : $callable: the callable wrapper
    :-->
   <xsl:function name="this:describe" as="xs:string">
      <xsl:param name="callable" as="map(*)"/>
      <xsl:sequence select='this:name($callable)'/>
   </xsl:function>

   <!--
    : this:annotation()
    : Get a particular annotation from the callable function.
    : 
    : $callable: the callable wrapper
    : $name: the annotation name
    :-->
   <xsl:function name="this:annotation" as="xs:string?">
      <xsl:param name="callable" as="map(*)"/>
      <xsl:param name="name" as="xs:QName"/>
      <xsl:sequence select='
      $callable($name)!string(.)
    '/>
   </xsl:function>

   <!--
    : this:annotations()
    : Get the annotation map.
    : 
    : $callable: the callable wrapper
    :-->
   <xsl:function name="this:annotations"
                 as="map(xs:QName, item()*)"
                >
      <xsl:param name="callable" as="map(*)"/>
      <xsl:sequence select='
       $callable=>map:remove("kind")=>map:remove("f")=>map:remove("describe")
    '/>
   </xsl:function>

   <!--
    : this:name()
    : Get the value of the name annotation
    :
    : $callable: the callable wrapper
    :-->
   <xsl:function name="this:name" as="xs:string">
      <xsl:param name="callable" as="map(*)"/>
      <xsl:sequence select='
      ($callable(xs:QName("art:name")),"[anon]")[1]
    '/>
   </xsl:function>

   <!--
    : this:f()
    : Get the callable function 
    :
    : $callable: the callable wrapper
    :-->
   <xsl:function name="this:f" as="function(*)">
      <xsl:param name="callable" as="map(*)"/>
      <xsl:sequence select='
      $callable("f")
    '/>
   </xsl:function>

   <!--
    : this:is-callable()
    : Item is a function or callable wrapper.
    :
    : $callable: the putative callable object
    :-->
   <xsl:function name="this:is-callable" as="xs:boolean">
      <xsl:param name="callable" as="item()"/>
      <xsl:sequence select='
     ($callable instance of function(*)) or
     ($callable instance of map(*) and $callable("kind")="callable")
    '/>
   </xsl:function>
   
   <!--
    : this:function()
    : If the argument is a function, return the function; if it is a callable wrapper,
    : return the function inside. Otherwise complain.
    :-->
   <xsl:function name="this:function" as="function(*)">
      <xsl:param name="callable" as="item()"/>
      <xsl:sequence select='
        if ($callable instance of map(*) and $callable("kind")="callable") then this:f($callable)
        else if ($callable instance of function(*)) then $callable
        else errors:error("ML-BADARGS", ("callable", $callable))
    '/>
   </xsl:function>

</xsl:package>

<!--callable.xsl-->
