<?xml version="1.0" encoding="UTF-8"?>
<!--
  : Complex numbers
  :
  : Copyright© Mary Holstege 2021-2023
  : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/core/complex"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:fn="http://www.w3.org/2005/xpath-functions"
             version="3.0"
             name="http://mathling.com/core/complex"
             exclude-result-prefixes="this xs fn math map array config errors util point">

  <xsl:expose component="variable" names="this:*" visibility="public"/>
  <xsl:expose component="function" names="this:*" visibility="public"/>

  <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>

  <xsl:variable name="this:i" as="map(xs:string,item()*)">
  <xsl:sequence select='this:complex(0, 1)'/>
  </xsl:variable>

  <xsl:variable name="this:one" as="map(xs:string,item()*)">
  <xsl:sequence select='this:complex(1, 0)'/>
  </xsl:variable>

  <!--
    : this:complex()
    : Construct a new complex number.
    :
    : $real: the real part of the number
    : $imaginary: the imaginary part of the number
    :-->
  <xsl:function name="this:complex" as="map(xs:string, item()*)">
    <xsl:param name="real" as="xs:double"/>
    <xsl:param name="imaginary" as="xs:double"/>
    <xsl:sequence select='
      point:point($real, $imaginary)
    '/>
  </xsl:function>

  <!--
    : this:as-complex()
    : Cast a real number to a complex.
    :
    : $real: the real part of the number
    :-->
  <xsl:function name="this:as-complex" as="map(xs:string, item()*)">
    <xsl:param name="real" as="xs:double"/>
    <xsl:sequence select='
      point:point($real, 0)
    '/>
  </xsl:function>

  <!--
    : this:is-real()
    : Test whether the number has no imaginary part. (Within ε). 
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:is-real" as="xs:boolean">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      abs(this:pi($z)) &lt; $config:ε
    '/>
  </xsl:function>

  <!--
    : this:as-real()
    : Cast a complex number to a double, if possible. Raises error if not.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:as-real" as="xs:double">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:is-real($z)) then this:pr($z)
      else errors:error("UTIL-CAST", ($z,"complex"))
    '/>
  </xsl:function>

  <!--
    : this:complex-rφd()
    : Construct a new complex number using radius and phase angle 
    : (polar repesentation re^iφ).
    :
    : $r: modulus of complex number
    : $φ: angle of complex number (degrees)
    :-->
  <xsl:function name="this:complex-rφd" as="map(xs:string, item()*)">
    <xsl:param name="r" as="xs:double"/>
    <xsl:param name="φ" as="xs:double"/>
    <xsl:sequence select='
      let $φR := util:radians($φ)
      return
        point:point($r * math:cos($φR), $r * math:sin($φR))
    '/>
  </xsl:function>

  <!--
    : this:complex-rφ()
    : Construct a new complex number using radius and phase angle 
    : (polar repesentation re^iφ).
    :
    : $r: modulus of complex number
    : $φ: angle of complex number (radians)
    :-->
  <xsl:function name="this:complex-rφ" as="map(xs:string, item()*)">
    <xsl:param name="r" as="xs:double"/>
    <xsl:param name="φ" as="xs:double"/>
    <xsl:sequence select='
      point:point($r * math:cos($φ), $r * math:sin($φ))
    '/>
  </xsl:function>

  <!--
    : this:pr()
    : Accessor for real part of the complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:pr" as="xs:double">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:px($z)
    '/>
  </xsl:function>

  <!--
    : this:pi()
    : Accessor for imaginary part of the complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:pi" as="xs:double">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:py($z)
    '/>
  </xsl:function>

  <!--
    : this:r()
    : Accessor for real part of the complex number, as integer.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:r" as="xs:integer">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:x($z)
    '/>
  </xsl:function>

  <!--
    : this:i()
    : Accessor for imaginary part of the complex number, as integer.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:i" as="xs:integer">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:y($z)
    '/>
  </xsl:function>

  <!--
    : this:add()
    : Add two complex numbers.
    :
    : $z1, $z2: the complex numbers to add
    :-->
  <xsl:function name="this:add" as="map(xs:string, item()*)">
    <xsl:param name="z1" as="map(xs:string, item()*)"/>
    <xsl:param name="z2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $z1=&gt;point:add($z2)
    '/>
  </xsl:function>

  <!--
    : this:sub()
    : Subtract one complex numbers from another.
    :
    : $z1: the first complex number
    : $z2: the complex number to subtract from the first complex number
    :-->
  <xsl:function name="this:sub" as="map(xs:string, item()*)">
    <xsl:param name="z1" as="map(xs:string, item()*)"/>
    <xsl:param name="z2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $z1=&gt;point:sub($z2)
    '/>
  </xsl:function>

  <!--
    : this:minus()
    : Negation
    :
    : $z1 the first complex number
    :-->
  <xsl:function name="this:minus" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $z=&gt;point:times(-1)
    '/>
  </xsl:function>

  <!--
    : this:multiply()
    : Multiply two complex numbers.
    :
    : $z1, $z2: the complex numbers to multiply
    :-->
  <xsl:function name="this:multiply" as="map(xs:string, item()*)">
    <xsl:param name="z1" as="map(xs:string, item()*)"/>
    <xsl:param name="z2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: (a + bi)*(c + di) = (ac - bd) + (ad + bc)i :)
      point:point(
        this:pr($z1)*this:pr($z2) - this:pi($z1)*this:pi($z2),
        this:pr($z1)*this:pi($z2) + this:pi($z1)*this:pr($z2)
      )
    '/>
  </xsl:function>

  <!--
    : this:pow()
    : Raise a complex number to the given (positive integer) power.
    : A convenience function for repeated multiplication. For general powers
    : use ppow()
    : 
    : $z: the complex number
    : $pow: the power to raise it to
    :-->
  <xsl:function name="this:pow" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:param name="pow" as="xs:integer"/>
    <xsl:sequence select='
      if (this:pi($z)=0) then this:complex(math:pow(this:pr($z), $pow), 0)
      else if ($pow = 0) then $this:one
      else if ($pow = 1) then $z
      else if ($pow = 2) then $z=>this:multiply($z)
      else if ($pow = 3) then $z=>this:multiply($z)=>this:multiply($z)
      else
        (: General: 
         : a + ib = r(cosφ + isinφ) r² = a² + b², φ = atan(b/a)
         : (a + ib)^N = r^N(cos(Nφ) + isin(Nφ))
         :)
        let $r := this:pr($z)*this:pr($z) + this:pi($z)*this:pi($z) return
        let $φ := math:atan2(this:pi($z), this:pr($z)) return
        let $r_N := math:pow($r,$pow)
        return
          this:complex($r_N*math:cos($pow*$φ), $r_N*math:sin($pow*$φ))
    '/>
  </xsl:function>

  <!--
    : this:ppow()
    : Principle power for non-integer powers.
    : $z: the complex number
    : $pow: the power to raise it to
    :-->
  <xsl:function name="this:ppow" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:param name="pow" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: a^b = e^(b log(a)) :)
      this:exp(this:log($z)=>this:multiply($pow))
    '/>
  </xsl:function>

  <!--
    : https://math.stackexchange.com/questions/1103102/how-to-raise-1-to-non-integer-powers 
    : http://www.milefoot.com/math/complex/functionsofi.htm 
    : http://en.wikipedia.org/wiki/Complex_number 
    :-->


  <!--
    : this:log()
    : Natural log of complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:log" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:complex(math:log(this:modulus($z)), this:angle($z)=>util:radians())
    '/>
  </xsl:function>

  <!--
    : this:sqrt()
    : Principal square root of complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:sqrt" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:pi($z) = 0) then (
        if (this:pr($z) >= 0) 
        then this:complex(math:sqrt(this:pr($z)), 0)
        else this:complex(0, math:sqrt(-this:pr($z)))
      ) else (
        let $r := this:modulus($z)
        return
          this:complex(
            math:sqrt(($r + this:pr($z)) div 2),
            util:sign(this:pi($z)) * math:sqrt(($r - this:pr($z)) div 2)
          )
      )
    '/>
  </xsl:function>

  <!--
    : this:cbrt()
    : cbrt()
    : Principal cubic root of a complex number. (Highest real part.)
    : Roots for n integer: for 0 ≤ k < n
    : z^(1/n) = r^(1/n)(cos((φ + 2kπ)/n) + i sin((φ + 2kπ)/n)) 
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:cbrt" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:pi($z) = 0) then (
        if (this:pr($z) >= 0)
        then this:complex(util:cbrt(this:pr($z)), 0)
        else this:complex(-util:cbrt(-this:pr($z)), 0)
      ) else (
        let $r := this:modulus($z) return
        let $φ := this:angle($z)=>util:radians() return
        let $rcbrt := util:cbrt($r) return
        let $k := 
          head(
            sort((0 to 2), (), function($k) {-math:cos(($φ + 2*$k*math:pi()) div 3)})
            (:
            for $k in 0 to 2
            order by math:cos(($φ + 2*$k*math:pi()) div 3) descending
            return $k
            :)
          )
        return (
          this:complex(
            $rcbrt * math:cos(($φ + 2*$k*math:pi()) div 3),
            $rcbrt * math:sin(($φ + 2*$k*math:pi()) div 3)
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:sin()
    : Sine of complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:sin" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: 
       : sin(x + iy) = sin(x)cosh(y) + i cos(x)sinh(y)
       :)
      this:complex(
        math:sin(this:pr($z))*util:cosh(this:pi($z)),
        math:cos(this:pr($z))*util:sinh(this:pi($z))
      )
    '/>
  </xsl:function>

  <!--
    : this:cos()
    : Cosine of complex number.
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:cos" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: 
       : cos(x + iy) = cos(x)cosh(y) - i sin(x)sinh(y)
       :)
      this:complex(
        math:cos(this:pr($z))*util:cosh(this:pi($z)),
        -math:sin(this:pr($z))*util:sinh(this:pi($z))
      )
    '/>
  </xsl:function>

  <!--
    : this:sinh()
    : Hyperbolic sine.
    :-->
  <xsl:function name="this:sinh" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: sinh(x + iy) = sinh(x)cos(y) + cosh(x)sin(y) :)
      this:complex(
        util:sinh(this:pr($z))*math:cos(this:pi($z)),
        util:cosh(this:pr($z))*math:sin(this:pi($z))
      )
    '/>
  </xsl:function>

  <!--
    : this:cosh()
    : Hyperbolic cosine.
    :-->
  <xsl:function name="this:cosh" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: cosh(x + iy) = cosh(x)cos(y) - sinh(x)sin(y) :)
      this:complex(
        util:cosh(this:pr($z))*math:cos(this:pi($z)),
        -util:sinh(this:pr($z))*math:sin(this:pi($z))
      )
    '/>
  </xsl:function>

  <!--
    : this:exp()
    : Exponential of z: e^z
    :
    : $z: the complex number
    :-->
  <xsl:function name="this:exp" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:complex(
        math:exp(this:pr($z)) * math:cos(this:pi($z)),
        math:exp(this:pr($z)) * math:sin(this:pi($z))
      )
    '/>
  </xsl:function>

  <!--
    : this:times()
    : Multiply a complex by a constant.
    :
    : $z: the complex number
    : $k: the constant to multiply by
    :-->
  <xsl:function name="this:times" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:param name="k" as="xs:double"/>
    <xsl:sequence select='
      $z=>point:times($k)
    '/>
  </xsl:function>

  <!--
    : this:plus()
    : Add a constant to a complex number. (i.e. add complex number ($k, 0))
    :
    : $z: the complex number
    : $k: the constant to add
    :-->
  <xsl:function name="this:plus" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:param name="k" as="xs:double"/>
    <xsl:sequence select='
      $z=>this:add(this:complex($k, 0))
    '/>
  </xsl:function>

  <!--
    : this:divide()
    : Divide two complex numbers.
    :
    : $z1, $z2: the complex numbers to divide
    :-->
  <xsl:function name="this:divide" as="map(xs:string, item()*)">
    <xsl:param name="z1" as="map(xs:string, item()*)"/>
    <xsl:param name="z2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: (a + bi)/(c + di) = ((ac + bd) + i(bc - ad))/(c² + d²) :)
      let $r2 := this:pr($z2)*this:pr($z2) + this:pi($z2)*this:pi($z2)
      return
        point:point(
          (this:pr($z1)*this:pr($z2) + this:pi($z1)*this:pi($z2)) div $r2,
          (this:pi($z1)*this:pr($z2) - this:pr($z1)*this:pi($z2)) div $r2
        )
    '/>
  </xsl:function>

  <!--
    : this:divide-by-i()
    : Divide complex numbers by i.
    :
    : $z
    :-->
  <xsl:function name="this:divide-by-i" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: (a + bi)/(c + di) = ((ac + bd) + i(bc - ad))/(c² + d²) :)
      (: (a + bi)/(0 + 1i) = ((0a + b1) + i(b0 - a1))/(0² + 1²) = b - ai :)
      point:point( this:pi($z), -this:pr($z) )
    '/>
  </xsl:function>

  <!--
    : this:modulus()
    : Modulus of a complex number. Given z = re^iφ, the r.
    :
    : $z: the complex numbers 
    :-->
  <xsl:function name="this:modulus" as="xs:double">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      math:sqrt(this:pr($z)*this:pr($z) + this:pi($z)*this:pi($z))
    '/>
  </xsl:function>

  <!--
    : this:angle()
    : The angle of the complex number. Given z = re^iφ, the φ in degrees.
    :
    : $z: the complex numbers 
    :-->
  <xsl:function name="this:angle" as="xs:double">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      math:atan2(this:pi($z), this:pr($z))=>util:degrees()=>util:remap-degrees()
    '/>
  </xsl:function>

  <!--
    : this:normalize()
    : Normalize the complex number so modulus is 1.
    :
    : $z: the complex number to normalize
    :-->
  <xsl:function name="this:normalize" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $l := this:modulus($z)
      return this:complex(this:pr($z) div $l, this:pi($z) div $l)
    '/>
  </xsl:function>

  <!--
    : this:is-gauss-prime()
    : Test whether complex number is gaussian prime; snaps to integral value.
    :
    : $z: complex number to test
    :-->
  <xsl:function name="this:is-gauss-prime" as="xs:boolean">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $a := this:r($z) return
      let $b := this:i($z)
      return (
        if ($a * $b != 0) then (
         util:is-prime($a*$a + $b*$b)
        ) else (
          let $c := abs($a + $b)
          return util:is-prime($c) and $c mod 4 = 3
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:conjugate()
    : Return the complex conjugate
    :
    : $z: complex number 
    :-->
  <xsl:function name="this:conjugate" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:complex(this:pr($z), -this:pi($z))
    '/>
  </xsl:function>

  <!--
    : this:reciprocal()
    : The reciprocal of z: 1 / z
    :
    : $z: complex number 
    :-->
  <xsl:function name="this:reciprocal" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $r := this:pr($z)*this:pr($z) + this:pi($z)*this:pi($z)
      return
       this:complex(this:pr($z) div $r, -this:pi($z) div $r)
    '/>
  </xsl:function>

  <!--
    : this:abs()
    : The absolute value of z
    :
    : $z: complex number 
    :-->
  <xsl:function name="this:abs" as="map(xs:string, item()*)">
    <xsl:param name="z" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:complex(abs(this:pr($z)), abs(this:pi($z)))
    '/>
  </xsl:function>

  <!--
    : this:quote()
    : Return a string value of the complex number
    :-->
  <xsl:function name="this:quote" as="xs:string">
    <xsl:param name="zs" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
      string-join(
        for $z in $zs return (
          let $r := this:pr($z) return
          let $i := this:pi($z)
          return
            if ($r = 0 and $i = 0) then "0"
            else if ($r = 0) then $i||"i"
            else if ($i = 0) then string($r)
            else if ($i &lt; 0) then $r||$i||"i"
            else $r||"+"||$i||"i"
        )
        ,
        " "
      )
    '/>
  </xsl:function>

  <!--
    : this:unquote()
    : Parse a string value of a complex number to produce that complex number.
    : Number is of the form 33+12i
    :
    : $zs: the string representation of the number
    :-->
  <xsl:function name="this:unquote" as="map(xs:string, item()*)">
    <xsl:param name="zs" as="xs:string"/>
    <xsl:sequence select='
      (: Cases:
       :  2
       : -2
       :     i
       :    3i
       :   -3i
       :  2+3i
       : -2+3i
       :  2-3i
       : -2-3i
       :  2+ i
       :  2- i
       : -2+ i
       : -2- i
       :)
      let $z := number($zs)
      return
        if ($z = $z) then this:complex($z, 0) (: 2 | -2 :)
        else if ($zs = "i") then this:complex(0, 1) (: i :)
        else (
          let $parts := analyze-string($zs, "[-+]?[0123456789.]+") return
          let $n-matches := count($parts/fn:match)
          return (
            if ($n-matches = 0) then
              if ($parts/fn:non-match=("i","+i"))
              then this:complex(0, 1)
              else if ($parts/fn:non-match="-i")
              then this:complex(0, -1)
              else errors:error("UTIL-BADCOMPLEX", $zs)
            else if ($n-matches = 1) then
              let $next := $parts/fn:match/following-sibling::fn:non-match[1]
              return
                if (empty($next))
                then this:complex(number($parts/fn:match), 0)
                else if ($next="+i")
                then this:complex(number($parts/fn:match), 1)
                else if ($next="-i")
                then this:complex(number($parts/fn:match), -1)
                else this:complex(0, number($parts/fn:match))
            else if ($n-matches = 2) then
              this:complex(number($parts/fn:match[1]), number($parts/fn:match[2]))
            else errors:error("UTIL-BADCOMPLEX", $zs)
          )
        )
    '/>
  </xsl:function>

  <!--
    : this:same()
    : Are the two complex numbers the same? (core properties only)
    :
    : $z1, $z2: the two numbers to compare
    :-->
  <xsl:function name="this:same" as="xs:boolean">
    <xsl:param name="z1" as="map(xs:string, item()*)"/>
    <xsl:param name="z2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:pr($z1)=this:pr($z2) and
      this:pi($z1)=this:pi($z2)
    '/>
  </xsl:function>

  <!--
    : this:decimal()
    :-->
  <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
    <xsl:param name="zs" as="map(xs:string, item()*)*"/>
    <xsl:param name="digits" as="xs:integer"/>
    <xsl:sequence select='
      point:decimal($zs, $digits)
    '/>
  </xsl:function>

</xsl:package>
<!--complex.xsl-->
