<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Module with functions producing random distributions of various kinds.
 :
 : This version is XQuery 3.1 compliant. It is not as performant as the
 : MarkLogic version because random-number-generator() just isn't as usable as
 : xdmp:random. This works through the ugly hack of relying on generate-id() 
 : on a constructed node to produce new seeds. That isn't entirely random
 : because whenever Saxon decides that one expression is the same as another
 : it will optimize it to produce the same result, but for the most part
 : this works ok. With Saxon-PE or Saxon-EE I use timestamp() which is more
 : reliable. Saxon-EE optimizer is often too clever regardless, and you get
 : repeated random numbers. Running with -opt:-l seems to fix this.
 : You may need more for some cases: -opt:-gl 
 :
 : Even generating a sequence of numbers off random-number-generator isn't
 : that great because you need to construct maps to hold both the next
 : generator function and the number, or use recursion in a way highly
 : likely to generate stack overflows. (The example in the spec suffers this
 : problem.)
 :
 : The problem is that for my purposes I need to be able to create multiple
 : random sequences in the same query that are different from the sequences
 : in subsequent episodes. Where the sequences are partitioned into separate
 : modules so even if I did the ugly thing of making every function everywhere
 : have to cart around and return both the values I cared about and the 
 : current randomizer (which XQuery is so very not good at), and used explicit
 : distinct seeds: whence those seeds? They need to be different based on 
 : global program context and different from each run of the program. So
 : current-dateTime() concatenated with a value from a value range that is
 : preassigned to each component module. Or: cart around yet another randomizer
 : to provide the seeds. It is ugly, ugly, ugly any way you slice it.
 :
 : Usage:
 : You can either use the explicit distribution functions here, e.g.
 : rand:normal(10.0, 2.0), or construct a distribution descriptor using
 : the distributions API and then pass it into a call to rand:randomize(), 
 : e.g. rand:randomize(dist:normal(10.0, 2.0))
 : The latter approach allows for more fine control over the behaviour of
 : the final result (e.g. you can set a minimum value, a post-multiplier, etc.)
 : and is useful when you have a randomization algorithm that is getting
 : reused a lot, or that you might want to change up without having to find
 : all the places where it is called in the code.
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:this="http://mathling.com/core/random"
             xmlns:art="http://mathling.com/art"
             xmlns:dist="http://mathling.com/type/distribution"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:config="http://mathling.com/core/config"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             exclude-result-prefixes="this art dist errors util config xs math map array"
             version="3.0"
             name="http://mathling.com/core/random">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/distribution" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>

   <!--
    : this:seed()
    : Function to hand out seeds: 
    : Ugly hack alert: browbeat r-n-g into using a new seed => new number
    : See top comment. If you start getting samey-samey sequences all I 
    : can suggest is to spring for Saxon-PE.
    :-->
   <xsl:function name="this:seed" as="xs:anyAtomicType" new-each-time="yes">
      <xsl:variable name="node">
        <xsl:element name="n"/>
      </xsl:variable>
      <xsl:sequence select='
      let $timestamp := function-lookup(QName("http://saxon.sf.net/", "timestamp"), 0)
      return (
        if (exists($timestamp)) then $timestamp()
        else (
          concat(current-dateTime(), generate-id($node))
        ) 
      )
      '/>
   </xsl:function>

   <!--
    : $this:SCALING: 
    : Determine the value of ε
    : ε used to handle the switch from exclusive to inclusive bounds.
    :-->
   <xsl:variable name="this:SCALING" as="xs:integer" select='100000'/>
   <xsl:variable name="this:ε" as="xs:double" visibility="private">
      <xsl:sequence select='
      (1 div $this:SCALING) cast as xs:double
      '/>
   </xsl:variable>

   <!--
    : $this:DECIMAL-DIGITS
    : Number of digits to keep in "decimal" casts
    :-->
   <xsl:variable name="this:DECIMAL-DIGITS" as="xs:integer" select="2"/>
   <xsl:variable name="this:DECIMAL-DIV" as="xs:integer">
      <xsl:sequence select='
      math:pow(10, $this:DECIMAL-DIGITS) cast as xs:integer
      '/>
   </xsl:variable>

   <!--
    : $this:π
    : I just like to read equations with π in them
    :-->
   <xsl:variable name="this:π" as="xs:double">
      <xsl:sequence select='math:pi()'/>
   </xsl:variable>

   <!--
    : $this:RESAMPLE-LIMIT
    : Number of resamplings will we try before we give up
    :-->
   <xsl:variable name="this:RESAMPLE-LIMIT" as="xs:integer" select='10'/>

   <!--
    : $this:STD-UNIFORM
    : Uniform distribution [0,1.0) 
    :-->
   <xsl:variable name="this:STD-UNIFORM" as="function(*)">
      <xsl:sequence select='
      dist:uniform(0.0, 1.0 - $this:ε)
      '/>
   </xsl:variable>

   <!--
    : $this:STD-NORMAL
    : Normal distribution mean=0 std=1
    :-->
   <xsl:variable name="this:STD-NORMAL" as="function(*)">
      <xsl:sequence select='
      dist:normal(0.0, 1.0)
     '/>
   </xsl:variable>

   <!--
    : this:flip()
    : Returns random Bernoulli distributed data as a boolean
    :
    : $percent: Percent that should come up true()
    :-->
   <xsl:function name="this:flip" as="xs:boolean" new-each-time="yes">
      <xsl:param name="percent" as="xs:double"/>
      <xsl:sequence select='
      100*random-number-generator(this:seed())("number") lt $percent
      '/>
   </xsl:function>

   <!--
    : this:constant()
    : Returns a constant value, either the average of $min and $max (if they
    : exist) or whichever one does exist. Failing everything else, 1.
    :
    : This is the degenerate case of uniform() and exists as a convenience 
    : for data-driven randomization in randomize().
    : 
    : $min: minimum value
    : $max: maximum value
    :-->
   <xsl:function name="this:constant" as="xs:double">
      <xsl:param name="min" as="xs:double?"/>
      <xsl:param name="max" as="xs:double?"/>
      <xsl:sequence select='
      if (exists($min) and exists($max)) then (
        if ($max lt $min) then errors:error("RAND-BADMINMAX",($min,$max))
        else if ($max = $min) then $min
        else ($max + $min) div 2
      ) else (
        ($min,$max,1)[1]
      )
      '/>
   </xsl:function>

   <!--
    : this:uniform()
    : Return random uniformly distributed data in the range.
    : Because the base randomizer is exclusive on the upper bound, add ε
    : and then cap. This creates a slight over-weighting of the max value,
    : depending on fineness of ε.
    :
    : When we are using integers, we need to avoid chopping off the top of
    : the range, or we lose. We can get it, but it is unlikely and for small
    : integer ranges spectacularly unlikely. So we add 1-ε instead of ε and
    : quantize at this level. You want to avoid this and get, say, floating
    : point values between 5 and 10 (inc), pass floating point arguments.
    : 
    : Base randomizer gives [0,1) we want [x,y]
    :   [0,1) * (ε+y-x)  => [0,ε+y-x) 
    :   [0,ε+y-x) + x    => [x, y+ε)
    :   min(([x, y+ε),y) => [x,y]
    : 
    : For integers:
    :   [0,1) * (1-ε+y-x)  => [0,1-ε+y-x) 
    :   [0,1-ε+y-x) + x    => [x, y+1-ε)
    :   [x,y+1-ε) idiv 1   => [x, y]
    :
    : $min: Lower bound (inclusive) of range
    : $max: Upper bound (inclusive) of range
    :-->
   <xsl:function name="this:uniform" as="xs:double" new-each-time="yes">
      <xsl:param name="min" as="xs:numeric"/>
      <xsl:param name="max" as="xs:numeric"/>
      <xsl:sequence select='
      if ($max lt $min) then errors:error("RAND-BADMINMAX",($min,$max))
      else if ($max = $min) then $min 
      else if (
        ($max instance of xs:integer) and ($min instance of xs:integer)
      ) then (
        let $random_0_1 := random-number-generator(this:seed())("number")
        return ((1 - $this:ε + $max - $min)*$random_0_1 + $min) idiv 1
      ) else (
        let $random_0_1 := random-number-generator(this:seed())("number")
        return min(( ($this:ε + $max - $min)*$random_0_1 + $min, $max ))
       )
      '/>
   </xsl:function>

   <!--
    : this:normal()
    : Return random normally distributed data.
    : 
    : $mean: Mean of range of values
    : $std: Standard deviation of range of values
    :-->
   <xsl:function name="this:normal" as="xs:double" new-each-time="yes">
      <xsl:param name="mean" as="xs:double"/>
      <xsl:param name="std" as="xs:double"/>
      <xsl:sequence select='
      $mean + this:gauss(random-number-generator(this:seed())) * $std
      '/>
   </xsl:function>

   <xsl:function name="this:gauss" as="xs:double" new-each-time="yes" visibility="private">
      <xsl:param name="randomizer" as="map(xs:string,item())"/>
      <xsl:variable name="u" as="xs:double" select='2 * $randomizer("number") - 1'/>
      <xsl:variable name="next" as="map(xs:string,item())" select='($randomizer("next"))()'/>
      <xsl:variable name="v" as="xs:double" select='2 * $next("number") - 1'/>
      <xsl:variable name="r" as="xs:double" select='$u * $u + $v * $v'/>
      <xsl:sequence select='
      if ($r = 0 or $r >= 1) then this:gauss(($next("next"))())
      else $u * math:sqrt(-2 * math:log($r) div $r)
      '/>
   </xsl:function>
   
   <!--
    : this:skewed()
    : Return random data in skewed normal distribution.
    :
    : $mean: Mean of distribution
    : $std: Standard deviation of distribution
    : $skew: Skew of distribution
    : 
    : Skewed defined by parameters α, ξ, and ω
    : α is skew 
    :   α = 0 => normal, α > 0 right skewed, α < 0 left skewed
    :      right skewed => long tail on right
    : ξ = location = shift along x
    : ω = scaling along y
    :   ω is positive 
    : 
    : mean = ξ + ω*δ*√(2/π) δ=α/√(1+α²) 
    : std = √(ω²*(1 - 2*δ²/π))
    :
    : So:
    : ω = std / √(1 - 2*δ²/π))
    : ξ = mean - ω*δ*√(2/π)
    :-->
   <xsl:function name="this:skewed" as="xs:double" new-each-time="yes">
      <xsl:param name="mean" as="xs:double"/>
      <xsl:param name="std" as="xs:double"/>
      <xsl:param name="skew" as="xs:double"/>
      <xsl:variable name="α" as="xs:double" select='$skew'/>
      <xsl:variable name="δ" as="xs:double" select='$α div math:sqrt(1 + $α*$α)'/>
      <xsl:variable name="ω" as="xs:double" select='$std div math:sqrt(1 - (2*$δ*$δ div $this:π))'/>
      <xsl:variable name="ξ" as="xs:double" select='$mean - $ω*$δ*math:sqrt(2 div $this:π)'/>
      <xsl:variable name="σ" as="xs:double" select='1 div math:sqrt(1 + $α*$α)'/>
      <xsl:variable name="u" as="xs:double" select='this:normal(0, 1)'/>
      <xsl:variable name="v" as="xs:double" select='this:normal(0, 1)'/>
      <xsl:variable name="u1" as="xs:double" select='$σ*($α*abs($u) + $v)'/>
      <xsl:sequence select='$ω * $u1 + $ξ'/>
   </xsl:function>

   <!--
    : this:binomial-poisson()
    : Return random data in a Poisson binomial distribution.
    : 
    : $probabilities: probabilities of the base Bernoulli variables
    :
    : μ = Σp[i]
    : σ = √(Σ(1 - p[i])p[i])
    :-->
   <xsl:function name="this:binomial-poisson" as="xs:double" new-each-time="yes">
      <xsl:param name="probabilities" as="xs:double*"/>
      <xsl:sequence select='
      sum(for $p in $probabilities return if (this:flip($p)) then 1 else 0)
      '/>
   </xsl:function>

   <!--
    : this:poisson()
    : Return random data in a Poisson distribution.
    : Using inverse transform sampling method. If λ is large and e^-λ underflows; 
    : this will not work well.
    : 
    : $mean: the lambda parameter
    :
    : μ = λ
    : σ = λ
    :-->
   <xsl:function name="this:poisson" as="xs:double" new-each-time="yes">
      <xsl:param name="mean" as="xs:double"/>
      <xsl:variable name="u" as="xs:double" select='this:uniform(0E0, 1E0)'/>
      <!--
          let $x := $x + 1
          let $p := $p * $mean div $x
          let $s := $s + $p
       -->
      <xsl:sequence select='
      util:while(
        function($x as xs:double, $p as xs:double, $s as xs:double) as xs:boolean {
          $u > $s
        },
        function($x as xs:double, $p as xs:double, $s as xs:double) as xs:double* {
          $x + 1, $p * $mean div ($x + 1), $s + ($p * $mean div ($x + 1))
        },
        0E0, math:exp(-$mean), math:exp(-$mean)
      )=>head()
      '/>
   </xsl:function>

   <!--
    : this:exponential()
    : Return random data in an exponential distribution.
    : 
    : $λ: the distribution parameter
    :
    : μ = 1/λ
    : 
    : Using inverse transform sampling -ln(U)/λ
    :-->
   <xsl:function name="this:exponential" as="xs:double" new-each-time="yes">
      <xsl:param name="λ" as="xs:double"/>
      <xsl:variable name="u" as="xs:double" select='this:uniform(0E0, 1E0)'/>
      <xsl:sequence select='
      -math:log($u) div $λ
      '/>
   </xsl:function>

   <!--
    : this:zipf()
    : Return random data in a Zipf distribution as an integer index from 1 to N
    : Note that this constructs the sum table anew each time: not very efficient.
    : You'd be better off constructing the distribution or just calling 
    : select-index() directly
    :
    : p(n)=(1/k^α)/sum(i=1 to n)(1/i^α)
    :    n = number of elements
    :    k = rank
    :    α = exponent
    :
    : $alpha: the α parameter, should be >=1
    :   For English words this by some estimates is around 1.6
    :   For city sizes 1.07
    : $n: number of Zipf values in range
    :-->
   <xsl:function name="this:zipf" as="xs:integer" new-each-time="yes">
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:variable name="zipf-weights" as="xs:double*" select='this:zipf-sums($alpha,$n+1)'/>
      <xsl:sequence select='
      this:select-index($zipf-weights, $n)
      '/>
   </xsl:function>

   <!--
    : this:shuffle()
    : Return the data items in a (uniformly) random order.
    : 
    : $data: Data items to shuffle
    :-->
   <xsl:function name="this:shuffle" as="item()*" new-each-time="yes">
      <xsl:param name="data" as="item()*"/>
      <xsl:sequence select='
      random-number-generator(this:seed())?permute($data)
      '/>
   </xsl:function>

   <!--
    : this:percent-sums()
    : Returns a sequence of cumulative probabilities, to be used
    : by select-index() to perform selections. This is to ensure we don't have to
    : keep recomputing the cumulative sums for each selection.
    :
    : To account for rounding, the last item is always forced to 1 and any sums
    : greater than 1 are also rounded down to 1. To get expected distributions, 
    : ensure that the input percentages sum to 100.
    :
    : Sums are returned in map:keys() order, so this assumes stability of that
    : function between calls to percent-sums() and calls to selection(). 
    : 
    : $weights a map from key to percent (expressed as integer 0 to 100)
    :-->
   <xsl:function name="this:percent-sums" as="xs:double*">
      <xsl:param name="weights" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      dist:percent-sums($weights)
      '/>
   </xsl:function>

   <!--
    : this:zipf-sums()
    : Returns a sequence of cumulative Zipf probabilities, to be used
    : by select-index() to perform selections. This is to ensure we don't have to
    : keep recomputing the cumulative sums for each selection.
    : 
    : p(k)=(1/k^α)/sum(i=1 to n)(1/i^α)
    :    n = number of elements
    :    k = rank
    :    α = exponent
    :
    : $alpha: the alpha parameter, should be >=1
    :   For English words this by some estimates is around 1.6
    :   For city sizes 1.07
    : $n: number to generate
    :-->
   <xsl:function name="this:zipf-sums" as="xs:double*">
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:sequence select='
      dist:zipf-sums($alpha, $n)
      '/>
   </xsl:function>

   <!--
    : this:select-index()
    : Return a rank from 1 to N of values distributed according to cumulative
    : probability sums. Use zipf-sums(), percent-sums(), etc. to compute those.
    :
    : $sums: cumulative probability sums, computed via zipf-sums() or percent-sums()
    : $n: maximum rank to return ($n should be <= count($sums))
    :-->
   <xsl:function name="this:select-index" as="xs:integer" new-each-time="yes">
      <xsl:param name="sums" as="xs:double*"/>
      <xsl:param name="n" as="xs:integer"/>
      <!--
       : Get value in (0,1) exclusive; underlying randomizer returns inclusive
       : values, so generate value within epsilon of bounds instead
       :-->
      <xsl:variable name="p" as="xs:double" select='this:uniform($this:ε, 1 - $this:ε)'/>
      <xsl:sequence select='
      util:rangeindex($sums[position()=(1 to $n)], $p, $n)
      '/>
   </xsl:function>

   <!--
    : this:selection()
    : Return a selection using data distributed per the cumulative probability sums.
    : 
    : $sums: cumulative probability sums, computed via zipf-sums(), percent-sums(), etc.
    : $values: values to select, in rank order
    :-->
   <xsl:function name="this:selection" as="item()" new-each-time="yes">
      <xsl:param name="sums" as="xs:double*"/>
      <xsl:param name="values" as="item()*"/>
      <xsl:variable name="n" as="xs:integer">
        <xsl:sequence select='min((count($sums),count($values))) cast as xs:integer'/>
      </xsl:variable>
      <xsl:variable name="k" as="xs:integer" select='this:select-index($sums, $n)'/>
      <xsl:sequence select='
      $values[$k]
      '/>
   </xsl:function>

   <!--
    : this:check-markov-matrix()
    : Check whether the matrix is a valid Markov matrix, raise error if it is not.
    :-->
   <xsl:function name="this:check-markov-matrix" as="empty-sequence()">
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:param name="matrix" as="xs:double*"/>
      <xsl:param name="is-cumulative" as="xs:boolean"/>
      <xsl:variable name="size" as="xs:integer" select='count($matrix)'/>
      <xsl:variable name="sqrt" as="xs:double" select='math:sqrt($size)'/>
      <!-- Matrix must be square -->
      <xsl:sequence select='
      if ($sqrt * $sqrt ne $size)
      then errors:error("RAND-NOTSQUARE", $size)
      else if ($dim ne ($sqrt cast as xs:integer))
      then errors:error("RAND-BADDIM", ($dim, $sqrt))
      else ()
      '/>
      <!-- Values must be probabilities -->
      <xsl:sequence select='
      for $i in 1 to $size return (
        if ($matrix[$i] lt 0 or $matrix[$i] gt 1)
        then errors:error("RAND-BADPROBABILITY", ($i, $matrix[$i]))
        else ()
      )
      '/>
     <!-- 
      : Each row should have a cumulative probability of 1 
      : If $is-cumulative, each row has been expressed as cumulative 
      : probabilities already (for efficiency) so we check p<1 vs sum(p)<1
      : but we did that check above already.
      :-->
      <xsl:variable name="sprobs" as="xs:double*">
        <xsl:sequence select='
        for $i in 1 to $dim return (
          sum($matrix[position()=(($i - 1)*$dim + 1 to ($i - 1) * $dim)])
        )
        '/>
      </xsl:variable>
      <xsl:sequence select='
      if ($is-cumulative) then () else (
        for $i in 1 to $dim return (
          if (util:twixt($sprobs[$i], 0, 1 + $this:ε)) then ()
          else errors:error("RAND-BADSUMPROB", ($i, $sprobs[$i]))
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:markov-sums()
    : Return a Markov probability matrix as a matrix where each row contains
    : the cumulative probability sums for that row. This allows the Markov
    : distribution selection to run more efficiently.
    :
    : Example:
    :   markov-sums(3, (
    :     0.2, 0.4, 0.4,
    :     0.5, 0.5, 0,
    :     0.4, 0.2, 0.4
    :   )) => (
    :     0.2, 0.6, 1,
    :     0.5, 1, 1,
    :     0.4, 0.6, 1
    :   )
    :
    :  To account for rounding, the last item in the row is always forced to 1
    :  and any sums greater than 1 are also rounded down to 1.
    :
    : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
    : $matrix: the input non-cumulative Markov matrix
    :-->
   <xsl:function name="this:markov-sums" as="xs:double*">
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:param name="matrix" as="xs:double*"/>
      <xsl:sequence select='
      dist:markov-sums($dim, $matrix)
      '/>
   </xsl:function>

   <!--
    : this:markov-percent-sums()
    : Return a Markov probability matrix as a matrix where each row contains
    : the cumulative probability sums for that row. This allows the Markov
    : distribution selection to run more efficiently.
    :
    : Example:
    :   markov-percent-sums(3, (
    :     20, 40, 40,
    :     50, 50, 0,
    :     40, 20, 40
    :   )) => (
    :     0.2, 0.6, 1,
    :     0.5, 1, 1,
    :     0.4, 0.6, 1
    :   )
    :
    :  To account for rounding, the last item in the row is always forced to 1
    :  and any sums greater than 1 are also rounded down to 1.
    :
    : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
    : $matrix: the input non-cumulative Markov matrix
    :-->
   <xsl:function name="this:markov-percent-sums" as="xs:double*">
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:param name="matrix" as="xs:integer*"/>
      <xsl:sequence select='
      dist:markov-percent-sums($dim, $matrix)
      '/>
   </xsl:function>

   <!--
    : this:markov()
    : Given a square matrix of Markov probabilities and a current state,
    : return a successor state with the probabilities as given. For efficiency,
    : the matrix should be given in the cumulative probability form, as created
    : by markov-sums() or markov-percent-sums()
    :
    : $current: Index of current state in the matrix
    : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
    : $matrix: Square matrix of cumulative probabilities
    :-->
   <xsl:function name="this:markov" as="xs:integer" new-each-time="yes">
      <xsl:param name="current" as="xs:integer"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:param name="matrix" as="xs:double*"/>
      <xsl:if test="not(util:twixt($current, 1, $dim))">
        <xsl:sequence select='errors:error("ML-BADARGS", ("current", $current))'/>
      </xsl:if>
      <xsl:variable name="row" as="xs:double*">
        <xsl:sequence select='
        $matrix[position()=(($current - 1) * $dim + 1 to $current * $dim)]
        '/>
      </xsl:variable>
      <xsl:sequence select='
      this:select-index($row, $dim)
      '/>
   </xsl:function>

   <!--
    : this:markov-selection()
    : Return a selection using Markov matrix distribution.
    : 
    : $current: index of current selection
    : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
    : $matrix: cumulative probability matrix, computed via markov-sums()
    : $values: values to select, in index order
    :-->
   <xsl:function name="this:markov-selection" as="item()" new-each-time="yes">
      <xsl:param name="current" as="xs:integer"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:param name="matrix" as="xs:double*"/>
      <xsl:param name="values" as="item()*"/>
      <xsl:variable name="k" as="xs:integer">
        <xsl:sequence select='
        this:markov($current, $dim, $matrix)
        '/>
      </xsl:variable>
      <!-- Make sure matrix and values are compatible -->
      <xsl:if test='not(util:twixt($k, 1, count($values)))'>
        <xsl:sequence select='errors:error("ML-BADARGS", ("values", $values))'/>
      </xsl:if>
      <xsl:sequence select='$values[$k]'/>
   </xsl:function>

   <!--
    : this:random-selection()
    : Return a random selection from a set of keys, where the key is selected
    : per a set of probabilities. A convenience wrapper (although less efficient,
    : because it recomputes the sums) for selection()
    : 
    : $weights: Map of probabilities (as integer percents) from key to weight
    :-->
   <xsl:function name="this:random-selection" as="xs:string" new-each-time="yes">
      <xsl:param name="weights" as="map(xs:string, item()*)"/>
      <xsl:variable name="sums" as="xs:double*" select='this:percent-sums($weights)'/>
      <xsl:variable name="values" as="item()*" select='$weights=>map:keys()'/>
      <xsl:sequence select='this:selection($sums, $values)'/>
   </xsl:function>

   <!--
    : this:random-assortment()
    : Select a subset of a set of keys, in (uniform) random order, where each key
    : is selected at most once, and is selected with probability given by set
    : of weights.
    :
    : $keys: Set of keys to choose from
    : $weights: Map of probabilities (as integer percents) from key to weight
    :-->
   <xsl:function name="this:random-assortment" as="xs:string*" new-each-time="yes">
      <xsl:param name="keys" as="xs:string*"/>
      <xsl:param name="weights" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:shuffle(
        for $key in $keys return (
          if (this:flip($weights($key))) then $key else ()
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:cast()
    : Cast/format a value to a particular type
    : Service function for randomize()
    : 
    : Casting to decimal is actually just rounding to $DECIMAL-DIGITS fractional
    : digits.
    : 
    : $value: Value to cast
    : $cast-as: One of "integer", "decimal", "boolean", "string", or "double"
    :-->
   <xsl:function name="this:cast" as="item()*">
      <xsl:param name="value" as="item()"/>
      <xsl:param name="cast-as" as="xs:string?"/>
      <xsl:sequence select='
      if ($cast-as = "integer") then $value cast as xs:integer
      else if ($cast-as = "decimal") then round($value * $this:DECIMAL-DIV) div $this:DECIMAL-DIV
      else if ($cast-as = "boolean") then boolean($value)
      else if ($cast-as = "string") then string($value)
      else $value
      '/>
   </xsl:function>

   <!--
    : this:range()
    : Ensure the value is in the range by changing value to minimum or maximum as
    : necessary. 
    : Service function for randomize()
    : Assumes min < max
    : 
    : $value: The value
    : $min: The lowest value (optional)
    : $max: The highest value (optional)
    :-->
   <xsl:function name="this:range" as="xs:numeric">
      <xsl:param name="value" as="xs:numeric"/>
      <xsl:param name="min" as="xs:numeric?"/>
      <xsl:param name="max" as="xs:numeric?"/>
      <xsl:variable name="low" as="xs:numeric">
        <xsl:sequence select='
        if (exists($min)) then max(($min,$value)) else $value
        '/>
      </xsl:variable>
      <xsl:sequence select='
      if (exists($max)) then min(($max,$low)) else $low
      '/>
   </xsl:function> 

   <!--
    : this:get-last()
    : Function available for callback use in fetch()
    : 
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   (see randomize())
    : $last: Previous value
    :-->
   <xsl:function name="this:get-last" as="item()*">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:sequence select='$last'/>
   </xsl:function>

   <!--
    : this:last-fraction()
    : Return a callback function for use in fetch() that returns the a
    : given fraction of previous value
    :
    : $centrality: the fraction to apply
    :-->
   <xsl:function name="this:last-fraction"
                 as="(function(map(xs:string, item()*), item()?) as item())">
     <xsl:param name="centrality" as="xs:double"/>
      <xsl:sequence select='
      (: %art:name("rand:last-fraction") :) function($algorithm as map(xs:string,item()*), $last as item()?) as item() {
        $centrality * $last
      }
      '/>
   </xsl:function>

   <!-- 
    : this:fetch()
    : Lookup the key in the map. If it is a two-arg function or a QName of
    : a two-arg function, call that function, passing the algorithm map and
    : the last value to it. Otherwise return the value for the key.
    :
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   (see randomize())
    : $key: Key to lookup in map
    : $last: Previous value
    :-->
   <xsl:function name="this:fetch" as="item()*" visibility="private"> 
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="key" as="xs:string"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:variable name="v" as="item()?" select='$algorithm($key)'/>
      <xsl:sequence select='
      if ($v instance of function(map(xs:string,item()*), item()?) as item())
      then (
        $v($algorithm, $last)
      ) else if ($v instance of xs:QName) then (
        (function-lookup($v, 2))($algorithm, $last)
      )
      else (
        $v
      )
      '/>
   </xsl:function>

   <!--
    : this:is-complex()
    : Helper function for randomize() to allow for optimization of case where
    : we're using a static algorithm with no use of a function
    :-->
   <xsl:function name="this:is-complex" as="xs:boolean" visibility="private">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($algorithm=>map:contains("is_complex")) then (
        $algorithm("is_complex")
      ) else (
        some $k in map:keys($algorithm)
        satisfies (
          ($algorithm($k) instance of xs:QName) or
          ($algorithm($k) instance of function(map(*), item()?) as item())
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:is-complex()
    : Helper function for randomize() to allow for optimization of case where
    : we're using a static algorithm with no use of a function
    :-->
   <xsl:function name="this:is-complex" as="xs:boolean" visibility="private">
      <xsl:param name="key" as="xs:string"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($algorithm($key) instance of xs:QName) or
      ($algorithm($key) instance of function(map(*), item()?) as item())
      '/>
   </xsl:function>

   <!--
    : modify-all()
    : Apply modifiers to all values. Assumes all key values are static.
    : Service routine for randomize()
    :-->
   <xsl:function name="this:modify-all" as="item()*" visibility="private">
      <xsl:param name="values" as="item()*"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $distribution := $algorithm("distribution") return
      let $min := $algorithm("min") return
      let $max := $algorithm("max") return
      let $pre-multiplier := $algorithm("pre-multiplier") return
      let $values :=
        if (exists($pre-multiplier))
        then $values!(. * $pre-multiplier)
        else $values
      return
      let $values :=
        for $value in $values return (
          if ($distribution = ("uniform","flip","constant") or
            util:twixt($value, $min, $max)
          ) then (
            $value
          ) else (
            let $truncation := $algorithm("truncation") return
            if ($truncation="ceiling") then this:range($value, $min, $max)
            else if ($truncation="drop") then ()
            else (: "resample" :) (
              let $new :=
                fold-left(
                  1 to $this:RESAMPLE-LIMIT, (),
                  function($v as item()?, $i as xs:integer) as item()? {
                    if (empty($v)) then (
                      let $v := this:raw-distribution(1, $last, $algorithm) return
                      let $v2 :=
                        if (exists($pre-multiplier))
                        then $v * $pre-multiplier
                        else $v
                      return
                        if (util:twixt($v2, $min, $max)) then $v2
                        else ()
                     ) else (
                      $v
                     )
                  }
                )
              return (
                if (empty($new)) then (
                  util:log("Gave up resampling "||util:quote($algorithm)),
                  this:range($value, $min, $max)
                ) else (
                  $new
                )
              )
            )
          )
        )
      return
      let $values :=
        let $post-multiplier := $algorithm("post-multiplier")
        return
          if (exists($post-multiplier))
          then $values!(. * $post-multiplier)
          else $values
      return
      let $values := 
        let $post-shift := $algorithm("post-shift")
        return
          if (exists($post-shift))
          then $values!(. + $post-shift)
          else $values
      return
      let $values :=
        let $cast-as := $algorithm("cast")
        return
          if (exists($cast-as))
          then $values!this:cast(., $cast-as)
          else $values
      return $values
      '/>
   </xsl:function>

   <!--
    : modify()
    : Apply modifiers to values. Assumes all key values may be dynamic.
    : Service routine for randomize()
    :-->
   <xsl:function name="this:modify" as="item()*" visibility="private">
      <xsl:param name="value" as="item()"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $distribution := $algorithm=>this:fetch("distribution",$last) return
      let $min := $algorithm=>this:fetch("min",$last) return
      let $max := $algorithm=>this:fetch("max",$last) return
      let $pre-multiplier := $algorithm=>this:fetch("pre-multiplier", $last) return
      let $value :=
        if (exists($pre-multiplier))
        then $value * $pre-multiplier
        else $value
      return
      let $value :=
        if ($distribution = ("uniform","flip","constant") or
          util:twixt($value, $min, $max)
        ) then (
          $value
        ) else (
          let $truncation := $algorithm=>this:fetch("truncation",$last)
          return
          if ($truncation="drop") then ()
          else if ($truncation="ceiling") then this:range($value, $min, $max)
          else (: "resample" :) (
            let $new :=
              fold-left(
                1 to $this:RESAMPLE-LIMIT, (),
                function($v as item()?, $i as xs:integer) as item()? {
                  if (empty($v)) then (
                    let $v := this:raw-distribution(1, $last, $algorithm) return
                    let $v2 :=
                      if (exists($pre-multiplier))
                      then $v * $pre-multiplier
                      else $v
                    return
                      if (util:twixt($v2, $min, $max)) then $v2
                      else ()
                   ) else (
                    $v
                   )
                }
              )
            return (
              if (empty($new)) then (
                trace((),"gave up resampling"),
                this:range($value, $min, $max)
              ) else (
                $new
              )
            )
          )
        )
      return
      let $value :=
        let $post-multiplier := $algorithm=>this:fetch("post-multiplier",$last)
        return
          if (exists($post-multiplier))
          then $value * $post-multiplier
          else $value
      return
      let $value :=
        let $post-shift := $algorithm=>this:fetch("post-shift",$last)
        return
          if (exists($post-shift))
          then $value + $post-shift
          else $value
      return
      let $value :=
        let $cast-as := $algorithm=>this:fetch("cast",$last)
        return
          if (exists($cast-as))
          then this:cast($value, $cast-as)
          else $value
      return $value
      '/>
   </xsl:function>

   <!--
    : raw-distribution()
    : Service function to randomize(). Creates the raw values from the 
    : distributions, without applying the modifications (min/max, multipliers,
    : casts)
    :
    : Assumes distribution is fixed (or distributions for multi-modal)
    : $N: How many values to return
    : $last: Previous value (needed for markov, available to key callback)
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   distribution: Distribution to use, one of "constant", "uniform", "normal",
    :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal",
    :     "binomial-poisson", "poisson", "exponential"
    :   min: mimumum value (uniform, constant)
    :   max: maximum value (uniform, constant)
    :   mean: mean of distribution (normal, skewed, poisson, exponential) (default=0)
    :   std: standard deviation of distribution (normal, skewed) (default=1)
    :   skew: skew of distribution (skewed) (default=0)
    :   p: probability (bernoulli, flip) (default=50)
    :   probabilities: probabilities (binomial-poisson) 
    :   sums: cumulative probability sums (zipf, markov, sums)
    :   alpha: alpha parameter (zipf) (needed if no sums) (default=1)
    :   limit: number of sums (zipf) (needed if no sums) (default=1000)
    :   start: index of starting symbol (markov) (default=uniform[1,dim])
    :   dim: size of each dimension of Markov matrix (markov)
    :   matrix: raw Markov matrix (used if sums not provided, not recommended)
    :   distributions: sequence of distributions to combine (multimodal)
    :-->
   <xsl:function name="this:raw-distribution" as="item()*" visibility="private" new-each-time="yes">
     <xsl:param name="N" as="xs:integer"/>
     <xsl:param name="last" as="item()?"/>
     <xsl:param name="algorithm" as="map(xs:string,item()*)"/>
     <xsl:choose>
       <xsl:when test="this:is-complex($algorithm)">
         <!--
          : Distribution is fixed, but we may alter any other parameter 
          : so need to feed $last as appropriate to the sequence
          : For multimodal distributions is also fixed
          :-->
         <xsl:variable name="distribution" as="xs:string">
           <xsl:sequence select='($algorithm("distribution"),"uniform")=>head()'/>
         </xsl:variable>
         <xsl:sequence select='
         if ($distribution="constant") then (
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $min := $algorithm=>this:fetch("min",$values[last()]) return
                 let $max := $algorithm=>this:fetch("max",$values[last()]) return
                 let $val := this:constant($min, $max)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         ) else if ($distribution="uniform") then
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $min := $algorithm=>this:fetch("min",$values[last()]) return
                 let $max := $algorithm=>this:fetch("max",$values[last()]) return
                 let $val := this:uniform($min, $max)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="normal") then
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $μ := ($algorithm=>this:fetch("mean",$last),0)[1] return
                 let $σ := ($algorithm=>this:fetch("std",$last),1)[1] return
                 let $val := this:normal($μ, $σ)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="skewed") then
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $μ := ($algorithm=>this:fetch("mean",$last),0)[1] return
                 let $σ := ($algorithm=>this:fetch("std",$last),1)[1] return
                 let $α := ($algorithm=>this:fetch("skew",$last),0)[1] return
                 let $val := this:skewed($μ, $σ, $α)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="bernoulli") then
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $p := ($algorithm=>this:fetch("p",$last),50)[1] return
                 let $val := if (this:flip($p)) then 1 else 0  
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="flip") then
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $p := ($algorithm=>this:fetch("p",$last),50)[1] return
                 let $val := this:flip($p)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="multimodal") then (
           (: multimodal has sequence of sub randomizers in distributions :)
           (: $last applies to each subrandomizer independently generally
            : except post-sum modifications apply to combined last
            :)
           let $distributions := $algorithm("distributions") return
           let $selector := $algorithm("selector") return
           (:
           let $numd := count($distributions) return
           let $flips := for $i in 1 to $N return this:cast(this:uniform(1, $numd),"integer") return
           :)
           let $flips := for $i in 1 to $N return this:randomize($selector) return
           (: RAW: v11, v12, ..., v1N, v21, ..., v2N, ...., vd1, ... vdN :)
           (: Run this way so $last applies :)
           let $raw :=
             for $subdistribution in $distributions
             return this:randomize($N, $last, $subdistribution)
           return
           let $result :=           
             for $i in 1 to $N
             return $raw[($flips[$i] - 1)*$N + $i] 
           return
             $result
         )
         else if ($distribution="binomial-poisson") then (
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $ps := $algorithm=>this:fetch("probabilities",$last) return
                 let $val := sum(for $p in $ps return if (this:flip($p)) then 1 else 0)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         )
         else if ($distribution="poisson") then (
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $λ := $algorithm=>this:fetch("mean",$last) return
                 let $val := this:poisson($λ)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         )
         else if ($distribution="exponential") then (
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $last := $values[last()] return
                 let $λ := 1 div $algorithm=>this:fetch("mean",$last) return
                 let $val := this:exponential($λ)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         )
         else if ($distribution="sums") then
           (: It is probably a bad idea to have non-static probability sums :)
           (: Precompute these and include them in the algorithm map :)
           let $fixed-sums :=
             if (this:is-complex("sums", $algorithm)) then ()
             else $algorithm("sums")
           return
           let $fixed-max-rank :=
             if (exists($fixed-sums)) then count($fixed-sums) else ()
           return
           let $result := 
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $sums :=
                   if (exists($fixed-sums)) then $fixed-sums
                   else $algorithm=>this:fetch("sums",$values[last()])
                 return
                 let $max-rank :=
                   if (exists($fixed-max-rank))
                   then $fixed-max-rank
                   else count($sums)
                 return
                 let $val := this:select-index($sums, $max-rank)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="zipf") then
           (: It is probably a bad idea to have non-static Zipf parameters
            : An even worse idea to have to recompute probability sums 
            : for each time through.
            : Precompute these and put them in the map or you will be sorry
            :)
           let $fixed-sums :=
             if (this:is-complex("sums", $algorithm)) then ()
             else (
               if (exists($algorithm("sums")))
               then $algorithm("sums")
               else (
                 if (this:is-complex("alpha", $algorithm) or
                     this:is-complex("limit", $algorithm)
                 ) then ()
                 else (
                   this:zipf-sums(
                     ($algorithm("alpha"),1)[1],
                     ($algorithm("limit"),1000)[1]
                   )
                 )
               )
             )
           return
           let $fixed-max-rank :=
             if (exists($fixed-sums)) then count($fixed-sums) else ()
           return
           let $result := 
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $sums :=
                   if (exists($fixed-sums)) then $fixed-sums
                   else if (exists($algorithm=>this:fetch("sums",$values[last()])))
                   then $algorithm=>this:fetch("sums",$values[last()])
                   else
                     this:zipf-sums(
                       ($algorithm=>this:fetch("alpha",$values[last()]),1)[1],
                       ($algorithm=>this:fetch("limit",$values[last()]),1000)[1]
                     )
                 return
                 let $max-rank :=
                   if (exists($fixed-max-rank))
                   then $fixed-max-rank
                   else count($sums)
                 return
                 let $val := this:select-index($sums, $max-rank)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         else if ($distribution="markov") then
           (: It is probably a bad idea to have non-static Markov parameters :)
           (: Precompute them and include them in the algorithm map :)
           let $fixed-sums :=
             if (this:is-complex("sums", $algorithm)) then ()
             else (
               if (exists($algorithm("sums")))
               then $algorithm("sums")
               else (
                 if (this:is-complex("matrix", $algorithm) or
                     this:is-complex("dim", $algorithm)
                 ) then ()
                 else (
                   let $matrix := $algorithm("matrix") return
                   let $dim := (
                     $algorithm("dim"),
                     math:sqrt(count($matrix)) cast as xs:integer
                   )[1]
                   return this:markov-sums($dim, $matrix)
                 )
               )
             )
           return
           let $fixed-dim :=
             if (this:is-complex("dim", $algorithm)) then ()
             else if ($algorithm=>map:contains("dim"))
             then $algorithm("dim")
             else if (exists($fixed-sums))
             then math:sqrt(count($fixed-sums)) cast as xs:integer
             else ()
           return
           let $start := ($last,$algorithm=>this:fetch("start",$last),this:cast(this:uniform(1,$fixed-dim),"integer"))[1]
           return
             fold-left(
               1 to $N, $start,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $sums :=
                   if (exists($fixed-sums)) then $fixed-sums
                   else if (exists($algorithm=>this:fetch("sums",$values[last()])))
                   then $algorithm=>this:fetch("sums",$values[last()])
                   else (
                     let $matrix := $algorithm=>this:fetch("matrix",$values[last()]) return
                     let $dim := (
                       $algorithm=>this:fetch("dim",$values[last()]),
                       math:sqrt(count($matrix)) cast as xs:integer
                     )[1]
                     return this:markov-sums($dim, $matrix)
                   )
                 return
                 let $dim := (
                   $fixed-dim,
                   $algorithm=>this:fetch("dim",$values[last()]),
                   math:sqrt(count($sums)) cast as xs:integer
                 )[1]
                 return
                 let $val := this:markov($values[last()],$dim,$sums)
                 return $val
               }
             )[position()>1] (: start is never empty :)
         else (
           let $result :=
             fold-left(
               1 to $N, $last,
               function($values as item()*, $i as xs:integer) as item()* {
                 $values
                 ,
                 let $min := $algorithm=>this:fetch("min",$values[last()]) return
                 let $max := $algorithm=>this:fetch("max",$values[last()]) return
                 let $val := this:constant($min, $max)
                 return $val
               }
             )
           return
             if (empty($last)) then $result else $result[position()>1]
         )
         '/>
          </xsl:when>
          <xsl:otherwise>
          <xsl:sequence select='
          (: Everything is fixed, just look up keys :)
          let $distribution := ($algorithm("distribution"),"uniform")[1] return
          let $min := $algorithm("min") (: OK to be empty :) return
          let $max := $algorithm("max") (: OK to be empty :)
          return (
            if ($distribution="constant") then
              for $i in 1 to $N
              return this:constant($min, $max)
            else if ($distribution="uniform") then 
              for $i in 1 to $N
              return this:uniform($min, $max)
            else if ($distribution="normal") then
              let $μ := ($algorithm("mean"),0)[1] return
              let $σ := ($algorithm("std"),1)[1] return
              for $i in 1 to $N
              return this:normal($μ, $σ)
            else if ($distribution="skewed") then
              let $μ := ($algorithm("mean"),0)[1] return
              let $σ := ($algorithm("std"),1)[1] return
              let $α := ($algorithm("skew"),0)[1] return
              for $i in 1 to $N
              return this:skewed($μ, $σ, $α)
            else if ($distribution="bernoulli") then
              let $p := ($algorithm("p"),50)[1] return
              for $i in 1 to $N
              return 
                if (this:flip($p)) then 1 else 0
            else if ($distribution="flip") then
              let $p := ($algorithm("p"),50)[1] return
              for $i in 1 to $N
              return this:flip($p)
            else if ($distribution="sums") then
              let $sums := $algorithm("sums") return
              let $max-rank := count($sums) return
              for $i in 1 to $N
              return this:select-index($sums, $max-rank)
            else if ($distribution="multimodal") then (
              let $distributions := $algorithm("distributions") return
              let $selector := $algorithm("selector") return
              let $flips := for $i in 1 to $N return this:randomize($selector)
              return
              (: let $numd := count($distributions) return
              let $flips := for $i in 1 to $N return this:cast(this:uniform(1, $numd),"integer")
              return
              :)
              (: RAW: v11, v12, ..., v1N, v21, ..., v2N, ...., vd1, ... vdN :)
              (: Run this way so that $last applies :)
              let $raw :=
                for $subdistribution in $distributions
                return this:randomize($N, $last, $subdistribution)
              return
              for $i in 1 to $N
              return (
                $raw[($flips[$i] - 1)*$N + $i]
              )
            )
            else if ($distribution="binomial-poisson") then
              let $ps := $algorithm("probabilities") return
              for $i in 1 to $N
              return this:binomial-poisson($ps)
            else if ($distribution="poisson") then
              let $λ := $algorithm("mean") return
              for $i in 1 to $N
              return this:poisson($λ)
            else if ($distribution="exponential") then
              let $λ := 1 div $algorithm("mean") return
              for $i in 1 to $N
              return this:exponential($λ)
            else if ($distribution="zipf") then
              let $sums :=
                if (exists($algorithm("sums")))
                then $algorithm("sums")
                else
                  this:zipf-sums(
                    ($algorithm("alpha"),1)[1],
                    ($algorithm("limit"),1000)[1]
                  )
              return
              let $max-rank := count($sums) return
              for $i in 1 to $N
              return this:select-index($sums, $max-rank)
            else if ($distribution="markov") then
              let $sums :=
                if (exists($algorithm("sums")))
                then $algorithm("sums")
                else (
                  let $matrix := $algorithm("matrix") return
                  let $dim := (
                    $algorithm("dim"),
                    math:sqrt(count($matrix)) cast as xs:integer
                  )[1]
                  return this:markov-sums($dim, $matrix)
                )
              return
              let $dim := (
                $algorithm("dim"),
                math:sqrt(count($sums)) cast as xs:integer
              )[1]
              return
              let $start := ($last,$algorithm("start"),this:cast(this:uniform(1,$dim),"integer"))[1]
              return
                fold-left(
                  1 to $N, $start,
                  function($z as item()*, $i as xs:integer) as item()* {
                    $z,
                    this:markov($z[last()],$dim,$sums)
                  }
                )[position()>1]
            else (
              for $i in 1 to $N return this:constant($min, $max)
            )
          )
          '/>
       </xsl:otherwise>
     </xsl:choose>
   </xsl:function>

   <!--
    : this:select-random()
    : Given a set of values and a randomizer (see randomize()), use the output
    : of randomize as an index into the value set, and return the selected values.
    : 
    : $N: How many values to return
    : $last: Previous value (needed for markov, available to key callback)
    : $algorithm: The algorithm to use, a map with various parameter keys
    : $values: The value set to choose from
    :-->
   <xsl:function name="this:select-random" as="item()*" new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="values" as="item()*"/>
      <xsl:sequence select='
      let $count := count($values) return
      let $indices := this:randomize($N, $last, $algorithm) return
      for $raw-index in $indices return
      let $index := max((min(($raw-index cast as xs:integer, $count)),1))
      return $values[$index]
      '/>
   </xsl:function>

   <!--
    : this:select-random()
    : Given a set of values and a randomizer (see randomize()), use the output
    : of randomize as an index into the value set, and return the selected values.
    : 
    : $N: How many values to return
    : $algorithm: The algorithm to use, a map with various parameter keys
    : $values: The value set to choose from
    :-->
   <xsl:function name="this:select-random" as="item()*" new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="values" as="item()*"/>
      <xsl:sequence select='
      let $count := count($values) return
      let $indices := this:randomize($N, (), $algorithm) return
      for $raw-index in $indices return
      let $index := max((min(($raw-index cast as xs:integer, $count)),1))
      return $values[$index]
      '/>
   </xsl:function>

   <!--
    : this:select-random()
    : Given a set of values and a randomizer (see randomize()), use the output
    : of randomize as an index into the value set, and return the selected values
    : 
    : $algorithm: The algorithm to use, a map with various parameter keys
    : $values: The value set to choose from
    :-->
   <xsl:function name="this:select-random" as="item()" new-each-time="yes">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="values" as="item()*"/>
      <xsl:sequence select='
      let $count := count($values) return
      let $raw-index := this:randomize($algorithm) return
      let $index := max((min(($raw-index cast as xs:integer, $count)),1))
      return $values[$index]
      '/>
   </xsl:function>

   <!--
    : this:select-random()
    : Given a set of values pick as (uniformly random) value from the
    : value set, and return it
    : 
    : $values: The value set to choose from
    :-->
   <xsl:function name="this:select-random" as="item()" new-each-time="yes">
      <xsl:param name="values" as="item()*"/>
      <xsl:sequence select='
      let $count := count($values) return
      let $algorithm :=
        dist:uniform(1, $count)=>dist:cast("integer")=>dist:is-complex(false())
      return
      let $raw-index := this:randomize($algorithm) return
      let $index := max((min(($raw-index cast as xs:integer, $count)),1))
      return $values[$index]
      '/>
   </xsl:function>

   <!--
    : this:randomize()
    : Return a sequence of random data distributed as per the algorithm.
    : 
    : $N: How many values to return
    : $last: Previous value (needed for markov, available to key callback)
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   distribution: Distribution to use, one of "constant", "uniform", "normal",
    :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal",
    :     "binomial-poisson", "poisson", "exponential"
    :   min: minimum value (optional)
    :   max: maximum value (optional)
    :   pre-multiplier: multiplier before min/max (optional)
    :   post-multiplier: multiplier after min/max (optional)
    :   post-shift: shift to add after min/max (optional)
    :   cast: cast type (optional)
    :   mean: mean of distribution (normal, skewed, poisson, exponential) (default=0)
    :   std: standard deviation of distribution (normal, skewed) (default=1)
    :   skew: skew of distribution (skewed) (default=0)
    :   p: probability (bernoulli, flip) (default=50)
    :   probabilities: probabilities (binomial-poisson)
    :   sums: cumulative probability sums (zipf, markov, sums)
    :   alpha: alpha parameter (zipf) (needed if no sums) (default=1)
    :   limit: number of sums (zipf) (needed if no sums) (default=1000)
    :   start: index of starting symbol (markov) (default=uniform[1,dim])
    :   dim: size of each dimension of Markov matrix (markov)
    :   matrix: raw Markov matrix (used if sums not provided, not recommended)
    :   distributions: sequence of distributions to combine (multimodal)
    :   truncation: how to keep value in [min,max] range (not constant/uniform)
    :     ceiling => ceiling/floor to min/max
    :     drop => toss out (so may return < $N)
    :     resample (default) => resample until an in-bound value is found or
    :       resample limit is reached ($RESAMPLE-LIMIT=10) then use ceiling
    :-->
   <xsl:function name="this:randomize" as="item()*" new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:is-complex($algorithm)) then (
        if (this:is-complex("distribution", $algorithm) or
            this:is-complex("distributions", $algorithm) or
            this:is-complex("selector", $algorithm)
        ) then (
          (: We change distributions, so we have to do this one value at a time :)
          let $result :=
            fold-left(
              1 to $N, $last,
              function($vals as item()*, $i as xs:integer) as item()* {
                $vals
                ,
                this:randomize(1, $vals[last()], $algorithm)
              }
            )
          return (
            if (empty($last)) then $result
            else $result[position()>1]
          )
        ) else (
          (: Distribution is fixed, but we may alter any other parameter 
           : so need to feed $last as appropriate to the sequence
           : For multimodal distributions is also fixed
           :)
          let $distribution := ($algorithm("distribution"),"uniform")[1] return
          let $values := this:raw-distribution($N, $last, $algorithm) return
          let $values :=
            fold-left(
              $values, $last,
              function($vals as item()*, $val as item()) as item()* {
                $vals
                ,
                this:modify($val, $vals[last()], $algorithm)
              }
            )
          return
            if (empty($last)) then $values
            else $values[position()>1]
        )
      ) else (
        (: Everything is fixed, just look up keys :)
        let $distribution := ($algorithm("distribution"),"uniform")[1] return
        let $values := this:raw-distribution($N, $last, $algorithm) return
        let $values := this:modify-all($values, $last, $algorithm)
        return (
          $values
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:randomize()
    : Generate one random value with given previous value.
    : See randomize($N, $last, $algorithm) above.
    :-->
   <xsl:function name="this:randomize" as="item()*" new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:randomize($N, (), $algorithm)
      '/>
   </xsl:function>

   <!--
    : this:randomize()
    : Generate one random value.
    : See randomize($N, $last, $algorithm) above.
    :-->
   <xsl:function name="this:randomize" as="item()*" new-each-time="yes">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:randomize(1, (), $algorithm)
      '/>
   </xsl:function>

   <!--
    : this:histogram()
    : Print out the distribution as a histogram.
    : 
    : $width: Size of each bucket (max - min)
    : $height: Number of values per star
    : $data: Data to analyze
    :-->
   <xsl:function name="this:histogram" as="xs:string*">
      <xsl:param name="width" as="xs:integer"/>
      <xsl:param name="height" as="xs:integer"/>
      <xsl:param name="data" as="xs:double*"/>
      <xsl:sequence select='
      let $buckets :=
        fold-left(
          $data, map {},
          function($buckets as map(xs:integer, xs:integer), $d as xs:double) as map(xs:integer,xs:integer) {
            let $key := round-half-to-even($d div $width, 0) cast as xs:integer
            return
            let $num := $buckets($key)
            return (
              if (exists($num))
              then $buckets=>map:put($key, $num+1)
              else $buckets=>map:put($key, 1)
            )
          }
        )
      return
      for $key in sort($buckets=>map:keys()) return
      let $num := $buckets($key)
      return (
        format-number(xs:integer($key)*$width," 00;-00")||" "||
        format-number($num,"000")||" "||
        string-join(for $i in 1 to $num idiv $height return "*","")
      )
      '/>
   </xsl:function>

   <!--
    : this:mean()
    : Return the mean of the distribution. Some distributions don't have a mean;
    : these will raise an error.
    : 
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   (see randomize())
    : $last: Previous value
    :-->
   <xsl:function name="this:mean" as="item()*">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:param name="last" as="item()?"/>
      <xsl:sequence select='
      let $distribution := $algorithm("distribution") return
      if ($distribution="constant") then this:randomize($algorithm)
      else if ($distribution="uniform") then
        let $min := $algorithm=>this:fetch("min",$last) (: OK to be empty :) return
        let $max := $algorithm=>this:fetch("max",$last) (: OK to be empty :)
        return
          if ($max &lt; $min) then errors:error("RAND-BADMINMAX",($min,$max))
          else ($max - $min) idiv 2
      else if ($distribution="normal") then ($algorithm=>this:fetch("mean",$last),0)[1]
      else if ($distribution="skewed") then ($algorithm=>this:fetch("mean",$last),0)[1]
      else if ($distribution="bernoulli") then
        let $p := ($algorithm=>this:fetch("p",$last),50)[1]
        return $p div 100
      else if ($distribution="flip") then
        let $p := ($algorithm=>this:fetch("p",$last),50)[1]
        return $p div 100
      else if ($distribution="binomial-poisson") then
        let $ps := $algorithm=>this:fetch("probabilities", $last)
        return sum($ps) div 100
      else if ($distribution="poisson") then $algorithm=>this:fetch("mean",$last)
      else if ($distribution="exponential") then $algorithm=>this:fetch("mean",$last)
      else errors:error("RAND-UNIMPLEMENTED", ($algorithm("distribution"), "mean"))
      '/>
   </xsl:function>

   <!--
    : this:mean()
    : Return the mean of the distribution. Some distributions don't have a mean;
    : these will raise an error.
    : 
    : $algorithm: The algorithm to use, a map with various parameter keys
    :   (see randomize())
    :-->
   <xsl:function name="this:mean" as="item()*">
      <xsl:param name="algorithm" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:mean($algorithm, ())
      '/>
   </xsl:function>

   <!--
    : this:id()
    : Construct an id of the form {$prefix}{hex-number in range 0 to $n}
    :-->
   <xsl:function name="this:id" as="xs:string" new-each-time="yes">
      <xsl:param name="prefix" as="xs:string"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:sequence select='
      $prefix||"-"||util:integer-to-hex(this:uniform(0,$n)=>this:cast("integer"))
      '/>
   </xsl:function>

   <!--
    : this:id()
    : Construct an id of the form {$prefix}{hex-number}
    :-->
   <xsl:function name="this:id" as="xs:string" new-each-time="yes">
      <xsl:param name="prefix" as="xs:string"/>
      <xsl:sequence select='
      $prefix||"-"||util:integer-to-hex(this:uniform(0,1E7)=>this:cast("integer"))
      '/>
   </xsl:function>

   <!--
    : this:mutual-prime-pair()
    : A random pair of mutually prime numbers less than the given maximum,
    : which has to be more than 2, or you can't find two primes.
    :-->
   <xsl:function name="this:mutual-prime-pair" as="xs:integer*" new-each-time="yes">
      <xsl:param name="max" as="xs:integer"/>
      <xsl:sequence select='
      if ($max le 2) then errors:error("ML-BADARGS", ("max", $max)) else (),
      let $base-selection :=
        if ($max > $util:PRIMES100[last()])
        then $util:CANVAS-PRIMES[. le $max]
        else $util:PRIMES100[. le $max]
      return
      let $q := this:select-random($base-selection) return
      let $r :=
        util:while(
          function($r as xs:integer) as xs:boolean {
            max(($q,$r)) mod min(($q,$r)) = 0
          },
          function($r as xs:integer) as xs:integer {
            $r + 1
          },
          this:select-random(2 to 2 * util:round(math:sqrt($q) + 1))
        )
      return ($q, $r)
      '/>
   </xsl:function>

   <!--
    : this:partition()
    : Make a partition (sum totalling to $sum) where no part is more than $max
    : Example: partition(10, 5) may be (4, 1, 2, 3) but not (6, 4)
    :-->
   <xsl:function name="this:partition" as="xs:integer*" new-each-time="yes">
      <xsl:param name="sum" as="xs:integer"/>
      <xsl:param name="max" as="xs:integer"/>
      <xsl:sequence select='
      if ($sum = 0 or $max = 0) then ()
      else if ($sum = 1) then 1
      else (
        let $limit := min(($sum, $max)) return
        let $first :=
          this:normal($limit div 2, $limit div 2)=>
            this:range(1, $limit)=>
            this:cast("integer")
        return (
          $first, this:partition($sum - $first, $max)
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:partition()
    : Make a partition (sum totalling to $sum) where no part is more than $max
    : and no part is less than $min
    : Example: partition(10, 5) may be (4, 2, 2, 2) but not (4, 1, 2, 3)
    :-->
   <xsl:function name="this:partition" as="xs:integer*" new-each-time="yes">
      <xsl:param name="sum" as="xs:integer"/>
      <xsl:param name="max" as="xs:integer"/>
      <xsl:param name="min" as="xs:integer"/>
      <xsl:sequence select='
      if ($sum = 0 or $max = 0) then ()
      else if ($sum = $min) then $min
      else (
        let $tentative := this:do-partition($sum, $max, $min) return
        let $tentative-sum := sum($tentative) return
        let $delta := $sum - $tentative-sum
        return (
          util:assert($delta >= 0, "Sum of tentative partition is too high"),
          if ($delta = 0) then $tentative else (
            fold-left(1 to $delta, $tentative,
              function($seq as xs:integer*, $i as xs:integer) as xs:integer* {
                let $pick :=
                  this:shuffle(
                    for $j in 1 to count($seq) return (
                      if ($seq[$j] &lt; $max) then $j else ()
                    )
                    (: for $x at $j in $seq where $x &lt; $max return $j :)
                  )=>head()
                return
                for $j in 1 to count($seq)
                return (
                  if ($j = $pick) then $seq[$j] + 1 else $seq[$j]
                )
              }
            )
          )
        )
      )
      '/>
   </xsl:function>

   <xsl:function name="this:do-partition" as="xs:integer*" new-each-time="yes" visibility="private">
      <xsl:param name="sum" as="xs:integer"/>
      <xsl:param name="max" as="xs:integer"/>
      <xsl:param name="min" as="xs:integer"/>
      <xsl:sequence select='
      if ($sum = 0 or $max = 0 or $min > $sum) then ()
      else if ($sum = $min) then $min
      else (
        let $limit := min(($sum, $max)) return
        let $first :=
          this:normal($limit div 2, $limit div 2)=>
            this:range($min, $limit)=>
            this:cast("integer")
        return (
          $first, this:do-partition($sum - $first, $max, $min)
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:n-partition()
    : Make a partition (sum) of $sum with exactly $n parts
    : It is possible to get unlucky and end up with a 0 partition here: we have
    : no min constraint. If you care about that, you'll need to adjust the sequence
    : afterwards.
    :-->
   <xsl:function name="this:n-partition" as="xs:integer*" new-each-time="yes">
      <xsl:param name="sum" as="xs:integer"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:sequence select='
      if ($n = 0) then ()
      else if ($n = 1) then $sum
      else if ($sum le 1) then (
        $sum, this:n-partition(0, $n - 1)
      )
      else (
        let $limit := max(($sum - 1, 1)) return
        let $first :=
          this:normal($limit div 2, $limit div 4)=>
            this:range(1, $limit)=>
            this:cast("integer")
        return (
          $first, this:n-partition($sum - $first, $n - 1)
        )
      )
      '/>
   </xsl:function>
</xsl:package>
<!--random.xsl-->
