<?xml version="1.0" encoding="UTF-8"?>
<!--
  : Roots of linear, quadratic, and cubic equations.
  :
  : Copyright© Mary Holstege 2021-2023
  : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/core/roots"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:poly="http://mathling.com/type/polynomial"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"             
             xmlns:z="http://mathling.com/core/complex"
             version="3.0"
             name="http://mathling.com/core/roots"
             exclude-result-prefixes="this xs math map array config errors util z poly">

  <xsl:expose component="variable" names="this:*" visibility="public"/>
  <xsl:expose component="function" names="this:*" visibility="public"/>

  <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/complex" package-version="*"/>
  <xsl:use-package name="http://mathling.com/type/polynomial" package-version="*"/>

  <!-- 
     : Generic solution to linear equation ax + b = 0
     : x = -b/a
     :-->

  <!--
    : this:linear-root()
    : Return the root of the linear equation of the form ax + b = 0
    : If a is 0 there is no solution unless b = 0 in which case there are
    : an infinite number of solutions, so callers should check on what to
    : do in that case.
    :-->
  <xsl:function name="this:linear-root" as="xs:double?">
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:sequence select='
      if ($a = 0) then ()
      else -$b div $a
    '/>
  </xsl:function>

  <!--
    : Generic solution to quadratic equation ax² + bx + c = 0
    : x = (-b ± √(b² - 4ac))/2a
    :-->

  <!--
    : this:quadratic-real-roots()
    : Return all the real-valued (within ε) roots of the quadratic equation 
    : of the form ax² + bx + c = 0
    :-->
  <xsl:function name="this:quadratic-real-roots" as="xs:double*">
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:param name="c" as="xs:double"/>
    <xsl:sequence select='
      (: If a = 0 this is a linear equation, use the simpler math :)
      if ($a = 0) then (
        this:linear-root($b, $c)
      ) else if ($b*$b >= 4*$a*$c) then (
        (-$b + math:sqrt($b*$b - 4*$a*$c)) div (2*$a),
        (-$b - math:sqrt($b*$b - 4*$a*$c)) div (2*$a)
      ) else (
        (: √ negative => not a real root :)
      )
    '/>
  </xsl:function>

  <!--
    : this:quadratic-roots()
    : Return all the roots of quadratic equation of the form ax² + bx + c = 0
    : Values returned as complex numbers.
    :-->
  <xsl:function name="this:quadratic-roots" as="map(xs:string, item()*)*">
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:param name="c" as="xs:double"/>
    <xsl:sequence select='
      (: If a = 0 this is a linear equation, use the simpler math :)
      if ($a = 0) then (
        this:linear-root($b, $c)!z:as-complex(.)
      ) else if ($b*$b >= 4*$a*$c) then (
        z:as-complex((-$b + math:sqrt($b*$b - 4*$a*$c)) div (2*$a)),
        z:as-complex((-$b - math:sqrt($b*$b - 4*$a*$c)) div (2*$a))
      ) else (
        let $sqrt := z:sqrt(z:as-complex($b*$b - 4*$a*$c))
        return (
          (z:as-complex(-$b)=>z:add($sqrt))=>z:times(1 div (2*$a)),
          (z:as-complex(-$b)=>z:sub($sqrt))=>z:times(1 div (2*$a))
        )
      )
    '/>
  </xsl:function>

  <!--
    : Per: https://en.wikipedia.org/wiki/Cubic_equation
    : Generic solution to cubic equation ax³ + bx² + cx + d = 0
    :
    : Δ0 = b² - 3ac
    : Δ1 = 2b³ - 9abc + 27a²d
    : C = ((Δ1 ± √(Δ1² - 4Δ0³))/2)^(1/3)
    : 
    : x[k] = -(1/(3a))(b + (ζ^k)C + Δ0/((ζ^k)C)) for k in 0 to 2
    : 
    : ζ = (-1 + √(-3))/2 = (-1 + i√3)/2 = -1/2 + i√3/2
    :-->

  <!--
    : this:cubic-real-roots()
    : Return all the real-valued (within ε) roots of cubic equation 
    : of the form ax³ + bx² + cx + d = 0
    :-->
  <xsl:function name="this:cubic-real-roots" as="xs:double*">
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:param name="c" as="xs:double"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:sequence select='
      (: If a = 0 this is a quadratic equation, use the simpler math :)
      if ($a = 0) then (
        this:quadratic-real-roots($b, $c, $d)
      ) else (
        let $p := $b*$b - 2*$a*$c return
        let $q := 2*$b*$b*$b - 9*$a*$b*$c + 27*$a*$a*$d
        return (
          if ($p = $q and $p = 0) then (
            -$b div (3*$a)
           ) else (
            for $root in this:cubic-roots($a, $b, $c, $d) return (
              if (z:is-real($root)) then $root=>z:as-real() else ()
            )
          )
        )
      )
    '/>
  </xsl:function>

  <xsl:function name="this:try-cbrt" as="xs:double" visibility="private">
    <xsl:param name="p" as="xs:double"/>
    <xsl:param name="q" as="xs:double"/>
    <xsl:param name="plus" as="xs:boolean"/>
    <xsl:try>
      <xsl:choose>
        <xsl:when test="$plus">
          <xsl:sequence select='util:cbrt(($q + math:sqrt($q*$q - 4*$p*$p*$p)) div 2)'/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:sequence select='util:cbrt(($q - math:sqrt($q*$q - 4*$p*$p*$p)) div 2)'/>
        </xsl:otherwise>
      </xsl:choose>
      <xsl:catch errors="*">
        <xsl:sequence select='xs:double("NaN")'/>
      </xsl:catch>
    </xsl:try>
  </xsl:function>
  
  <!--
    : this:cubic-roots()
    : Return all the roots of cubic equation of the form ax³ + bx² + cx + d = 0
    : Values returned as complex numbers.
    :-->
  <xsl:function name="this:cubic-roots" as="map(xs:string, item()*)*">
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:param name="c" as="xs:double"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:sequence select='
      (: If a = 0 this is a quadratic equation, use the simpler math :)
      if ($a = 0) then (
        this:quadratic-roots($b, $c, $d)
      ) else (
        let $p := $b*$b - 3*$a*$c return
        let $q := 2*$b*$b*$b - 9*$a*$b*$c + 27*$a*$a*$d return
        let $qz := z:as-complex($q) return
        let $pz := z:as-complex($p)
        return if ($p = $q and $p = 0) then z:as-complex(-$b div (3*$a)) else (
          let $ζ0 := z:as-complex(1) return
          let $ζ1 := z:complex(-1 div 2, math:sqrt(3) div 2) return
          let $ζ2 := z:multiply($ζ1, $ζ1) return
          (: r = cbrt((p ± sqrt(q² - 4p³))/2) :)
          let $r :=
            let $val := this:try-cbrt($p, $q, true())
              (:
              try {
                util:cbrt(
                  ($q + math:sqrt($q*$q - 4*$p*$p*$p)) div 2
                )
              } catch * {xs:double("NaN")}
              :)
            return (
              if ($val=$val) then z:as-complex($val) (: i.e. not NaN :)
              else (
                z:cbrt(
                  $qz=>z:add(
                    z:sqrt(
                      z:multiply($qz,$qz)=>
                        z:sub(
                         z:multiply(z:multiply(z:times($pz, 4), $pz), $pz)
                        )
                    )
                  )=>z:times(0.5)
                )
              )
            )
          return
          let $r :=
            if (z:pr($r)=0 and z:pi($r)=0) then (
              let $val := this:try-cbrt($p, $q, false())
                (:
                try {
                  util:cbrt(
                    ($q - math:sqrt($q*$q - 4*$p*$p*$p)) div 2
                  )
                } catch * {xs:double("NaN")}
                :)
              return (
                if ($val=$val) then z:as-complex($val) (: i.e. not NaN :)
                else (
                  z:cbrt(
                    $qz=>z:sub(
                      z:sqrt(
                        z:multiply($qz,$qz)=>
                          z:sub(
                           z:multiply(z:multiply(z:times($pz, 4), $pz), $pz)
                          )
                      )
                    )=>z:times(0.5)
                  )
                )
              )
            ) else $r
          return 
          let $rζ0 := $r=>z:multiply($ζ0) return
          let $rζ1 := $r=>z:multiply($ζ1) return
          let $rζ2 := $r=>z:multiply($ζ2) return
          let $bz := z:as-complex($b) return
          (: x = (-1/(3a))(b + (ζ^k)r + p/((ζ^k)r) for k=0 to 2 :)
          for $rζk in ($rζ0, $rζ1, $rζ2) return (
            ($bz=>z:add($rζk)=>z:add( $pz=>z:divide($rζk) ))=>z:times(-1 div (3*$a))
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:real-roots()
    :-->
  <xsl:function name="this:real-roots" as="xs:double*">
    <xsl:param name="polynomial" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $cfs := poly:coefficients($polynomial) return
      let $order :=  poly:order($polynomial)
      return
        if ($order = 1) then this:linear-root($cfs[1], $cfs[2])
        else if ($order = 2) then this:quadratic-real-roots($cfs[1], $cfs[2], $cfs[3])
        else if ($order = 3) then this:cubic-real-roots($cfs[1], $cfs[2], $cfs[3], $cfs[4])
        else errors:error("UTIL-CANNOT", ("real-roots", $polynomial))
    '/>
  </xsl:function>

  <!--
    : this:roots()
    :-->
  <xsl:function name="this:roots" as="map(xs:string, item()*)*">
    <xsl:param name="polynomial" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $cfs := poly:coefficients($polynomial) return
      let $order := poly:order($polynomial)
      return
        if ($order = 1) then this:linear-root($cfs[1], $cfs[2])!z:as-complex(.)
        else if ($order = 2) then this:quadratic-roots($cfs[1], $cfs[2], $cfs[3])
        else if ($order = 3) then this:cubic-roots($cfs[1], $cfs[2], $cfs[3], $cfs[4])
        else errors:error("UTIL-CANNOT", ("roots", $polynomial))
    '/>
  </xsl:function>

</xsl:package>
<!--roots.xsl-->
