<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Affine transformation support: the matrices plus point-level application.
 : Other kinds of region will call into this.
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/affine"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric/affine"
             exclude-result-prefixes="this xs math map array errors config util point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   
<!--
 :======================================================================
 : Affine transformations
 : The operations are allowed for all the region types, but they really
 : don't do the "right" thing for some of them, particularly boxes/spaces,
 : ellipses, arcs, and ellipsoids. For example, you can rotate a box,
 : but you don't get a rotated box out of the end of it, because boxes
 : here are axis-aligned boxes of space between minimum and maximum x-y
 : bounds. I allow the operation to proceed in these naive ways because
 : it allows for interesting variations in my art, and allows affine
 : transformations that preserve region types. If you want geometrically
 : correct transformations of the overall shapes, make a polygon from
 : the source region and transform that. For boxes you can make a polygon
 : from the edges, for ellipses you can use interpolate-ellipse() to 
 : get a set of points to construct a polygon from; similarly for arcs.
 : You can scale and translate rectangles as is just fine. It is the
 : rotations and shears that will fail. 
 : 
 : 3D solids will general do weird things when you apply 2D transforms to them
 : and ellipsoids are badly supported because I need to add rotation angles
 : to the representation.
 :
 : The basic rule of this API is that you get out the same kind of thing
 : you put in; so the cases where this can't be done you get "wrong"
 : results. Exception: generally you get generic polyhedra if you do anything
 : to a regular solid, because (naturally!) it is no longer regular so 
 : you can't stick with the simpler representation.
 :======================================================================
 :-->
 
<!--
 :============================================================
 : Matrix set-up, composition, etc.
 :============================================================
 :-->

<!--
 : 2D: M times [x y 1]
 : Identity:
 : | 1 0 0 |   
 : | 0 1 0 |
 : | 0 0 1 |
 :
 : Translation:
 : | 1 0 cx |   
 : | 0 1 cy |
 : | 0 0  1 |
 :
 : Reflection:
 : | -1  0 0 |   
 : |  0 -1 0 |
 : |  0  0 1 |
 : 
 : Scale:
 : | cx  0 0 |   
 : |  0 cy 0 |
 : |  0  0 1 |
 :
 : Rotate:
 : | cos(θ) -sin(θ) 0 |   
 : | sin(θ)  cos(θ) 0 |
 : |  0           0 1 |
 :
 : Shear
 : |  1 cx 0 |   
 : | cy  1 0 |
 : |  0  0 1 |
 :
 : 3D: M times [x y z 1]
 : Identity:
 : | 1 0 0 0 |
 : | 0 1 0 0 |
 : | 0 0 1 0 |
 : | 0 0 0 1 |
 :
 : Translation:
 : | 1 0 0 cx |   
 : | 0 1 0 cy |
 : | 0 0 1 cz |
 : | 0 0 0  1 |
 :
 : Scale:
 : | cx  0  0  0 |   
 : |  0 cy  0  0 |
 : |  0  0 cz  0 |
 : |  0  0  0  1 |
 :
 : Rotate about X:
 : | 1      0       0 0 |
 : | 0 cos(θ) -sin(θ) 0 |   
 : | 0 sin(θ)  cos(θ) 0 |
 : | 0      0       0 1 |
 :
 : Rotate about Y:
 : |  cos(θ) 0 sin(θ) 0 |
 : |       0 1      0 0 |   
 : | -sin(θ) 0 cos(θ) 0 |
 : | 0      0       0 1 |
 :
 : Rotate about Z:
 : | cos(θ) -sin(θ) 0 0 |
 : | sin(θ)  cos(θ) 0 0 |
 : |      0       0 1 0 |   
 : |      0       0 0 1 |   
 :
 : roll-pitch-yaw: Rx then Ry then Rz
 :
 : Shear
 : |   1 hxy hxz 0 |   
 : | hyx   1 hyz 0 |   
 : | hzx hzy   1 0 |   
 : | 0     0   0 1 |
 :
 : Perspective: translate relative to camera, then rotate 
 : roll, pitch, yaw for camera angles then 
 :
 : Project 
 : | 1 0 dx/dz 0 | 
 : | 0 1 dy/dz 0 |
 : | 0 0  1/dz 0 |
 : | 0 0     0 1 |
 : (dx, dy, dz) = from camera translations
 :-->

   <!--
    : this:affine2()
    : Apply the 2D transformation matrix to the point
    : 
    :     | a b c |      | px |        | a px + b py + c |
    : T = | d e f |  P = | py |  TxP = | d px + e py + f |
    :     | 0 0 1 |      |  1 |        |               1 |
    :
    : $p: coordinates of point
    : $m: matrix values (a,b,c,d,e,f)
    :-->
   <xsl:function name="this:affine2" as="xs:double*">
      <xsl:param name="p" as="xs:double*"/>
      <xsl:param name="m" as="xs:double*"/>
      <xsl:sequence select='
      util:assert(count($m)=6, "Expected 6 values for 2D matrix"),
      $m[1]*$p[1] + $m[2]*$p[2] + $m[3],
      $m[4]*$p[1] + $m[5]*$p[2] + $m[6],
      $p[position() &gt; 2] (: If you passed in z/w coordinates, leave them be :)
      '/>
   </xsl:function>
   
   <!--
    : this:affine3()
    : Apply the 3D transformation matrix to the point
    : 
    :     | a b c d |      | px |        | a px + b py + c pz + d |
    : T = | e f g h |  P = | py |  TxP = | e px + f py + g pz + h |
    :     | i j k l |      | pz |        | i px + j py + k pz + l |
    :     | 0 0 0 1 |      | 1  |        |                      1 |
    :
    : $p: coordinates of point
    : $m: matrix values (a,b,c,d,e,f,h,i,j,k,l)
    :-->
   <xsl:function name="this:affine3" as="xs:double*">
      <xsl:param name="p" as="xs:double*"/>
      <xsl:param name="m" as="xs:double*"/>
      <xsl:sequence select='
      util:assert(count($m)=12, "Expected 12 values for 3D matrix"),
      $m[1]*$p[1] + $m[2]*$p[2] + $m[3]*($p[3],0)[1] + $m[4],
      $m[5]*$p[1] + $m[6]*$p[2] + $m[7]*($p[3],0)[1] + $m[8],
      $m[9]*$p[1] + $m[10]*$p[2] + $m[11]*($p[3],0)[1] + $m[12],
      $p[position() &gt; 3] (: If you passed in w coordinate, leave it be :)
      '/>
   </xsl:function>
   
   <!--
    : this:compose2()
    : Compose two transformation matrixes: T1 x T2 ==
    : Apply T2 then T1
    : 
    :      | a b c |       | g h i |          | ag+bj ah+bk ai+bl+c |
    : T1 = | d e f |  T2 = | j k l |  T1xT2 = | dg+ej dh+ek di+el+f |
    :      | 0 0 1 |       | 0 0 1 |          | 0     0     1       |
    :-->
   <xsl:function name="this:compose2" as="xs:double*">
      <xsl:param name="m1" as="xs:double*"/>
      <xsl:param name="m2" as="xs:double*"/>
      <xsl:sequence select='
      util:assert(count($m1)=6, "Expected 6 values for 2D matrix"),
      util:assert(count($m2)=6, "Expected 6 values for 2D matrix"),

      $m1[1]*$m2[1] + $m1[2]*$m2[4],
      $m1[1]*$m2[2] + $m1[2]*$m2[5],
      $m1[1]*$m2[3] + $m1[2]*$m2[6] + $m1[3],

      $m1[4]*$m2[1] + $m1[5]*$m2[4],
      $m1[4]*$m2[2] + $m1[5]*$m2[5],
      $m1[4]*$m2[3] + $m1[5]*$m2[6] + $m1[6]
      '/>
   </xsl:function>
   
   <!--
    : this:compose3()
    : Compose two transformation matrixes: T1 x T2
    : 
    :      | a b c d |       | m n o p |          | am+bq+cu an+br+cv ao+bs+cw ap+bt+cx+d |
    : T1 = | e f g h |  T2 = | q r s t |  T1xT2 = | em+fq+gu en+fr+gv eo+fs+gw ep+ft+gx+h |
    :      | i j k l |       | u v w x |          | im+jq+ku in+jr+kv io+js+kw ip+jt+kx+l |
    :      | 0 0 0 1 |       | 0 0 0 1 |          |        0        0        0          1 |
    :-->
   <xsl:function name="this:compose3" as="xs:double*">
      <xsl:param name="m1" as="xs:double*"/>
      <xsl:param name="m2" as="xs:double*"/>
      <xsl:sequence select='
      util:assert(count($m1)=12, "Expected 12 values for 2D matrix"),
      util:assert(count($m2)=12, "Expected 12 values for 2D matrix"),

      $m1[1]*$m2[1] + $m1[2]*$m2[5] + $m1[3]*$m2[9],
      $m1[1]*$m2[2] + $m1[2]*$m2[6] + $m1[3]*$m2[10],
      $m1[1]*$m2[3] + $m1[2]*$m2[7] + $m1[3]*$m2[11],
      $m1[1]*$m2[4] + $m1[2]*$m2[8] + $m1[3]*$m2[12] + $m1[4],

      $m1[5]*$m2[1] + $m1[6]*$m2[5] + $m1[7]*$m2[9],
      $m1[5]*$m2[2] + $m1[6]*$m2[6] + $m1[7]*$m2[10],
      $m1[5]*$m2[3] + $m1[6]*$m2[7] + $m1[7]*$m2[11],
      $m1[5]*$m2[4] + $m1[6]*$m2[8] + $m1[7]*$m2[12] + $m1[8],

      $m1[9]*$m2[1] + $m1[10]*$m2[5] + $m1[11]*$m2[9],
      $m1[9]*$m2[2] + $m1[10]*$m2[6] + $m1[11]*$m2[10],
      $m1[9]*$m2[3] + $m1[10]*$m2[7] + $m1[11]*$m2[11],
      $m1[9]*$m2[4] + $m1[10]*$m2[8] + $m1[11]*$m2[12] + $m1[12]
      '/>
   </xsl:function>
   
   <!--
    : this:identity2()
    :-->
   <xsl:function name="this:identity2" as="xs:double*">
      <xsl:sequence select='
      1, 0, 0,
      0, 1, 0
      '/>
   </xsl:function>
   
   <!--
    : this:identity3()
    :-->
   <xsl:function name="this:identity3" as="xs:double*">
      <xsl:sequence select='
      1, 0, 0, 0,
      0, 1, 0, 0,
      0, 0, 1, 0
      '/>
   </xsl:function>
   
   <!--
    : this:translation2()
    :-->
   <xsl:function name="this:translation2" as="xs:double*">
      <xsl:param name="tx" as="xs:double"/>
      <xsl:param name="ty" as="xs:double"/>
      <xsl:sequence select='
      1, 0, $tx,
      0, 1, $ty
      '/>
   </xsl:function>
   
   <!--
    : this:translation3()
    :-->
   <xsl:function name="this:translation3" as="xs:double*">
      <xsl:param name="tx" as="xs:double"/>
      <xsl:param name="ty" as="xs:double"/>
      <xsl:param name="tz" as="xs:double"/>
      <xsl:sequence select='
      1, 0, 0, $tx,
      0, 1, 0, $ty,
      0, 0, 1, $tz
      '/>
   </xsl:function>
   
   <!--
    : this:scaling2()
    :-->
   <xsl:function name="this:scaling2" as="xs:double*">
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      $sx, 0, 0,
      0, $sy, 0
      '/>
   </xsl:function>
   
   <!--
    : this:scaling3()
    :-->
   <xsl:function name="this:scaling3" as="xs:double*">
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:sequence select='
      $sx, 0, 0, 0,
      0, $sy, 0, 0,
      0, 0, $sz, 0
      '/>
   </xsl:function>
   
   <!--
    : this:scaling2()
    : scale with respect to an origin: translate, scale, translate back 
    :-->
   <xsl:function name="this:scaling2" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      if (point:same($center, $point:ORIGIN)) then this:scaling2($sx, $sy)
      else (
        this:compose2(
          this:translation2(point:px($center), point:py($center)),
          this:compose2(
            this:scaling2($sx, $sy),
            this:translation2(-point:px($center), -point:py($center))
          )
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scaling3()
    :-->
   <xsl:function name="this:scaling3" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:sequence select='
      if (point:same($center, $point:ORIGIN3D)) then this:scaling3($sx, $sy, $sz)
      else (
        this:compose3(
          this:translation3(point:px($center), point:py($center), point:pz($center)),
          this:compose3(
            this:scaling3($sx, $sy, $sz),
            this:translation3(-point:px($center), -point:py($center), -point:pz($center))
          )
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:reflection2()
    : reflect about a center: translate to 0, reflect, translate back to center 
    :-->
   <xsl:function name="this:reflection2" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:compose2(
        this:translation2(point:px($center), point:py($center)),
        this:compose2(
          this:scaling2(-1, -1),
          this:translation2(-point:px($center), -point:py($center))
         )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:reflection3()
    :-->
   <xsl:function name="this:reflection3" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:compose3(
        this:translation3(point:px($center), point:py($center), point:pz($center)),
        this:compose3(
          this:scaling3(-1, -1, -1),
          this:translation3(-point:px($center), -point:py($center), -point:pz($center))
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:reflection2()
    : reflect about a line: translate to start, rotate, scale, rotate back, translate back
    :-->
   <xsl:function name="this:reflection2" as="xs:double*">
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (point:same($start, $end)) then this:reflection2($start) else (
        let $dx := point:px($end) - point:px($start) return
        let $dy := point:py($end) - point:py($start) return
        let $d := math:sqrt($dx*$dx + $dy*$dy) return
        let $sinAlpha := $dy div $d return
        let $cosAlpha := $dx div $d
        return (
          this:compose2(
            this:translation2(point:px($start), point:py($start)),
            this:compose2(
              this:rotation-helper($sinAlpha, $cosAlpha),
              this:compose2(
                this:scaling2(1,-1),
                this:compose2(
                  this:rotation-helper(-$sinAlpha, $cosAlpha),
                  this:translation2(-point:px($start), -point:py($start))
                )
              )
            )
          )
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotation2()
    :-->
   <xsl:function name="this:rotation2" as="xs:double*">
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      let $θ :=
        if ($config:CANVAS-GEOMETRY)
        then util:radians(-$degrees)
        else util:radians($degrees)
      return (
        this:rotation-helper(math:sin($θ), math:cos($θ))
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotation-helper()
    : This version is useful as an intermediary for reflections 
    :-->
   <xsl:function name="this:rotation-helper" as="xs:double*">
      <xsl:param name="sinθ" as="xs:double"/>
      <xsl:param name="cosθ" as="xs:double"/>
      <xsl:sequence select='
      $cosθ, -$sinθ, 0,
      $sinθ,  $cosθ, 0
      '/>
   </xsl:function>
   
   <!--
    : this:rotation3()
    :-->
   <xsl:function name="this:rotation3" as="xs:double*">
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:sequence select='
      let $rollθ := util:radians($roll-degrees) return
      let $pitchθ := util:radians($pitch-degrees) return
      let $yawθ := util:radians($yaw-degrees) return
      let $rx := (
        1, 0, 0, 0,
        0, math:cos($rollθ), -math:sin($rollθ), 0,
        0, math:sin($rollθ), math:cos($rollθ), 0
      )  return
      let $ry := (
        math:cos($pitchθ), 0, math:sin($pitchθ), 0,
        0, 1, 0, 0,
        -math:sin($pitchθ), 0, math:cos($pitchθ), 0
      ) return
      let $rz := (
        math:cos($yawθ), -math:sin($yawθ), 0, 0,
        math:sin($yawθ), math:cos($yawθ), 0, 0,
        0, 0, 1, 0
      )
      return this:compose3($rz, this:compose3($ry, $rx))
      '/>
   </xsl:function>
   
   <!--
    : this:rotation2()
    : rotate about a center: translate, rotate, translate back
    :-->
   <xsl:function name="this:rotation2" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      if (point:same($center, $point:ORIGIN)) then this:rotation2($degrees)
      else (
        this:compose2(
          this:translation2(point:px($center), point:py($center)),
          this:compose2(
            this:rotation2($degrees),
            this:translation2(-point:px($center), -point:py($center))
          )
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotation3()
    :-->
   <xsl:function name="this:rotation3" as="xs:double*">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:sequence select='
      if (point:same($center, $point:ORIGIN3D)) then (
        this:rotation3($roll-degrees, $pitch-degrees, $yaw-degrees)
      ) else (
        this:compose3(
          this:translation3(point:px($center), point:py($center), point:pz($center)),
          this:compose3(
            this:rotation3($roll-degrees, $pitch-degrees, $yaw-degrees),
            this:translation3(-point:px($center), -point:py($center), -point:pz($center))
          )
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:shearing2()
    :-->
   <xsl:function name="this:shearing2" as="xs:double*">
      <xsl:param name="sxy" as="xs:double"/>
      <xsl:param name="syx" as="xs:double"/>
      <xsl:sequence select='
      1, $syx, 0,
      $sxy, 1, 0
      '/>
   </xsl:function>
   
   <!--
    : this:skewing2()
    : You should avoid angles near 90, 270 => shear approaches +/- infinity 
    :-->
   <xsl:function name="this:skewing2" as="xs:double*">
      <xsl:param name="sx-degrees" as="xs:double"/>
      <xsl:param name="sy-degrees" as="xs:double"/>
      <xsl:sequence select='
      1, math:tan(util:radians($sx-degrees)), 0,
      math:tan(util:radians($sy-degrees)), 1, 0
      '/>
   </xsl:function>
   
   <!--
    : this:shearing3()
    :-->
   <xsl:function name="this:shearing3" as="xs:double*">
      <xsl:param name="sxy" as="xs:double"/>
      <xsl:param name="syx" as="xs:double"/>
      <xsl:param name="sxz" as="xs:double"/>
      <xsl:param name="szx" as="xs:double"/>
      <xsl:param name="syz" as="xs:double"/>
      <xsl:param name="szy" as="xs:double"/>
      <xsl:sequence select='
      1, $sxy, $sxz, 0,
      $syx, 1, $syz, 0,
      $szx, $szy, 1, 0
      '/>
   </xsl:function>
   
   <!--
    : this:project3()
    :-->
   <xsl:function name="this:project3" as="xs:double*">
      <xsl:param name="dx" as="xs:double"/>
      <xsl:param name="dy" as="xs:double"/>
      <xsl:param name="dz" as="xs:double"/>
      <xsl:sequence select='
      if ($dz = 0) then errors:error("GEOM-BADPROJECT")
      else (
        1, 0, $dx div $dz, 0,  (: shear x to z :)
        0, 1, $dy div $dz, 0,  (: shear y to z :)
        0, 0, 1 div $dz, 0     (: scale z :)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:homogeneous-to-cartesian()
    : Perspective matrix gives us homogeneous coordinates: need to map back to cartesian x-y 
    :-->
   <xsl:function name="this:homogeneous-to-cartesian" as="xs:double*">
      <xsl:param name="coordinates" as="xs:double*"/>
      <xsl:sequence select='
      $coordinates[1] div $coordinates[3],
      $coordinates[2] div $coordinates[3]
      '/>
   </xsl:function>
   
   <!--
    : this:perspective3()
    :-->
   <xsl:function name="this:perspective3" as="xs:double*">
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:compose3(
        this:project3(point:px($display), point:py($display), point:pz($display)),
        this:compose3(
          this:rotation3($roll-degrees, $pitch-degrees, $yaw-degrees),
          this:translation3(-point:px($camera), -point:py($camera), -point:pz($camera))
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:perspective3()
    :-->
   <xsl:function name="this:perspective3" as="xs:double*">
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:compose3(
        this:project3(point:px($display), point:py($display), point:pz($display)),
        this:translation3(-point:px($camera), -point:py($camera), -point:pz($camera))
      )
      '/>
   </xsl:function>
   
   <!--
    : this:isometric3()
    : Isometric projection: two rotations and an orthographic projection
    : Rotate by β about x, rotate by α about y, scale(1,1,0)
    : β = +-/45°, α=asin(1/√3)=atan(1/√2)
    :-->
   <xsl:function name="this:isometric3" as="xs:double*">
      <xsl:param name="positive" as="xs:boolean"/>
      <xsl:sequence select='
      this:compose3(
        this:scaling3(1, 1, 0),
        this:rotation3(
          if ($positive) then 45 else -45,             (: roll about x :)
          util:degrees(math:asin(1 div math:sqrt(3))), (: pitch about y :)
          0                                            (: no yaw about z :)
        )
      )
      '/>
   </xsl:function>
   
<!--
  :============================================================
  : Public-facing API
  : You can use the lower-level one, e.g. if you want to chain a bunch
  : of transforms and only want to calculate the matrix once. Otherwise
  : these are more usable.
  :============================================================
  :-->

  <!--
    : this:apply-matrix2()
    : Apply a 2D affine matrix to the points
    : 
    : $points: the points to apply the affine transformation to
    : $affine2: the 2D matrix (6 values)
    :-->
   <xsl:function name="this:apply-matrix2" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="affine2" as="xs:double*"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), $affine2)
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:translate()
    : Translate the points by the given amount.
    : 
    : $points: points to translate
    : $x: how much to translate in the x direction
    : $y: how much to translate in the y direction
    :-->
   <xsl:function name="this:translate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point),this:translation2($x,$y))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:translate()
    : Translate the points by the given amount.
    : 
    : $points: points to translate
    : $x: how much to translate in the x direction
    : $y: how much to translate in the y direction
    : $z: how much to translate in the z direction
    :-->
   <xsl:function name="this:translate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point),this:translation3($x,$y,$z))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scale()
    : Scale the points by the given amount.
    : 
    : $points: points to scale
    : $scaling: how much to scale the point by (all dimensions)
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="scaling" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        $point=&gt;point:times($scaling)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scale()
    : Scale the points by the given amount.
    : 
    : $points: points to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:scaling2($sx, $sy))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scale3D()
    : Scale the points by the given amount.
    : 
    : $points: points to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $sz: how much to scale the point by in the z dimension
    :-->
   <xsl:function name="this:scale3D" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:scaling3($sx, $sy, $sz))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scale()
    : Scale the points by the given amount, but keep centered on a particular
    : point. Normally scaling will shift points closer to the origin. This
    : allows you do shift then closer to a different center of scaling instead,
    : so you can, for example, scale a circle about its own center, keeping it
    : in place.
    : 
    : $points: points to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $center: center of scaling
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:scaling2($center, $sx, $sy))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:scale3D()
    : Scale the points by the given amount, but keep centered on a particular
    : point. Normally scaling will shift points closer to the origin. This
    : allows you do shift then closer to a different center of scaling instead,
    : so you can, for example, scale a circle about its own center, keeping it
    : in place.
    : 
    : $points: points to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $sz: how much to scale the point by in the z dimension
    : $center: center of scaling
    :-->
   <xsl:function name="this:scale3D" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:scaling3($center, $sx, $sy, $sz))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotate()
    : Rotate the points by the given amount.
    : 
    : $points: points to rotate
    : $degrees: how many degrees to rotate the point
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:rotation2($degrees))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotate()
    : Rotate the points by the given amount.
    : 
    : $points: points to rotate
    : $roll-degrees: how many degrees to rotate the point forwards and back (x)
    : $pitch-degrees: how many degrees to rotate the point up and down (y)
    : $yaw-degrees: how many degrees to rotate the point left and right (z)
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:rotation3($roll-degrees, $pitch-degrees, $yaw-degrees))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotate()
    : Rotate the points by the given amount around a center of rotation.
    : Normally rotation is relative to the origin.
    : 
    : $points: points to rotate
    : $degrees: how many degrees to rotate the point 
    : $center: center of rotation
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:rotation2($center, $degrees))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:rotate()
    : Rotate the points by the given amount around a center of rotation.
    : Normally rotation is relative to the origin.
    : 
    : $points: points to rotate
    : $roll-degrees: how many degrees to rotate the point forwards and back (x)
    : $pitch-degrees: how many degrees to rotate the point up and down (y)
    : $yaw-degrees: how many degrees to rotate the point left and right (z)
    : $center: center of rotation
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:rotation3($center, $roll-degrees, $pitch-degrees, $yaw-degrees))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:shear()
    : Shear the points by the given amount.
    : You may wish to think of this as skewing angles instead: 
    : math:tan(util:radians($skew-degrees)) for $xy or $yx 
    : But note that tan approaches infinity near 90°/270°
    : 
    : $points: points to shear.
    : $xy: how much to shear x relative to y
    : $yx: how much to shear y relative to x
    :-->
   <xsl:function name="this:shear" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="xy" as="xs:double"/>
      <xsl:param name="yx" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:shearing2($xy, $yx))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:shear()
    : Shear the points by the given amount. 
    : You may wish to think of this as skewing angles instead: 
    : math:tan(math:radians($skew-degrees)) for $xy or $yx 
    : But note that tan approaches infinity near 90°/270°
    : 
    : $points: points to shear.
    : $xy: how much to shear x relative to y
    : $yx: how much to shear y relative to x
    : $xz: how much to shear x relative to z
    : $zx: how much to shear z relative to x
    : $yz: how much to shear y relative to z
    : $zy: how much to shear z relative to y
    :-->
   <xsl:function name="this:shear" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="xy" as="xs:double"/>
      <xsl:param name="yx" as="xs:double"/>
      <xsl:param name="xz" as="xs:double"/>
      <xsl:param name="zx" as="xs:double"/>
      <xsl:param name="yz" as="xs:double"/>
      <xsl:param name="zy" as="xs:double"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:shearing3($xy, $yx, $xz, $zx, $yz, $zy))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:reflect()
    : Reflect the points across the center.
    : 
    : $points: points to reflect.
    : $center: center of reflection.
    :-->
   <xsl:function name="this:reflect" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        if (point:max-dimension(($point,$center))=2) then (
          point:vector(
            this:affine2(point:pcoordinates($point), this:reflection2($center))
          )=&gt;point:with-properties($point)
        ) else (
          point:vector(
            this:affine3(point:coordinates($point), this:reflection3($center))
          )=&gt;point:with-properties($point)
        )
      )
      '/>
   </xsl:function>
   
   <!--
    : this:reflect()
    : Reflect the points across the line.
    : 
    : $points: points to reflect.
    : $start: start of reflection line
    : $end: end of reflection line
    :-->
   <xsl:function name="this:reflect" as="map(xs:string, item()*)">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine2(point:pcoordinates($point), this:reflection2($start,$end))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:perspective()
    : Perform a perspective transformation on the points, rotated relative to 
    : camera angles. Generally speaking camera location is at negative or zero
    : Z and the display surface is at some positive Z.
    : Not, strictly speaking, an affine transformation, because of the remapping
    : of the points back to 2D.
    : 
    : $points: points to compute perspective on.
    : $camera: location of the camera
    : $roll-degrees: camera angle (x)
    : $pitch-degrees: camera angle (y)
    : $yaw-degrees: camera angle (z)
    : $display: display surface relative to camera
    :-->
   <xsl:function name="this:perspective" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:homogeneous-to-cartesian(
            this:affine3(point:pcoordinates($point), this:perspective3($camera, $roll-degrees, $pitch-degrees, $yaw-degrees, $display))
          )
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
   
   <!--
    : this:perspective()
    : Perform a perspective transformation on the points. Generally speaking
    : camera location is at negative or zero Z and the display surface is at
    : some positive Z.
    : Not, strictly speaking, an affine transformation, because of the remapping
    : of the points back to 2D.
    : 
    : $points: points to compute perspective on.
    : $camera: location of the camera
    : $display: display surface relative to camera
    :-->
   <xsl:function name="this:perspective" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:perspective($points, $camera, 0, 0, 0, $display)
      '/>
   </xsl:function>
   
   <!--
    : this:isometric()
    : Perform a isometric projection of the points.
    : 
    : $points: points to compute isometric projection on
    : $positive: tilt up or down?
    :-->
   <xsl:function name="this:isometric" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="positive" as="xs:boolean"/>
      <xsl:sequence select='
      for $point in $points return (
        point:vector(
          this:affine3(point:pcoordinates($point), this:isometric3($positive))
        )=&gt;point:with-properties($point)
      )
      '/>
   </xsl:function>
</xsl:package>

<!--affine.xsl-->
