<?xml version="1.0" encoding="UTF-8"?>
<!--
  : Axis-aligned boxes and blocks: the basics
  : Split into separate library to avoid tangled imports
  :
  : Copyright© Mary Holstege 2020-2023
  : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/box"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric/box"
             exclude-result-prefixes="this xs math map array config errors util point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>

   <xsl:variable name="this:RESERVED" as="xs:string*">,
      <xsl:sequence select='("kind", "min-x", "min-y", "max-x", "max-y", "min-z", "max-z") '/>
   </xsl:variable>

<!--
  :======================================================================
  : Constructors
  :======================================================================
  :-->

   <!--
    : this:rectangle()
    : Construct an axis-aligned rectangle in x/y space.
    :   $x-min: Minimum on x-axis
    :   $y-min: Minimim on y-axis
    :   $x-max: Maximum on x-axis
    :   $y-max: Maximim on y-axis
    :-->
   <xsl:function name="this:rectangle" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:sequence select='
      (: Instead of tossing an error: rearrange the axes to meet the
       : constraints. This is so we do not just toss errors if some fool
       : does something like shear a rectangle :)
      if ($x-min &gt; $x-max) then (
        this:rectangle($x-max, $y-min, $x-min, $y-max)
      ) else if ($y-min &gt; $y-max) then (
        this:rectangle($x-min, $y-max, $x-max, $y-min)
      ) else (
        map {
          "kind": "box",
          "min-x": $x-min,
          "max-x": $x-max,
          "min-y": $y-min,
          "max-y": $y-max
        }
      )
     '/>
   </xsl:function>

   <!--
    : this:block()
    : block()
    : Construct an axis-aligned rectanglular block in x/y/z space.
    :   $x-min: Minimum on x-axis
    :   $y-min: Minimim on y-axis
    :   $z-min: Minimim on y-axis
    :   $x-max: Maximum on x-axis
    :   $y-max: Maximim on y-axis
    :   $z-max: Maximim on y-axis
    :-->
   <xsl:function name="this:block" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="z-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:param name="z-max" as="xs:double"/>
      <xsl:sequence select='
      (: Instead of tossing an error: rearrange the axes to meet the
       : constraints. This is so we do not just toss errors if some fool
       : does something like shear a rectangle :)
      if ($x-min &gt; $x-max) then (
        this:block($x-max, $y-min, $z-min, $x-min, $y-max, $z-max)
      ) else if ($y-min &gt; $y-max) then (
        this:block($x-min, $y-max, $z-min, $x-max, $y-min, $z-max)
      ) else (
        map {
          "kind": "block",
          "min-x": $x-min,
          "max-x": $x-max,
          "min-y": $y-min,
          "max-y": $y-max,
          "min-z": $z-min,
          "max-z": $z-max
        }
      )
     '/>
   </xsl:function>

   <!--
    : this:rectangle()
    :-->
   <xsl:function name="this:rectangle" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:rectangle(
        point:px($pt-min), point:py($pt-min), point:px($pt-max), point:py($pt-max)
      )
     '/>
   </xsl:function>

   <!--
    : this:block()
    :-->
   <xsl:function name="this:block" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:block(
        point:px($pt-min), point:py($pt-min), point:pz($pt-min),
        point:px($pt-max), point:py($pt-max), point:pz($pt-max)
      )
     '/>
   </xsl:function>

   <!--
    : this:box()
    :-->
   <xsl:function name="this:box" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:sequence select='
      this:rectangle($x-min, $y-min, $x-max, $y-max)
     '/>
   </xsl:function>

   <!--
    : this:box()
    :-->
   <xsl:function name="this:box" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:rectangle($pt-min, $pt-max)
     '/>
   </xsl:function>

<!--
  :======================================================================
  : Accessors
  :======================================================================
  :-->

   <!--
    : this:min-point()
    : Accessor for minimal point
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-point" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($box("kind")="block")
      then point:point(this:min-px($box), this:min-py($box), this:min-pz($box))
      else point:point(this:min-px($box), this:min-py($box))
     '/>
   </xsl:function>

   <!--
    : this:max-point()
    : Accessor for maximal point
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-point" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($box("kind")="block")
      then point:point(this:max-px($box), this:max-py($box), this:max-pz($box))
      else point:point(this:max-px($box), this:max-py($box))
     '/>
   </xsl:function>

   <!--
    : this:min-px()
    : Accessor for raw minimum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-px" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("min-x")
     '/>
   </xsl:function>

   <!--
    : this:max-px()
    : Accessor for raw maximum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-px" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("max-x")
      '/>
   </xsl:function>

   <!--
    : this:min-py()
    : Accessor for raw minimum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-py" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("min-y")
      '/>
   </xsl:function>

   <!--
    : this:max-py()
    : Accessor for raw maximum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-py" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("max-y")
     '/>
   </xsl:function>

   <!--
    : this:min-pz()
    : Accessor for raw minimum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-pz" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($box("min-z"),0)[1]
     '/>
   </xsl:function>

   <!--
    : this:max-pz()
    : Accessor for raw maximum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-pz" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($box("max-z"),0)[1]
     '/>
   </xsl:function>

   <!--
    : this:min-x()
    : Accessor for snapped minimum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-x" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even($box("min-x")) cast as xs:integer
     '/>
   </xsl:function>

   <!--
    : this:max-x()
    : Accessor for snapped maximum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-x" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even($box("max-x")) cast as xs:integer
     '/>
   </xsl:function>

   <!--
    : this:min-y()
    : Accessor for snapped minimum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-y" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even($box("min-y")) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:max-y()
    : Accessor for snapped maximum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-y" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even($box("max-y")) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:min-z()
    : Accessor for snapped minimum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-z" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even(($box("min-z"),0)[1]) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:max-z()
    : Accessor for snapped maximum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-z" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      round-half-to-even(($box("max-z"),0)[1]) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic geo functions
    :
    : $box: the rectangle
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("kind")
      '/>
   </xsl:function>

   <!--
    : this:width()
    : Width of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:width" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:max-px($box) - this:min-px($box)
      '/>
   </xsl:function>

   <!--
    : this:height()
    : Height of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:height" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:max-py($box) - this:min-py($box)
      '/>
   </xsl:function>

   <!--
    : this:depth()
    : Depth of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:depth" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:max-pz($box) - this:min-pz($box)
      '/>
   </xsl:function>

   <!--
    : this:edge()
    : The edge padding for a canvas space
    :-->
   <xsl:function name="this:edge" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
     0
     '/>
   </xsl:function>

   <!--
    : this:vertices()
    : Get all the region's points.
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($region("kind")="block") then (
        point:point(this:min-px($region), this:min-py($region), this:min-pz($region)),
        point:point(this:min-px($region), this:min-py($region), this:max-pz($region)),
        point:point(this:min-px($region), this:max-py($region), this:min-pz($region)),
        point:point(this:min-px($region), this:max-py($region), this:max-pz($region)),
        point:point(this:max-px($region), this:min-py($region), this:min-pz($region)),
        point:point(this:max-px($region), this:min-py($region), this:max-pz($region)),
        point:point(this:max-px($region), this:max-py($region), this:min-pz($region)),
        point:point(this:max-px($region), this:max-py($region), this:max-pz($region))
      ) else (
        point:point(this:min-px($region), this:min-py($region)),
        point:point(this:min-px($region), this:max-py($region)),
        point:point(this:max-px($region), this:max-py($region)),
        point:point(this:max-px($region), this:min-py($region))
      )
      '/>
   </xsl:function>

   <!--
    : this:points()
    : Get 2D points from the region
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($region("kind")="block") then (
        this:vertices($region)!point:as-dimension(.,2) (: will have duplicates! :)
      ) else (
        (: box vertices are 2D already :)
        this:vertices($region)
      )
      '/>
   </xsl:function>

   <!--
    : this:center()
   : Point in center of the box
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($box)="block") then (
        point:point(
          (this:min-px($box) + this:max-px($box)) div 2,
          (this:min-py($box) + this:max-py($box)) div 2,
          (this:min-pz($box) + this:max-pz($box)) div 2
        )
      ) else (
        point:point(
          (this:min-px($box) + this:max-px($box)) div 2,
          (this:min-py($box) + this:max-py($box)) div 2
        )
      )
     '/>
   </xsl:function>

   <!--
    : this:surface-area()
    : Full area of a box (all faces for block)
    :-->
   <xsl:function name="this:surface-area" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($box)="block") then (
        2 * this:width($box) * this:height($box) +
        2 * this:width($box) * this:depth($box) +
        2 * this:height($box) * this:depth($box)
      ) else (
        this:width($box) * this:height($box)
      )
     '/>
   </xsl:function>

<!--
  :======================================================================
  : Property management
  :======================================================================
  :-->

   <!--
    : this:property-map()
    : Return the annotation properties of the box as a map. Check whether this
    : is actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:exclude($region, $this:RESERVED)
     '/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the box as a map. 
    : Check whether this is actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($region=&gt;map:keys())[not(. = $this:RESERVED)]
     '/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the box with some new properties and return the new box.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($region, util:exclude($properties,$this:RESERVED))
     '/>
   </xsl:function>

<!--
  :======================================================================
  : Box operations
  :======================================================================
  :-->

 <!--
    : this:snap()
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $rectangle in $rectangles return (
        util:merge-into($rectangle,
          if ($rectangle("kind")="block") then (
            this:block(
              this:min-x($rectangle), this:min-y($rectangle), this:min-z($rectangle),
              this:max-x($rectangle), this:max-y($rectangle), this:max-z($rectangle)
            )
          )
          else (: box :) (
            this:rectangle(
              this:min-x($rectangle), this:min-y($rectangle),
              this:max-x($rectangle), this:max-y($rectangle)
            )
          )
        )
      )
     '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $rectangle in $rectangles return (
        util:merge-into($rectangle,
          if ($rectangle("kind") = "block") then (
            this:block(
              util:decimal(this:min-px($rectangle), $digits),
              util:decimal(this:min-py($rectangle), $digits),
              util:decimal(this:min-pz($rectangle), $digits),
              util:decimal(this:max-px($rectangle), $digits),
              util:decimal(this:max-py($rectangle), $digits),
              util:decimal(this:max-pz($rectangle), $digits)
            )
          )
          else (: box :)  (
            this:rectangle(
              util:decimal(this:min-px($rectangle), $digits),
              util:decimal(this:min-py($rectangle), $digits),
              util:decimal(this:max-px($rectangle), $digits),
              util:decimal(this:max-py($rectangle), $digits)
            )
          )
        )
      )
     '/>
   </xsl:function>

   <!--
    : this:quote()
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $rectangle in $rectangles return (
          if ($rectangle("kind") = "block") then (
            "[["||this:min-px($rectangle)||","||this:min-py($rectangle)||","||this:min-pz($rectangle)||","||
                  this:max-px($rectangle)||","||this:max-py($rectangle)||","||this:max-pz($rectangle)||"]]"
          )
          else (: box :) (
            "["||this:min-px($rectangle)||","||this:min-py($rectangle)||","||
                 this:max-px($rectangle)||","||this:max-py($rectangle)||"]"
          )
        )
        ,
        " "
      )
     '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for rectangles, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one rectangle
    : $other: the rectangle to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:min-px($this)=this:min-px($other) and
      this:min-py($this)=this:min-py($other) and
      this:max-px($this)=this:max-px($other) and
      this:max-py($this)=this:max-py($other) and
      ($this("kind")="box" or (
        this:min-pz($this)=this:min-pz($other) and
        this:max-pz($this)=this:max-pz($other)
      ))
     '/>
   </xsl:function>

   <!--
    : this:mutate()
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $region in $regions return (
        util:merge-into($region,
          if ($region("kind") = "block") then (
            this:block(
              $mutate(this:min-point($region)),
              $mutate(this:max-point($region))
            )
          )
          else (: box :) (
            this:rectangle(
              $mutate(this:min-point($region)),
              $mutate(this:max-point($region))
            )
          )
        )
      )
     '/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      util:twixt(point:px($point), this:min-px($box) - $tolerance, this:max-px($box) + $tolerance) and
      util:twixt(point:py($point), this:min-py($box) - $tolerance, this:max-py($box) + $tolerance) and
      ($box("kind")="box" or
        util:twixt(point:pz($point), this:min-pz($box) - $tolerance, this:max-pz($box) + $tolerance)
      )
     '/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:twixt(point:px($point), this:min-px($box), this:max-px($box)) and
      util:twixt(point:py($point), this:min-py($box), this:max-py($box)) and
      ($box("kind")="box" or
       util:twixt(point:pz($point), this:min-pz($box), this:max-pz($box))
      )
     '/>
   </xsl:function>

   <!--
    : this:intersects-box()
    : Box intersects some other region kind
    :-->
   <xsl:function name="this:intersects-box" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: No overlap is maximum of box &lt; minimum of other or vice versa :)
      (this:max-px($box) &gt;= this:min-px($other)) and
      (this:min-px($box) &lt;= this:max-px($other)) and
      (this:max-py($box) &gt;= this:min-py($other)) and
      (this:min-py($box) &lt;= this:max-py($other)) and
      ($box("kind")="box" or (
        (this:max-pz($box) &gt;= this:min-pz($other)) and
        (this:min-pz($box) &lt;= this:max-pz($other))
      ))
     '/>
   </xsl:function>

   <!--
    : this:intersects-box()
    :-->
   <xsl:function name="this:intersects-box" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      (: No overlap is maximum of box &lt; minimum of other or vice versa :)
      (this:max-px($box) &gt;= this:min-px($other) - $tolerance) and
      (this:min-px($box) &lt;= this:max-px($other) + $tolerance) and
      (this:max-py($box) &gt;= this:min-py($other) - $tolerance) and
      (this:min-py($box) &lt;= this:max-py($other) + $tolerance) and
      ($box("kind")="box" or (
        (this:max-pz($box) &gt;= this:min-pz($other) - $tolerance) and
        (this:min-pz($box) &lt;= this:max-pz($other) + $tolerance)
      ))
     '/>
   </xsl:function>
</xsl:package>
<!--box.xsl-->
