<?xml version="1.0" encoding="UTF-8"?>
<!--
  : Edges: including arcs, quads, and cubics
  : Edges are decorated with properties which art infrastructure may care
  : about for rendering purposes.
  : 
  : Properties of note:
  :   r: radius of arc for arc edges; default is none
  :   width: width for stroke; default=d
  :   label: label for this edge
  :   class, colour, opacity: direct styling overrides
  :
  : Copyright© Mary Holstege 2020-2023
  : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/edge"
             xmlns:affine="http://mathling.com/geometric/affine"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/box"
             xmlns:config="http://mathling.com/core/config"
             xmlns:ellipse="http://mathling.com/geometric/ellipse"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:roots="http://mathling.com/core/roots"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/geometric/edge"
             exclude-result-prefixes="this xs math map array config errors util v roots point affine box ellipse">

  <xsl:expose component="variable" names="this:*" visibility="public"/>
  <xsl:expose component="function" names="this:*" visibility="public"/>

  <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/roots" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/affine" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/box" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>

  <xsl:variable name="this:precision" as="xs:integer">
    <xsl:sequence select='math:log10(1 div $config:ε) cast as xs:integer'/>
  </xsl:variable>

  <xsl:variable name="this:EDGE-RESERVED" as="xs:string*">
    <xsl:sequence select='("kind", "u", "v", "d")'/>
  </xsl:variable>


  <!--
    : this:edge()
    : Make an edge
    :
    : $from: starting point
    : $to: ending point
    : $d: weight of the edge (used by spanning graph as distance)
    : $properties: additional edge properties
    :-->
  <xsl:function name="this:edge" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:param name="properties" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "edge",
          "u": $from,
          "v": $to,
          "d": $d
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:edge()
    : Make an edge
    : 
    : $from: starting point
    : $to: ending point
    : $d: the weight of the edge
    :-->
  <xsl:function name="this:edge" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:sequence select='
      map {
        "kind": "edge",
        "u": $from,
        "v": $to,
        "d": $d
      }
    '/>
  </xsl:function>

  <!--
    : this:edge()
    : Make an edge
    : 
    : $from: starting point
    : $to: ending point
    :-->
  <xsl:function name="this:edge" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:edge($from, $to, 0)
    '/>
  </xsl:function>

  <!--
    : this:kind()
    :-->
  <xsl:function name="this:kind" as="xs:string">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $edge("kind")
    '/>
  </xsl:function>

  <!--
    : this:start()
    : Accessor for starting point of edge
    :
    : $edge: The edge
    :-->
  <xsl:function name="this:start" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then $edge("u")
      else if ($kind = "quad") then $edge("u")
      else if ($kind = "cubic") then $edge("u")
      else if ($kind = "arc") then (
        let $start := $edge("start")
        return
          if (exists($start)) then $start
          else (
            let $circle := $edge("circle") return 
            let $center := ellipse:center($circle) return
            let $r := ellipse:radius($circle) return
            let $from := util:radians($edge("from-angle"))
            return (
              point:point(
                point:px($center)+math:cos($from)*$r,
                point:py($center)+math:sin($from)*$r
              )
            )
          )
        )
      else if ($kind = "ellipse-arc") then $edge("start")
      else errors:error("GEOM-BADREGION", ($edge, "start"))
    '/>
  </xsl:function>

  <!--
    : this:start()
    : Settor for starting point of edge
    :
    : $edge: The edge
    : $start: Starting point of edge
    :-->
  <xsl:function name="this:start" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="start" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then $edge=>map:put("u", $start)
      else if ($kind = "quad") then $edge=>map:put("u", $start)
      else if ($kind = "cubic") then $edge=>map:put("u", $start)
      else if ($kind = "arc") then this:as-point-arc($edge)=>map:put("start", $start)
      else if ($kind = "ellipse-arc") then $edge=>map:put("start", $start)
      else errors:error("GEOM-BADREGION", ($edge, "start"))
    '/>
  </xsl:function>

  <!--
    : this:end()
    : Accessor for ending point of edge
    :
    : $edge: The edge
    :-->
  <xsl:function name="this:end" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then $edge("v")
      else if ($kind = "quad") then $edge("v")
      else if ($kind = "cubic") then $edge("v")
      else if ($kind = "arc") then (
        let $end := $edge("end")
        return
          if (exists($end)) then $end
          else (
            let $circle := $edge("circle") return
            let $center := ellipse:center($circle) return
            let $r := ellipse:radius($circle) return
            let $to := util:radians($edge("to-angle"))
            return 
              point:point(
                point:px($center)+math:cos($to)*$r,
                point:py($center)+math:sin($to)*$r
              )
          )
      )
      else if ($kind = "ellipse-arc") then $edge("end")
      else errors:error("GEOM-BADREGION", ($edge, "end"))
    '/>
  </xsl:function>

  <!--
    : this:weight()
    : Accessor for the weight of an edge
    :
    : $edge: The edge
    :-->
  <xsl:function name="this:weight" as="xs:double">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($edge) = ("edge","cubic","quad"))
      then ($edge("d"),0)[1]
      else 0
    '/>
  </xsl:function>

  <!--
    : $this:ARC-RESERVED
    :-->
  <xsl:variable name="this:ARC-RESERVED" as="xs:string*">
    <xsl:sequence select='
      ("kind", "circle", "start", "end", "from-angle", "to-angle", "flipped", "large")
    '/>
  </xsl:variable>

  <!--
    : this:arc()
    : Construct an arc (e.g. hyperbolic edge)
    : Note: calculations on $start and $end need to be correct: 
    : use hyperbolic:arc-through() to get them right; this is really an
    : internal contructor function. Ad hoc construction of arc should use 
    : arc-by-angle().
    :
    : $center: center of curvature
    : $radius: radius of circle
    : $start: starting point
    : $end: ending point
    : $flipped: indicates sweep
    : $large: indicated large arc
    :-->
  <xsl:function name="this:arc" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="start" as="map(xs:string, item()*)"/>
    <xsl:param name="end" as="map(xs:string, item()*)"/>
    <xsl:param name="flipped" as="xs:boolean"/>
    <xsl:param name="large" as="xs:boolean"/>
    <xsl:sequence select='
      map {
        "kind": "arc",
        "circle": ellipse:circle($center, $radius),
        "start": $start,
        "end": $end,
        "flipped": $flipped,
        "large": $large
      }
    '/>
  </xsl:function>

  <!--
    : this:arc()
    :-->
  <xsl:function name="this:arc" as="map(xs:string, item()*)*">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="start" as="map(xs:string, item()*)"/>
    <xsl:param name="end" as="map(xs:string, item()*)"/>
    <xsl:param name="flipped" as="xs:boolean"/>
    <xsl:sequence select='
      this:arc($center, $radius, $start, $end, $flipped, false())
    '/>
  </xsl:function>

  <!--
    : this:arc()
    :-->
  <xsl:function name="this:arc" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="start" as="map(xs:string, item()*)"/>
    <xsl:param name="end" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:arc($center, $radius, $start, $end, false(), false())
    '/>
  </xsl:function>

  <!--
    : this:arc-by-angle()
    : Construct an arc
    :
    : $center: center of curvature
    : $radius: radius of circle
    : $θ: starting angle (degrees)
    : $θ2: ending angle (degrees)
    : $flipped: indicates sweep
    : $large: indicates large arc
    :-->
  <xsl:function name="this:arc-by-angle" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="θ" as="xs:double"/>
    <xsl:param name="θ2" as="xs:double"/>
    <xsl:param name="flipped" as="xs:boolean"/>
    <xsl:param name="large" as="xs:boolean"/>
    <xsl:sequence select='
      map {
        "kind": "arc",
        "circle": ellipse:circle($center, $radius),
        "from-angle": $θ,
        "to-angle": $θ2,
        "flipped": $flipped,
        "large": $large
      }
    '/>
  </xsl:function>

  <!--
    : this:arc-by-angle()
    :-->
  <xsl:function name="this:arc-by-angle" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="θ" as="xs:double"/>
    <xsl:param name="θ2" as="xs:double"/>
    <xsl:param name="flipped" as="xs:boolean"/>
    <xsl:sequence select='
      this:arc-by-angle($center, $radius, $θ, $θ2, $flipped, false())
    '/>
  </xsl:function>

  <!--
    : this:arc-by-angle()
    :-->
  <xsl:function name="this:arc-by-angle" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="radius" as="xs:double"/>
    <xsl:param name="θ" as="xs:double"/>
    <xsl:param name="θ2" as="xs:double"/>
    <xsl:sequence select='
      this:arc-by-angle($center, $radius, $θ, $θ2, false(), false())
    '/>
  </xsl:function>

  <!--
    : this:arc-circle()
    : Center and radius of curvature for this arc
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-circle" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $arc("circle")
    '/>
  </xsl:function>

  <!--
    : this:arc-center()
    : Center of this arc
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-center" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($arc)="ellipse-arc")
      then $arc("ellipse")=>ellipse:center()
      else $arc("circle")=>ellipse:center()
    '/>
  </xsl:function>

  <!--
    : this:arc-radius()
    : Radius of curvature of this arc
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-radius" as="xs:double">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $arc("circle")=>ellipse:radius()
    '/>
  </xsl:function>

  <!--
    : this:arc-angles()
    : Starting and ending angles for the arc (degrees), for arc-by-angle arcs
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-angles" as="xs:double*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      ($arc("from-angle"),$arc("to-angle"))
    '/>
  </xsl:function>

  <!--
    : this:arc-ends()
    : Starting and ending points for the arc, for arc-by-point arcs
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-ends" as="map(xs:string, item()*)*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      ($arc("start"),$arc("end"))
    '/>
  </xsl:function>

  <!--
    : this:arc-flipped()
    : Sweep for the arc
    :   $arc: The arc
    :-->
  <xsl:function name="this:arc-flipped" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: Default is false :)
      if ($arc("flipped")) then true() else false()
    '/>
  </xsl:function>

  <!--
    : this:arc-large()
    :-->
  <xsl:function name="this:arc-large" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: Default is false :)
      if ($arc("large")) then true() else false()
    '/>
  </xsl:function>

  <!--
    : this:as-angle-arc()
    : Shift arc edge from the end-point representation to the angle representation.
    : 
    : $arc: input arc
    :-->
  <xsl:function name="this:as-angle-arc" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $points := $arc=>this:arc-ends()
      return (
        if (exists($points)) then (
          let $circle := $arc=>this:arc-circle() return
          let $center := ellipse:center($circle) return
          let $α := this:angle($center, $points[1]) return
          let $ω := this:angle($center, $points[2])
          return
            util:merge-into(
              this:property-map($arc),
              this:arc-by-angle(
                $center,
                ellipse:radius($circle),
                $α,
                if ($ω - $α > 360) then 360 - $ω else $ω,
                $arc=>this:arc-flipped(),
                $arc=>this:arc-large()
              )
            )
        ) else (
          $arc
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:as-point-arc()
    : Shift arc edge from the angle representation to the end-point representation.
    : 
    : $arc: input arc
    :-->
  <xsl:function name="this:as-point-arc" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $points := $arc=>this:arc-ends()
      return (
        if (exists($points)) then (
          $arc
        ) else (
          let $circle := $arc=>this:arc-circle()
          return (
            util:merge-into(
              this:property-map($arc),
              this:arc(
                ellipse:center($circle),
                ellipse:radius($circle),
                $arc=>this:start(),
                $arc=>this:end(),
                $arc=>this:arc-flipped(),
                $arc=>this:arc-large()
              )
            )
          )
        )
      )
    '/>
  </xsl:function>

  <!-- ellipse arc: right now all you can do with this draw it -->

  <xsl:variable name="this:ELLIPSE-ARC-RESERVED" as="xs:string*">
    <xsl:sequence select='
      ("kind", "ellipse", "start", "end", "flipped", "large")
    '/>
  </xsl:variable>

  <!--
    : this:ellipse-arc()
    :-->
  <xsl:function name="this:ellipse-arc" as="item()*">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="a" as="xs:double"/>
    <xsl:param name="b" as="xs:double"/>
    <xsl:param name="rotation" as="xs:double"/>
    <xsl:param name="start" as="map(xs:string, item()*)"/>
    <xsl:param name="end" as="map(xs:string, item()*)"/>
    <xsl:param name="flipped" as="xs:boolean"/>
    <xsl:param name="large" as="xs:boolean"/>
    <xsl:sequence select='
      map {
        "kind": "ellipse-arc",
        "ellipse": ellipse:ellipse($center, $a, $b, $rotation),
        "start": $start,
        "end": $end,
        "flipped": $flipped,
        "large": $large
      }
    '/>
  </xsl:function>

  <!--
    : this:arc-ellipse()
    :-->
  <xsl:function name="this:arc-ellipse" as="map(xs:string, item()*)">
    <xsl:param name="ellipse-arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      $ellipse-arc("ellipse")
    '/>
  </xsl:function>

  <xsl:variable name="this:QUAD-RESERVED" as="xs:string*">
    <xsl:sequence select='
      ($this:EDGE-RESERVED, "c1")'/>
  </xsl:variable>

  <!--
    : this:quad()
    : Construct a quadratic edge.
    : 
    : $from: starting point
    : $to: ending point
    : $control: control point
    : $d: edge weight
    : $properties: additional properties
    :-->
  <xsl:function name="this:quad" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:param name="properties" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "quad",
          "u": $from,
          "v": $to,
          "c1": $control,
          "d": $d
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:quad()
    :-->
  <xsl:function name="this:quad" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:sequence select='
      map {
        "kind": "quad",
        "u": $from,
        "v": $to,
        "c1": $control,
        "d": $d
      }
    '/>
  </xsl:function>

  <!--
    : this:quad()
    :-->
  <xsl:function name="this:quad" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:quad($from, $to, $control, 0)
    '/>
  </xsl:function>

  <!--
    : this:controls()
    : Get control points from a quadratic or cubic edge.
    :-->
  <xsl:function name="this:controls" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "quad") then $edge("c1")
      else if ($kind = "cubic") then ($edge("c1"),$edge("c2"))
      else errors:error("GEOM-BADREGION", ($edge, "controls"))
    '/>
  </xsl:function>

  <xsl:variable name="this:CUBIC-RESERVED" as="xs:string*">
    <xsl:sequence select='
      ($this:EDGE-RESERVED, "c1", "c2")'/>
  </xsl:variable>

  <!--
    : this:cubic()
    : Construct a cubic edge.
    : 
    : $from: starting point
    : $to: ending point
    : $control1: first control point
    : $control2: second control point
    : $d: edge weight
    : $properties: additional properties
    :-->
  <xsl:function name="this:cubic" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control1" as="map(xs:string, item()*)"/>
    <xsl:param name="control2" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:param name="properties" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "cubic",
          "u": $from,
          "v": $to,
          "c1": $control1,
          "c2": $control2,
          "d": $d
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:cubic()
    :-->
  <xsl:function name="this:cubic" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control1" as="map(xs:string, item()*)"/>
    <xsl:param name="control2" as="map(xs:string, item()*)"/>
    <xsl:param name="d" as="xs:double"/>
    <xsl:sequence select='
      map {
        "kind": "cubic",
        "u": $from,
        "v": $to,
        "c1": $control1,
        "c2": $control2,
        "d": $d
      }
    '/>
  </xsl:function>

  <!--
    : this:cubic()
    :-->
  <xsl:function name="this:cubic" as="map(xs:string, item()*)">
    <xsl:param name="from" as="map(xs:string, item()*)"/>
    <xsl:param name="to" as="map(xs:string, item()*)"/>
    <xsl:param name="control1" as="map(xs:string, item()*)"/>
    <xsl:param name="control2" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:cubic($from, $to, $control1, $control2, 0)
    '/>
  </xsl:function>

  <!--
    : this:points()
    : Get 2D points from the region
    :-->
  <xsl:function name="this:points" as="map(xs:string, item()*)*">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:vertices($region)!point:as-dimension(.,2)
    '/>
  </xsl:function>

  <!--
    : this:vertices()
    : Get all the region's points.
    :-->
  <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "edge") then ($region=>this:start(), $region=>this:end())
      else if ($kind = "arc") then ($region=>this:start(), $region=>this:end())
      else if ($kind = "quad") then ($region=>this:start(), $region=>this:end())
      else if ($kind = "cubic") then ($region=>this:start(), $region=>this:end())
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
  </xsl:function>

  <!--
    : this:to-edges()
    : Convert a sequence of points into a sequence of edges.
    :
    : $points: the sequence of points
    :-->
  <xsl:function name="this:to-edges" as="map(xs:string, item()*)*">
    <xsl:param name="points" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
      for $i in 2 to count($points) return this:edge($points[$i - 1], $points[$i])
      (: for $pt at $i in tail($points) return this:edge($points[$i], $pt) :)
    '/>
  </xsl:function>

  <!--
    : this:to-edges()
    : Convert a sequence of points into a sequence of edges.
    :
    : $points: the sequence of points
    : $properties: edge properties
    :-->
  <xsl:function name="this:to-edges" as="map(xs:string, item()*)*">
    <xsl:param name="points" as="item()*"/>
    <xsl:param name="properties" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $d := ($properties("d"),0)[1] (: Special case for d :) return
      for $i in 2 to count($points) return this:edge($points[$i - 1], $points[$i], $d, $properties) 
      (: for $pt at $i in tail($points)
      return this:edge($points[$i], $pt, $d, $properties) :)
    '/>
  </xsl:function>

  <!--
    :======================================================================
    : Property management
    :======================================================================
    :-->

  <!--
    : this:property-map()
    : Return the annotation properties of the edge as a map. Check whether this
    : is actually an edge.
    : 
    : $region: the region
    :-->
  <xsl:function name="this:property-map" as="map(xs:string, item()*)">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "edge") then util:exclude($region,$this:EDGE-RESERVED)
      else if ($kind = "arc") then util:exclude($region,$this:ARC-RESERVED)
      else if ($kind = "quad") then util:exclude($region, $this:QUAD-RESERVED)
      else if ($kind = "cubic") then util:exclude($region, $this:CUBIC-RESERVED)
      else if ($kind = "ellipse-arc") then util:exclude($region, $this:ELLIPSE-ARC-RESERVED)
      else map {}
    '/>
  </xsl:function>

  <!--
    : this:properties()
    : Return the names of the annotation properties of the edge.
    : Check whether this is actually an edge
    : 
    : $region: the region
    :-->
  <xsl:function name="this:properties" as="xs:string*">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "edge") then ($region=>map:keys())[not(. = $this:EDGE-RESERVED)]
      else if ($kind = "arc") then ($region=>map:keys())[not(. = $this:ARC-RESERVED)]
      else if ($kind = "quad") then ($region=>map:keys())[not(. = $this:QUAD-RESERVED)]
      else if ($kind = "cubic") then ($region=>map:keys())[not(. = $this:CUBIC-RESERVED)]
      else if ($kind = "ellipse-arc") then ($region=>map:keys())[not(. = $this:ELLIPSE-ARC-RESERVED)]
      else ()
    '/>
  </xsl:function>

  <!--
    : this:with-properties()
    : Annotate the edge with some new properties and return the new edge.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually an edge.
    : 
    : $region: the region
    :-->
  <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:param name="properties" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "edge") then util:merge-into($region, util:exclude($properties,$this:EDGE-RESERVED))
      else if ($kind = "arc") then util:merge-into($region, util:exclude($properties,$this:ARC-RESERVED))
      else if ($kind = "quad") then util:merge-into($region, util:exclude($properties, $this:QUAD-RESERVED))
      else if ($kind = "cubic") then util:merge-into($region, util:exclude($properties, $this:CUBIC-RESERVED))
      else if ($kind = "ellipse-arc") then util:merge-into($region, util:exclude($properties,$this:ELLIPSE-ARC-RESERVED))
      else errors:error("GEOM-BADREGION", ($region, "with-properties"))
    '/>
  </xsl:function>

  <!--
    :======================================================================
    : Operations
    :======================================================================:
    :-->

  <!--
    : this:snap()
    : Snap the coordinates of the points, returning the edges with snapped
    : (i.e. integer) points coordinates.
    :
    : $edges the edges
    :-->
  <xsl:function name="this:snap" as="map(xs:string, item()*)*">
    <xsl:param name="edges" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
      for $edge in $edges return let $kind := this:kind($edge) return
      if ($kind = "edge") then
        this:edge(
          point:snap(this:start($edge)),
          point:snap(this:end($edge)),
          this:weight($edge),
          $edge
        )
      else if ($kind = "quad") then
        this:quad(
          point:snap(this:start($edge)),
          point:snap(this:end($edge)),
          point:snap(this:controls($edge)[1]),
          this:weight($edge),
          $edge
        )
      else if ($kind = "cubic") then
        let $controls := this:controls($edge)
        return
          this:cubic(
            point:snap(this:start($edge)),
            point:snap(this:end($edge)),
            point:snap($controls[1]),
            point:snap($controls[2]),
            this:weight($edge),
            $edge
          )
      else if ($kind = "arc") then
        let $circle := $edge=>this:arc-circle() return
        let $points := $edge=>this:arc-ends() return
        let $angles := $edge=>this:arc-angles()
        return
          if (empty($points)) then (
            util:assert(exists($angles),"Arc must have points or angles"),
            $edge=>
              map:put("circle", ellipse:snap($circle))
          ) else (
            $edge=>
               map:put("circle", ellipse:snap($circle))=>
               map:put("start", point:snap($points[1]))=>
               map:put("end", point:snap($points[2]))
          )
      else if ($kind = "ellipse-arc") then
        let $ellipse := $edge=>this:arc-ellipse() return
        let $points := $edge=>this:arc-ends()
        return
          $edge=>
             map:put("ellipse", ellipse:snap($ellipse))=>
             map:put("start", point:snap($points[1]))=>
             map:put("end", point:snap($points[2]))
      else $edge
    '/>
  </xsl:function>

  <!--
    : this:decimal()
    : Perform decimal rounding on all the point coordinates (see util:decimal).
    :
    : $edges: the edges to round
    : $digits: how many digits after the decimal point to keep
    :-->
  <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
    <xsl:param name="edges" as="map(xs:string, item()*)*"/>
    <xsl:param name="digits" as="xs:integer"/>
    <xsl:sequence select='
      for $edge in $edges return let $kind := this:kind($edge) return
      if ($kind = "edge") then
        this:edge(
          point:decimal(this:start($edge), $digits),
          point:decimal(this:end($edge), $digits),
          this:weight($edge),
          $edge
        )
      else if ($kind = "quad") then
        this:quad(
          point:decimal(this:start($edge), $digits),
          point:decimal(this:end($edge), $digits),
          point:decimal(this:controls($edge)[1], $digits),
          this:weight($edge),
          $edge
        )
      else if ($kind = "cubic") then
        let $controls := this:controls($edge)
        return
          this:cubic(
            point:decimal(this:start($edge), $digits),
            point:decimal(this:end($edge), $digits),
            point:decimal($controls[1], $digits),
            point:decimal($controls[2], $digits),
            this:weight($edge),
            $edge
          )
      else if ($kind = "arc") then
        let $circle := $edge=>this:arc-circle() return
        let $points := $edge=>this:arc-ends() return
        let $angles := $edge=>this:arc-angles()
        return
          if (empty($points)) then (
            util:assert(exists($angles),"Arc must have points or angles"),
            $edge=>
              map:put("circle", ellipse:decimal($circle, $digits))
          ) else (
            $edge=>
               map:put("circle", ellipse:decimal($circle, $digits))=>
               map:put("start", point:decimal($points[1], $digits))=>
               map:put("end", point:decimal($points[2], $digits))
          )
      else if ($kind = "ellipse-arc") then
        let $ellipse := $edge=>this:arc-ellipse() return
        let $points := $edge=>this:arc-ends()
        return
          $edge=>
             map:put("ellipse", ellipse:decimal($ellipse, $digits))=>
             map:put("start", point:decimal($points[1], $digits))=>
             map:put("end", point:decimal($points[2], $digits))
      else $edge
    '/>
  </xsl:function>

  <!--
    : this:quote()
    : Return a string value for the edges, suitable for debugging.
    : 
    : $edges: the edge sequence to quote
    :-->
  <xsl:function name="this:quote" as="xs:string">
    <xsl:param name="edges" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
      string-join(
        for $edge in $edges return let $kind := this:kind($edge) return
        if ($kind = "edge") then
          point:quote($edge("u"))||":"||point:quote($edge("v"))
        else if ($kind = "quad") then
          point:quote($edge("u"))||":"||point:quote($edge("v"))||"&lt;"||point:quote($edge("c1"))||">"
        else if ($kind = "cubic") then
          point:quote($edge("u"))||":"||point:quote($edge("v"))||"&lt;"||point:quote($edge("c1"))||";"||point:quote($edge("c2"))||">"
        else if ($kind = "arc") then
          "("||
            ellipse:quote($edge("circle"))||
            "&lt;"||
              point:quote($edge("start"))||
              $edge("from-angle")||":"||
              point:quote($edge("end"))||
              $edge("to-angle")||
            ">"||
            (if ($edge("large")) then "+l1" else "+l0")||
            (if ($edge("flipped")) then "+f1" else "+f0")||
          ")"
        else if ($kind = "ellipse-arc") then
          "("||
            ellipse:quote($edge("ellipse"))||
            "&lt;"||
              point:quote($edge("start"))||
              point:quote($edge("end"))||
            ">"||
            (if ($edge("large")) then "+l1" else "+l0")||
            (if ($edge("flipped")) then "+f1" else "+f0")||
          ")"
        else errors:quote($edge)
        ,
        " "
      )
    '/>
  </xsl:function>

  <!--
    : this:same()
    : Equality comparison for edges, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one edge
    : $other: the edge to compare it to
    :-->
  <xsl:function name="this:same" as="xs:boolean">
    <xsl:param name="this" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $this-kind := this:kind($this) return
      let $other-kind := this:kind($other)
      return
      (
        ($this-kind=$other-kind) and (
        if ($this-kind = "arc") then
          (this:arc-large($this) = this:arc-large($other)) and
          (this:arc-flipped($this) = this:arc-flipped($other)) and
          ellipse:same($this("circle"),$other("circle")) and
          (
            if (empty($this("start"))) then empty($other("start"))
            else if (empty($other("start"))) then false()
            else point:same($this("start"), $other("start"))
          ) and
          (
            if (empty($this("end"))) then empty($other("end"))
            else if (empty($other("end"))) then false()
            else point:same($this("end"), $other("end"))
          ) and (
            if (empty($this("from-angle"))) then empty($other("from-angle"))
            else if (empty($other("from-angle"))) then false()
            else $this("from-angle")=$other("from-angle")
          ) and (
            if (empty($this("to-angle"))) then empty($other("to-angle"))
            else if (empty($other("to-angle"))) then false()
            else $this("to-angle")=$other("to-angle")
          )
        else if ($this-kind = "edge") then
          point:same($this("u"),$other("u")) and
          point:same($this("v"),$other("v"))
        else if ($this-kind = "quad") then
          point:same($this("u"),$other("u")) and
          point:same($this("v"),$other("v")) and
          point:same($this("c1"),$other("c1"))
        else if ($this-kind = "cubic") then
          point:same($this("u"),$other("u")) and
          point:same($this("v"),$other("v")) and
          point:same($this("c1"),$other("c1")) and
          point:same($this("c2"),$other("c2"))
        else if ($this-kind = "ellipse-arc") then
          (this:arc-large($this) = this:arc-large($other)) and
          (this:arc-flipped($this) = this:arc-flipped($other)) and
          ellipse:same($this("ellipse"),$other("ellipse")) and
          point:same($this("start"), $other("start")) and 
          point:same($this("end"), $other("end"))
        else deep-equal($this,$other)
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:mutate()
    : Run a function over a sequence of edges to produce a new sequence of
    : edges. The function maps points to points.
    : 
    : $regions: input sequence of edges 
    : $mutate: function that takes a point as an argument and returns a new point
    :-->
  <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
    <xsl:param name="regions" as="map(xs:string, item()*)*"/>
    <xsl:param name="mutate"
               as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
    <xsl:sequence select='
      for $edge in $regions 
      return (
        let $kind := this:kind($edge) return
        if ($kind = "edge") then (
          this:edge(
            $mutate($edge=>this:start()),
            $mutate($edge=>this:end()),
            $edge=>this:weight(),
            $edge
          )
        )
        else if ($kind = "cubic") then (
          let $controls := ($edge=>this:controls())!$mutate(.)
          return
            this:cubic(
              $mutate($edge=>this:start()),
              $mutate($edge=>this:end()),
              $controls[1],
              $controls[2],
              $edge=>this:weight(),
              $edge
            )
        )
        else if ($kind = "quad") then (
          let $control := $mutate(this:controls($edge)[1])
          return
            this:quad(
              $mutate($edge=>this:start()),
              $mutate($edge=>this:end()),
              $control,
              $edge=>this:weight(),
              $edge
            )
        )
        else if ($kind = "arc") then (
          let $points := $edge=>this:arc-ends() return
          let $angles := $edge=>this:arc-angles()
          return (
            if (exists($points)) then (
              this:arc(
                $mutate($edge=>this:arc-center()),
                $edge=>this:arc-radius(),
                $mutate($edge=>this:start()),
                $mutate($edge=>this:end()),
                $edge=>this:arc-flipped(),
                $edge=>this:arc-large()
              )
            ) else (
              this:arc-by-angle(
                $mutate($edge=>this:arc-center()),
                $edge=>this:arc-radius(),
                $angles[1],
                $angles[2],
                $edge=>this:arc-flipped(),
                $edge=>this:arc-large()
              )
            )
          )=>this:with-properties($edge)
        )
        else $edge
      )
    '/>
  </xsl:function>

  <!--
    : this:reverse()
    : Reverse the direction of the edge.
    :-->
  <xsl:function name="this:reverse" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then
        this:edge(
          this:end($edge),
          this:start($edge),
          this:weight($edge),
          $edge
        )
      else if ($kind = "quad") then
        this:quad(
          this:end($edge),
          this:start($edge),
          this:controls($edge)[1],
          this:weight($edge),
          $edge
        )
      else if ($kind = "cubic") then
        this:cubic(
          this:end($edge),
          this:start($edge),
          this:controls($edge)[2],
          this:controls($edge)[1],
          this:weight($edge),
          $edge
        )
      else if ($kind = "arc") then (
        let $circle := $edge=>this:arc-circle() return
        let $points := $edge=>this:arc-ends() return
        let $angles := $edge=>this:arc-angles()
        return (
          if (exists($points)) then (
            this:arc(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $points[2],
              $points[1],
              not($edge=>this:arc-flipped()),
              $edge=>this:arc-large()
            )
          ) else (
            this:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[2],
              $angles[1],
              not($edge=>this:arc-flipped()),
              $edge=>this:arc-large()
            )
          )
        )
      )
      else $edge
    '/>
  </xsl:function>

  <!--:======================================================================
    : Edge calculations
    :======================================================================
    :-->

  <!--
    : this:angle()
    : Bearing of the edge between start and end points.
    : 
    : $edge: the edge
    :-->
  <xsl:function name="this:angle" as="xs:double">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:angle(this:start($edge), this:end($edge))
    '/>
  </xsl:function>

  <!--
    : this:angle()
    : Compute the angle (azimuth) from one point to the next, in degrees
    : Return 0 if points are the same
    :
    : $last: previous point; use (0,0) if no previous
    : $curr: current point
    :-->
  <xsl:function name="this:angle" as="xs:double">
    <xsl:param name="last" as="map(xs:string, item()*)?"/>
    <xsl:param name="curr" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $this := $curr=>point:as-dimension(2) return
      let $prev := point:as-dimension(($last,$point:ORIGIN)[1],2)
      return
        if (point:same($prev,$this) or (point:distance($prev,$this) &lt; $config:ε)) then 0
        else if (point:px($this)=point:px($prev)) then (
          if (point:py($prev) > point:py($this))
          then 270 (: remapped -90 :)
          else 90
        ) else (
          util:remap-degrees(util:degrees(
            (math:pi() div 2) - math:atan2(point:px($this) - point:px($prev), point:py($this) - point:py($prev))
          ))
        )
    '/>
  </xsl:function>

  <!--
    : this:inclination()
    : Compute the inclination angle from one point in the next, in degrees
    : Return 90 if the points are the same
    :
    : $last: previous point; use (0,0,0) if no previous
    : $curr: current point
    :-->
  <xsl:function name="this:inclination" as="xs:double">
    <xsl:param name="last" as="map(xs:string, item()*)?"/>
    <xsl:param name="curr" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $curr := $curr=>point:as-dimension(3) return
      let $prev := (($last,$point:ORIGIN3D)[1])=>point:as-dimension(3) return
      let $d := point:distance($prev,$curr)
      return
        if (point:same($prev,$curr) or ($d &lt; $config:ε)) then 90
        else (
          util:remap-degrees(util:degrees(
            math:acos( (point:pz($curr) - point:pz($prev)) div $d )
          ))
        )
    '/>
  </xsl:function>

  <!--
    : this:linear-point()
    : Compute the point on the edge the given fraction from the start
    :
    : $edge: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
  <xsl:function name="this:linear-point" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $from := $edge=>this:start() return
      let $to := $edge=>this:end()
      return (
        point:map2(function($a,$b) {$a + ($b - $a)*$t}, $from, $to)
      )
    '/>
  </xsl:function>

  <!--
    : this:quad-point()
    : Compute the point on the quad edge the given fraction from the start
    : Midpoint of quad edge is quad-point(0.5) 
    : B(t)=(1-t)²P0 + 2(1-t)tP1 + t²P2, 0 <= t <= 1 
    :
    : $quadratic: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
  <xsl:function name="this:quad-point" as="map(xs:string, item()*)">
    <xsl:param name="quadratic" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $from := $quadratic=>this:start() return
      let $to := $quadratic=>this:end() return
      let $controls := $quadratic=>this:controls()
      return (
        point:map3(
          function($a, $b, $c) {
            (1 - $t)*(1 - $t)*$a +
            2*(1 - $t)*$t*$c + 
            $t * $t * $b
          },
          $from, $to, $controls[1]
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:cubic-point()
    : Compute the point on the cubic edge the given fraction from the start
    : Midpoint of cubic edge is cubic-point(0.5) 
    : B(t)=(1-t)³P0 + 3(1-t)²tP1 + 3(1-t)t²P2 + t³P3, 0 <= t <= 1 
    :
    : $cubic: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
  <xsl:function name="this:cubic-point" as="map(xs:string, item()*)">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $from := $cubic=>this:start() return
      let $to := $cubic=>this:end() return
      let $controls := $cubic=>this:controls()
      return (
        point:map4(
          function($a, $b, $c, $d) {
            (1 - $t)*(1 - $t)*(1 - $t)*$a +
            3*(1 - $t)*(1 - $t)*$t*$c +
            3*(1 - $t)*$t*$t*$d +
            $t*$t*$t*$b
          },
          $from, $to, $controls[1], $controls[2]
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:arc-point()
    : Compute the point on the arc edge the given fraction from the start
    : Midpoint of arc edge is arc-point(0.5) 
    : Will normalize arc properly.
    :
    : $arc: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
  <xsl:function name="this:arc-point" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $arc := this:normalize-arc($arc) return
      let $center := $arc=>this:arc-center() return
      let $radius := $arc=>this:arc-radius() return
      let $angles := $arc=>this:arc-angles() return
      let $from-angle := $angles[1] return
      let $to-angle := $angles[2] return
      let $da := $to-angle - $from-angle
      return point:destination($center, $from-angle + $t*$da, $radius)
    '/>
  </xsl:function>

  <!--
    : this:length()
    : Travel distance along edge.
    : Cubic is an approximation.
    :-->
  <xsl:function name="this:length" as="xs:double">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "arc") then this:arc-length($edge)
      else if ($kind = "quad") then this:quad-length($edge)
      else if ($kind = "cubic") then this:cubic-length($edge) (: approximation by 10 slices :)
      else if ($kind = "edge") then this:linear-length($edge)
      else errors:error("GEOM-BADREGION", ($edge, "length"))
    '/>
  </xsl:function>

  <!--
    : this:linear-length()
    : Crow-flies distance between end points of an edge. Same a travel distance
    : along the edge if it is a linear edge.
    :
    : $edge: the edge
    :-->
  <xsl:function name="this:linear-length" as="xs:double">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:distance($edge=>this:start(), $edge=>this:end())
    '/>
  </xsl:function>

  <!--
    : this:quad-length()
    : Length of travel along a quadratic edge.
    : 
    : $quad: the quadratic edge
    : 
    : Derivation, see: https://members.loria.fr/SHornus/quadratic-arc-length.html
    :
    : B(t) = (1-t)²P0 + 2(1-t)tP1 + t²P2
    : Reformulate to: P(t) = At² + 2Bt + C where
    : A = P0 - P1 + P2 - P1
    : B = P1 - P0
    : C = P0
    : let F = P2 - P1, so A = F - B
    : length = 
    :   (|F|(A · F) - |B|(A · B)) / |A|² + 
    :   (|B|²/|A| - (A · B)²/|A|³) X (log(|A||F| + A · F) - log(|A||B| + A · B))
    :-->
  <xsl:function name="this:quad-length" as="xs:double">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $P0 := this:start($quad) return
      let $P1 := this:controls($quad)[1] return
      let $P2 := this:end($quad) return
      let $F := point:sub($P2, $P1) return
      let $B := point:sub($P1, $P0) return
      let $A := point:sub($F, $B) return
      let $nA := point:magnitude($A) return
      let $nB := point:magnitude($B) return
      let $nF := point:magnitude($F) return
      let $A.B := point:dot($A, $B) return
      let $A.F := point:dot($A, $F)
      return (
        (: Possible error cases
       : nA=0 => F=B => start, c0, end are colinear => a straight edge
       :
       : nA*nF + A.F = 0 => nA*nF = -A.F => log(nA*nF + A.F) = -INF
       : nA*nF= √(Σai*ai)√(Σfi*fi) A.F=Σ(ai*fi)
       : A = F - B => nA*nF = √(Σ(fi-bi)*(fi-bi))√(Σfi*fi) A.F=Σ((fi-bi)*fi) => det(B,F) = 0 =>
       :   B and F are colinear => a straight edge again
       :
       : nA*nB + A.B = 0 => log(nA*nA + A.B) = -INF => nA*nB = -A.B => nA*nA*nB*nB = A.B*A.B => 
       :    (nB*nB / nA) - (A.B*A.B)/(nA*nA*nA) = 0 => treat 0*-INF as 0
       :)
        if ($nA*$nA = 0) then (
          point:distance($P0, $P2)
        ) else if ($nA*$nB + $A.B = 0) then (
          ($nF * $A.F - $nB * $A.B) div ($nA*$nA)
        ) else if ($nA * $nF + $A.F &lt;= 0) then (
          point:distance($P0, $P2)
        ) else if ($nA * $nB + $A.B &lt;= 0) then (
          point:distance($P0, $P2)
        ) else (
          (
            ($nF * $A.F - $nB * $A.B) div ($nA*$nA)
          ) +
          (
            (($nB*$nB div $nA) - ($A.B*$A.B) div ($nA*$nA*$nA)) *
            (math:log($nA*$nF + $A.F) - math:log($nA*$nB + $A.B))
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:cubic-length()
    : Travel length along a cubic edge. There is no general closed form, 
    : so this is approximated by interpolating the edge with line segments.
    : 
    : $cubic: the cubic edge
    : $n: number of interpolations to use
    :-->
  <xsl:function name="this:cubic-length" as="xs:double">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="n" as="xs:integer"/>
    <xsl:sequence select='
      let $ts := util:linspace($n, 0, 1)  return
      let $pts := for $t in $ts return this:cubic-point($cubic, $t)
      return
        sum(
          for $i in 2 to count($pts) return point:distance($pts[$i - 1], $pts[$i])
          (: for $pt at $i in tail($pts) return point:distance($pts[$i], $pt) :)
        )
    '/>
  </xsl:function>

  <!--
    : this:cubic-length()
    : Travel length along a cubic edge. There is no general closed form, 
    : so this is approximated by interpolating the edge with 10 line segments.
    : 
    : $cubic: the cubic edge
    :-->
  <xsl:function name="this:cubic-length" as="xs:double">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:cubic-length($cubic, 10)
    '/>
  </xsl:function>

  <!--
    : this:arc-length()
    : Travel distance along arc 
    :-->
  <!-- For 360 d=2πR; for α d=2πRα/360 i.e. the αth fraction of the total -->
  <xsl:function name="this:arc-length" as="xs:double">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $r := $arc=>this:arc-radius() return
      let $α := this:arc-extent($arc)
      return 2 * math:pi() * $r * $α div 360
    '/>
  </xsl:function>

  <!--
    : this:arc-extent()
    : Degrees swept out by an arc edge.
    :
    : $arc: the arc edge
    :-->
  <xsl:function name="this:arc-extent" as="xs:double">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $center := $arc=>this:arc-center() return
      let $points := $arc=>this:arc-ends() return
      let $angles := $arc=>this:arc-angles()
      return (
        if (exists($points)) then (
          util:remap-degrees(abs(this:angle($center, $points[2]) - this:angle($center, $points[1])))
        ) else (
          util:remap-degrees(abs($angles[2] - $angles[1]))
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:slice()
    : Create two edges at the given fraction of the source edge. Cubic edges
    : will be sliced into two cubic edges, quadratric edges will be slices into
    : two quadratic edges, etc.
    :
    : If t is at end, get same edge back.
    :
    : $edge: the edge to cut
    : $t: fraction of edge [0,1]
    :-->
  <xsl:function name="this:slice" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then this:slice-linear($edge, $t)
      else if ($kind = "arc") then this:slice-arc($edge, $t)
      else if ($kind = "quad") then this:slice-quad($edge, $t)
      else if ($kind = "cubic") then this:slice-cubic($edge, $t)
      else errors:error("GEOM-BADREGION", ($edge, "slice"))
    '/>
  </xsl:function>

  <!--
    : this:slice()
    : Slice the edge between two cut points. Will end up with three edges in
    : general: [start, p(start-t)], [p(start-t), p(end-t)], [p(end-t), end]
    : start-t <= end-t
    :
    : If start-t=end-t this is same as slice(edge, start-t)
    : If start-t or end-t is at an end will get one less edge
    :
    : Cubics are cut into cubics, quads into quads, straight edges into straight
    : edges
    :
    : $edge: the edge to cut
    : $start-t: fraction of edge [0,1]
    : $end-t: fraction of edge[0,1]
    :-->
  <xsl:function name="this:slice" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="start-t" as="xs:double"/>
    <xsl:param name="end-t" as="xs:double"/>
    <xsl:sequence select='
      if ($start-t > $end-t) then errors:error("GEOM-BADT", ($start-t, $end-t)) else (),
      if ($start-t = $end-t) then this:slice($edge, $start-t)
      else if ($start-t = 1) then $edge (: start-t=end-t=1 :)
      else if ($start-t = 0) then this:slice($edge, $end-t)
      else (
        (: 
       : end-t needs to be adjusted for the second slice:
       : Effective end-t should be fraction of length(slice) not of length(edge)
       : e * l - l1 = x * l2; l1 = s * l, l2 = l - l1 = l - s * l
       : x = (el - sl) / (s - sl) = (e - s) / (1 - s)
       :)
        let $slice := this:slice($edge, $start-t) return
        let $eff-end-t := ($end-t - $start-t) div (1 - $start-t)
        return ($slice[1], this:slice($slice[2], $eff-end-t))
      )
    '/>
  </xsl:function>

  <!--
    : this:chop()
    : Slice the edge by chopping it at the given cut points.
    :
    : Cubics are cut into cubics, quads into quads, straight edges into straight
    : edges
    :
    : $edge: the edge to cut
    : $ts: fraction of edge [0,1]
    :-->
  <xsl:function name="this:chop" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="ts" as="xs:double*"/>
    <xsl:sequence select='
      let $ts := distinct-values(sort($ts)) return
      
      let $eff-ts :=
        (: Adjust ts:
       : Need to adjust t2
       : Effective t2 should be fraction of length(slice)
       : e * l - l1 = x * l2; l1 = s * l, l2 = l - l1 = l - s * l
       : x = (el - sl) / (s - sl) = (e - s) / (1 - s)
       : However, the t1 is itself an adjustment
       : So t4 needs to be adjusted to t3 which is adjusted to t2
       : which is adjusted to t1 and so on. But we need to adjust 
       : them all for each go round:
       : [  t1  t2  t3 ]
       : [  t1][  t2′  t3′]
       : [  t1][  t2′][  t3′′]
       :)
        fold-left(2 to count($ts) - 1, $ts,
          function($new-ts, $it) {
            $new-ts[position() &lt; $it - 1],
            fold-left($new-ts[position() >= $it], $new-ts[position() = $it - 1],
              function($new-ts, $t2) {
                $new-ts,
                let $t1 := $new-ts[last()]
                return ($t2 - $t1) div (1 - $t1)
              }
            )
          }
        )
      return
      let $split1 := this:slice($edge, head($eff-ts)) return
      let $splits := (
        fold-left(tail($eff-ts), array{$split1},
          function($splits, $t) {
            $splits,
            array {
              if (empty($splits[last()]) or array:size($splits[last()]) &lt; 2) then ()
              else this:slice($splits[last()](2), $t)
            }
          }
        )
      )
      return (
        for $split in $splits return $split?*
      )
    '/>
  </xsl:function>

  <!--
    : this:slice-linear()
    : Create two straight edges at the given fraction of the source edge.
    :
    : If t is at end, get same edge back.
    :
    : $this: the straight edge to cut
    : $t: fraction of edge [0,1]
    :-->
  <xsl:function name="this:slice-linear" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      if ($t = (0,1)) then $edge else
      let $p1 := this:start($edge) return
      let $p2 := this:end($edge) return
      let $pt := this:linear-point($edge, $t)
      return (
        this:edge($p1, $pt),
        this:edge($pt, $p2)
      )
    '/>
  </xsl:function>

  <!--
    : this:slice-quad()
    : Create two quadratic edges at the given fraction of the source quadratic.
    :
    : If t is at end, get same edge back.
    :
    : $quad: the quad edge to cut
    : $t: fraction of edge [0,1]
    :-->
  <xsl:function name="this:slice-quad" as="map(xs:string, item()*)*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      if ($t = (0,1)) then $quad else
      let $p1 := this:start($quad) return
      let $p2 := this:controls($quad)[1] return
      let $p3 := this:end($quad) return
    
      let $p12 := point:add($p1, point:sub($p2, $p1)=>point:times($t)) return
      let $p23 := point:add($p2, point:sub($p3, $p2)=>point:times($t)) return
    
      let $p123 := point:add($p12, point:sub($p23, $p12)=>point:times($t))
                    
      return (
        (: Avoid creating quads that are really lines :)
        if (point:same($p1, $p2, $config:ε) or point:same($p2, $p3, $config:ε))
        then this:slice-linear(this:edge($p1, $p3), $t)
        else (
          if (point:same($p1, $p12, $config:ε) or point:same($p12, $p123, $config:ε)) then (
            this:edge($p1, $p123)
          ) else (
            this:quad($p1, $p123, $p12)
          ),
          if (point:same($p123, $p23, $config:ε) or point:same($p23, $p3, $config:ε)) then (
            this:edge($p123, $p3)
          ) else (
            this:quad($p123, $p3, $p23)
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:slice-cubic()
    : Create two cubic edges at the given fraction of the source cubic.
    :
    : If t is at end, get same edge back.
    :
    : $cubic: the cubic edge to cut
    : $t: fraction of edge [0,1]
    :-->
  <xsl:function name="this:slice-cubic" as="map(xs:string, item()*)*">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      if ($t = (0,1)) then $cubic else
      let $p1 := this:start($cubic) return
      let $p2 := this:controls($cubic)[1] return
      let $p3 := this:controls($cubic)[2] return
      let $p4 := this:end($cubic) return
    
      let $p12 := point:add($p1, point:sub($p2, $p1)=>point:times($t)) return
      let $p23 := point:add($p2, point:sub($p3, $p2)=>point:times($t)) return
      let $p34 := point:add($p3, point:sub($p4, $p3)=>point:times($t)) return
    
      let $p123 := point:add($p12, point:sub($p23, $p12)=>point:times($t)) return
      let $p234 := point:add($p23, point:sub($p34, $p23)=>point:times($t)) return
    
      let $p1234 := point:add($p123, point:sub($p234, $p123)=>point:times($t))
    
      return (
        if (point:same($p1, $p2, $config:ε) or point:same($p2, $p3, $config:ε))
        then this:slice-quad(this:quad($p1, $p3, $p4), $t)
        else if (point:same($p4, $p3, $config:ε))
        then this:slice-quad(this:quad($p1, $p2, $p4), $t)
        else (
          (: Avoid creating "cubics" that are really quads or lines :)
          if (point:same($p12, $p123, $config:ε)) then (
            if (point:same($p1, $p12, $config:ε) or point:same($p123, $p1234, $config:ε)) then (
              (: Really a line: make a straight edge :)
              this:edge($p1, $p1234)
            ) else (
              (: Really a quad: make a quad edge :)
              this:quad($p1, $p1234, $p12)
            )
          ) else (
            this:cubic($p1, $p1234, $p12, $p123)
          ),
          if (point:same($p234, $p34, $config:ε)) then (
            if (point:same($p1234, $p234, $config:ε) or point:same($p34, $p4, $config:ε)) then (
              (: Really a line: make a straight edge :)
              this:edge($p1234, $p4)
            ) else (
              (: Really a quad: make a quad edge :)
              this:quad($p1234, $p4, $p234)
            )
          ) else (
            this:cubic($p1234, $p4, $p234, $p34)
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:slice-arc()
    : Create two arc edges at the given fraction of the source arc.
    : If t is at end, get same edge back.
    : Normalizes the arc
    :
    : $arc: the arc edge to cut
    : $t: fraction of edge [0,1]
    :-->
  <xsl:function name="this:slice-arc" as="map(xs:string, item()*)*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      if ($t = (0,1)) then $arc else
      let $arc := this:normalize-arc($arc) return
      let $center := $arc=>this:arc-center() return
      let $radius := $arc=>this:arc-radius() return
      let $angles := $arc=>this:arc-angles() return
      let $from-angle := $angles[1] return
      let $to-angle := $angles[2] return
      let $da := $to-angle - $from-angle return
      let $at := $from-angle + $t*$da return
      let $larges :=
        if ($arc=>this:arc-large()) then (
          util:remap-degrees($at - $from-angle) > 180,
          util:remap-degrees($to-angle - $at) > 180
        ) else (
          false(), false()
        )
      return
      let $a1 := this:arc-by-angle($center, $radius, util:remap-degrees($from-angle), util:remap-degrees($at), false(), $larges[1]) return
      let $a2 := this:arc-by-angle($center, $radius, util:remap-degrees($at), util:remap-degrees($to-angle), false(), $larges[2])
      return (
        if (point:same(this:start($arc), this:end($arc), $config:ε) and not(this:arc-large($arc))) then (
          this:slice-linear(this:edge(this:start($arc), this:end($arc)), $t)
        ) else (
          if (not($larges[1]) and point:same(this:start($a1), this:end($a1), $config:ε)) then (
            this:edge(this:start($a1), this:end($a1))
          ) else (
            $a1
          ),
          if (not($larges[2]) and point:same(this:start($a2), this:end($a2), $config:ε)) then (
            this:edge(this:start($a2), this:end($a2))
          ) else (
            $a2
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:midpoint()
    : Return midpoint between two points
    :
    : $a: one point
    : $b: the other
    :-->
  <xsl:function name="this:midpoint" as="map(xs:string, item()*)">
    <xsl:param name="a" as="map(xs:string, item()*)"/>
    <xsl:param name="b" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      point:midpoint($a, $b)
    '/>
  </xsl:function>

  <!--
    : this:midpoint()
    : Center of the edge
    : Note: if you use on an arc, the arc must be normalized (edge:normalize-arc)
    :-->
  <xsl:function name="this:midpoint" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then point:midpoint(this:start($edge), this:end($edge))
      else if ($kind = "quad") then this:quad-point($edge, 0.5)
      else if ($kind = "cubic") then this:cubic-point($edge, 0.5)
      else if ($kind = "arc") then this:arc-point($edge, 0.5)
      else errors:error("GEOM-BADREGION", ($edge, "midpoint"))
    '/>
  </xsl:function>

  <!--
    : this:arc-midpoint()
    : Center of an arc. Note: arcs are treated as 2D, so output is 2D point;
    : we only look at x and y coordinates in the calculations.
    :
    : $center: center of the circle defining the arc
    : $a: start of arc
    : $p: end of arc
    :-->
  <xsl:function name="this:arc-midpoint" as="map(xs:string, item()*)">
    <xsl:param name="center" as="map(xs:string, item()*)"/>
    <xsl:param name="a" as="map(xs:string, item()*)"/>
    <xsl:param name="b" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $r := point:distance($center, $a) return
      let $lm := this:midpoint($a, $b) return
      let $v := point:sub($lm, $center) return
      let $lenv := math:sqrt(point:px($v)*point:px($v) + point:py($v)*point:py($v))
      return
        if (point:orientation($a, $center, $b)="counterclockwise") then (
          point:sub(
            $center,
            point:point($r*point:px($v) div $lenv, $r*point:py($v) div $lenv)
          )
        ) else (
          point:add(
            $center,
            point:point($r*point:px($v) div $lenv, $r*point:py($v) div $lenv)
          )
       )
    '/>
  </xsl:function>

  <!--
    : this:edge-point()
    : Compute the point on the edge the given fraction from the start
    :
    : $edge: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
  <xsl:function name="this:edge-point" as="map(xs:string, item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "edge") then this:linear-point($edge, $t)
      else if ($kind = "arc") then this:arc-point($edge, $t)
      else if ($kind = "quad") then this:quad-point($edge, $t)
      else if ($kind = "cubic") then this:cubic-point($edge, $t)
      else errors:error("GEOM-BADREGION", ($edge, "edge-point"))
    '/>
  </xsl:function>

  <!--
    : this:point-distance()
    : Shortest distance from point to linear edge
    :-->
  <xsl:function name="this:point-distance" as="xs:double">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="pt" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (point:same(this:start($edge), this:end($edge))) then (
        point:distance($pt, this:start($edge))
      ) else (
        let $v := this:start($edge) return
        let $w := this:end($edge) return
        let $l := this:linear-length($edge) return
        let $lsquared := $l*$l return
        let $w_sub_v := point:sub($w, $v) return
        let $t := max((0, min((1, point:dot(point:sub($pt, $v), $w_sub_v) div $lsquared))))
        return (
          point:distance($pt,
            point:point(
              point:px($v) + $t * point:px($w_sub_v),
              point:py($v) + $t * point:py($w_sub_v)
            )
          )
        )
      )
    '/>
  </xsl:function>

    <!--
    : this:closest-point()
    : Closest point on linear edge to the given point
    :-->
  <xsl:function name="this:closest-point" as="map(xs:string,item()*)">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="pt" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    let $start := this:start($edge) return
    let $end := this:end($edge)
    return (
      if (point:px($start)=point:px($end))
      then point:point(point:px($start), point:py($pt))
      else if (point:py($start)=point:py($end))
      then point:point(point:px($pt), point:py($start))
      else (
        let $m1 := (point:py($end) - point:px($start)) div (point:px($end) - point:px($start)) return
        let $m2 := -1 div $m1 return
        let $x := ($m1*point:px($start) - $m2*point:px($pt) + point:py($pt) - point:py($start)) div ($m1 - $m2) return
        let $y := $m2*($x - point:px($pt)) + point:py($pt)
        return (
          point:point($x, $y)
        )
      )
    )
    '/>
  </xsl:function>

  <!--
    : this:on-segment-colinear()
    : Given edge and colinear point, check whether point is on that edge
    :-->
  <xsl:function name="this:on-segment-colinear" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="p" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      util:twixt(point:px($p),
        min(this:vertices($edge)!point:px(.)) - $tolerance,
        max(this:vertices($edge)!point:px(.)) + $tolerance
      )
      and
      util:twixt(point:py($p),
        min(this:vertices($edge)!point:py(.)) - $tolerance,
        max(this:vertices($edge)!point:py(.)) + $tolerance
      )
    '/>
  </xsl:function>

  <!--
    : this:on-segment-colinear()
    :-->
  <xsl:function name="this:on-segment-colinear" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="p" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:on-segment-colinear($edge, $p, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:on-segment()
    : Given edge and arbitary point, check whether point is on that edge
    :-->
  <xsl:function name="this:on-segment" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="p" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:on-segment($edge, $p, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:on-segment()
    :-->
  <xsl:function name="this:on-segment" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="p" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      (
        (point:orientation(this:start($edge), this:end($edge), $p)="colinear") or
        (this:point-distance($edge, $p) &lt; $tolerance)
      )
      and (
        util:twixt(point:px($p),
          min(this:vertices($edge)!point:px(.)) - $tolerance,
          max(this:vertices($edge)!point:px(.)) + $tolerance
        )
      )
      and (
        util:twixt(point:py($p),
          min(this:vertices($edge)!point:py(.)) - $tolerance,
          max(this:vertices($edge)!point:py(.)) + $tolerance
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:linear-t()
    : Compute the t value of the point wrt the linear edge.
    :-->
  <xsl:function name="this:linear-t" as="xs:double*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:on-segment-colinear($edge, $point)) then (
        point:distance(this:start($edge), $point) div this:length($edge)
      )
      else () 
    '/>
  </xsl:function>

  <!--
    : this:on-quad()
    : True if the point is on the quadractic edge.
    :
    : B(t)=(1-t)²P0 + 2(1-t)tP1 + t²P2, 0 <= t <= 1 
    : B(t)=(P0 - 2P1 + P2)t² + (2P1 - 2P0)t + P0
    : If Q is on quad then solve (P0 - 2P1 + P2)t² + (2P1 - 2P0)t + P0 - Q = 0
    : i.e. (-b ± √((b² - 4ac))/2a where a=(P0 - 2P1 + P2), b=(2P1 - 2P0), c = (P0 - Q)
    : If t in range then you're good
    : 
    : We're using equality within ε here, because floating point equality is 
    : problematic. Alternatively: you could compute minimum distance and 
    : make sure it is within tolerance.
    :-->
  <xsl:function name="this:on-quad" as="xs:boolean">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      exists(this:quad-t($quad, $point, $tolerance))
    '/>
  </xsl:function>

  <!--
    : this:on-quad()
    :-->
  <xsl:function name="this:on-quad" as="xs:boolean">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:on-quad($quad, $point, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:quad-t()
    : Compute the t value of the point wrt the quadractic edge.
    :
    : B(t)=(1-t)²P0 + 2(1-t)tP1 + t²P2, 0 <= t <= 1 
    : B(t)=(P0 - 2P1 + P2)t² + (2P1 - 2P0)t + P0
    : If Q is on quad then solve (P0 - 2P1 + P2)t² + (2P1 - 2P0)t + P0 - Q = 0
    : i.e. (-b ± √((b² - 4ac))/2a where a=(P0 - 2P1 + P2), b=(2P1 - 2P0), c = (P0 - Q)
    : If t in range then you're good
    : 
    : We're using equality within ε here, because floating point equality is 
    : problematic. Alternatively: you could compute minimum distance and 
    : make sure it is within tolerance.
    :-->
  <!-- XYZZY TODO: how should tolerance be handled, really? -->
  <xsl:function name="this:quad-t" as="xs:double*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      if (point:same($point, this:start($quad), $tolerance)) then 0.0
      else if (point:same($point, this:end($quad), $tolerance)) then 1.0
      else (
        let $P0 := this:start($quad) return
        let $P1 := this:controls($quad)[1] return
        let $P2 := this:end($quad) return
        let $ax := point:px($P0) - 2*point:px($P1) + point:px($P2) return
        let $bx := 2*point:px($P1) - 2*point:px($P0) return
        let $cx := point:px($P0) - point:px($point) return
        let $x-all-zero := every $x in ($ax, $bx, $cx) satisfies abs($x) &lt; $config:ε return
        let $ay := point:py($P0) - 2*point:py($P1) + point:py($P2) return
        let $by := 2*point:py($P1) - 2*point:py($P0) return
        let $cy := point:py($P0) - point:py($point) return
        let $y-all-zero := every $y in ($ay, $by, $cy) satisfies abs($y) &lt; $config:ε
        return (
          if ($x-all-zero and $y-all-zero) then (
            (: 
           : Quad is really a point
           : If this works it was caught above
           :)
          ) else if ($x-all-zero) then (
            (: 
           : Quad is really a straight line with all xs the same
           : If this works point x must be same, then use ty
           :)
            if (abs(point:px($point) - point:px($P0)) &lt; $tolerance) then (
              let $ty := roots:quadratic-real-roots($ay, $by, $cy)[util:twixt(., -$config:ε, 1 + $config:ε)] return
              for $t in $ty return util:clamp($t, 0.0, 1.0)
             ) else (
             )
          ) else if ($y-all-zero) then (
            (: 
           : Quad is really a straight line with all ys the same
           : If this works point y must be same, then use tx
           :)
            if (abs(point:py($point) - point:py($P0)) &lt; $tolerance) then (
              let $tx := roots:quadratic-real-roots($ax, $bx, $cx)[util:twixt(., -$config:ε, 1 + $config:ε)] return
              for $t in $tx return util:clamp($t, 0.0, 1.0)
             ) else (
             )
          ) else (
            (: 
           : Normal case: match tx with ty
           :)
            let $tx := roots:quadratic-real-roots($ax, $bx, $cx)[util:twixt(., -$config:ε, 1 + $config:ε)] return
            let $ty := roots:quadratic-real-roots($ay, $by, $cy)[util:twixt(., -$config:ε, 1 + $config:ε)] return
            let $ts := $tx[some $t in $ty satisfies abs(. - $t) &lt; $config:ε] return
            for $t in $ts return util:clamp($t, 0.0, 1.0)
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:quad-t()
    :-->
  <xsl:function name="this:quad-t" as="xs:double*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:quad-t($quad, $point, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:on-cubic()
    : True if the point is on the cubic edge.
    : B(t)=(1-t)³P0 + 3(1-t)²tP1 + 3(1-t)t²P2 + t³P3, 0 <= t <= 1 
    : B(t)=(P3 - 3P2 + P1 - P0)t³ + (3P2 - 6P1)t² + (3P1)t + P0
    : If Q is on cubic then solve 
    : (P3 - 3P2 + P1 - P0)t³ + (3P2 - 6P1)t² + (3P1)t + P0 - Q = 0
    : i.e. cubic roots
    :
    : If t in range then you're good
    : 
    : We're using equality within ε here, because floating point equality is 
    : problematic. Alternatively: you could compute minimum distance and 
    : make sure it is within tolerance.
    :-->
  <xsl:function name="this:on-cubic" as="xs:boolean">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      exists(this:cubic-t($cubic, $point, $tolerance))
    '/>
  </xsl:function>

  <!--
    : this:on-cubic()
    :-->
  <xsl:function name="this:on-cubic" as="xs:boolean">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:on-cubic($cubic, $point, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:cubic-t()
    : Compute the t value of the point wrt the cubic edge.
    : B(t)=(1-t)³P0 + 3(1-t)²tP1 + 3(1-t)t²P2 + t³P3, 0 <= t <= 1 
    : B(t)=(-P0 + 3P1 - 3P2 + P3)t³ + (3P0 - 6P1 + 3P2)t² + (-3P0 + 3P1)t + P0
    : If Q is on cubic then solve 
    : (-P0 + 3P1 - 3P2 + P3)t³ + (3P0 - 6P1 + 3P2)t² + (-3P0 + 3P1)t + (P0 - Q)
    : i.e. cubic roots
    :
    : If t in range then you're good
    : 
    : We're using equality within ε here, because floating point equality is 
    : problematic. Alternatively: you could compute minimum distance and 
    : make sure it is within tolerance.
    :-->
  <!-- XYZZY TODO: how should tolerance be handled, really? -->
  <xsl:function name="this:cubic-t" as="xs:double*">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      if (point:same($point, this:start($cubic), $tolerance)) then 0.0
      else if (point:same($point, this:end($cubic), $tolerance)) then 1.0
      else (
        let $P0 := this:start($cubic) return
        let $P1 := this:controls($cubic)[1] return
        let $P2 := this:controls($cubic)[2] return
        let $P3 := this:end($cubic) return
        let $ax := -point:px($P0) + 3*point:px($P1) - 3*point:px($P2) + point:px($P3) return
        let $bx := 3*point:px($P0) - 6*point:px($P1) + 3*point:px($P2) return
        let $cx := -3*point:px($P0) + 3*point:px($P1) return
        let $dx := point:px($P0) - point:px($point) return
        let $x-all-zero := every $x in ($ax, $bx, $cx, $dx) satisfies abs($x) &lt; $config:ε return
        let $ay := -point:py($P0) + 3*point:py($P1) - 3*point:py($P2) + point:py($P3) return
        let $by := 3*point:py($P0) - 6*point:py($P1) + 3*point:py($P2) return
        let $cy := -3*point:py($P0) + 3*point:py($P1) return
        let $dy := point:py($P0) - point:py($point) return
        let $y-all-zero := every $y in ($ay, $by, $cy, $dy) satisfies abs($y) &lt; $config:ε 
        return (
          if ($x-all-zero and $y-all-zero) then (
            (: 
           : Cubic is really a point
           : If this works it was caught above
           :)
          ) else if ($x-all-zero) then (
            (: 
           : Cubic is really a straight line with all xs the same
           : If this works point x must be same, then use ty
           :)
            if (abs(point:px($point) - point:px($P0)) &lt; $tolerance) then (
              let $ty := roots:cubic-real-roots($ay, $by, $cy, $dy)[util:twixt(., -$config:ε, 1 + $config:ε)] return
              for $t in $ty return util:clamp($t, 0.0, 1.0)
             ) else (
             )
          ) else if ($y-all-zero) then (
            (: 
           : Cubic is really a straight line with all ys the same
           : If this works point y must be same, then use tx
           :)
            if (abs(point:py($point) - point:py($P0)) &lt; $tolerance) then (
              let $tx := roots:cubic-real-roots($ax, $bx, $cx, $dx)[util:twixt(., -$config:ε, 1 + $config:ε)] return
              for $t in $tx return util:clamp($t, 0.0, 1.0)
             ) else (
             )
          ) else (
            (: 
           : Normal case: match tx with ty
           :)
            let $tx := roots:cubic-real-roots($ax, $bx, $cx, $dx)[util:twixt(., -$config:ε, 1 + $config:ε)] return
            let $ty := roots:cubic-real-roots($ay, $by, $cy, $dy)[util:twixt(., -$config:ε, 1 + $config:ε)] return
            let $ts := $tx[some $t in $ty satisfies abs(. - $t) &lt; $config:ε] return
            for $t in $ts return util:clamp($t, 0.0, 1.0)
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:cubic-t()
    :-->
  <xsl:function name="this:cubic-t" as="xs:double*">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:cubic-t($cubic, $point, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:on-arc()
    : Is the point on the arc (within ε)?
    : Note: will normalize the arc
    :-->
  <xsl:function name="this:on-arc" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      let $arc := this:normalize-arc($arc) return
      let $center := $arc=>this:arc-center() return
      let $radius := $arc=>this:arc-radius() return
      let $angles := $arc=>this:arc-angles() return
      let $from-angle := $angles[1] return
      let $to-angle := $angles[2]
      return (
        abs(point:distance($center, $point) - $radius) &lt; $tolerance and
        (
          util:twixt(this:angle($center, $point), $from-angle, $to-angle) or
          util:twixt(this:angle($center, $point)+360, $from-angle, $to-angle) 
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:on-arc()
    :-->
  <xsl:function name="this:on-arc" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:on-arc($arc, $point, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:normalize-arc()
    : A lot of arc operations need us to know the actual arc center and
    : starting/ending angles. For large and flipped arcs that's get 
    : problematic. Rather than repeat that code everywhere, encapsulate
    : it here. The normalized arc will always run in sweep order and may or
    : may not be large.
    :
    : $arc: source arc
    :-->
  <xsl:function name="this:normalize-arc" as="map(xs:string, item()*)">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $arc := this:as-angle-arc($arc) return
      let $flipped := $arc=>this:arc-flipped() return
      let $large := $arc=>this:arc-large() return
      let $angles := ($arc=>this:arc-angles())!util:remap-degrees(.) return
      let $delta := $angles[2] - $angles[1] return
      let $swap :=
        ($large and
          ($delta &lt; -180 or ($delta > 0 and $delta &lt; 180))
        ) or
        (not($large) and
          (($delta &lt; 0 and abs($delta) &lt; 180) or ($delta > 180))
        ) return
      let $adjust-from :=
        ($large and $delta > 0 and $delta &lt; 180) or
        (not($large) and $delta > 180) return
      let $adjust-to :=
        ($large and $delta &lt; 0 and abs($delta) &lt;= 180) or
        (not($large) and $delta &lt;= -180) return
      let $adjusted-from :=
        if ($adjust-from) then $angles[1] + 360 else $angles[1] return
      let $adjusted-to :=
        if ($adjust-to) then $angles[2] + 360 else $angles[2] return
      let $from-angle :=
        if ($swap) then $adjusted-to else $adjusted-from return
      let $to-angle := 
        if ($swap) then $adjusted-from else $adjusted-to return
      let $center := $arc=>this:arc-center() return
      let $radius := $arc=>this:arc-radius() return
      let $from := point:destination($center, $from-angle, $radius) return
      let $to := point:destination($center, $to-angle, $radius) return
      let $flipped-center :=
        if ($swap) then (
          if ($flipped) then $center else affine:reflect($center, $from, $to)
        ) else (
          if ($flipped) then affine:reflect($center, $from, $to) else $center
        )
      return (
        util:assert($from-angle &lt;= $to-angle, "Normalized angles wonky"),
        util:assert((abs($to-angle - $from-angle) > 180) = $large, "Large flag wonky"),
        this:arc-by-angle(
          $flipped-center,
          $radius,
          $from-angle,
          $to-angle,
          false(),
          $large
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:arc-t()
    : Compute the t value of the point wrt the arc edge.
    : Will normalize the arc
    :-->
  <xsl:function name="this:arc-t" as="xs:double*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (point:same($point, this:start($arc))) then 0.0
      else if (point:same($point, this:end($arc))) then 1.0
      else (
        let $arc := this:normalize-arc($arc) return
        let $center := $arc=>this:arc-center() return
        let $radius := $arc=>this:arc-radius() return
        let $angles := $arc=>this:arc-angles() return
        let $from-angle := $angles[1] return
        let $to-angle := $angles[2] return
        let $point-angle := this:angle($center, $point)
        return (
          if (abs(point:distance($center, $point) - $radius) &lt; $config:ε and
              (
                util:twixt($point-angle, $from-angle, $to-angle) or
                util:twixt($point-angle + 360, $from-angle, $to-angle)
              )
          ) then (
            ($point-angle - $from-angle) div ($to-angle - $from-angle)
          ) else ()
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:edge-t()
    :-->
  <xsl:function name="this:edge-t" as="xs:double*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="point" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind = "quad") then this:quad-t($edge, $point)
      else if ($kind = "cubic") then this:cubic-t($edge, $point)
      else if ($kind = "edge") then this:linear-t($edge, $point)
      else if ($kind = "arc") then this:arc-t($edge, $point)
      else errors:error("GEOM-NOTIMPLEMENTED", ("edge-t", this:kind($edge)))
    '/>
  </xsl:function>

  <!--:======================================================================
    : Operations
    :======================================================================
    :-->

  <!--
    : this:bounding-box()
    : Minimum box surrounding the region. Some approximation for non-linear
    : edges.
    :-->
  <xsl:function name="this:bounding-box" as="map(xs:string, item()*)">
    <xsl:param name="regions" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
      let $boxes := (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind = "edge") then
          let $pts := this:points($region)
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        else if ($kind = "quad") then ( 
          (: P0 = start, P1 = control, P2 = end :)
          (: B(t) = (1-t)²·P0 + 2(1-t)t·P1 + t²·P2 :)
          (: B′(t) = a(1-t) + b where a = 2(B-A), b=2(C-B) :)
          (: A=P0 B=P1 C=P2 :)
          (: B′(t) = 0 where t = -a/(b-a) :)
          let $P0 := this:start($region) return
          let $P1 := this:controls($region)[1] return
          let $P2 := this:end($region) return
          let $T := 
            point:map3(
              function($p0, $p1, $p2) {
                let $A := $p0 return
                let $B := $p1 return
                let $C := $p2 return
                let $a := 2*($B - $A) return
                let $b := 2*($C - $B)
                return
                  if ($a = $b) then xs:double("INF")
                  else -$a div ($b - $a)
              },
              $P0, $P1, $P2
            ) return
          let $pts := (
            $P0, $P2, 
            if (util:twixt(point:px($T), 0, 1))
            then $region=>this:quad-point(point:px($T))
            else (),
            if (util:twixt(point:py($T), 0, 1))
            then $region=>this:quad-point(point:py($T))
            else ()
          )
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        )
        else if ($kind = "cubic") then (
          (: Cubic P1 to P4 with controls P2, P3 :)
          (: B(t) = (1-t)³·P1 + 3(1-t)²t·P2 + 3(1-t)t²·P3 + t³·P4 :)
          (: B′(t) = (1-t)²·A + 2(1-t)t·B + t²·C :)
          (: A=3(P2-P1), B=3(P3-P2), C=3(P4-P3) :)
          (: B′(t) = at² + bt + c = 0 => t = (-b ± √(b² - 4ac))/2a :)
          (: a = A - 2B + C = 3(-P1 + 3P2 - 3P3 + P4) :) 
          (: b = 2(B - A) = 6(P1 - 2P2 + P3) :) 
          (: c = A = 3(P2 - P1) :)
          let $P1 := this:start($region) return
          let $P2 := this:controls($region)[1] return
          let $P3 := this:controls($region)[2] return
          let $P4 := this:end($region) return
          let $ts := (
            let $p1 := point:px($P1) return
            let $p2 := point:px($P2) return
            let $p3 := point:px($P3) return
            let $p4 := point:px($P4) return
            let $a := 3*(-$p1 + 3*$p2 - 3*$p3 + $p4) return
            let $b := 6*($p1 - 2*$p2 + $p3) return
            let $c := 3*($p2 - $p1)
            return roots:quadratic-real-roots($a, $b, $c),
            let $p1 := point:py($P1) return
            let $p2 := point:py($P2) return
            let $p3 := point:py($P3) return
            let $p4 := point:py($P4) return
            let $a := 3*(-$p1 + 3*$p2 - 3*$p3 + $p4) return
            let $b := 6*($p1 - 2*$p2 + $p3) return
            let $c := 3*($p2 - $p1)
            return roots:quadratic-real-roots($a, $b, $c)
          )  return
          let $pts := (
            $P1, $P4,
            for $t in $ts return (
              if (util:twixt($t, 0, 1))
              then $region=>this:cubic-point($t)
              else ()
            )
          )
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        )
        else if ($kind = "arc") then (
          let $arc := this:normalize-arc($region) return
          let $center := $arc=>this:arc-center() return
          let $radius := $arc=>this:arc-radius() return
          let $angles := $arc=>this:arc-angles() return
          let $from-angle := $angles[1] return
          let $to-angle := $angles[2] return
          let $from := $arc=>this:start() return
          let $to := $arc=>this:end()
          (: If the arc is less than 180° we can do better than the whole
         : circle&#x27;s bounding box by computing the intersection of the
         : tangent lines and finding the point from that to center at r
         : Tangent line to p has slope angle(center,p)+90
         : Equation y = mx + b => p[y] = mp[x] + b => b = p[y] - mp[x]
         : Intersection of a1x + b1y + c1 = 0, a2x + b2y + c2 = 0
         : is ((b1c2 - b2c1)/(a1b2 - a2b1), (c1a2 - c2a1)/(a1b2 - a2b1))
         : a1=m b1=-1 c1=p[y]-mp[x]
         : a2=n b2=-1 c2=q[y]-mq[x]
         :)
          return (
            if ($from-angle = $to-angle) then (
              box:box($from, $from)
            ) else if ($to-angle - $from-angle &lt; 180) then (
              let $m := util:radians($from-angle - 90) return
              let $a1 := $m return
              let $b1 := -1 return
              let $c1 := point:py($from) - $m*point:px($from) return
              let $n := util:radians($to-angle - 90) return
              let $a2 := $n return
              let $b2 := -1 return
              let $c2 := point:py($to) - $n*point:px($to) return
              let $intersection := (
                util:assert($a1*$b2 - $a2*$b1 != 0, "Non intersecting lines"),
                point:point(
                  ($b1*$c2 - $b2*$c1) div ($a1*$b2 - $a2*$b1),
                  ($c1*$a2 - $c2*$a1) div ($a1*$b2 - $a2*$b1)
                )
              ) return
              let $pts := (
                $from, $to,
                point:destination($center, this:angle($center, $intersection), $radius)
              )
              return
                box:box(
                  min($pts!point:px(.)), min($pts!point:py(.)),
                  max($pts!point:px(.)), max($pts!point:py(.))
                )
            ) else (
              ellipse:bounding-box(this:arc-circle($arc))
            )
          )
        )
        else errors:error("GEOM-BADREGION", ($region, "bounding-box"))
      )
      return (
        if (empty($boxes)) then box:box(0,0,0,0)
        else if (empty(tail($boxes))) then $boxes
        else (
          let $pts := $boxes!this:points(.)
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        )
      )
    '/>
  </xsl:function>

  <!--
    :======================================================================
    : Intersections 
    :======================================================================
    :-->

  <!--
    : this:line-intersection()
    : Point of intersection, if any between lines defined as extension of
    : two edges.
    : Assumes linear edges: so wrong for other kinds of edges
    :-->
  <xsl:function name="this:line-intersection" as="map(xs:string, item()*)?">
    <xsl:param name="this" as="map(xs:string, item()*)"/>
    <xsl:param name="that" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      (: edge: ((x1,y1), (x2,y2)) => line Ax + By = C
     : A = y2 - y1
     : B = x1 - x2
     : C = Ax1 + By1
     : det = A1 * B2 - A2 * B1
     : => ( (B2 * C1 - B1 * C2) / det, (A1 * C2 - A2 * C1) / det
     :)
    
      (: A = y2 - y1 ; Ax = A(this), Ay = A(that) :)
      let $A := (
          point:py(this:end($this)) - point:py(this:start($this)),
          point:py(this:end($that)) - point:py(this:start($that))
      )
      return
      (: B = x1 - x2 ; Bx = B(this), By = B(that) :)
      let $B := (
          point:px(this:start($this)) - point:px(this:end($this)),
          point:px(this:start($that)) - point:px(this:end($that))
      )
      return
      (: C = Ax1 + By1 ; Cx = C(this), Cy = C(that) :)
      let $C := (
        v:px($A) * point:px(this:start($this)) + v:px($B) * point:py(this:start($this)),
        v:py($A) * point:px(this:start($that)) + v:py($B) * point:py(this:start($that))
      )
      return
      let $det := v:determinant($A, $B)
      return (
        if ($det = 0) then () else (
          point:point(
            -v:determinant($B, $C) div $det,
             v:determinant($A, $C) div $det
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:intersection()
    : Point of intersection, if any between two edges
    :-->
  <xsl:function name="this:intersection" as="map(xs:string, item()*)*">
    <xsl:param name="this" as="map(xs:string, item()*)"/>
    <xsl:param name="that" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      let $kind := this:kind($this) return
      if ($kind = "edge") then this:linear-intersection($this, $that)
      else if ($kind = "quad") then this:quad-intersection($this, $that)
      else if ($kind = "cubic") then this:cubic-intersection($this, $that)
      else if ($kind = "arc") then this:arc-intersection($this, $that)
      else errors:error("GEOM-BADREGION", ($this, "intersection"))
    '/>
  </xsl:function>

  <!--
    : this:linear-intersection()
    : Point of intersection, if any between linear edge and some other edge
    :-->
  <xsl:function name="this:linear-intersection" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($edge) != "edge") then errors:error("GEOM-BADREGION", ($edge, "linear-intersection"))
      else ( let $kind := this:kind($other) return
        if ($kind = "edge") then (
          let $p := this:line-intersection($edge, $other)
          return (
            if (empty($p)) then ()
            else if (this:on-segment-colinear($edge, $p) and this:on-segment-colinear($other, $p))
            then $p
            else ()
          )
        )
        else if ($kind = "quad") then this:quad-intersection($other, $edge)
        else if ($kind = "cubic") then this:cubic-intersection($other, $edge)
        else if ($kind = "arc") then this:arc-intersection($other, $edge)
        else errors:error("GEOM-BADREGION", ($edge, "linear-intersection"))
      )
    '/>
  </xsl:function>

  <!--
    : this:quad-edge-intersection-ts()
    :-->
  <xsl:function name="this:quad-edge-intersection-ts" as="xs:double*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      if (this:kind($quad) != "quad") then errors:error("GEOM-BADREGION", ($quad, "quad-edge-intersection-ts"))
      else if (this:kind($other) != "edge") then errors:error("GEOM-BADREGION", ($quad, "quad-edge-intersection-ts"))
      else (
        (: https://pomax.github.io/bezierinfo/ :)
        let $angle := this:angle($other) return
        let $rotated-edge :=
          this:mutate($other,
            function($p) {affine:rotate($p, $angle, this:start($other))}
          )
        return
        let $translation := point:py(this:start($rotated-edge)) return
        let $translated-edge :=
          this:mutate($rotated-edge,
            function($p) {affine:translate($p, 0, -$translation)}
          )
        return
        let $translated-quad :=
          this:mutate($quad,
            function($p) {affine:rotate($p, $angle, this:start($other))=>affine:translate(0,-$translation)}
          )
        return
        let $P0 := this:start($translated-quad) return
        let $P1 := this:controls($translated-quad)[1] return
        let $P2 := this:end($translated-quad) return
        let $ts := (
          let $p0 := point:px($P0) return
          let $p1 := point:px($P1) return
          let $p2 := point:px($P2) return
          let $a := $p0 - 2*$p1 + $p2 return
          let $b := 2*$p1 - 2*$p0 return
          let $c := $p0
          return roots:quadratic-real-roots($a, $b, $c)
          ,
          let $p0 := point:py($P0) return
          let $p1 := point:py($P1) return
          let $p2 := point:py($P2) return
          let $a := $p0 - 2*$p1 + $p2 return
          let $b := 2*$p1 - 2*$p0 return
          let $c := $p0
          return roots:quadratic-real-roots($a, $b, $c)
        )
        return (
          distinct-values(
            for $t in $ts return (
              if (
                util:twixt($t, 0, 1) and (
                  let $qp := this:quad-point($translated-quad, $t)
                  return
                    abs(point:py($qp)) &lt; $tolerance and
                    util:twixt(point:px($qp), point:px(this:start($translated-edge)), point:px(this:end($translated-edge)))
                )
              ) then $t else ()
            )
          )
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:quad-edge-intersection-ts()
    :-->
  <xsl:function name="this:quad-edge-intersection-ts" as="xs:double*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:quad-edge-intersection-ts($quad, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:quad-intersection()
    : Point of intersection, if any between quad edge and some other edge
    :-->
  <xsl:function name="this:quad-intersection" as="map(xs:string, item()*)*">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($quad) != "quad") then errors:error("GEOM-BADREGION", ($quad, "quad-intersection"))
      else ( let $kind := this:kind($other) return
        if ($kind = "edge") then (
          let $good-ts := this:quad-edge-intersection-ts($quad, $other) return
          for $t in $good-ts return this:quad-point($quad, $t)
        )
        else if ($kind = "quad") then (
          if (box:intersects-box(this:bounding-box($quad), this:bounding-box($other))) then (
            if (this:quad-length($quad) &lt; $config:ε and
                this:quad-length($other) &lt; $config:ε)
            then (
              this:start($quad)
            ) else (
              let $pts :=
                let $quadsubs := this:slice-quad($quad, 0.5) return
                let $othersubs := this:slice-quad($other, 0.5) return
                for $subquad in $quadsubs return
                for $subother in $othersubs
                return this:intersection($subquad, $subother)
              return
              let $dpts := for $p in $pts return point:decimal($p, $this:precision)
              return
              for $i in 1 to count($dpts) return (
                if (every $j in 1 to $i - 1 satisfies (not(point:same($dpts[$i], $dpts[$j]))))
                then $pts[$i] else ()
              )
            )
          ) else (
          )
        )
        else if ($kind = "cubic") then this:cubic-intersection($other, $quad)
        else errors:error("GEOM-BADREGION", ($other, "quad-intersection"))
      )
    '/>
  </xsl:function>

  <!--
    : this:cubic-intersection()
    : Point of intersection, if any between cubic edge and some other edge
    :-->
  <xsl:function name="this:cubic-intersection" as="map(xs:string, item()*)*">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($cubic) != "cubic") then errors:error("GEOM-BADREGION", ($cubic, "cubic-intersection"))
      else ( let $kind := this:kind($other) return
        if ($kind = "edge") then (
          if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other))) then (
            if (this:linear-length($other) &lt; $config:ε and
                this:cubic-length($cubic, 4) &lt; $config:ε)
            then this:start($cubic)
            else (
              let $pts := 
                let $cubicsubs := this:slice-cubic($cubic, 0.5) return
                let $othersubs := this:slice-linear($other, 0.5) return
                for $subcubic in $cubicsubs return
                for $subother in $othersubs
                return this:intersection($subcubic, $subother)
              return
              let $dpts := for $p in $pts return point:decimal($p, $this:precision)
              return
              for $i in 1 to count($dpts) return (
                if (every $j in 1 to $i - 1 satisfies (not(point:same($dpts[$i], $dpts[$j]))))
                then $pts[$i] else ()
              )
            )
          ) else (
          )
        )
        else if ($kind = "quad") then (
          if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other))) then (
            if (this:quad-length($other) &lt; $config:ε and
                this:cubic-length($cubic, 4) &lt; $config:ε)
            then this:start($cubic)
            else (
              let $pts :=
                let $cubicsubs := this:slice-cubic($cubic, 0.5) return
                let $othersubs := this:slice-quad($other, 0.5) return
                for $subcubic in $cubicsubs return
                for $subother in $othersubs
                return this:intersection($subcubic, $subother)
              return
              let $dpts := for $p in $pts return point:decimal($p, $this:precision)
              return
              for $i in 1 to count($dpts) return (
                if (every $j in 1 to $i - 1 satisfies (not(point:same($dpts[$i], $dpts[$j]))))
                then $pts[$i] else ()
              )
            )
          ) else (
          )
        )
        else if ($kind = "cubic") then (
          if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other))) then (
            if (this:cubic-length($cubic, 4) &lt; $config:ε and
                this:cubic-length($other, 4) &lt; $config:ε)
            then this:start($cubic)
            else (
              let $pts :=
                let $cubicsubs := this:slice-cubic($cubic, 0.5) return
                let $othersubs := this:slice-cubic($other, 0.5) return
                for $subcubic in $cubicsubs return
                for $subother in $othersubs
                return this:intersection($subcubic, $subother)
              return
              let $dpts := for $p in $pts return point:decimal($p, $this:precision)
              return
              for $i in 1 to count($dpts) return (
                if (every $j in 1 to $i - 1 satisfies (not(point:same($dpts[$i], $dpts[$j]))))
                then $pts[$i] else ()
              )
            )
          ) else (
          )
        )
        else errors:error("GEOM-BADREGION", ($other, "cubic-intersection"))
      )
    '/>
  </xsl:function>

  <!--
    : this:arc-intersection()
    :-->
  <xsl:function name="this:arc-intersection" as="map(xs:string, item()*)*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      if (this:kind($arc) != "arc") then errors:error("GEOM-BADREGION", ($arc, "arc-intersection"))
      else ( let $kind := this:kind($other) return
        if ($kind = "edge") then (
          let $arc := this:normalize-arc($arc) return
          (: https://math.stackexchange.com/questions/228841/how-do-i-calculate-the-intersections-of-a-straight-line-and-a-circle :)
          (: line y=mx + c; m = (y2-y1)/(x2-x1), c=y1 :)
          let $r := this:arc-radius($arc) return
          let $p := point:px(this:arc-center($arc)) return
          let $q := point:py(this:arc-center($arc))  return
          let $x1 := point:px(this:start($other)) return
          let $y1 := point:py(this:start($other)) return
          let $x2 := point:px(this:end($other)) return
          let $y2 := point:py(this:end($other)) return
          let $pts := (
             if (abs($x1 - $x2) &lt; $config:ε) then ( (: vertical line :)
               let $k := $x1 return
               let $A := 1 return
               let $B := -2*$q return
               let $C := $p*$p + $q*$q - $r*$r - 2*$k*$p + $k*$k
               return (
                 if ($B*$B >= 4*$A*$C) then (
                   for $y in roots:quadratic-real-roots($A, $B, $C)
                   return point:point($k, $y)
                 ) else ()
               )
             ) else (
               let $m := ($y2 - $y1) div ($x2 - $x1) return
               let $c := $y1 - $m*$x1 return
               let $A := $m*$m + 1 return
               let $B := 2*($m*$c - $m*$q - $p) return
               let $C := $q*$q - $r*$r + $p*$p - 2*$c*$q + $c*$c
               return (
                 if ($B*$B >= 4*$A*$C) then (
                   for $x in roots:quadratic-real-roots($A, $B, $C)
                   return point:point($x, $m*$x + $c)
                 ) else ()
               )
             )
          )
          return (
            for $pt in $pts return (
              if (this:on-segment-colinear($other, $pt) and this:on-arc($arc, $pt))
              then $pt else ()
            )
          )
        )
        else if ($kind = "quad") then (
          errors:error("GEOM-NOTIMPLEMENTED", ("arc-intersection", $kind))
        )
        else if ($kind = "cubic") then (
          errors:error("GEOM-NOTIMPLEMENTED", ("arc-intersection", $kind))
        )
        else if ($kind = "arc") then (
          errors:error("GEOM-NOTIMPLEMENTED", ("arc-intersection", $kind))
        ) 
        else errors:error("GEOM-BADREGION", ($other, "arc-intersection"))
      )
    '/>
  </xsl:function>

  <!--
    : this:edge-intersects-edge()
    : linear edge to linear edge intersection 
    :-->
  <xsl:function name="this:edge-intersects-edge" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:edge-intersects-edge($edge, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:edge-intersects-edge()
    :-->
  <xsl:function name="this:edge-intersects-edge" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      (: See https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/ :)
      let $o1 := point:orientation(this:start($edge), this:end($edge), this:start($other)) return
      let $o2 := point:orientation(this:start($edge), this:end($edge), this:end($other)) return
      let $o3 := point:orientation(this:start($other), this:end($other), this:start($edge)) return
      let $o4 := point:orientation(this:start($other), this:end($other), this:end($edge))
      return (
        ($o1 != $o2 and $o3 != $o4) or
        (: $other.start colinear with and on segment $edge :)
        ($o1="colinear" and this:on-segment-colinear($edge, this:start($other), $tolerance)) or
        (: $other.end colinear with and on segment $edge :)
        ($o2="colinear" and this:on-segment-colinear($edge, this:end($other), $tolerance)) or
        (: $edge.start colinear with and on segment $other :)
        ($o3="colinear" and this:on-segment-colinear($other, this:start($edge), $tolerance)) or
        (: $edge.end colinear with and on segment $other :)
        ($o4="colinear" and this:on-segment-colinear($other, this:end($edge), $tolerance))
      )
    '/>
  </xsl:function>

  <!--
    : this:linear-intersects()
    : Edge (straight) intersects some other region kind
    :-->
  <xsl:function name="this:linear-intersects" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind = "point") then this:on-segment($edge, $other, $tolerance)
      else if ($kind = "edge") then this:edge-intersects-edge($edge, $other, $tolerance) 
      else if ($kind = "quad") then this:quad-intersects($other, $edge, $tolerance)
      else if ($kind = "cubic") then this:cubic-intersects($other, $edge, $tolerance)
      else if ($kind = "arc") then this:arc-intersects($other, $edge, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "linear-intersects"))
    '/>
  </xsl:function>

  <!--
    : this:linear-intersects()
    :-->
  <xsl:function name="this:linear-intersects" as="xs:boolean">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:linear-intersects($edge, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:quad-intersects()
    : Quad edge intersects some other region kind
    :-->
  <xsl:function name="this:quad-intersects" as="xs:boolean">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind = "point") then this:on-quad($quad, $other, $tolerance)
      else if ($kind = "edge") then exists(this:quad-edge-intersection-ts($quad, $other, $tolerance))
      else if ($kind = "quad") then (
        if (box:intersects-box(this:bounding-box($quad), this:bounding-box($other), $tolerance)) then (
          if (this:quad-length($quad) &lt; $config:ε and
              this:quad-length($other) &lt; $config:ε)
          then true()
          else (
            let $quadsubs := this:slice-quad($quad, 0.5) return
            let $othersubs := this:slice-quad($other, 0.5)
            return (
              some $subquad in $quadsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subquad, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "cubic") then this:cubic-intersects($other, $quad, $tolerance)
      else if ($kind = "arc") then this:arc-intersects($other, $quad, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "quad-intersects"))
    '/>
  </xsl:function>

  <!--
    : this:quad-intersects()
    :-->
  <xsl:function name="this:quad-intersects" as="xs:boolean">
    <xsl:param name="quad" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:quad-intersects($quad, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:cubic-intersects()
    : Cubic edge intersects some other region kind
    :-->
  <xsl:function name="this:cubic-intersects" as="xs:boolean">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind = "point") then (
        this:on-cubic($cubic, $other, $tolerance)
      )
      else if ($kind = "edge") then (
        if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other), $tolerance)) then (
          if (this:linear-length($other) &lt; $tolerance and
              this:cubic-length($cubic, 4) &lt; $tolerance)
          then true()
          else (
            let $cubicsubs := this:slice-cubic($cubic, 0.5) return
            let $othersubs := this:slice-linear($other, 0.5)
            return (
              some $subcubic in $cubicsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subcubic, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "quad") then (
        if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other), $tolerance)) then (
          if (this:quad-length($other) &lt; $tolerance and
              this:cubic-length($cubic, 4) &lt; $tolerance)
          then true()
          else (
            let $cubicsubs := this:slice-cubic($cubic, 0.5) return
            let $othersubs := this:slice-quad($other, 0.5)
            return (
              some $subcubic in $cubicsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subcubic, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "cubic") then (
        if (box:intersects-box(this:bounding-box($cubic), this:bounding-box($other), $tolerance)) then (
          if (this:cubic-length($cubic, 4) &lt; $tolerance and
              this:cubic-length($other, 4) &lt; $tolerance)
          then true()
          else (
            let $cubicsubs := this:slice-cubic($cubic, 0.5) return
            let $othersubs := this:slice-cubic($other, 0.5)
            return (
              some $subcubic in $cubicsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subcubic, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "arc") then this:arc-intersects($other, $cubic, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "cubic-intersects"))
    '/>
  </xsl:function>

  <!--
    : this:cubic-intersects()
    :-->
  <xsl:function name="this:cubic-intersects" as="xs:boolean">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:cubic-intersects($cubic, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:arc-intersects()
    : Arc edge intersects some other region kind
    :-->
  <xsl:function name="this:arc-intersects" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      if (point:same(this:start($arc), this:end($arc), $tolerance)) then (
        if (this:arc-large($arc)) then (
          this:edge-intersects(this:arc-circle(this:normalize-arc($arc)), $other, $tolerance)
        ) else (
          this:edge-intersects(this:start($arc), $other, $tolerance)
        )
      ) else let $kind := this:kind($other) return
      if ($kind = "point") then this:on-arc($arc, $other, $tolerance)
      else if ($kind = "edge") then
        let $arc := this:normalize-arc($arc) return
        (: https://math.stackexchange.com/questions/228841/how-do-i-calculate-the-intersections-of-a-straight-line-and-a-circle :)
        (: line y=mx + x; m = (y2-y1)/(x2-x1), c=y1 :)
        let $r := this:arc-radius($arc) return
        let $p := point:px(this:arc-center($arc)) return
        let $q := point:py(this:arc-center($arc))  return
        let $x1 := point:px(this:start($other)) return
        let $y1 := point:py(this:start($other)) return
        let $x2 := point:px(this:end($other)) return
        let $y2 := point:py(this:end($other)) return
        let $pts := 
           if (abs($x1 - $x2) &lt; $config:ε) then ( (: vertical line :)
             let $k := $x1 return
             let $A := 1 return
             let $B := -2*$q return
             let $C := $p*$p + $q*$q - $r*$r - 2*$k*$p + $k*$k 
             return (
               if ($B*$B >= 4*$A*$C) then (
                 for $y in roots:quadratic-real-roots($A, $B, $C)
                 return point:point($k, $y)
               ) else ()
             )
           ) else (
             let $m := ($y2 - $y1) div ($x2 - $x1) return
             let $c := $y1 - $m*$x1 return
             let $A := $m*$m + 1 return
             let $B := 2*($m*$c - $m*$q - $p) return
             let $C := $q*$q - $r*$r + $p*$p - 2*$c*$q + $c*$c
             return (
               if ($B*$B >= 4*$A*$C) then (
                 for $x in roots:quadratic-real-roots($A, $B, $C)
                 return point:point($x, $m*$x + $c)
               ) else ()
             )
           )
        return (
          some $pt in $pts 
          satisfies this:on-segment-colinear($other, $pt, $tolerance) and
                    this:on-arc($arc, $pt, $tolerance)
        )
      else if ($kind = "quad") then (
        if (box:intersects-box(this:bounding-box($arc), this:bounding-box($other), $tolerance)) then (
          if (this:arc-length($arc) &lt; $tolerance and
              this:quad-length($other) &lt; $tolerance)
          then (
            true()
          ) else (
            let $arcsubs := this:slice-arc($arc, 0.5) return
            let $othersubs := this:slice-quad($other, 0.5)
            return (
              some $subarc in $arcsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subarc, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "cubic") then (
        if (box:intersects-box(this:bounding-box($arc), this:bounding-box($other), $tolerance)) then (
          if (this:arc-length($arc) &lt; $tolerance and
              this:cubic-length($other, 4) &lt; $tolerance)
          then true()
          else (
            let $arcsubs := this:slice-arc($arc, 0.5) return
            let $othersubs := this:slice-cubic($other, 0.5)
            return (
              some $subarc in $arcsubs
              satisfies (
                some $subother in $othersubs
                satisfies this:edge-intersects($subarc, $subother, $tolerance)
              )
            )
          )
        ) else (
          false()
        )
      )
      else if ($kind = "arc") then
        let $arc := this:normalize-arc($arc)
        return (
          ellipse:ellipse-intersects-ellipse(this:arc-circle($arc), this:arc-circle($other), $tolerance) and
          (
            (: https://stackoverflow.com/questions/3349125/circle-circle-intersection-points :)
            let $d := point:distance(this:arc-center($arc), this:arc-center($other))
            return if ($d &lt;= $tolerance) then (
              some $pt in (this:start($other), this:end($other))
              satisfies this:on-arc($arc, $pt, $tolerance)
            ) else (
              let $r1 := this:arc-radius($arc) return
              let $p1 := this:arc-center($arc) return
              let $x1 := point:px($p1) return
              let $y1 := point:py($p1) return
              let $r2 := this:arc-radius($other) return
              let $p2 := this:arc-center($other) return
              let $x2 := point:px($p2) return
              let $y2 := point:py($p2) return
              let $l := ($r1*$r1 - $r2*$r2 + $d*$d) div (2*$d) return
              let $h := math:sqrt($r1*$r1 - $l*$l) return
              let $dp := $p2=>point:sub($p1) return
              let $dx := $x2 - $x1 return
              let $dy := $y2 - $y1 return
              let $mid := $p1=>point:add($dp=>point:times($l div $d)) return
              let $pts := (
                $mid=>point:add($dp=>point:times($h div $d)),
                $mid=>point:sub($dp=>point:times($h div $d))
              )
              return (
                some $pt in $pts 
                satisfies this:on-arc($arc, $pt, $tolerance) and
                          this:on-arc($other, $pt, $tolerance)
              )
            )
          )
        )
      else errors:error("GEOM-BADREGION", ($other, "arc-intersects"))
    '/>
  </xsl:function>

  <!--
    : this:arc-intersects()
    :-->
  <xsl:function name="this:arc-intersects" as="xs:boolean">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:arc-intersects($arc, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:edge-intersects()
    :-->
  <xsl:function name="this:edge-intersects" as="xs:boolean">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:param name="tolerance" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "point") then this:edge-intersects($other, $region, $tolerance)
      else if ($kind = "edge") then this:linear-intersects($region, $other, $tolerance)
      else if ($kind = "quad") then this:quad-intersects($region, $other, $tolerance)
      else if ($kind = "cubic") then this:cubic-intersects($region, $other, $tolerance)
      else if ($kind = "arc") then this:arc-intersects($region, $other, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "edge-intersects"))
    '/>
  </xsl:function>

  <!--
    : this:edge-intersects()
    :-->
  <xsl:function name="this:edge-intersects" as="xs:boolean">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:param name="other" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:edge-intersects($region, $other, $config:tolerance)
    '/>
  </xsl:function>

  <!--
    : this:tangent()
    : Return the tangent vector to the given point on the edge (2D)
    :-->
  <xsl:function name="this:tangent" as="map(xs:string, item()*)">
    <xsl:param name="region" as="map(xs:string, item()*)"/>
    <xsl:param name="t" as="xs:double"/>
    <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "edge") then (
        this:end($region)=>point:sub(this:start($region))=>point:normalize()
      )
      else if ($kind = "quad") then (
        (: 
       : Derivative:
       : P0 = start, P1 = control, P2 = end 
       : B(t) = (1-t)²·P0 + 2(1-t)t·P1 + t²·P2 
       : B′(t) = (1-t)·A + t·B
       : B′(t) = a(1-t) + b where a = 2(B-A), b=2(C-A)
       : A=P0 B=P1
       :)
        let $P0 := this:start($region) return
        let $P1 := this:controls($region)[1] return
        let $P2 := this:end($region)
        return (
          point:map3(
            function($p0, $p1, $p2) {
              let $A := $p0 return
              let $B := $p1 return
              let $C := $p2 return
              let $a := 2*($B - $A) return
              let $b := 2*($C - $B)
              return $a*(1 - $t) + $b
            },
            $P0, $P1, $P2
          )=>point:normalize()
        )
      )
      else if ($kind = "cubic") then (
        (: 
       : Derivative 
       : Cubic P1 to P4 with controls P2, P3 
       : B(t) = (1-t)³·P1 + 3(1-t)²t·P2 + 3(1-t)t²·P3 + t³·P2 
       : B′(t) = (1-t)²·A + 2(1-t)t·B + t²·C 
       : A=3(P2-P1), B=3(P3-P2), C=3(P4-P3)
       : B′(t) = at² + bt + c = 0
       : a = A - 2B + C = 3(-P1 + 3P2 - 3P3 + P4)
       : b = 2(B - A) = 6(P1 - 2P2 + P3)
       : c = A = 3(P2 - P1)
       :)
        let $P1 := this:start($region) return
        let $P2 := this:controls($region)[1] return
        let $P3 := this:controls($region)[2] return
        let $P4 := this:end($region)
        return (
          point:map4(
            function($p1, $p2, $p3, $p4) {
              let $a := 3*(-$p1 + 3*$p2 - 3*$p3 + $p4) return
              let $b := 6*($p1 - 2*$p2 + $p3) return
              let $c := 3*($p2 - $p1) 
              return $a*$t*$t + $b*$t + $c
            },
            $P1, $P2, $P3, $P4
          )
        )=>point:normalize()
      )
      else errors:error("GEOM-NOTIMPLEMENTED", ("tangent", this:kind($region)))
    '/>
  </xsl:function>

  <!--
    :======================================================================
    : Interpolations
    :======================================================================
    :-->

  <!--
    : this:interpolate-edge-using()
    :-->
  <xsl:function name="this:interpolate-edge-using" as="map(xs:string, item()*)*">
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
    <xsl:sequence select='
      let $from := this:start($edge) return
      let $to := this:end($edge) return
      let $d := point:max-dimension(($from, $to)) return
      for $t in sort($divisions($edge))
      return (
        point:map2(
          function ($a, $b) {
            $a + $t * ($b - $a)
          },
          $from,
          $to,
          $d
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-edge()
    : Interpolate a linear edge.
    : 
    : $n: number of interpolating points
    : $edge: the edge
    : $exclusive: exclude end point? default=false
    :-->
  <xsl:function name="this:interpolate-edge" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:param name="exclusive" as="xs:boolean"/>
    <xsl:sequence select='
      this:interpolate-edge-using($edge,
        function($region as map(xs:string,item()*)) as xs:double* {
          util:linspace($n, 0, 1, $exclusive)
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-edge()
    :-->
  <xsl:function name="this:interpolate-edge" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="edge" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:interpolate-edge($n, $edge, false())
    '/>
  </xsl:function>

  <!--
    : this:interpolate-quadratic-using()
    :-->
  <xsl:function name="this:interpolate-quadratic-using" as="map(xs:string, item()*)*">
    <xsl:param name="quadratic" as="map(xs:string, item()*)"/>
    <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
    <xsl:sequence select='
      let $from := $quadratic=>this:start() return
      let $to := $quadratic=>this:end() return
      let $controls := $quadratic=>this:controls() return
      let $d := point:max-dimension(($from, $to, $controls[1])) return
      for $t in sort($divisions($quadratic))
      return (
        point:map3(
          function($a,$b,$c) {
            (1 - $t)*(1 - $t)*$a +
            2*(1 - $t)*$t*$c +
            $t*$t*$b
          },
          $from, $to, $controls[1],
          $d
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-quadratic()
    : Interpolate a quadratic edge.
    : 
    : B(t)=(1-t)²P0 + 2(1-t)tP1 + t²P2, 0 <= t <= 1 
    :
    : $n: number of interpolating points
    : $quadratic the edge
    : $exclusive: exclude end point? default=false
    :-->
  <xsl:function name="this:interpolate-quadratic" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="quadratic" as="map(xs:string, item()*)"/>
    <xsl:param name="exclusive" as="xs:boolean"/>
    <xsl:sequence select='
      this:interpolate-quadratic-using($quadratic,
        function($region as map(xs:string,item()*)) {
          util:linspace($n, 0, 1, $exclusive)
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-quadratic()
    :-->
  <xsl:function name="this:interpolate-quadratic" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="quadratic" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:interpolate-quadratic($n, $quadratic, false())
    '/>
  </xsl:function>

  <!--
    : this:interpolate-cubic-using()
    :-->
  <xsl:function name="this:interpolate-cubic-using" as="map(xs:string, item()*)*">
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
    <xsl:sequence select='
      let $from := $cubic=>this:start() return
      let $to := $cubic=>this:end() return
      let $controls := $cubic=>this:controls() return
      let $d := point:max-dimension(($from, $to, $controls[1], $controls[2])) return
      for $t in sort($divisions($cubic))
      return (
        point:map4(
          function($a, $b, $c, $d) {
            (1 - $t)*(1 - $t)*(1 - $t)*$a +
            3*(1 - $t)*(1 - $t)*$t*$c +
            3*(1 - $t)*$t*$t*$d +
            $t*$t*$t*$b
          },
          $from, $to, $controls[1], $controls[2],
          $d
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-cubic()
    : Interpolate a cubic edge.
    : 
    : B(t)=(1-t)³P0 + 3(1-t)²tP1 + 3(1-t)t²P2 + t³P3, 0 <= t <= 1 
    :
    : $n: number of interpolating points
    : $cubic the edge
    : $exclusive: exclude end point? default=false
    :-->
  <xsl:function name="this:interpolate-cubic" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:param name="exclusive" as="xs:boolean"/>
    <xsl:sequence select='
      this:interpolate-cubic-using($cubic,
        function($region as map(xs:string,item()*)) {
          util:linspace($n, 0, 1, $exclusive)
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-cubic()
    :-->
  <xsl:function name="this:interpolate-cubic" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="cubic" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:interpolate-cubic($n, $cubic, false())
    '/>
  </xsl:function>

  <!--
    : this:interpolate-arc-using()
    : angle arc: break into sub-angles
    : point arc: point on circle => dest(center, angle, radius)
    : We only do angle interpolations; so convert to angle representation
    :
    : However: the center and start/end to use depends on large and flipped
    : We reflect around the line between start/end points when we drawing on
    : the arc from the other circle through the points
    : 
    : Remapped degrees run from 0 to 360
    : We need to adjust the angles to account for running backwards and
    : for crossing 0
    : If flipped then reflect, unless we swapped, in which case reflect if we
    : aren't flipped.
    :
    : Cases:
    : from=200 to=10  large => to-from = -190  swap; 10 to 200; reverse
    : from=200 to=40  large => to-from = -160  to+=360; 200 to 400
    : from=10  to=200 large => to-from = 190   10 to 200
    : from=40  to=200 large => to-from = 160   swap; from+=360; 200 to 400; reverse
    :
    : from=200 to=10  !large => to-from = -190 to+=360; 200 to 370 
    : from=200 to=40  !large => to-from = -160 swap; 40 to 200; reverse
    : from=10  to=200 !large => to-from = 190  swap; from+=360; 200 to 370; reverse
    : from=40  to=200 !large => to-from = 160  40 to 200
    :-->
  <xsl:function name="this:interpolate-arc-using" as="map(xs:string, item()*)*">
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
    <xsl:sequence select='
      let $arc := this:normalize-arc($arc) return
      let $center := $arc=>this:arc-center() return
      let $radius := $arc=>this:arc-radius() return
      let $angles := $arc=>this:arc-angles() return
      let $from-angle := $angles[1] return
      let $to-angle := $angles[2] return
      let $da := $to-angle - $from-angle
      return (
        for $t in sort($divisions($arc))
        return (
          point:destination($center, $from-angle + $t * $da, $radius)
        )
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-arc()
    : Interpolate a arc edge.
    :
    : $n: number of interpolating points
    : $arc the edge
    : $exclusive: exclude end point? default=false
    :-->
  <xsl:function name="this:interpolate-arc" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:param name="exclusive" as="xs:boolean"/>
    <xsl:sequence select='
      this:interpolate-arc-using($arc,
        function($region as map(xs:string,item()*)) {
          util:linspace($n, 0, 1, $exclusive)
        }
      )
    '/>
  </xsl:function>

  <!--
    : this:interpolate-arc()
    :-->
  <xsl:function name="this:interpolate-arc" as="map(xs:string, item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="arc" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
      this:interpolate-arc($n, $arc, false())
    '/>
  </xsl:function>

  <xsl:function name="this:interpolate-using" as="map(xs:string,item()*)*">
    <xsl:param name="edges" as="map(xs:string,item()*)*"/>
    <xsl:param name="divisions" as="function(item()) as xs:double*"/>
    <xsl:sequence select='
    for $edge in $edges return
    let $kind := this:kind($edge) return
    if ($kind = "edge") then this:interpolate-edge-using($edge, $divisions)
    else if ($kind = "arc") then this:interpolate-arc-using($edge, $divisions)
    else if ($kind = "quad") then this:interpolate-quadratic-using($edge, $divisions)
    else if ($kind = "cubic") then this:interpolate-cubic-using($edge, $divisions)
    else $edge
    '/>
  </xsl:function>

  <!--
   : interpolate()
   : Interpolate the edges; return the interpolated points.
   :
   : $n: number of points of interpolation for each edge
   : $regions: the set of regions
   : $exclusive: include end point? default=false
   :-->
  <xsl:function name="this:interpolate" as="map(xs:string,item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="edges" as="map(xs:string,item()*)*"/>
    <xsl:param name="exclusive" as="xs:boolean"/>
    <xsl:sequence select='
    this:interpolate-using($edges,
      if ($exclusive) then (
        function($edge as map(xs:string,item()*)) {
          util:linspace($n, 0, 1, true())
        }
      ) else (
        function($edge as map(xs:string,item()*)) {
          util:linspace($n, 0, 1)
        }
      )
    )
    '/>
  </xsl:function>

  <xsl:function name="this:interpolate" as="map(xs:string,item()*)*">
    <xsl:param name="n" as="xs:integer"/>
    <xsl:param name="edges" as="map(xs:string,item()*)*"/>
    <xsl:sequence select='this:interpolate($n, $edges, false())'/>
  </xsl:function>
  
  <xsl:function name="this:map-command" as="xs:string">
    <xsl:param name="kind" as="xs:string?"/>
    <xsl:param name="relative" as="xs:string?"/>
    <xsl:sequence select='
      if ($relative="relative") then (
        if ($kind="goto") then "m"
        else if ($kind="close") then "z"
        else if ($kind="line") then "l"
        else if ($kind="hr") then "h"
        else if ($kind="vr") then "v"
        else if ($kind="cubic") then "c"
        else if ($kind="smooth_cubic") then "s"
        else if ($kind="quad") then "q"
        else if ($kind="smooth_quad") then "t"
        else if ($kind="arc") then "a"
        else if ($kind="ellipse-arc") then "a"
        else (: line :) "l"
      )
      else ( (: absolute :)
        if ($kind="goto") then "M"
        else if ($kind="close") then "Z"
        else if ($kind="line") then "L"
        else if ($kind="hr") then "H"
        else if ($kind="vr") then "V"
        else if ($kind="cubic") then "C"
        else if ($kind="smooth_cubic") then "S"
        else if ($kind="quad") then "Q"
        else if ($kind="smooth_quad") then "T"
        else if ($kind="arc") then "A"
        else if ($kind="ellipse-arc") then "A"
        else (: line :) "L"
      )
      '/>
  </xsl:function>

  <xsl:function name="this:translate-edge" as="xs:string">
    <xsl:param name="edge" as="map(xs:string,item()*)"/>
    <xsl:sequence select='
      string-join((
        this:map-command(($edge("variety"),$edge("kind"))[1],"absolute"),
        let $kind := this:kind($edge) return
        if ($kind="arc") then (
          (: XYZZY should swap start and end for large to get proper intent :)
          this:arc-radius($edge), (: rx :)
          this:arc-radius($edge), (: ry :)
          0,                      (: rotate :)
          if (this:arc-large($edge)) then 1 else 0, (: large-arc :)
          if (this:arc-flipped($edge)) then 0 else 1 (: sweep :)
        )
        else if ($kind="ellipse-arc") then (
          $edge("rx"), (: rx :)
          $edge("ry"), (: ry :)
          0,           (:rotate :)
          if (this:arc-large($edge)) then 1 else 0, (: large-arc :)
          if (this:arc-flipped($edge)) then 0 else 1 (: sweep :)
        )
        else if ($kind="quad") then (
          point:px(this:controls($edge)[1]), (: c1.x :)
          point:py(this:controls($edge)[1])  (: c1.y :)
        )
        else if ($kind="cubic") then (
          point:px(this:controls($edge)[1]), (: c1.x :)
          point:py(this:controls($edge)[1]), (: c1.y :)
          point:px(this:controls($edge)[2]), (: c2.x :)
          point:py(this:controls($edge)[2])  (: c2.y :)
        )
        else ()
        ,
        point:px(this:end($edge)),
        point:py(this:end($edge))
      )!string(.)," ")
      '/>
  </xsl:function>

   <xsl:function name="this:draw-edge" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="end" as="map(xs:string,item()*)"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="variety" as="xs:string"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:edge kind="{$variety}"
       x="{point:px($start)}" y="{point:py($start)}"
       x2="{point:px($end)}" y2="{point:py($end)}"
     >
       <xsl:if test='point:max-dimension(($start, $end)) > 2'>
         <xsl:attribute name="z" select='point:pz($start)'/>
         <xsl:attribute name="z2" select='point:pz($end)'/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:if test='$d gt 0 and not($style-properties=>map:contains("width"))'>
         <xsl:attribute name="width" select="$d"/>
       </xsl:if>
     </art:edge>
   </xsl:function>

   <xsl:function name="this:draw-quad" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="end" as="map(xs:string,item()*)"/>
     <xsl:param name="control" as="map(xs:string,item()*)"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:edge kind="quad"
       x="{point:px($start)}" y="{point:py($start)}"
       c1x="{point:px($control)}" c1y="{point:py($control)}"
       x2="{point:px($end)}" y2="{point:py($end)}"
     >
       <xsl:if test='point:max-dimension(($start, $end)) > 2'>
         <xsl:attribute name="z" select='point:pz($start)'/>
         <xsl:attribute name="z2" select='point:pz($end)'/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:if test='$d gt 0 and not($style-properties=>map:contains("width"))'>
         <xsl:attribute name="width" select="$d"/>
       </xsl:if>
     </art:edge>
   </xsl:function>

   <xsl:function name="this:draw-cubic" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="end" as="map(xs:string,item()*)"/>
     <xsl:param name="controls" as="map(xs:string,item()*)*"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:edge kind="quad"
       x="{point:px($start)}" y="{point:py($start)}"
       c1x="{point:px($controls[1])}" c1y="{point:py($controls[1])}"
       c2x="{point:px($controls[2])}" c2y="{point:py($controls[2])}"
       x2="{point:px($end)}" y2="{point:py($end)}"
     >
       <xsl:if test='point:max-dimension(($start, $end)) > 2'>
         <xsl:attribute name="z" select='point:pz($start)'/>
         <xsl:attribute name="z2" select='point:pz($end)'/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:if test='$d gt 0 and not($style-properties=>map:contains("width"))'>
         <xsl:attribute name="width" select="$d"/>
       </xsl:if>
     </art:edge>
   </xsl:function>

   <xsl:function name="this:draw-arc" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="end" as="map(xs:string,item()*)"/>
     <xsl:param name="r" as="xs:double"/>
     <xsl:param name="flipped" as="xs:boolean"/>
     <xsl:param name="large" as="xs:boolean"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:edge kind="arc"
       x="{point:px($start)}" y="{point:py($start)}"
       x2="{point:px($end)}" y2="{point:py($end)}"
       r="{$r}"
     >
       <!-- no zs on arcs -->
       <xsl:if test='$flipped'>
         <xsl:attribute name="flipped" select='"true"'/>
       </xsl:if>
       <xsl:if test='$large'>
         <xsl:attribute name="large" select='"true"'/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:if test='$d gt 0 and not($style-properties=>map:contains("width"))'>
         <xsl:attribute name="width" select="$d"/>
       </xsl:if>
     </art:edge>
   </xsl:function>

   <xsl:function name="this:draw-ellipse-arc" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="end" as="map(xs:string,item()*)"/>
     <xsl:param name="rx" as="xs:double"/>
     <xsl:param name="ry" as="xs:double"/>
     <xsl:param name="flipped" as="xs:boolean"/>
     <xsl:param name="large" as="xs:boolean"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:edge kind="ellipse-arc"
       x="{point:px($start)}" y="{point:py($start)}"
       x2="{point:px($end)}" y2="{point:py($end)}"
       rx="{$rx}" ry="{$ry}"
     >
       <!-- no zs on arcs -->
       <xsl:if test='$flipped'>
         <xsl:attribute name="flipped" select='"true"'/>
       </xsl:if>
       <xsl:if test='$large'>
         <xsl:attribute name="large" select='"true"'/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:if test='$d gt 0 and not($style-properties=>map:contains("width"))'>
         <xsl:attribute name="width" select="$d"/>
       </xsl:if>
     </art:edge>
   </xsl:function>
   
   <xsl:function name="this:draw-stroke" as="element()" visibility="private">
     <xsl:param name="item" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:stroke-path>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:sequence select='$drawing("draw:draw")($item, map {}, $drawing)'/>
     </art:stroke-path>
   </xsl:function>
   
   <xsl:function name="this:draw-svg-path" as="element()" visibility="private">
     <xsl:param name="path" as="xs:string"/>
     <xsl:param name="d" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
      <svg:path d="{$path}">
        <xsl:sequence select='util:as-attributes($style-properties)'/>
        <xsl:if test='$d gt 0 and not($style-properties=>map:contains("stroke-width"))'>
          <xsl:attribute name="stroke-width" select="$d"/>
        </xsl:if>
      </svg:path>
   </xsl:function>

   <xsl:variable name="this:draw-edge-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("variety","type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-edge(
                this:start($item)=>point:snap(), this:end($item)=>point:snap(), 
                this:weight($item),
                ($item=>map:get("variety"),"line")[1],
                $style-properties
              )
            )
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("variety","type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-edge(
                this:start($item), this:end($item),
                this:weight($item),
                ($item=>map:get("variety"),"line")[1],
                $style-properties
              )
            )
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-quad-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-quad(
                this:start($item)=>point:snap(), this:end($item)=>point:snap(), 
                this:controls($item)=>point:snap(),
                this:weight($item),
                $style-properties
              )
            )
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-quad(
                this:start($item), this:end($item), this:controls($item),
                this:weight($item),
                $style-properties
              )
            )
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-cubic-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-cubic(
                this:start($item)=>point:snap(), this:end($item)=>point:snap(), 
                this:controls($item)=>point:snap(),
                this:weight($item),
                $style-properties
              )
            )
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),("type")))) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item=>map:remove("type"), $style-properties, $drawing)
            ) else (
              this:draw-cubic(
                this:start($item), this:end($item), this:controls($item),
                this:weight($item),
                $style-properties
              )
            )
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-arc-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),"type"))) return
          let $center := $item=>this:arc-center() return
          let $r := $item=>this:arc-radius() return
          let $points := $item=>this:arc-ends() return
          let $angles := $item=>this:arc-angles() return
          let $start :=
            if (exists($points)) then (
              $points[1]
            ) else (
              let $from := util:radians($angles[1])
              return
              point:point(
                point:px($center)+math:cos($from)*$r,
                point:py($center)+math:sin($from)*$r
              )
            )
          return
          let $end :=
            if (exists($points)) then (
              $points[2]
            ) else (
              let $to := util:radians($angles[2])
              return
              point:point(
                point:px($center)+math:cos($to)*$r,
                point:py($center)+math:sin($to)*$r
              )
            )
          return
          let $d := this:weight($item)  return
          let $flipped := ($item=>this:arc-flipped(),false())[1] return
          let $large := ($item=>this:arc-large(),false())[1] return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item, $style-properties, $drawing)
            ) else (
              this:draw-arc(
                $start=>point:snap(), $end=>point:snap(),
                round-half-to-even($r),
                $flipped, $large,
                $d,
                $style-properties
              )
            )
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),"type"))) return
          let $center := $item=>this:arc-center() return
          let $r := $item=>this:arc-radius() return
          let $points := $item=>this:arc-ends() return
          let $angles := $item=>this:arc-angles() return
          let $start :=
            if (exists($points)) then (
              $points[1]
            ) else (
              let $from := util:radians($angles[1])
              return
              point:point(
                point:px($center)+math:cos($from)*$r,
                point:py($center)+math:sin($from)*$r
              )
            )
          return
          let $end :=
            if (exists($points)) then (
              $points[2]
            ) else (
              let $to := util:radians($angles[2])
              return
              point:point(
                point:px($center)+math:cos($to)*$r,
                point:py($center)+math:sin($to)*$r
              )
            )
          return
          let $d := this:weight($item) return
          let $flipped := ($item=>this:arc-flipped(),false())[1] return
          let $large := ($item=>this:arc-large(),false())[1] return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item, $style-properties, $drawing)
            ) else (
              this:draw-arc(
                $start, $end,
                $r,
                $flipped, $large,
                $d,
                $style-properties
              )
            )
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-ellipse-arc-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),"type"))) return
          let $start := $item=>this:start()=>point:snap() return
          let $end := $item=>this:end()=>point:snap() return
          let $d := $item=>this:weight() return
          let $flipped := ($item=>this:arc-flipped(),false())[1] return
          let $large := ($item=>this:arc-large(),false())[1] return
          let $ellipse := $item=>this:arc-ellipse() return
          let $rotation := ellipse:rotation($ellipse) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item, $style-properties, $drawing)
            ) else (
              (: Note: not handling rotations here: handled in main draw() function, sorta :)
              this:draw-ellipse-arc(
                $start, $end, 
                round-half-to-even(ellipse:rx($ellipse)), round-half-to-even(ellipse:ry($ellipse)), 
                $flipped, $large,
                $d,
                $style-properties
              )
            )
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($properties, util:exclude(this:property-map($item),"type"))) return
          let $start := $item=>this:start() return
          let $end := $item=>this:end() return
          let $d := $item=>this:weight() return
          let $flipped := ($item=>this:arc-flipped(),false())[1] return
          let $large := ($item=>this:arc-large(),false())[1] return
          let $ellipse := $item=>this:arc-ellipse() return
          let $rotation := ellipse:rotation($ellipse) return
          let $type := ($item=>map:get("type"),"")[1]
          return (
            if ($type="stroke") then (
              this:draw-stroke($item, $style-properties, $drawing)
            ) else (
              (: Note: not handling rotations here: handled in main draw() function, sorta :)
              this:draw-ellipse-arc(
                $start, $end, 
                ellipse:rx($ellipse), ellipse:ry($ellipse), 
                $flipped, $large,
                $d,
                $style-properties
              )
            )
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-svg-edge-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAW-SNAPPED) then (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(util:exclude(this:property-map($item),("variety"))))) return
          let $path := 
            this:map-command("goto","absolute")||" "||
            point:x(this:start($item))||" "||
            point:y(this:start($item))||" "||
            this:translate-edge($item)
          return (
            this:draw-svg-path($path, this:weight($item), $style-properties)
          )
        }
      ) else (
        function(
          $item as map(xs:string,item()*),
          $properties as map(xs:string,item()*),
          $drawing as map(xs:string,function(*)?)
        ) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(util:exclude(this:property-map($item),("variety"))))) return
          let $path := 
            this:map-command("goto","absolute")||" "||
            point:px(this:start($item))||" "||
            point:py(this:start($item))||" "||
            this:translate-edge($item)
          return (
            this:draw-svg-path($path, this:weight($item), $style-properties)
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='
     if ($config:DRAWING-METHOD="art") then (
       let $kind := this:kind($item) return
       if ($kind="edge") then $this:draw-edge-impl($item, $properties, $drawing)
       else if ($kind="quad") then $this:draw-quad-impl($item, $properties, $drawing)
       else if ($kind="cubic") then $this:draw-cubic-impl($item, $properties, $drawing)
       else if ($kind="arc") then $this:draw-arc-impl($item, $properties, $drawing)
       else if ($kind="ellipse-arc") then $this:draw-ellipse-arc-impl($item, $properties, $drawing)
       else ()
     ) else (
       $this:draw-svg-edge-impl($item, $properties, $drawing)
     )
     '/>
   </xsl:function>

 </xsl:package>

<!--edge.xsl-->
