<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Graphs: with indices for finding connected edges
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/graph"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/geometric/graph"
             exclude-result-prefixes="this xs math map array config errors util point edge">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "points", "edges", "from-index", "to-index")'/>
   </xsl:variable>

   <!--
    : this:graph()
    : Make a graph structure
    :   $points: vertices of the graph
    :   $edges: edges of the graph (edge or one of its subtypes)
    :   $properties: additional properties that will drive rendering
    : The points and edges should be unique, but we don't check that
    : There should be only one edge between each pair of points
    : Properties of significance:
    :   edge-mode: one of "graph", "circles", "sand", "brush", "stroke"; kind 
    :     of stroke default=graph
    :   no-vertices: boolean; render vertices? default=true
    :   no-edges: boolean; render edges? default=true
    :   labels: map from point to label
    :   class, colour, transparency: direct styling overrides
    :
    : { 
    :    kind: "graph",
    :    points: (pt1, pt2, ...)
    :    edges: (edge1, edge2, ...)
    :    from-index: {
    :      u1: {
    :        v1: edge(u1,v1)
    :        v2: edge(u2,v2)                    
    :       ...
    :      },
    :     ...
    :    },
    :    to-index: {
    :      v1: {
    :        u1: edge(u1,v1),
    :        u2: edge(u2,v1)
    :       ...
    :      },
    :     ...
    :    }
    : }
    :-->
   <xsl:function name="this:graph" as="map(xs:string, item()*)">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $from-index :=
        fold-left($edges, map {},
          function($index as map(xs:string,item()*), $edge as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $u-key := point:quote(edge:start($edge)) return
            let $v-key := point:quote(edge:end($edge)) return
            let $u-entry := ($index($u-key), map{})[1]
            return $index=>map:put($u-key, $u-entry=>map:put($v-key, $edge))
          }
        )
      return
      let $to-index :=
        fold-left($edges, map {},
          function($index as map(xs:string,item()*), $edge as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $u-key := point:quote(edge:start($edge)) return
            let $v-key := point:quote(edge:end($edge)) return
            let $v-entry := ($index($v-key),map{})[1]
            return $index=>map:put($v-key, $v-entry=>map:put($u-key, $edge))
          }
        )
      return
        util:merge-into($properties,
          map {
            "kind": "graph",
            "points": $points,
            "edges": $edges,
            "from-index": $from-index,
            "to-index": $to-index
          }
        )
    '/>
   </xsl:function>

   <!--
    : this:graph()
    :-->
   <xsl:function name="this:graph" as="map(xs:string, item()*)">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:graph($points, $edges, map {})
    '/>
   </xsl:function>

   <!--
    : this:graph()
    :-->
   <xsl:function name="this:graph" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:graph(
        point:distinct(($edges!edge:start(.), $edges!edge:end(.))),
        $edges,
        map {}
      )
    '/>
   </xsl:function>

   <!--
    : this:edges()
    :-->
   <xsl:function name="this:edges" as="map(xs:string, item()*)*">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $graph("edges")
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $graph("points")
    '/>
   </xsl:function>

   <!--
    : this:points()
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $graph("points")!point:as-dimension(.,2)
    '/>
   </xsl:function>

   <!--
    : this:has-edge()
    :-->
   <xsl:function name="this:has-edge" as="xs:boolean">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      exists(
        $graph("from-index")=>
          map:get(point:quote(edge:start($edge)))=>
            map:get(point:quote(edge:end($edge))))
    '/>
   </xsl:function>

   <!--
    : this:edges-from()
    :-->
   <xsl:function name="this:edges-from" as="map(xs:string, item()*)*">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:map-entries($graph("from-index")(point:quote($point)))
    '/>
   </xsl:function>

   <!--
    : this:edges-to()
    :-->
   <xsl:function name="this:edges-to" as="map(xs:string, item()*)*">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:map-entries($graph("to-index")(point:quote($point)))
    '/>
   </xsl:function>

   <!--
    : this:add-edge()
    : Add an edge to a graph. Note: we're not checking that these points exist
    : in the graph; add them first if they are new
    :   $graph: graph we're adding an edge to
    :   $edge: edge we're adding
    :-->
   <xsl:function name="this:add-edge" as="map(xs:string, item()*)">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $u-key := point:quote(edge:start($edge)) return
      let $v-key := point:quote(edge:end($edge)) return
      let $u-entry := ($graph("from-index")($u-key),map {})[1] return
      let $v-entry := ($graph("to-index")($v-key),map {})[1]
      return (
        $graph=>
          map:put("edges",
            ($graph("edges"), $edge)
          )=>
          map:put("from-index",
            $graph("from-index")=>map:put($u-key, $u-entry=>map:put($v-key, $edge))
          )=>
          map:put("to-index",
            $graph("to-index")=>map:put($v-key, $v-entry=>map:put($u-key, $edge))
          )
      )
    '/>
   </xsl:function>

   <!--
    : this:add-edges()
    :-->
   <xsl:function name="this:add-edges" as="map(xs:string, item()*)">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $from-index :=
        fold-left($edges, $graph("from-index"),
          function($index as map(xs:string,item()*), $edge as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $u-key := point:quote(edge:start($edge)) return
            let $v-key := point:quote(edge:end($edge)) return
            let $u-entry := ($index($u-key), map{})[1]
            return $index=>map:put($u-key, $u-entry=>map:put($v-key, $edge))
          }
        )
      return
      let $to-index :=
        fold-left($edges, $graph("to-index"),
          function($index as map(xs:string,item()*), $edge as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $u-key := point:quote(edge:start($edge)) return
            let $v-key := point:quote(edge:end($edge)) return
            let $v-entry := ($index($v-key),map{})[1]
            return $index=>map:put($v-key, $v-entry=>map:put($u-key, $edge))
          }
        )
      return (
        $graph=>
          map:put("edges", ($graph("edges"), $edges))=>
          map:put("from-index", $from-index)=>
          map:put("to-index", $to-index)
      )
    '/>
   </xsl:function>

   <!--
    : this:add-points()
    :-->
   <xsl:function name="this:add-points" as="map(xs:string, item()*)">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      $graph=>map:put("points", point:distinct(($graph("points"), $points)))
    '/>
   </xsl:function>

   <!--
    : this:add-node-label()
    :-->
   <xsl:function name="this:add-node-label" as="map(xs:string, item()*)">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="label" as="item()"/>
      <xsl:sequence select='
      let $labels := $graph("labels")
      return (
        if (exists($labels)) then (
          $labels=>map:put(point:quote($point), $label)
        ) else (
          $graph=>map:put("labels", map {point:quote($point): $label})
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:add-node-labels()
    :-->
   <xsl:function name="this:add-node-labels" as="map(xs:string, item()*)">
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="labels" as="item()*"/>
      <xsl:sequence select='
      if (count($points) != count($labels)) then (
        errors:error("GEOM-BADLABELS", (count($points), count($labels)))
      ) else (
        let $graph-labels :=
          fold-left(1 to count($points), ($graph("labels"),map{})[1],
            function($graph-labels as map(xs:string, item()*), $i as xs:integer) as map(xs:string,item()*) {
              $graph-labels=>map:put(point:quote($points[$i]), $labels[$i])
            }
          )
        return
          $graph=>map:put("labels", $graph-labels)
      )
    '/>
   </xsl:function>

   <!--
    : this:merge-into()
    :-->
   <xsl:function name="this:merge-into" as="map(xs:string, item()*)">
      <xsl:param name="old" as="map(xs:string, item()*)"/>
      <xsl:param name="new" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $points1 := $old=>this:vertices() return
      let $points := point:distinct(($points1, ($new=>this:vertices()))) return
      let $edges1 := $old=>this:edges() return
      let $properties := util:merge-into($old=>this:property-map(), $new=>this:property-map())
      return(
        this:graph($points, $old=>this:edges(), $properties)=>
          this:add-edges($new=>this:edges())
      )
    '/>
   </xsl:function>

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:exclude($region,$this:RESERVED)
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($region=>map:keys())[not(. = $this:RESERVED)]
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($region, util:exclude($properties,$this:RESERVED))
    '/>
   </xsl:function>

   <!--
    : this:snap()
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="graphs" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $graph in $graphs return
      if ($graph("kind") = "graph") then (
        let $points := point:snap($graph=>this:vertices()) return
        let $edges := edge:snap($graph=>this:edges())
        return this:graph($points, $edges, $graph=>this:property-map())
      ) else $graph
    '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="graphs" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $graph in $graphs return
      if ($graph("kind") = "graph") then (
        let $points := point:decimal($graph=>this:vertices(), $digits) return
        let $edges := edge:decimal($graph=>this:edges(), $digits)
        return this:graph($points, $edges, $graph=>this:property-map())
      ) else $graph
    '/>
   </xsl:function>

   <!--
    : this:quote()
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="graphs" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $graph in $graphs return
        if ($graph("kind") = "graph") then "GRAPH["||edge:quote($graph=>this:edges())||"]"
        else errors:quote($graph)
        ,
        " "
      )
    '/>
   </xsl:function>

   <!--
    : this:same()
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($this("kind")=$other("kind")) and (
        if ($this("kind") = "graph") then (
          count($this=>this:vertices())=count($other=>this:vertices()) and
          count($this=>this:edges())=count($other=>this:edges()) and
          (
            let $this-vertices := $this=>this:vertices() return
            let $o-vertices := $other=>this:vertices()
            return (
              every $i in 1 to count($this-vertices)
              satisfies point:same($this-vertices[$i], $o-vertices[$i])
            )
          ) and
          (
            let $this-edges := $this=>this:edges() return
            let $o-edges := $other=>this:edges()
            return (
              every $i in 1 to count($this-edges)
              satisfies edge:same($this-edges[$i], $o-edges[$i])
            )
          )
        ) else deep-equal($this, $other)
      )
    '/>
   </xsl:function>

   <!--
    : this:mutate()
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $region in $regions return
      if ($region("kind") = "graph") then (
        let $edges := edge:mutate($region=>this:edges(), $mutate) return
        let $points := point:mutate($region=>this:vertices(), $mutate)
        return this:graph($points, $edges, $region=>this:property-map())
      ) else $region
    '/>
   </xsl:function>


   <xsl:function name="this:draw-graph" as="element()" visibility="private">
     <xsl:param name="vertices" as="map(xs:string,item()*)*"/>
     <xsl:param name="edges" as="map(xs:string,item()*)*"/>
     <xsl:param name="labels" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:graph>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:choose>
         <xsl:when test='exists($labels)'>
           <xsl:for-each select="$vertices">
             <xsl:variable name="pt" select="." as="map(xs:string,item()*)"/>
             <xsl:variable name="label" select="$labels(point:quote($pt))"/>
             <art:vertex x="{point:x($pt)}" y="{point:y($pt)}">
               <xsl:if test="$pt=>point:dimension()>2">
                 <xsl:attribute name="z" select="point:z($pt)"/>
               </xsl:if>
               <xsl:sequence select='util:as-attributes($style-properties)'/>
               <xsl:if test="exists($label) and $label ne ''">
                 <xsl:attribute name="label" select="util:quote($label)"/>
               </xsl:if>
             </art:vertex>
           </xsl:for-each>
         </xsl:when>
         <xsl:otherwise>
         <xsl:sequence select='$drawing("draw:draw")($vertices, map{}, $drawing)'/>
         </xsl:otherwise>
       </xsl:choose>
       <xsl:sequence select='$drawing("draw:draw")($edges, map {}, $drawing)'/>
     </art:graph>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into($properties, util:exclude(this:property-map($item),"labels"))
            return (
              this:draw-graph(
                this:vertices($item)!point:snap(.),  this:edges($item), 
                $item("labels"), $style-properties, $drawing
              )
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into($properties, util:exclude(this:property-map($item),"labels"))
            return (
              this:draw-graph(
                this:vertices($item), this:edges($item), 
                $item("labels"), $style-properties, $drawing
              )
            )
          }
        )
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(util:exclude(this:property-map($item),("labels","no-vertices","no-edges"))))) return
          let $no-vertices := ($item("no-vertices"),true())[1] return
          let $no-edges := ($item("no-edges"),true())[1]
          return (
            if ($no-edges) then () else $drawing("draw:draw")(this:edges($item), $style-properties, $drawing),
            if ($no-vertices) then () else $drawing("draw:draw")(this:vertices($item), $style-properties, $drawing)
          )
        }
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>
<!--graph.xsl-->
