<?xml version="1.0" encoding="UTF-8"?>
<!--
  : Path and polygon (closed path) objects
  :
  : Copyright© Mary Holstege 2020-2023
  : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/path"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/box"
             xmlns:config="http://mathling.com/core/config"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/geometric/path"
             exclude-result-prefixes="this xs math map array errors config util point edge box">

  <xsl:expose component="variable" names="this:*" visibility="public"/>
  <xsl:expose component="function" names="this:*" visibility="public"/>

  <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/box" package-version="*"/>

  <xsl:variable name="this:RESERVED" as="xs:string*">
    <xsl:sequence select='("kind", "edges", "edge-ts")'/>
    </xsl:variable>

    <!--
      : this:path()
      : Construct a path
      :
      : $edges: edges of the path, in order. Edges could be of any kind.
      : $properties: additional properties
      :-->
    <xsl:function name="this:path" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "path",
          "edges": $edges
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:path()
    :-->
   <xsl:function name="this:path" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      map {
        "kind": "path",
        "edges": $edges
      }
    '/>
   </xsl:function>

   <!--
    : this:kind()
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $region("kind")
    '/>
   </xsl:function>

   <!--
    : this:edges()
    :-->
   <xsl:function name="this:edges" as="map(xs:string, item()*)*">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $path("edges")
    '/>
   </xsl:function>

   <!--
    : this:points()
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()=>point:as-dimension(2)
        ,
        ($region=>this:edges())[last()]=>edge:end()=>point:as-dimension(2)
      )
      else if ($kind) then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()=>point:as-dimension(2)
        ,
        ($region=>this:edges())[last()]=>edge:end()=>point:as-dimension(2)
      )
      else errors:error("GEOM-BADREGION", ($region, "points"))
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()
        ,
        ($region=>this:edges())[last()]=>edge:end()
      )
      else if ($kind = "polygon") then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()
        ,
        ($region=>this:edges())[last()]=>edge:end() (: Includes closing vertex=starting vertex :)
      )
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="close" as="xs:boolean"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()
        ,
        ($region=>this:edges())[last()]=>edge:end()
      )
      else if ($kind = "polygon") then (
        for $edge in $region=>this:edges()
        return $edge=>edge:start()
        ,
        if ($close) then ($region=>this:edges())[last()]=>edge:end() else ()
      )
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
   </xsl:function>

   <!--
    : this:as-polygon()
    : as-polygon()
    : Turn a path into a polygon. 
    :
    : $region: the path
    :-->
   <xsl:function name="this:as-polygon" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then this:polygon($region=>this:edges(), $region=>this:property-map())
      else if ($kind = "polygon") then $region
      else errors:error("GEOM-BADREGION", ($region, "as-polygon"))
    '/>
   </xsl:function>

   <!--
    : this:polygon()
    : Construct a polygon from edges
    :   $edges: edges of the polygon
    :-->
   <xsl:function name="this:polygon" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:polygon($edges, map {})
    '/>
   </xsl:function>

   <!--
    : this:polygon()
    :-->
   <xsl:function name="this:polygon" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (empty($edges)) then (
        util:merge-into($properties,
          map {
            "kind": "polygon",
            "edges": ()
          }
        )
      ) else if ($edges[1]("kind")="point") then (
        (: Creating polygon from sequence of points :)
        (: An error, which we try to recover from :)
        this:polygon(
          (
            edge:to-edges($edges),
            if (not(point:same($edges[1],$edges[last()])))
            then edge:edge($edges[last()], $edges[1])
            else () 
          ) 
          ,
          $properties
        )
      ) else (
        util:merge-into($properties,
          map {
            "kind": "polygon",
            "edges": (
              $edges, (
                if (not(point:same(edge:start($edges[1]),edge:end($edges[last()])))) then (
                  edge:edge(edge:end($edges[last()]), edge:start($edges[1]))
                ) else ()
              )
            )
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:rectangle()
    :-->
   <xsl:function name="this:rectangle" as="map(xs:string, item()*)">
      <xsl:param name="min-pt" as="map(xs:string, item()*)"/>
      <xsl:param name="max-pt" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $z := avg((point:pz($min-pt),point:pz($max-pt))) return
      let $minmaxpt := point:point(point:px($min-pt), point:py($max-pt), $z) return
      let $maxminpt := point:point(point:px($max-pt), point:py($min-pt), $z)
      return
      this:polygon((
        edge:edge($min-pt, $minmaxpt),
        edge:edge($minmaxpt, $max-pt),
        edge:edge($max-pt, $maxminpt),
        edge:edge($maxminpt, $min-pt)
      ))
    '/>
   </xsl:function>

   <!--
    : this:triangle()
    :-->
   <xsl:function name="this:triangle" as="map(xs:string, item()*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:polygon((
        edge:edge($a, $b),
        edge:edge($b, $c),
        edge:edge($c, $a)
      ))
    '/>
   </xsl:function>

   <!--
    : this:triangle()
    :-->
   <xsl:function name="this:triangle" as="map(xs:string, item()*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:polygon((
        edge:edge($a, $b),
        edge:edge($b, $c),
        edge:edge($c, $a)
      ), $properties)
    '/>
   </xsl:function>

   <!--
    : this:compute-edge-ts()
    :-->
   <xsl:function name="this:compute-edge-ts" as="xs:double*" visibility="private">
      <xsl:param name="path" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      let $edges := this:edges($path) return
      let $edge-lengths := $edges!edge:length(.) return
      let $partial-lengths :=
        fold-left(
          tail($edge-lengths),
          head($edge-lengths),
          function($lengths as xs:double*, $edge-length as xs:double) as xs:double* {
            $lengths,
            $lengths[last()] + $edge-length
          }
        )
      return
      let $path-length := $partial-lengths[last()]
      return $partial-lengths!(. div $path-length) 
    '/>
   </xsl:function>

   <!--
    : this:edge-ts()
    : Accessor for the edge t values: fractions of the path length that
    : edge edge represents. Used for optimized path interpolations.
    : 
    : $path: the path
    :-->
   <xsl:function name="this:edge-ts" as="xs:double*">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($path=>map:contains("edge-ts")) then $path("edge-ts")
      else this:compute-edge-ts($path) (: Warning: SLOW! :)
    '/>
   </xsl:function>

   <!--
    : this:with-edge-ts()
    : Return the path with edge t values computed.
    : Recommended for optimized path interpolations.
    : 
    : $path: the path
    :-->
   <xsl:function name="this:with-edge-ts" as="map(xs:string, item()*)">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $path=>
        map:put("edge-ts", this:compute-edge-ts($path))
    '/>
   </xsl:function>

   <!--
    : this:edge-intersections()
    : Points of intersection, if any between path and edge
    :-->
   <xsl:function name="this:edge-intersections" as="map(xs:string, item()*)*">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="line" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $edge in this:edges($this) return edge:intersection($edge, $line)
    '/>
   </xsl:function>

<!--
  :======================================================================
  : Property management
  :======================================================================
  :-->
   
   <!--
    : this:property-map()
    : Return the annotation properties of the region as a map. Check whether this
    : is actually a path or polygon.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then util:exclude($region, $this:RESERVED)
      else if ($kind = "polygon") then util:exclude($region, $this:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the region.
    : Check whether this is actually a path or polygon.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then ($region=>map:keys())[not(. = $this:RESERVED)]
      else if ($kind="polygon") then ($region=>map:keys())[not(. = $this:RESERVED)]
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the region with some new properties and return the new region.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually a path or polygon.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "path") then 
        util:merge-into($region, util:exclude($properties,$this:RESERVED))
      else if ($kind = "polygon") then
        util:merge-into($region, util:exclude($properties,$this:RESERVED))
      else errors:error("GEOM-BADREGION", ($region, "with-properties"))
    '/>
   </xsl:function>

<!--
  :======================================================================
  : Operations
  :======================================================================
  :-->

   <!--
    : this:snap()
    : Snap the coordinates of the points, returning the region with snapped
    : (i.e. integer) points coordinates.
    :
    : $regions: the regions
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind = "path") then
        this:path(edge:snap(this:edges($region)), $region)
      else if ($kind = "polygon") then
        this:polygon(edge:snap(this:edges($region)), $region)
      else $region
    '/>
   </xsl:function>

   <!--
       : this:decimal()
       : Perform decimal rounding on all the point coordinates (see util:decimal).
       :
       : $regions: the regions to round
       : $digits: how many digits after the decimal point to keep
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind = "path") then
        this:path(edge:decimal(this:edges($region), $digits), $region)
      else if ($kind = "polygon") then
        this:polygon(edge:decimal(this:edges($region), $digits), $region)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:quote()
    : Return a string value for the path, suitable for debugging.
    : 
    : $paths: the path sequence to quote
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="paths" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $path in $paths return let $kind := this:kind($path) return
        if ($kind = "path") then
          "("||edge:quote($path=>this:edges())||")"
        else if ($kind = "polygon") then
          "("||edge:quote($path=>this:edges())||")"
        else errors:quote($path)
        ,
        " "
      )
    '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for paths, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one path or polygon
    : $other: the path or polygon to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $this-kind := this:kind($this) return
      let $other-kind := this:kind($other)
      return
      (
        (
          (($this-kind="polygon") and ($other-kind="polygon")) or
          (($this-kind="path") and ($other-kind="path"))
        ) and
        (
          let $this-edges := $this=>this:edges() return
          let $other-edges := $other=>this:edges()
          return (
            every $i in 1 to count($this-edges)
            satisfies edge:same($this-edges[$i], $other-edges[$i])
          )
        )
      ) or (
        $this-kind=$other-kind and deep-equal($this,$other)
      )
    '/>
   </xsl:function>

   <!--
    : this:mutate()
    : Run a function over a sequence of regions to produce a new sequence of
    : regions. The function maps points to points.
    : 
    : $regions: input sequence of regions
    : $mutate: function that takes a point as an argument and returns a new point
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate" as="(function(item()) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind = "path") then
        let $edges := edge:mutate($region=>this:edges(), $mutate)
        return this:path($edges, $region)
      else if ($kind = "polygon") then
        let $edges := edge:mutate($region=>this:edges(), $mutate)
        return this:polygon($edges, $region)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:reverse()
    : Reverse the direction of the path.
    : 
    : $path: the path to reverse
    :-->
   <xsl:function name="this:reverse" as="map(xs:string, item()*)">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($path) return
      if ($kind = "path") then
        this:path(
          for $edge in reverse(this:edges($path)) return edge:reverse($edge),
          $path
        )
      else if ($kind = "polygon") then
        this:polygon(
          for $edge in reverse(this:edges($path)) return edge:reverse($edge),
          $path
        )
      else $path
    '/>
   </xsl:function>

   <!--
    : this:start()
    : First point of path.
    :
    : $path: the path
    :-->
   <xsl:function name="this:start" as="map(xs:string, item()*)">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:edges($path)[1]=>edge:start()
    '/>
   </xsl:function>

   <!--
    : this:end()
    : Last point of path.
    :
    : $path: the path
    :-->
   <xsl:function name="this:end" as="map(xs:string, item()*)">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:edges($path)[last()]=>edge:end()
    '/>
   </xsl:function>

   <!--
    : this:area()
    : Area of polygon. Assumes vertices are ordered counterclockwise.
    : Will fail for certain twisty polygons. 
    :-->
   <xsl:function name="this:area" as="xs:double">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($polygon) return
      if ($kind = "polygon") then (
        let $points := this:vertices($polygon) return
        let $n := count($points) - 1
        return (
          if ($n = 1 or $n = 2) then 0
          else if ($n = 3) then (
            (: Heron&#x27;s formula :)
            let $a := point:distance($points[1], $points[2]) return
            let $b := point:distance($points[2], $points[3]) return
            let $c := point:distance($points[3], $points[1]) return
            let $s := sum(($a, $b, $c)) div 2
            return math:sqrt($s * ($s - $a) * ($s - $b) * ($s - $c))
          )
          else (
            abs( (: Sign flips for canvas geometry :)
              sum(
                for $i in 1 to $n return (
                  point:px($points[$i])*point:py($points[$i+1]) -
                  point:px($points[$i+1])*point:py($points[$i])
                )
              )
            ) div 2E0
          )
        )
      )
      else errors:error("GEOM-BADREGION", ($polygon, "area"))
    '/>
   </xsl:function>

   <!--
    : this:length()
    :-->
   <xsl:function name="this:length" as="xs:double">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      sum(
        for $edge in this:edges($path) return edge:length($edge)
      )
    '/>
   </xsl:function>

   <!--
    : this:polygon-center()
    :-->
   <xsl:function name="this:polygon-center" as="map(xs:string, item()*)">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $n := count(this:edges($polygon))
      return
      if ($n = 0) then $point:ORIGIN (: GIGO :)
      else if ($n = 1) then this:edges($polygon)[1]=>edge:edge-point(0.5)
      else if ($n = 2) then this:edges($polygon)[1]=>edge:edge-point(0.5)
      else if ($n = 3) then
        let $edges := this:edges($polygon)
        return (
          (
            edge:intersection(
              edge:edge(edge:start($edges[1]), $edges[2]=>edge:edge-point(0.5)),
              edge:edge(edge:start($edges[2]), $edges[3]=>edge:edge-point(0.5))
            ),          
            $edges[1]=>edge:edge-point(0.5)  (: In case of colinearity :)
          )=>head()
        )
      else if ($n = 4) then
        let $points := this:points($polygon)
        return (
          (
            edge:intersection(
              edge:edge($points[1], $points[3]),
              edge:edge($points[2], $points[4])
            ),
            this:edges($polygon)[1]=>edge:edge-point(0.5)  (: In case of colinearity :)
          )=>head()
        )
      else (
        (: This formula is actually only good for non-intersecting
         : straight-edge 2D polygons, so it is going to be a dubious approximation
         : for other cases.
         :)
        let $points := this:points($polygon) return
        (: Note: 1 more point than edges start=end :)
        let $area :=
          sum(
            for $i in 1 to $n return (
              point:px($points[$i])*point:py($points[$i+1]) -
              point:px($points[$i+1])*point:py($points[$i])
            )
          ) div 2
        return (
          point:point(
            sum(
              for $i in 1 to $n return (
                (point:px($points[$i]) + point:px($points[$i+1]))*
                (
                  point:px($points[$i])*point:py($points[$i+1]) -
                  point:px($points[$i+1])*point:py($points[$i])
                )
              )            
            ) div (6*$area),
            sum(
              for $i in 1 to $n return (
                (point:py($points[$i]) + point:py($points[$i+1]))*
                (
                  point:px($points[$i])*point:py($points[$i+1]) -
                  point:px($points[$i+1])*point:py($points[$i])
                )
              )
            ) div (6*$area)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:path-point()
    : Find a point at fraction $t between start and finish
    :
    : First find the edge that encompasses that fraction of the path
    : The find the edge t and use it to get the point
    : Suppose edge covers [0.3, 0.5] of path and $t = 0.4 => $edge-t = 0.5
    : ($t - $start-t) div ($end-t - $start-t)
    :
    : If you are going to compute a bunch of these, augment the path using
    : path:with-edge-ts()
    :
    : $path: the path
    : $t: fraction of path [0,1]
    :-->
   <xsl:function name="this:path-point" as="map(xs:string, item()*)">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      let $edges := this:edges($path) return
      let $edge-ts := $path=>this:edge-ts() return
      let $edge-ix := util:rangeindex($edge-ts, $t) return
      let $start-t := if ($edge-ix = 1) then 0 else $edge-ts[$edge-ix - 1] return
      let $end-t := $edge-ts[$edge-ix] return
      let $edge-t := (
        (: start-t = end-t if one of the edges has zero length, which really
         : should&#x27;t happen. If it does, just take the start of the edge
         :)
        if ($start-t = $end-t)
        then 0
        else ($t - $start-t) div ($end-t - $start-t)
      )
      return (
        edge:edge-point($edges[$edge-ix], $edge-t)
      )
    '/>
   </xsl:function>

   <!--
    : this:slice()
    : Create two paths at the given fraction of the source path. If we
    : need to slice an edge, it will be sliced per slice-edge()
    :
    : $path: the path to cut
    : $t: fraction of edge [0,1]
    :-->
   <xsl:function name="this:slice" as="map(xs:string, item()*)*">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      if (empty(this:edges($path))) then ()
      else if (count(this:edges($path)) = 1)
      then (
        for $edge in edge:slice(this:edges($path), $t)
        return this:path($edge)
      ) else (
        let $edges := this:edges($path) return
        let $edge-ts := $path=>this:edge-ts() return
        let $edge-ix := util:rangeindex($edge-ts, $t) return
        let $start-t := (if ($edge-ix = 1) then 0 else $edge-ts[$edge-ix - 1]) return
        let $end-t := $edge-ts[$edge-ix] return
        let $edge-t :=
          (: start-t = end-t if one of the edges has zero length, which really
           : should&#x27;t happen. If it does, just take the start of the edge
           :)
          if ($start-t = $end-t)
          then 0
          else ($t - $start-t) div ($end-t - $start-t)
        return
        let $slices := edge:slice($edges[$edge-ix], $edge-t) return
        let $edges1 := (
          $edges[position() &lt; $edge-ix],
          if ($edge-t=0) then ()
          else if ($edge-t=1) then $slices
          else $slices[1]
        ) return
        let $edges2 := (
          if ($edge-t=0) then $slices
          else if ($edge-t=1) then ()
          else $slices[2],
          $edges[position() > $edge-ix]
        )
        return (
          if (exists($edges1)) then this:path($edges1) else (),
          if (exists($edges2)) then this:path($edges2) else ()
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:slice()
    : Slice the path between two cut points. Will end up with three paths in
    : general: [start, p(start-t)], [p(start-t), p(end-t)], [p(end-t), end]
    : start-t <= end-t
    :
    : If start-t=end-t this is same as slice(path, start-t)
    : If start-t or end-t is at an end will get one less edge
    :
    : $path: the path to cut
    : $start-t: fraction of path [0,1]
    : $end-t: fraction of path [0,1]
    :-->
   <xsl:function name="this:slice" as="map(xs:string, item()*)*">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:param name="start-t" as="xs:double"/>
      <xsl:param name="end-t" as="xs:double"/>
      <xsl:sequence select='
      if ($start-t > $end-t) then errors:error("GEOM-BADT", ($start-t, $end-t)) else (),
      if (empty(this:edges($path))) then ()
      else if (count(this:edges($path))=1) then (
        for $edge in edge:slice(this:edges($path), $start-t, $end-t)
        return this:path($edge)
      ) else (
        if ($start-t = 1) then $path (: start-t=end-t=1 :)
        else if ($start-t = $end-t) then this:slice($path, $start-t)
        else if ($start-t = 0) then this:slice($path, $end-t)
        else (
          (: 
           : end-t needs to be adjusted for the second slice:
           : Effective end-t should be fraction of length(slice) not of length(path)
           : e * l - l1 = x * l2; l1 = s * l, l2 = l - l1 = l - s * l
           : x = (el - sl) / (s - sl) = (e - s) / (1 - s)
           :)
          let $slice := this:slice($path, $start-t) return
          let $eff-end-t := ($end-t - $start-t) div (1 - $start-t)
          return ($slice[1], this:slice($slice[2]=>this:with-edge-ts(), $eff-end-t))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:path-t()
    : Determine the path t-value of the given point.
    :
    : $path: the path 
    : $point: the point on the path
    :-->
   <xsl:function name="this:path-t" as="xs:double">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $path := $path=>this:with-edge-ts() return
      let $edges := this:edges($path) return
      let $edge-ts := $path=>this:edge-ts() return
      let $edge-ix :=
        (
          for $i in 1 to count($edges) return
          let $edge := $edges[$i] return
          let $intersects := $edge=>edge:edge-intersects($point)
          return (
            if ($intersects) then $i else ()
          )
        )=>head()
      return
      let $edge-t := $edges[$edge-ix]=>edge:edge-t($point) return
      let $start-t := if ($edge-ix = 1) then 0.0 else $edge-ts[$edge-ix - 1] return
      let $end-t := $edge-ts[$edge-ix]
      return (
        $start-t + $edge-t * ($end-t - $start-t)
      )
    '/>
   </xsl:function>

   <!--
    : this:tangent()
    : Return the tangent vector to the given point on the path (2D)
    :-->
   <xsl:function name="this:tangent" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      let $edges := this:edges($region) return
      let $edge-ts := $region=>this:edge-ts() return
      let $edge-ix := util:rangeindex($edge-ts, $t) return
      let $start-t := if ($edge-ix = 1) then 0 else $edge-ts[$edge-ix - 1] return
      let $end-t := $edge-ts[$edge-ix] return
      let $edge-t :=
        (: start-t = end-t if one of the edges has zero length, which really
         : should&#x27;t happen. If it does, just take the start of the edge
         :)
        if ($start-t = $end-t)
        then 0
        else ($t - $start-t) div ($end-t - $start-t)
      return (
        edge:tangent($edges[$edge-ix], $edge-t)
      )
    '/>
   </xsl:function>

   <!--
    : this:bounding-box()
    : Minimum box surrounding the region. Some approximation for non-linear
    : edges.
    :-->
   <xsl:function name="this:bounding-box" as="map(xs:string, item()*)">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $boxes := (
        for $region in $regions return
        if (this:kind($region) = ("path", "polygon")) then (
          let $pts := 
            if (every $edge in this:edges($region) satisfies this:kind($edge)="edge")
            then (
              this:points($region)
            ) else (
              let $boxes := this:edges($region)!edge:bounding-box(.)
              return $boxes!box:points(.)
            )
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        ) else (
          errors:error("GEOM-BADREGION", ($region, "bounding-box"))
        )
      )
      return (
        if (empty($boxes)) then box:box(0,0,0,0)
        else if (empty(tail($boxes))) then $boxes
        else (
          let $pts := $boxes!box:points(.)
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:polygon-contains-point()
    :-->
<!-- TODO: better tolerance handling -->
<!-- TODO: 3D -->
   <xsl:function name="this:polygon-contains-point" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      (: Special cases: point is a vertex or lies on a axis edge
       : count point as in. Non-axis edges below with winding numbers.
       :)
      (some $edge in this:edges($polygon) satisfies (
        let $v0 := edge:start($edge) return
        let $v1 := edge:end($edge)
        return
          point:same($point, $v0, $tolerance) or
          point:same($point, $v1, $tolerance) or
          (
            point:px($v0)=point:px($v1) and
            point:px($v0)=point:px($point) and
            util:twixt(point:py($point), point:py($v0), point:py($v1))
          ) or
          (
            point:py($v0)=point:py($v1) and
            point:py($v0)=point:py($point) and
            util:twixt(point:px($point), point:px($v0), point:px($v1))
          )
      ))
      or
      fold-left(
        this:edges($polygon),
        0,
        function($wn as xs:integer, $edge as map(xs:string,item()*)) as xs:integer {
          let $v0 := edge:start($edge) return
          let $v1 := edge:end($edge) return
          let $orientation := point:orientation($point, $v0, $v1)
          return (
            if (point:py($v0) &lt;= point:py($point)) then (
              (: point is above start of edge :)
              if (point:py($v1) > point:py($point)) then (
                (: point is below end of edge (i.e. between start and end) :)
                (: increment winding number if it is to left :)
                if ($orientation="counterclockwise")
                then $wn + 1
                else $wn
              ) else (
                $wn
              )
            ) else (
              (: point is below start of edge :)
              if (point:py($v1) &lt;= point:py($point)) then (
                (: point is above end of edge (i.e. between start and end) :)
                (: decrement winding number if it is to right :)
                if (not($orientation="counterclockwise"))
                then $wn - 1
                else $wn
              ) else (
                $wn
              )
            )
          )
        }
      ) != 0
    '/>
   </xsl:function>

   <!--
    : this:polygon-contains-point()
    :-->
   <xsl:function name="this:polygon-contains-point" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:polygon-contains-point($polygon, $point, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:simplify()
    : Douglas-Peucker simplification. Given a path returns a path, given a 
    : polygon returns a polygon.
    : https://karthaus.nl/rdp/ via Wikipedia
    :
    : $path-or-polygon: either a path, or a polygon
    : $ε: error
    :-->
   <xsl:function name="this:simplify" as="map(xs:string, item()*)">
      <xsl:param name="path-or-polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="ε" as="xs:double"/>
      <xsl:sequence select='
      let $pts := this:vertices($path-or-polygon) return
      let $n := count($pts)
      return (
        if ($n &lt;= 2) then $path-or-polygon
        else let $kind := this:kind($path-or-polygon) return
        if ($kind = "polygon") then this:polygon(edge:to-edges(this:simplify-points($pts, $ε)))
        else if ($kind = "path") then this:path(edge:to-edges(this:simplify-points($pts, $ε)))
        else $path-or-polygon
      )
    '/>
   </xsl:function>

   <!--
    : this:simplify-points()
    : Workhorse of simplify(): recursive Douglas-Peucker
    :-->
   <xsl:function name="this:simplify-points" as="map(xs:string, item()*)*">
      <xsl:param name="pts" as="map(xs:string, item()*)*"/>
      <xsl:param name="ε" as="xs:double"/>
      <xsl:sequence select='
      let $line := edge:edge($pts[1], $pts[last()]) return
      let $find-max :=
        fold-left(2 to count($pts) - 1, array{0, 1},
          function($data as array(xs:numeric), $i as xs:integer) as array(xs:numeric) {
            let $d := $line=>edge:point-distance($pts[$i])
            return (
              if ($d > $data(1)) then array {$d, $i} else $data
            )
          }
        )
      return
      let $dmax := $find-max(1) return
      let $index := $find-max(2)
      return (
        if ($dmax > $ε) then (
          this:simplify-points($pts[position() &lt;= $index], $ε)[position() &lt; last()],
          this:simplify-points($pts[position() > $index], $ε)
        ) else (
          $pts[1], $pts[last()]
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:chaikin-points()
    : Workhorse of Chaikin's corner-cutting smoothing algorithm
    : Don't use high iterations: you'll run out of heap and it doesn't really change things
    : much. Closed sequences will cut the corner from last to first. Open sequences
    : will leave the first and last points alone.
    :
    : $points: sequence of input points
    : $closed: closed or open?
    : $t: cutting ratio [0, 0.5]
    : $iterations: number of iterations
    :-->
   <xsl:function name="this:chaikin-points"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="closed" as="xs:boolean"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:param name="iterations" as="xs:integer"/>
      <xsl:sequence select='
      if (not(util:twixt($t, 0.0, 0.5))) then errors:error("ML-BADARGS", ("t", $t)) else
      if ($closed) then (
        (: Closed loop: be sure to cut corner between last and first as well :)
        fold-left(1 to $iterations, $points,
          function($points as map(xs:string,item()*)*, $i as xs:integer) as map(xs:string,item()*)* {
            if (empty($points[2])) then $points else
            let $n := count($points) return
            for $i in 1 to $n return (
              (: 
               : e.g. for t=0.25 (default)
               : p[i]*0.75 + p[i+1]*0.25
               : p[i]*0.25 + p[i+1]*0.75
               :)
              $points[$i]=>point:times(1 - $t)=>point:add($points[util:modix($i + 1, $n)]=>point:times($t)),
              $points[$i]=>point:times($t)=>point:add($points[util:modix($i + 1, $n)]=>point:times(1 - $t))
            )
          }
        )
      ) else (
        (: Open: leave start and end points alone :)
        fold-left(1 to $iterations, $points,
          function($points as map(xs:string,item()*)*, $i as xs:integer) as map(xs:string,item()*)* {
            if (empty($points[2])) then $points else
            let $n := count($points) return (
              $points[1],
              $points[1]=>point:times($t)=>point:add($points[2]=>point:times(1 - $t))
              ,
              for $i in 2 to $n - 2 return (
                (: 
                 : e.g. for t=0.25 (default)
                 : p[i]*0.75 + p[i+1]*0.25
                 : p[i]*0.25 + p[i+1]*0.75
                 :)
                $points[$i]=>point:times(1 - $t)=>point:add($points[$i + 1]=>point:times($t)),
                $points[$i]=>point:times($t)=>point:add($points[$i + 1]=>point:times(1 - $t))
              ),
              $points[$n - 1]=>point:times(1 - $t)=>point:add($points[$n]=>point:times($t)),
              $points[$n]
            )
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:chaikin()
    : Chaikin's corner-cutting algorithm.
    :
    : $region: path or polygon
    : $t: cutting ratio [0, 0.5]
    : $iterations: number of iterations
    :-->
   <xsl:function name="this:chaikin" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:param name="iterations" as="xs:integer"/>
      <xsl:sequence select='
      if (this:kind($region) = "polygon") then (
        this:polygon(edge:to-edges(
          this:chaikin-points(this:vertices($region), true(), $t, $iterations)
        ))
      ) else (
        this:path(edge:to-edges(
          this:chaikin-points(this:vertices($region), false(), $t, $iterations)
        ))
      )
    '/>
   </xsl:function>

   <!--
    : this:chaikin()
    :-->
   <xsl:function name="this:chaikin" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      this:chaikin($region, $t, 1)
    '/>
   </xsl:function>

   <!--
    : this:chaikin()
    :-->
   <xsl:function name="this:chaikin" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:chaikin($region, 0.25, 1)
    '/>
   </xsl:function>

   <xsl:function name="this:draw-stroke-path" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="edges" as="map(xs:string,item()*)*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:stroke-path>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:choose>
         <xsl:when test='$config:DENSE-EDGES'>
           <xsl:variable name="path" as="xs:string">
             <xsl:sequence select='
             edge:map-command("goto","absolute")||" "||
             point:px($start)||" "||
             point:py($start)||" "||
             string-join(
               for $edge in $edges return edge:translate-edge($edge),
               " "
             )
             '/>
           </xsl:variable>
           <xsl:attribute name="d" select="$path"/>
         </xsl:when>
         <xsl:otherwise>
         <xsl:sequence select='$drawing("draw:draw")($edges, map {}, $drawing)'/>
         </xsl:otherwise>
       </xsl:choose>
     </art:stroke-path>
   </xsl:function>

   <xsl:function name="this:draw-path" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="edges" as="map(xs:string,item()*)*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:path x="{point:px($start)}" y="{point:py($start)}">
       <xsl:if test='$start=>point:dimension()>2'>
         <xsl:attribute name="z" select="point:pz($start)"/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:choose>
         <xsl:when test='$config:DENSE-EDGES'>
           <xsl:variable name="path" as="xs:string">
             <xsl:sequence select='
             edge:map-command("goto","absolute")||" "||
             point:px($start)||" "||
             point:py($start)||" "||
             string-join(
               for $edge in $edges return edge:translate-edge($edge),
               " "
             )
             '/>
           </xsl:variable>
           <xsl:attribute name="d" select="$path"/>
         </xsl:when>
         <xsl:otherwise>
         <xsl:sequence select='$drawing("draw:draw")($edges, map {}, $drawing)'/>
         </xsl:otherwise>
       </xsl:choose>
     </art:path>
   </xsl:function>

      <xsl:function name="this:draw-polygon" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="edges" as="map(xs:string,item()*)*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:polygon>
       <xsl:if test='$start=>point:dimension()>2'>
         <xsl:attribute name="z" select="point:pz($start)"/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:choose>
         <xsl:when test='$config:DENSE-EDGES'>
           <xsl:variable name="path" as="xs:string">
             <xsl:sequence select='
             edge:map-command("goto","absolute")||" "||
             point:px($start)||" "||
             point:py($start)||" "||
             string-join(
               for $edge in $edges return edge:translate-edge($edge),
               " "
             )
             '/>
           </xsl:variable>
           <xsl:attribute name="d" select="$path"/>
         </xsl:when>
         <xsl:otherwise>
         <xsl:sequence select='$drawing("draw:draw")($edges, map {}, $drawing)'/>
         </xsl:otherwise>
       </xsl:choose>
     </art:polygon>
   </xsl:function>

   <xsl:function name="this:draw-svg-path" as="element()" visibility="private">
     <xsl:param name="start" as="map(xs:string,item()*)"/>
     <xsl:param name="edges" as="map(xs:string,item()*)*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <svg:path>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
       <xsl:variable name="path" as="xs:string">
         <xsl:sequence select='
         edge:map-command("goto","absolute")||" "||
         point:px($start)||" "||
         point:py($start)||" "||
         string-join(
           for $edge in $edges return edge:translate-edge($edge),
           " "
         )
         '/>
       </xsl:variable>
       <xsl:attribute name="d" select="$path"/>
     </svg:path>
   </xsl:function>

   <xsl:variable name="this:draw-path-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item), "type"))) return
            let $edges := this:snap($item)=>this:edges() return
            let $start := $edges[1]=>edge:start() return
            let $type := ($item=>map:get("type"),"")[1] 
            return (
              if ($type="stroke") then (
                this:draw-stroke-path($start, $edges, $style-properties, $drawing)
              ) else (
                this:draw-path($start, $edges, $style-properties, $drawing)
              )
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item), "type"))) return
            let $edges := $item=>this:edges() return
            let $start := $edges[1]=>edge:start() return
            let $type := ($item=>map:get("type"),"")[1]
            return (
              if ($type="stroke") then (
                this:draw-stroke-path($start, $edges, $style-properties, $drawing)
              ) else (
                this:draw-path($start, $edges, $style-properties, $drawing)
              )
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              this:draw-svg-path(
                point:snap(this:start($item)), 
                this:edges(this:snap($item)), 
                $style-properties
              )
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              this:draw-svg-path(this:start($item), this:edges($item), $style-properties)
            )
          }
        )
      )
      '/>
   </xsl:variable>

   <xsl:variable name="this:draw-polygon-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item), "type"))) return
            let $edges := this:snap($item)=>this:edges() return
            let $start := $edges[1]=>edge:start() return
            let $type := ($item=>map:get("type"),"")[1] 
            return (
              if ($type="stroke") then (
                this:draw-stroke-path($start, $edges, $style-properties, $drawing)
              ) else (
                this:draw-polygon($start, $edges, $style-properties, $drawing)
              )
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item), "type"))) return
            let $edges := $item=>this:edges() return
            let $start := $edges[1]=>edge:start() return
            let $type := ($item=>map:get("type"),"")[1]
            return (
              if ($type="stroke") then (
                this:draw-stroke-path($start, $edges, $style-properties, $drawing)
              ) else (
                this:draw-polygon($start, $edges, $style-properties, $drawing)
              )
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              this:draw-svg-path(
                point:snap(this:start($item)), 
                this:edges(this:snap($item)), 
                $style-properties
              )
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            if (empty(this:edges($item))) then () else
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              this:draw-svg-path(this:start($item), this:edges($item), $style-properties)
            )
          }
        )
      )
      '/>
   </xsl:variable>

   <xsl:function name="this:draw" as="item()">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='
      let $kind := this:kind($item) return
      if ($kind="path") then $this:draw-path-impl($item, $properties, $drawing)
      else if ($kind="polygon") then $this:draw-polygon-impl($item, $properties, $drawing)
      else ()
     '/>
   </xsl:function>
</xsl:package>

<!--path.xsl-->
