<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Point objects; 2D and 3D and 4D
 :
 : Points have two kinds of accessors: "precision" accessors that return the
 : raw double coordinates, and basic accessors which snap the coordinates
 : to integer values.
 :
 : Points are maps of coordinates which may be decorated with other 
 : properties (e.g. for rendering). There are also methods for dealing with
 : the point coordinates as sequences: these exist mainly for efficiency in 
 : certain cases (see noise modules and core/vector).
 :
 : Copyright© Mary Holstege 2020-2022
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:this="http://mathling.com/geometric/point"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:config="http://mathling.com/core/config"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:art="http://mathling.com/art"
             xmlns:svg="http://www.w3.org/2000/svg"
             exclude-result-prefixes="this config util errors xs math map array"
             version="3.0"
             name="http://mathling.com/geometric/point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <xsl:variable name="this:ORIGIN" as="map(xs:string,item()*)">
      <xsl:sequence select="this:point(0,0)"/>
   </xsl:variable>
   <xsl:variable name="this:ORIGIN2D" as="map(xs:string,item()*)">
      <xsl:sequence select="this:point(0,0)"/>
   </xsl:variable>
   <xsl:variable name="this:ORIGIN3D" as="map(xs:string,item()*)">
      <xsl:sequence select="this:point(0,0,0)"/>
   </xsl:variable>
   <xsl:variable name="this:ORIGIN4D" as="map(xs:string,item()*)">
      <xsl:sequence select="this:point(0,0,0,0)"/>
   </xsl:variable>

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "x", "y", "z", "w", "dim")'/>
   </xsl:variable>

<!--
 :======================================================================
 : Constructors
 :======================================================================
 :-->

   <!--
    : this:point()
    : Construct a 2D point
    :
    : $x: x coordinate
    : $y: y coordinate
    :-->
   <xsl:function name="this:point" as="map(xs:string, item()*)">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "point",
        "x": $x,
        "y": $y,
        "dim": 2
      }
      '/>
   </xsl:function>

   <!--
    : this:point()
     : Construct a 3D point
     :
     : $x: x coordinate
     : $y: y coordinate
     : $z: z coordinate
    :-->
   <xsl:function name="this:point" as="map(xs:string, item()*)">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "point",
        "x": $x,
        "y": $y,
        "z": $z,
        "dim": 3
      }
      '/>
   </xsl:function>

   <!--
    : this:point()
    : Construct a 4D point
    :
    : $x: x coordinate
    : $y: y coordinate
    : $z: z coordinate
    : $w: w coordinate
    :-->
   <xsl:function name="this:point" as="map(xs:string, item()*)">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:param name="w" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "point",
        "x": $x,
        "y": $y,
        "z": $z,
        "w": $w,
        "dim": 4
      }
      '/>
   </xsl:function>

   <!--
    : this:vector()
    : Construct a point of some dimension from a series of coordinates.
    : Useful for genericizing some of the noise code. It is an error if the
    : number of coordinates is not 2, 3, or 4. Coordinates are given in the
    : (x, y, z w) order.
    :
    : $coordinates: the coordinates
    :-->
   <xsl:function name="this:vector" as="map(xs:string, item()*)">
      <xsl:param name="coordinates" as="xs:double*"/>
      <xsl:sequence select='
      let $dim := count($coordinates) return
      if ($dim=2) then this:point($coordinates[1], $coordinates[2])
      else if ($dim=3) then this:point($coordinates[1], $coordinates[2], $coordinates[3])
      else if ($dim=4) then this:point($coordinates[1], $coordinates[2], $coordinates[3], $coordinates[4])
      else errors:error("GEOM-UNKDIM", (count($coordinates), $coordinates))
      '/>
   </xsl:function>

   <!--
    : this:as-dimension()
    : Reconstitute a point as a point with the given number of dimensions.
    : If we are reducing dimensions, just remove excess coordinates. 
    : If we are expanding dimensions, fill in with 0.
    :
    : Examples: 
    :   point:as-dimension(point:point(1,2), 3) == point:point(1,2,0)
    :   point:as-dimension(point:point(1,2,3), 3) == point:point(1,2)
    :
    : $point: source point
    : $dim: dimension to cast point to
    :-->
   <xsl:function name="this:as-dimension" as="map(xs:string, item()*)">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:sequence select='
      let $v := this:vector(this:pcoordinates($point, $dim))
      return
        if (empty(this:properties($point))) then $v
        else util:merge-into($point, $v)
      '/>
   </xsl:function>


<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:dimension()
    : Return the dimension of the point.
    :
    : $point: the point
    :-->
   <xsl:function name="this:dimension" as="xs:integer">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (
        $point("dim"),
        if ($point=>map:contains("w")) then 4
        else if ($point=>map:contains("z")) then 3
        else if ($point=>map:contains("y")) then 2
        else 0 (: Dunno :)
      )[1]
      '/>
   </xsl:function>

   <!--
    : this:max-dimension()
    : Convenience method to return the maximum dimension of a set of points.
    : 
    : $points: sequence of points
    :-->
   <xsl:function name="this:max-dimension" as="xs:integer">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      max(for $pt in $points return this:dimension($pt))
      '/>
   </xsl:function>

   <!--
    : this:x()
    : Accessor for the snapped x coordinate
    :
    : $point: the point
    :-->
   <xsl:function name="this:x" as="xs:integer">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($point("x") = xs:double("INF")) then $util:UINT64_MAX 
      else if ($point("x") = xs:double("-INF")) then -$util:UINT64_MAX 
      else round-half-to-even($point("x")) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:y()
    : Accessor for the snapped y coordinate
    :
    : $point: the point
    :-->
   <xsl:function name="this:y" as="xs:integer">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($point("y") = xs:double("INF")) then $util:UINT64_MAX 
      else if ($point("y") = xs:double("-INF")) then -$util:UINT64_MAX 
      else round-half-to-even($point("y")) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:z()
    : Accessor for the snapped z coordinate; return 0 for 2D point
    :
    : $point: the point
    :-->
   <xsl:function name="this:z" as="xs:integer">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $z := head(($point("z"),0))
      return
        if ($z = xs:double("INF")) then $util:UINT64_MAX 
        else if ($z = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($z) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:w()
    : Accessor for the snapped w coordinate; return 0 for 2D or 3D point
    :
    : $point: the point
    :-->
   <xsl:function name="this:w" as="xs:integer">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $w := head(($point("w"),0))
      return 
        if ($w = xs:double("INF")) then $util:UINT64_MAX 
        else if ($w = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($w) cast as xs:integer
      '/>
   </xsl:function>

   <!--
    : this:px()
    : Accessor for the raw precision x coordinate
    :
    : $point: the point
    :-->
   <xsl:function name="this:px" as="xs:double">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='$point("x")'/>
   </xsl:function>

   <!--
    : this:py()
    : Accessor for the raw precision y coordinate
    :
    : $point: the point
    :-->
   <xsl:function name="this:py" as="xs:double">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='$point("y")'/>
   </xsl:function>

   <!--
    : this:pz()
    : Accessor for the raw precision z coordinate; returns 0 for 2D point
    :
    : $point: the point
    :-->
   <xsl:function name="this:pz" as="xs:double">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='head(($point("z"), 0))'/>
   </xsl:function>

   <!--
    : this:pw()
    : Accessor for the raw precision w coordinate; returns 0 for 2D or 3D point
    :
    : $point: the point
    :-->
   <xsl:function name="this:pw" as="xs:double">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='head(($point("w"), 0))'/>
   </xsl:function>

   <!--
    : this:coordinates()
    : Return snapped coordinates of the point as a sequence.
    : 
    : $point: the point
    :-->
   <xsl:function name="this:coordinates" as="xs:integer*">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := this:dimension($point)
      return (
        if ($d lt 2 or $d gt 4)
        then errors:error("GEOM-UNKDIM", ($d, $point))
        else (),
        if ($d ge 2) then (this:x($point), this:y($point)) else (),
        if ($d ge 3) then this:z($point) else (),
        if ($d ge 4) then this:w($point) else ()
      )
      '/>
   </xsl:function>

   <!--
    : this:coordinates()
    : Return snapped coordinates of the point as a sequence, treating the
    : point as having a given dimension. 
    : 
    : $point: the point
    : $d: desired dimension
    :-->
   <xsl:function name="this:coordinates" as="xs:integer*">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      if ($d = 0) then this:coordinates($point)
      else (
        if ($d lt 2 or $d gt 4)
        then errors:error("GEOM-UNKDIM", ($d, $point))
        else (),
        if ($d ge 2) then (this:x($point), this:y($point)) else (),
        if ($d ge 3) then this:z($point) else (),
        if ($d ge 4) then this:w($point) else ()
      )
      '/>
   </xsl:function>

   <!--
    : this:pcoordinates()
    : Return raw precision coordinates of the point as a sequence.
    : 
    : $point: the point
    :-->
   <xsl:function name="this:pcoordinates" as="xs:double*">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := this:dimension($point)
      return (
        if ($d lt 2 or $d gt 4)
        then errors:error("GEOM-UNKDIM", ($d, $point))
        else (),
        if ($d ge 2) then (this:px($point), this:py($point)) else (),
        if ($d ge 3) then this:pz($point) else (),
        if ($d ge 4) then this:pw($point) else ()
      )
      '/>
   </xsl:function>

   <!--
    : this:pcoordinates()
    : Return raw precision coordinates of the point as a sequence, treating the
    : point as having a given dimension. 
    : 
    : $point: the point
    : $d: the desired dimension
    :-->
   <xsl:function name="this:pcoordinates" as="xs:double*">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      if ($d = 0) then this:pcoordinates($point)
      else (
        if ($d lt 2 or $d gt 4)
        then errors:error("GEOM-UNKDIM", ($d, $point))
        else (),
        if ($d ge 2) then (this:px($point), this:py($point)) else (),
        if ($d ge 3) then this:pz($point) else (),
        if ($d ge 4) then this:pw($point) else ()
      )
      '/>
   </xsl:function>

   <!--
    : this:pcoordinate()
    : Return the given raw precision coordinates of the point.
    : 
    : $point: the point
    : $ix: index of the coordinate, e.g. 1=1st, etc.
    :-->
   <xsl:function name="this:pcoordinate" as="xs:double">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="ix" as="xs:integer"/>
      <xsl:sequence select='
      if ($ix = 1) then this:px($point)
      else if ($ix = 2) then this:py($point)
      else if ($ix = 3) then this:pz($point)
      else if ($ix = 4) then this:pw($point)
      else errors:error("GEOM-UNKDIM", ($ix, $point))
      '/>
   </xsl:function>

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic
    : geo functions
    :
    : $point: the point
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='$point("kind")'/>
   </xsl:function>

<!--
 :======================================================================
 : Property management
 :======================================================================
 :-->

   <!--
    : this:property-map()
    : Return the annotation properties of the point as a map. Check whether this
    : is actually a point.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($region)="point") 
      then util:exclude($region, $this:RESERVED)
      else map {}
      '/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the point.
    : Check whether this is actually a point.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($region)="point")
      then ($region=>map:keys())[not(. = $this:RESERVED)]
      else ()
      '/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the point with some new properties and return the new point.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually a point.
    : 
    : $region: the region
    : $properties: the properties
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($region)="point")
      then util:merge-into($region, util:exclude($properties,$this:RESERVED))
      else errors:error("GEOM-BADREGION", ($region, "with-properties"))
      '/>
   </xsl:function>

<!--
 :======================================================================
 : Point operations
 :======================================================================
 :-->

   <!--
    : this:midpoint()
    : Return midpoint between two points
    :
    : $a: one point
    : $b: the other
    :-->
   <xsl:function name="this:midpoint" as="map(xs:string, item()*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:map2(
        function ($ca as xs:double, $cb as xs:double) as xs:double {
          ($ca + $cb) div 2
        },
        $a,
        $b
      )
      '/>
   </xsl:function>

   <!--
    : this:magnitude()
    : Magnitude of the point as vector from origin.
    : 
    : $v: the point
    :-->
   <xsl:function name="this:magnitude" as="xs:double">
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      math:sqrt(
        this:px($v)*this:px($v) +
        this:py($v)*this:py($v) + 
        this:pz($v)*this:pz($v) +
        this:pw($v)*this:pw($v)
      )
      '/>
   </xsl:function>

   <!--
    : this:dot()
    : Compute the dot product of two points
    : 
    : $p1: one point
    : $p2: another point
    :-->
   <xsl:function name="this:dot" as="xs:double">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:px($p1)*this:px($p2) +
      this:py($p1)*this:py($p2) + 
      this:pz($p1)*this:pz($p2) +
      this:pw($p1)*this:pw($p2)
      '/>
   </xsl:function>

   <!--
    : this:dot2()
    : Compute the dot product of a point with itself
    : 
    : $p: the point
    :-->
   <xsl:function name="this:dot2" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:px($p)*this:px($p) +
      this:py($p)*this:py($p) + 
      this:pz($p)*this:pz($p) +
      this:pw($p)*this:pw($p)
      '/>
   </xsl:function>

   <!--
    : this:determinant()
    : Compute the determinant of two 2D points.
    :
    : $u: one point
    : $v: another point
    :-->
   <xsl:function name="this:determinant" as="xs:double">
      <xsl:param name="u" as="map(xs:string, item()*)"/>
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $u=>this:px()*$v=>this:py() - $u=>this:py()*$v=>this:px()
      '/>
   </xsl:function>

   <!--
    : this:cross()
    : Compute the cross product of two 3D points.
    :
    : $u: one point
    : $v: another point
    :-->
   <xsl:function name="this:cross" as="map(xs:string, item()*)">
      <xsl:param name="u" as="map(xs:string, item()*)"/>
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:point(
        $u=>this:py()*$v=>this:pz() - $u=>this:pz()*$v=>this:py(),
        $u=>this:pz()*$v=>this:px() - $u=>this:px()*$v=>this:pz(),
        $u=>this:px()*$v=>this:py() - $u=>this:py()*$v=>this:px()
      )
      '/>
   </xsl:function>

   <!--
    : this:times()
    : Scale every coordinate by a constant, returning the new point.
    :
    : $u: point to scale
    : $k: constant multiplier
    :-->
   <xsl:function name="this:times" as="map(xs:string, item()*)">
      <xsl:param name="u" as="map(xs:string, item()*)"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:map(function ($c) {$k * $c}, $u)
      '/>
   </xsl:function>

   <!--
    : this:add()
    : Add two points, returning the new point.
    : 
    : $u: one point
    : $v: the other point
    :-->
   <xsl:function name="this:add" as="map(xs:string, item()*)">
      <xsl:param name="u" as="map(xs:string, item()*)"/>
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {$a + $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:sub()
    : Subtract one point from another, returning the new point.
    : 
    : $u: one point
    : $v: the other point
    :-->
   <xsl:function name="this:sub" as="map(xs:string, item()*)">
      <xsl:param name="u" as="map(xs:string, item()*)"/>
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {$a - $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:minus()
    : Negate the coordinates of the point, returning the new point.
    : 
    : $v: the point
    :-->
   <xsl:function name="this:minus" as="map(xs:string, item()*)">
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:map(function($a) {-$a}, $v)
      '/>
   </xsl:function>

   <!--
    : this:normalize()
    : Normalize a vector represented as a point (scale to unit vector), returning
    : the new point.
    :
    : $v: the point
    :-->
   <xsl:function name="this:normalize" as="map(xs:string, item()*)">
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := this:magnitude($v)
      return
        if ($l = 0) then $v
        else this:map(function ($c) {$c div $l}, $v)
      '/>
   </xsl:function>

   <!--
    : this:perpendicular()
    : Compute the perpendicular vector for a 2D vector represented as a point.
    : (Rotated by 90 degrees.)
    : 
    : $v: the point
    :-->
   <xsl:function name="this:perpendicular" as="map(xs:string, item()*)">
      <xsl:param name="v" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:point($v=>this:py(), -($v=>this:px()))
      '/>
   </xsl:function>

   <!--
    : this:perpendiculars()
    : Compute both perpendicular vectors for a 2D vector represented as a point.
    : Rotated by 90 degrees and by -90 degrees, in that order.
    : 
    : $v: the point
    :-->
   <xsl:function name="this:perpendiculars" as="map(xs:string,item()*)*">
      <xsl:param name="v" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      this:point(this:py($v), -this:px($v)),
      this:point(-this:py($v), this:px($v))
      '/>
   </xsl:function>

   <!--
    : this:perpendiculars-with-z()
    : Compute perpendicular vectors for a 3D vector represented as a point.
    : Rotated in y by 90 degrees and by -90 degrees, in that order, with z
    : rotated in the corresponding direction. For spline/smoothing.
    : 
    : $v: the point
    :-->
   <xsl:function name="this:perpendiculars-with-z" as="map(xs:string,item()*)*">
      <xsl:param name="v" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      this:point(this:py($v), -this:px($v), this:pz($v)),
      this:point(-this:py($v), this:px($v), -this:pz($v))
      '/>
   </xsl:function>

   <!--
    : this:snap()
    : Snap the coordinates of the points, returning the points with snapped
    : (i.e. integer) coordinates.
    :
    : $points: the points
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $point in $points return (
        this:map(
          function ($c) {
            if ($c = xs:double("INF")) then $util:UINT64_MAX 
            else if ($c = xs:double("-INF")) then -$util:UINT64_MAX 
            else round-half-to-even($c) cast as xs:integer
          },
          $point
        )
      )
      '/>
   </xsl:function>
 
   <!--
    : this:floor()
    : Return points where all the coordinates have been set to the floor of
    : the original points' values. Differs from snap() for negative coordinates.
    :
    : $points: the points
    :-->
   <xsl:function name="this:floor" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $point in $points return (
        this:map(floor#1, $point)
      )
      '/>
   </xsl:function>

   <!--
    : this:decimal()
    : Perform decimal rounding on all the coordinates (see util:decimal).
    :
    : $points: the points to round
    : $digits: how many digits after the decimal point to keep
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $point in $points return (
        this:map(function($c) {util:decimal($c, $digits)}, $point)
      )
      '/>
   </xsl:function>

   <!--
    : this:quote()
    : Return a string value for the points, suitable for debugging.
    : 
    : $points: the point sequence to quote
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $point in $points return (
          string-join(this:pcoordinates($point)!string(.),",")
        )
        ,
        " "
      )
      '/>
   </xsl:function>

   <!--
    : this:valid()
    : False if a coordinate is INF or NaN
    :-->
   <xsl:function name="this:valid" as="xs:boolean">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      every $c in this:pcoordinates($point)
      satisfies $c=$c and not($c=(xs:double("INF"),xs:double("-INF")))
      '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for points, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one point
    : $other: the point to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:px($this)=this:px($other) and
      this:py($this)=this:py($other) and
      this:pz($this)=this:pz($other) and
      this:pw($this)=this:pw($other)
      '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for points, ignoring annotation properties.
    : Return true() if they have equal coordinates within the given tolerance.
    : 
    : $this: one point
    : $other: the point to compare it to
    : $tolerance: allowable slop
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      abs(this:px($this)-this:px($other)) lt $tolerance and
      abs(this:py($this)-this:py($other)) lt $tolerance and
      abs(this:pz($this)-this:pz($other)) lt $tolerance and
      abs(this:pw($this)-this:pw($other)) lt $tolerance
      '/>
   </xsl:function>

   <!--
    : this:mutate()
    : Run a function over a sequence of points to produce a new sequence of
    : points. 
    : 
    : $regions: input sequence of points 
    : $mutate: function that takes a point as an argument and returns a new point
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate" as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='$regions!$mutate(.)'/>
   </xsl:function>

   <!--
    : this:distinct()
    : Return the points in the sequence that have distinct coordinates from
    : each other.
    : 
    : $points: the sequence of points
    :-->
   <xsl:function name="this:distinct" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $sorted :=
        sort($points, (), function($pt) {this:px($pt), this:py($pt), this:pz($pt), this:pw($pt)})
      return (
        head($sorted), 
        for $i in 2 to count($sorted)
        return (
          if (not(this:same($sorted[$i - 1], $sorted[$i])))
          then $sorted[$i]
          else ()
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:unquote()
    : Parse a string representation of a point and produce a point.
    : The string consists of the coordinates in x,y,z,w order with commas and 
    : optional spaces in between.
    : Example "3.4, 1" == point:point(3.4, 1)
    :
    : $string: the input string to parse
    :-->
   <xsl:function name="this:unquote" as="map(xs:string, item()*)">
      <xsl:param name="string" as="xs:string"/>
      <xsl:sequence select='
      this:vector(tokenize($string, "[ ]*,[ ]*")!xs:double(.))
      '/>
   </xsl:function>

   <!--
    : this:map()
    : Map a function over the coordinates of a point, returning the new point.
    : 
    : $f: function from coordinate to coordinate
    : $point: input point
    :-->
   <xsl:function name="this:map" as="map(xs:string, item()*)">
      <xsl:param name="f" as="(function(xs:double) as xs:double)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $v := this:vector(for-each(this:pcoordinates($point), $f))
      return (
        if (empty(this:properties($point))) then $v
        else util:merge-into($point, $v)
      )
      '/>
   </xsl:function>

   <!--
    : this:map()
    : Map a function over the coordinates of a point, returning the new point.
    : Treat the point as having a particular dimension.
    : 
    : $f: function from coordinate to coordinate
    : $point: input point
    : $d: the effective dimension of the point
    :-->
   <xsl:function name="this:map" as="map(xs:string, item()*)">
      <xsl:param name="f" as="(function(xs:double) as xs:double)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      let $v := this:vector(for-each(this:pcoordinates($point, $d), $f))
      return (
        if (empty(this:properties($point))) then $v
        else util:merge-into($point, $v)
      )
      '/>
   </xsl:function>

   <!--
    : this:map2()
    : Map a function over the coordinates of two points, pair by pair, returning
    : the new point. That is, apply function to the x coordinates, then the
    : y coordinates, and so on.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    :-->
   <xsl:function name="this:map2" as="map(xs:string, item()*)">
      <xsl:param name="f" as="(function(xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := this:max-dimension(($p1, $p2))
      return
        this:vector(
          util:zip($f, this:pcoordinates($p1, $d), this:pcoordinates($p2, $d))
        )
      '/>
   </xsl:function>

   <!--
    : this:map2()
    : Map a function over the coordinates of two points, pair by pair, returning
    : the new point. That is, apply function to the x coordinates, then the
    : y coordinates, and so on. Treat the point as having a particular dimension.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    : $d: the effective dimension of the points
    :-->
   <xsl:function name="this:map2" as="map(xs:string, item()*)">
      <xsl:param name="f" as="(function(xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      this:vector(
        util:zip($f, this:pcoordinates($p1, $d), this:pcoordinates($p2, $d))
      )
      '/>
   </xsl:function>

   <!--
    : this:map3()
    : Map a function over the coordinates of three points, triple by triple, 
    : returning the new point. That is, apply function to the x coordinates, 
    : then the y coordinates, and so on.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    : $p3: third input point
    :-->
   <xsl:function name="this:map3" as="map(xs:string, item()*)">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:param name="p3" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := this:max-dimension(($p1, $p2, $p3))
      return
        this:vector(
          util:zip(
            function ($a, $fc) {$fc($a)},
            this:pcoordinates($p3, $d),
            util:zip(
              function($a, $b) { $f($a, $b, ?) },
              this:pcoordinates($p1, $d), this:pcoordinates($p2, $d)
            )
          )
        )
      '/>
   </xsl:function>

   <!--
    : this:map3()
    : Map a function over the coordinates of three points, triple by triple, 
    : returning the new point. That is, apply function to the x coordinates, 
    : then the y coordinates, and so on. Treat the point as having a 
    : particular dimension.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    : $p3: third input point
    : $d: the effective dimension of the points
    :-->
   <xsl:function name="this:map3" as="map(xs:string, item()*)">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:param name="p3" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      this:vector(
        util:zip(
          function ($a, $fc) {$fc($a)},
          this:pcoordinates($p3, $d),
          util:zip(
            function($a, $b) { $f($a, $b, ?) },
            this:pcoordinates($p1, $d), this:pcoordinates($p2, $d)
          )
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:map4()
    : Map a function over the coordinates of four points, coordinate by
    : coordinate, returning the new point. That is, apply function to the x
    : coordinates, then the y coordinates, and so on.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    : $p3: third input point
    : $p4: fourth input point
    :-->
   <xsl:function name="this:map4" as="map(xs:string, item()*)">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:param name="p3" as="map(xs:string, item()*)"/>
      <xsl:param name="p4" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $dim := this:max-dimension(($p1, $p2, $p3))
      return
        this:vector(
          util:zip(
            function($d, $fd) {$fd($d)},
            this:pcoordinates($p4, $dim),
            util:zip(
              function ($c, $fc) {$fc($c, ?)},
              this:pcoordinates($p3, $dim),
              util:zip(
                function($a, $b) { $f($a, $b, ?, ?) },
                this:pcoordinates($p1, $dim), this:pcoordinates($p2, $dim)
              )
            )
          )
        )
      '/>
   </xsl:function>

   <!--
    : this:map4()
    : Map a function over the coordinates of four points, coordinate by
    : coordinate, returning the new point. That is, apply function to the x
    : coordinates, then the y coordinates, and so on. Treat the point as having a 
    : particular dimension.
    : 
    : $f: function from coordinate pairs to coordinate
    : $p1: one input point
    : $p2: another input point
    : $p3: third input point
    : $p4: fourth input point
    : $dim: the effective dimension of the points
    :-->
   <xsl:function name="this:map4" as="map(xs:string, item()*)">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:param name="p3" as="map(xs:string, item()*)"/>
      <xsl:param name="p4" as="map(xs:string, item()*)"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:sequence select='
      this:vector(
        util:zip(
          function($d, $fd) {$fd($d)},
          this:pcoordinates($p4, $dim),
          util:zip(
            function ($c, $fc) {$fc($c, ?)},
            this:pcoordinates($p3, $dim),
            util:zip(
              function($a, $b) { $f($a, $b, ?, ?) },
              this:pcoordinates($p1, $dim), this:pcoordinates($p2, $dim)
            )
          )
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:distance()
    : Euclidean distance between points
    :-->
   <xsl:function name="this:distance" as="xs:double">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := this:max-dimension(($p1, $p2))
      return (
        math:sqrt(
          sum(
            util:zip(
              function($c1 as xs:double, $c2 as xs:double) as xs:double {
                ($c1 - $c2)*($c1 - $c2)
              },
              this:pcoordinates($p1, $d),
              this:pcoordinates($p2, $d)
            )
          )
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:angle()
    : Compute the angle (azimuth) from one point to the next, in degrees
    : Return 0 if points are the same
    :
    : $last: previous point; use (0,0) if no previous
    : $curr: current point
    :-->
   <xsl:function name="this:angle" as="xs:double">
      <xsl:param name="last" as="map(xs:string, item()*)?"/>
      <xsl:param name="curr" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $this := $curr=>this:as-dimension(2) return
      let $prev := this:as-dimension(($last,$this:ORIGIN)[1],2)
      return
        if (this:same($prev, $this) or (this:distance($prev,$this) lt $config:ε)) then 0
        else if (this:px($this)=this:px($prev)) then (
          if (this:py($prev) > this:py($this))
          then 270 (: remapped -90 :)
          else 90
        ) else (
          util:remap-degrees(util:degrees(
            (math:pi() div 2) - math:atan2(this:px($this) - this:px($prev), this:py($this) - this:py($prev))
          ))
        )
      '/>
   </xsl:function>

   <!--
    : this:inclination()
    : Compute the inclination angle from one point in the next, in degrees
    : Return 90 if the points are the same
    :
    : $last: previous point; use (0,0,0) if no previous
    : $curr: current point
    :-->
   <xsl:function name="this:inclination" as="xs:double">
      <xsl:param name="last" as="map(xs:string, item()*)?"/>
      <xsl:param name="curr" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $curr := $curr=>this:as-dimension(3) return
      let $prev := (($last,$this:ORIGIN3D)[1])=>this:as-dimension(3) return
      let $d := this:distance($prev,$curr)
      return
        if (this:same($prev,$curr) or ($d lt $config:ε)) then 90
        else (
          util:remap-degrees(util:degrees(
            math:acos( (this:pz($curr) - this:pz($prev)) div $d )
          ))
        )
      '/>
   </xsl:function>

   <!--
    : this:destination()
    : Point at a particular distance and direction.
    : 
    : $point: starting point
    : $degrees: bearing from point
    : $length: how far along bearing to travel
    :-->
   <xsl:function name="this:destination" as="map(xs:string, item()*)">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:param name="length" as="xs:double"/>
      <xsl:sequence select='
      let $angle := util:radians($degrees)
      return
        this:point(
          this:px($point) + math:cos($angle)*$length,
          this:py($point) + math:sin($angle)*$length
        )
      '/>
   </xsl:function>

   <!--
    : this:destination()
    : Point (3D) at a particular distance, azimuth, and inclination.
    : 
    : $this: starting point
    : $azimuth_degrees: angle of azimuth from point
    : $inclination_degrees: angle of inclination from point
    : $length: how far along bearing to travel
    :-->
   <xsl:function name="this:destination" as="map(xs:string, item()*)">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="azimuth_degrees" as="xs:double"/>
      <xsl:param name="inclination_degrees" as="xs:double"/>
      <xsl:param name="length" as="xs:double"/>
      <xsl:sequence select='
      let $azimuth := util:radians($azimuth_degrees) return
      let $inclination := util:radians($inclination_degrees) return
      (: Avoid having imprecision on sin/cos introduce non-zero Z values :)
      let $cos_inclination := math:cos($inclination) return
      let $cos_inclination :=
        if (abs($cos_inclination) lt $config:ε) then 0 else $cos_inclination
      return
      let $sin_inclination := math:sin($inclination) return
      let $sin_inclination := 
        if (abs($sin_inclination) lt $config:ε) then 0 else $sin_inclination
      return
      if (this:dimension($point)=4) then
        this:point(
          this:px($point) + $length*math:cos($azimuth)*$sin_inclination,
          this:py($point) + $length*math:sin($azimuth)*$sin_inclination,
          this:pz($point) + $length*$cos_inclination,
          this:pw($point)
        )
      else
        this:point(
          this:px($point) + $length*math:cos($azimuth)*$sin_inclination,
          this:py($point) + $length*math:sin($azimuth)*$sin_inclination,
          this:pz($point) + $length*$cos_inclination
        )
      '/>
   </xsl:function>

   <!--
    : this:orientation()
    : Orientation of triplet of 2D points p-q-r
    : "colinear"          all on the same line
    : "clockwise"         slope of p-q > slope q-r
    : "counterclockwise"  slope of p-q < slope q-r
    :
    : slope p-q: (qy - py)/(qx - px) 
    : slope q-r: (ry - qy)/(rx - qx)
    : (qy - py)/(qx - px) > (ry - qy)/(rx - qx) =>
    : (qy - py)/(qx - px) - (ry - qy)/(rx - qx) > 0
    :
    : $p, $q, $r: the points
    :-->
   <xsl:function name="this:orientation" as="xs:string">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $test := (this:py($q) - this:py($p)) * (this:px($r) - this:px($q)) -
                   (this:px($q) - this:px($p)) * (this:py($r) - this:py($q))
      return
        if ($test = 0) then "colinear"
        else if ($test > 0) then "clockwise"
        else "counterclockwise"
      '/>
   </xsl:function>

   <!--
    : this:centroid()
    :-->
   <xsl:function name="this:centroid" as="map(xs:string, item()*)">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $maxd := this:max-dimension($points) return
      let $n := count($points) return
      let $coords :=
        for $d in 1 to $maxd return (
          sum(
            for $p in $points return $p=>this:pcoordinate($d)
          ) div $n
        )
      return this:vector($coords)
      '/>
   </xsl:function>

   <xsl:function name="this:draw-px" as="element()" visibility="private">
     <xsl:param name="x" as="xs:numeric"/>
     <xsl:param name="y" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:px x="{$x}" y="{$y}">
       <xsl:sequence select='util:as-attributes($style-properties)'/>
     </art:px>
   </xsl:function>
   
   <xsl:function name="this:draw-dot" as="element()" visibility="private">
     <xsl:param name="x" as="xs:numeric"/>
     <xsl:param name="y" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:dot x="{$x}" y="{$y}">
       <xsl:sequence select='util:as-attributes($style-properties)'/>
     </art:dot>
   </xsl:function>

   <xsl:function name="this:draw-polyline" as="element()" visibility="private">
     <xsl:param name="x" as="xs:numeric"/>
     <xsl:param name="y" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:variable name="size" as="xs:numeric">
       <xsl:sequence select='
         if (exists($style-properties("r"))) then $style-properties("r") else 1
         '/>
     </xsl:variable>
     <svg:polyline points="{$x},{$y} {$x+$size},{$y} {$x+$size},{$y+$size} {$x},{$y+1} {$x},{$y}">
       <xsl:if test='empty($style-properties("stroke-width"))'>
         <xsl:attribute name="stroke-width" select="0.1"/>
       </xsl:if>
       <xsl:sequence select='
       util:as-attributes(util:exclude($style-properties,("r")))
       '/>
     </svg:polyline>
   </xsl:function>

   <xsl:function name="this:draw-circle" as="element()" visibility="private">
     <xsl:param name="x" as="xs:numeric"/>
     <xsl:param name="y" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <svg:circle cx="$x" cy="$y">
       <xsl:if test='empty($style-properties("r"))'>
         <xsl:attribute name="r" select="3"/>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
     </svg:circle>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item)))
            return (
              if ($config:POINT-IS-PIXEL) 
              then this:draw-px(this:x($item), this:y($item), $style-properties)
              else this:draw-dot(this:x($item), this:y($item), $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item)))
            return (
              if ($config:POINT-IS-PIXEL)
              then this:draw-px(this:px($item), this:py($item), $style-properties)
              else this:draw-dot(this:px($item), this:py($item), $style-properties)
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              if ($config:POINT-IS-PIXEL) 
              then this:draw-polyline(this:x($item), this:y($item), $style-properties)
              else this:draw-circle(this:x($item), this:y($item), $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              if ($config:POINT-IS-PIXEL) 
              then this:draw-polyline(this:px($item), this:py($item), $style-properties)
              else this:draw-circle(this:px($item), this:py($item), $style-properties)
            )
          }
        )
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>
<!--point.xsl-->
