<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Axis-aligned rectangles and blocks
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:box="http://mathling.com/geometric/box"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:space="http://mathling.com/type/space"
             xmlns:this="http://mathling.com/geometric/rectangle"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/geometric/rectangle"
             exclude-result-prefixes="this xs math map array errors util config point box edge space">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/box" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/space" package-version="*"/>

<!--
 :======================================================================
 : Constructors
 :======================================================================
 :-->

   <!--
    : this:rectangle()
    : Construct an axis-aligned rectangle in x/y space.
    :   $x-min: Minimum on x-axis
    :   $y-min: Minimim on y-axis
    :   $x-max: Maximum on x-axis
    :   $y-max: Maximim on y-axis
    :-->
   <xsl:function name="this:rectangle" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:sequence select='
      box:rectangle($x-min, $y-min, $x-max, $y-max)
    '/>
   </xsl:function>

   <!--
    : this:block()
    : Construct an axis-aligned rectangle in x/y/z space.
    :   $x-min: Minimum on x-axis
    :   $y-min: Minimim on y-axis
    :   $z-min: Minimim on z-axis
    :   $x-max: Maximum on x-axis
    :   $y-max: Maximim on y-axis
    :   $z-max: Maximim on z-axis
    :-->
   <xsl:function name="this:block" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="z-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:param name="z-max" as="xs:double"/>
      <xsl:sequence select='
      box:block($x-min, $y-min, $z-min, $x-max, $y-max, $z-max)
    '/>
   </xsl:function>

   <!--
    : this:rectangle()
    :-->
   <xsl:function name="this:rectangle" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      box:rectangle($pt-min, $pt-max)
    '/>
   </xsl:function>

   <!--
    : this:block()
    :-->
   <xsl:function name="this:block" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      box:block($pt-min, $pt-max)
    '/>
   </xsl:function>

   <!--
    : this:box()
    :-->
   <xsl:function name="this:box" as="map(xs:string, item()*)">
      <xsl:param name="x-min" as="xs:double"/>
      <xsl:param name="y-min" as="xs:double"/>
      <xsl:param name="x-max" as="xs:double"/>
      <xsl:param name="y-max" as="xs:double"/>
      <xsl:sequence select='
      box:rectangle($x-min, $y-min, $x-max, $y-max)
    '/>
   </xsl:function>

   <!--
    : this:box()
    :-->
   <xsl:function name="this:box" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      box:box($pt-min, $pt-max)
    '/>
   </xsl:function>

   <!--
    : this:as-box()
    :-->
   <xsl:function name="this:as-box" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "box") then $region
      else if ($kind = "block") then box:box(box:min-point($region), box:max-point($region))
      else if ($kind = "space") then
        util:merge-into(
          util:exclude($region, $space:RESERVED),
          this:box(this:min-point($region), this:max-point($region))
        )
      else errors:error("GEOM-BADREGION", ($region, "as-box"))
    '/>
   </xsl:function>

   <!--
    : this:as-space()
    :-->
   <xsl:function name="this:as-space" as="map(xs:string, item()*)">
      <xsl:param name="pt-min" as="map(xs:string, item()*)"/>
      <xsl:param name="pt-max" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into(
        $properties, 
        space:space(
          abs(point:px($pt-max) - point:px($pt-min)),
          abs(point:py($pt-max) - point:py($pt-min)),
          this:edge($properties)
        )
      )
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:min-point()
    : Accessor for minimal point
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-point" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-point($box)
      else if ($kind = "block") then box:min-point($box)
      else if ($kind = "space") then point:point(this:min-px($box), this:min-py($box))
      else errors:error("GEOM-BADREGION", ($box, "min-point"))
    '/>
   </xsl:function>

   <!--
    : this:max-point()
    : Accessor for maximal point
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-point" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-point($box)
      else if ($kind = "block") then box:max-point($box)
      else if ($kind = "space") then point:point(this:max-px($box), this:max-py($box))
      else errors:error("GEOM-BADREGION", ($box, "max-point"))
    '/>
   </xsl:function>

   <!--
    : this:min-px()
    : Accessor for raw minimum x-axis of point. 
 :   $box: rectangle
    :-->
   <xsl:function name="this:min-px" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-px($box)
      else if ($kind = "block") then box:min-px($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-px"))
    '/>
   </xsl:function>

   <!--
    : this:max-px()
    : Accessor for raw maximum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-px" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-px($box)
      else if ($kind = "block") then box:max-px($box)
      else if ($kind = "space") then $box("width")
      else errors:error("GEOM-BADREGION", ($box, "max-px"))
    '/>
   </xsl:function>

   <!--
    : this:min-py()
    : Accessor for raw minimum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-py" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-py($box)
      else if ($kind = "block") then box:min-py($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-py"))
    '/>
   </xsl:function>

   <!--
    : this:max-py()
    : Accessor for raw maximum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-py" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-py($box)
      else if ($kind = "block") then box:max-py($box)
      else if ($kind = "space") then $box("height")
      else errors:error("GEOM-BADREGION", ($box, "max-py"))
    '/>
   </xsl:function>

   <!--
    : this:min-pz()
    : Accessor for raw minimum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-pz" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-pz($box)
      else if ($kind = "block") then box:min-pz($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-pz"))
    '/>
   </xsl:function>

   <!--
    : this:max-pz()
    : Accessor for raw maximum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-pz" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-pz($box)
      else if ($kind = "block") then box:max-pz($box)
      else if ($kind = "space") then ($box("depth"),0)[1]
      else errors:error("GEOM-BADREGION", ($box, "max-pz"))
    '/>
   </xsl:function>

   <!--
    : this:min-x()
    : Accessor for snapped minimum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-x" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-x($box)
      else if ($kind = "block") then box:min-x($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-x"))
    
    '/>
   </xsl:function>

   <!--
    : this:max-x()
    : Accessor for snapped maximum x-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-x" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-x($box)
      else if ($kind = "block") then box:max-x($box)
      else if ($kind = "space") then round-half-to-even($box("width")) cast as xs:integer
      else errors:error("GEOM-BADREGION", ($box, "max-x"))
    '/>
   </xsl:function>

   <!--
    : this:min-y()
    : Accessor for snapped minimum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-y" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-y($box)
      else if ($kind = "block") then box:min-y($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-y"))
    '/>
   </xsl:function>

   <!--
    : this:max-y()
    : Accessor for snapped maximum y-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-y" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-y($box)
      else if ($kind = "block") then box:max-y($box)
      else if ($kind = "space") then round-half-to-even($box("height")) cast as xs:integer
      else errors:error("GEOM-BADREGION", ($box, "max-y"))
    '/>
   </xsl:function>

   <!--
    : this:min-z()
    : Accessor for snapped minimum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:min-z" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:min-z($box)
      else if ($kind = "block") then box:min-z($box)
      else if ($kind = "space") then 0
      else errors:error("GEOM-BADREGION", ($box, "min-z"))
    '/>
   </xsl:function>

   <!--
    : this:max-z()
    : Accessor for snapped maximum z-axis of point. 
    :   $box: rectangle
    :-->
   <xsl:function name="this:max-z" as="xs:integer">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:max-z($box)
      else if ($kind = "block") then box:max-z($box)
      else if ($kind = "space") then round-half-to-even(($box("depth"),0)[1]) cast as xs:integer
      else errors:error("GEOM-BADREGION", ($box, "max-z"))
    '/>
   </xsl:function>

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $box: the rectangle
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $box("kind")
    '/>
   </xsl:function>

   <!--
    : this:width()
    : Width of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:width" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:width($box)
      else if ($kind = "block") then box:width($box)
      else if ($kind = "space") then space:width($box)
      else errors:error("GEOM-BADREGION", ($box, "width"))
    '/>
   </xsl:function>

   <!--
    : this:height()
    : Height of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:height" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:height($box)
      else if ($kind = "block") then box:height($box)
      else if ($kind = "space") then space:height($box)
      else errors:error("GEOM-BADREGION", ($box, "height"))
    '/>
   </xsl:function>

   <!--
    : this:depth()
    : Depth of rectangle
    :   $box: rectangle
    :-->
   <xsl:function name="this:depth" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:depth($box)
      else if ($kind = "block") then box:depth($box)
      else if ($kind = "space") then ($box("depth"),0)[1]
      else errors:error("GEOM-BADREGION", ($box, "depth"))
    '/>
   </xsl:function>

   <!--
    : this:diagonal()
    : Length of diagonal (x-y)
    :   $box: rectangle
    :-->
   <xsl:function name="this:diagonal" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $w := this:width($box) return
      let $h := this:height($box)
      return math:sqrt($w*$w + $h*$h)
    '/>
   </xsl:function>

   <!--
    : this:full-diagonal()
    : Length of diagonal from min-point to max-point (which may be 3D)
    :   $box: rectangle
    :-->
   <xsl:function name="this:full-diagonal" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:distance(this:min-point($box), this:max-point($box))
    '/>
   </xsl:function>

   <!--
    : this:center()
    : Point in center of the box
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($box)="block") then (
        point:point(
          (this:min-px($box) + this:max-px($box)) div 2,
          (this:min-py($box) + this:max-py($box)) div 2,
          (this:min-pz($box) + this:max-pz($box)) div 2
        )
      ) else (
        point:point(
          (this:min-px($box) + this:max-px($box)) div 2,
          (this:min-py($box) + this:max-py($box)) div 2
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:area()
    : Area of a box (x-y)
    :-->
   <xsl:function name="this:area" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:width($box) * this:height($box)
    '/>
   </xsl:function>

   <!--
    : this:edge()
    : The edge padding for a canvas space
    :-->
   <xsl:function name="this:edge" as="xs:double">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:edge($box)
      else if ($kind = "block") then box:edge($box)
      else if ($kind = "space") then ($box=>map:get("edge"),0)[1]
      else errors:error("GEOM-BADREGION", ($box, "edge"))
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    : Get all the region's points.
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "box") then box:vertices($region)
      else if ($kind = "block") then box:vertices($region)
      else if ($kind = "space") then (
        point:point(this:min-px($region), this:min-py($region)),
        point:point(this:min-px($region), this:max-py($region)),
        point:point(this:max-px($region), this:max-py($region)),
        point:point(this:max-px($region), this:min-py($region))
      )
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
   </xsl:function>

   <!--
    : this:points()
    : Get 2D points from the region
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "box") then box:points($region)
      else if ($kind = "block") then box:points($region)
      else if ($kind = "space") then this:vertices($region) (: 2D already :)
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
   </xsl:function>

   <!--
    : this:edges()
    :   Get the edges of a region
    : $region: The region
    :-->
   <xsl:function name="this:edges" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind = "box") then (
        let $vertices := this:vertices($region)
        return (
          for $i in 2 to count($vertices) return edge:edge($vertices[$i - 1], $vertices[$i]),
          edge:edge($vertices[last()], $vertices[1]) (: close loop :)
        )
      )
      else if ($kind = "block") then (
        let $vertices := this:vertices($region) return
        let $face-indexes := (
          [1, 3, 4, 2], [1, 2, 6, 5], [1, 5, 7, 3],
          [8, 4, 3, 7], [8, 6, 5, 7], [8, 4, 2, 6]
        ) return
        let $pairs := 
          fold-left($face-indexes, (),
            function($pairs as array(xs:integer)*, $faceindex as array(xs:integer)) {
              $pairs,
              let $vixes := array:flatten($faceindex)
              return (
                for $i in 2 to count($vixes) return (
                  if (
                    every $pair in $pairs satisfies
                      (not($pair(1)=$vixes[$i] and $pair(2)=$vixes[$i - 1]) and
                       not($pair(1)=$vixes[$i - 1] and $pair(2)=$vixes[$i])
                      )
                  ) then array { $vixes[$i], $vixes[$i - 1] } else ()
                ),
                if (
                  every $pair in $pairs satisfies
                    (not($pair(1)=$vixes[1] and $pair(2)=$vixes[last()]) and
                     not($pair(1)=$vixes[last()] and $pair(2)=$vixes[1])
                    )
                ) then array {$vixes[1], $vixes[last()]} else ()
                (:
                for $vix at $i in (tail($vixes), head($vixes))
                where every $pair in $pairs
                satisfies (not($pair(1)=$vix and $pair(2)=$vixes[$i]) and
                           not($pair(1)=$vixes[$i] and $pair(2)=$vix))
                return array{ $vix, $vixes[$i] }
                :)
              )
            }
          )
        return
        for $pair in $pairs
        return edge:edge($vertices[$pair(1)], $vertices[$pair(2)])
      )
      else if ($kind = "space") then (
        let $vertices := this:vertices($region)
        return (
          for $i in 2 to count($vertices) return edge:edge($vertices[$i - 1], $vertices[$i]),
          edge:edge($vertices[last()], $vertices[1]) (: close loop :)
        )
      )
      else errors:error("GEOM-BADREGION", ($region, "edges"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property management
 :======================================================================
 :-->

   <!--
    : this:property-map()
    : Return the annotation properties of the box as a map. Check whether this
    : is actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return 
      if ($kind = "box") then box:property-map($region)
      else if ($kind = "block") then box:property-map($region)
      else if ($kind = "space") then util:exclude($region, $space:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the box as a map. 
    : Check whether this is actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return 
      if ($kind = "box") then box:properties($region)
      else if ($kind = "block") then box:properties($region)
      else if ($kind = "space") then ($region=>map:keys())[not(. = $space:RESERVED)]
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the box with some new properties and return the new box.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually a box.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return 
      if ($kind = "box") then box:with-properties($region, $properties)
      else if ($kind = "block") then box:with-properties($region, $properties)
      else if ($kind = "space") then
        util:merge-into($region, util:exclude($properties,$space:RESERVED))
      else errors:error("GEOM-BADREGION", ($region, "with-properties"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Box operations
 :======================================================================
 :-->

   <!--
    : this:snap()
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $rectangle in $rectangles return let $kind := this:kind($rectangle) return
      if ($kind = "box") then box:snap($rectangle)
      else if ($kind = "block") then box:snap($rectangle)
      else if ($kind = "space") then
        util:merge-into($rectangle, 
          space:space(
            round-half-to-even(space:width($rectangle)) cast as xs:integer,
            round-half-to-even(space:height($rectangle)) cast as xs:integer,
            round-half-to-even(space:edge($rectangle)) cast as xs:integer
          )
        )
      else $rectangle
    '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $rectangle in $rectangles return let $kind := this:kind($rectangle) return
      if ($kind = "box") then box:decimal($rectangle, $digits)
      else if ($kind = "block") then box:decimal($rectangle, $digits)
      else if ($kind = "space") then
        util:merge-into($rectangle,
          space:space(
            util:decimal(space:width($rectangle), $digits),
            util:decimal(space:height($rectangle), $digits),
            util:decimal(space:edge($rectangle), $digits)
          )
        )
      else $rectangle
    '/>
   </xsl:function>

   <!--
    : this:quote()
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="rectangles" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $rectangle in $rectangles return let $kind := this:kind($rectangle) return
        if ($kind = "box") then box:quote($rectangles)
        else if ($kind = "block") then box:quote($rectangles)
        else if ($kind = "space") then
          "["||this:min-px($rectangle)||","||this:min-py($rectangle)||","||
               this:max-px($rectangle)||","||this:max-py($rectangle)||"]"
        else errors:quote($rectangle)
        ,
        " "
      )
    '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for rectangles, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one rectangle
    : $other: the rectangle to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:min-px($this)=this:min-px($other) and
      this:min-py($this)=this:min-py($other) and
      this:max-px($this)=this:max-px($other) and
      this:max-py($this)=this:max-py($other) and
      ($this("kind")!="block" or (
        this:min-pz($this)=this:min-pz($other) and
        this:max-pz($this)=this:max-pz($other)
      ))
    '/>
   </xsl:function>

   <!--
    : this:mutate()
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind = "box") then box:mutate($region, $mutate)
      else if ($kind = "block") then box:mutate($region, $mutate)
      else if ($kind = "space") then
        this:as-space(
          $mutate(this:min-point($region)),
          $mutate(this:max-point($region)),
          $region
        )
      else $region
    '/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:contains-point($box, $point, $tolerance)
      else if ($kind = "block") then box:contains-point($box, $point, $tolerance)
      else (
        util:twixt(point:px($point), this:min-px($box) - $tolerance, this:max-px($box) + $tolerance) and
        util:twixt(point:py($point), this:min-py($box) - $tolerance, this:max-py($box) + $tolerance)
      )
    '/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($box) return
      if ($kind = "box") then box:contains-point($box, $point)
      else if ($kind = "block") then box:contains-point($box, $point)
      else (
        util:twixt(point:px($point), this:min-px($box), this:max-px($box)) and
        util:twixt(point:py($point), this:min-py($box), this:max-py($box))
      )
    '/>
   </xsl:function>

   <!--
    : this:intersects-box()
    : Box intersects some other box
    :-->
   <xsl:function name="this:intersects-box" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: No overlap is maximum of box &lt; minimum of other or vice versa :)
      (this:max-px($box) >= this:min-px($other)) and
      (this:min-px($box) &lt;= this:max-px($other)) and
      (this:max-py($box) >= this:min-py($other)) and
      (this:min-py($box) &lt;= this:max-py($other)) and
      ($box("kind")!="block" or (
        (this:max-pz($box) >= this:min-pz($other)) and
        (this:min-pz($box) &lt;= this:max-pz($other))
      ))
    '/>
   </xsl:function>

   <!--
    : this:intersects-box()
    :-->
   <xsl:function name="this:intersects-box" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      (: No overlap is maximum of box &lt; minimum of other or vice versa :)
      (this:max-px($box) >= this:min-px($other) - $tolerance) and
      (this:min-px($box) &lt;= this:max-px($other) + $tolerance) and
      (this:max-py($box) >= this:min-py($other) - $tolerance) and
      (this:min-py($box) &lt;= this:max-py($other) + $tolerance) and
      ($box("kind")!="block" or (
        (this:max-pz($box) >= this:min-pz($other) - $tolerance) and
        (this:min-pz($box) &lt;= this:max-pz($other) + $tolerance)
      ))
    '/>
   </xsl:function>

   <xsl:function name="this:draw-box" as="element()" visibility="private">
     <xsl:param name="x1" as="xs:numeric"/>
     <xsl:param name="y1" as="xs:numeric"/>
     <xsl:param name="x2" as="xs:numeric"/>
     <xsl:param name="y2" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:box x1="{$x1}" y1="{$y1}" x2="{$x2}" y2="{$y2}">
       <xsl:sequence select='util:as-attributes($style-properties)'/>
     </art:box>
   </xsl:function>

   <xsl:function name="this:draw-polyline" as="element()" visibility="private">
     <xsl:param name="x1" as="xs:numeric"/>
     <xsl:param name="y1" as="xs:numeric"/>
     <xsl:param name="x2" as="xs:numeric"/>
     <xsl:param name="y2" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <svg:polyline points="{$x1},{$y1} {$x2},{$y1} {$x2},{$y2} {$x1},{$y2} {$x1},{$y1}">
       <xsl:sequence select='util:as-attributes($style-properties)'/>
     </svg:polyline>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item)))
            return (
              this:draw-box(this:min-x($item), this:min-y($item), this:max-x($item), this:max-y($item), $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item)))
            return (
              this:draw-box(this:min-px($item), this:min-py($item), this:max-px($item), this:max-py($item), $style-properties)
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            return (
              this:draw-polyline(this:min-x($item), this:min-y($item), this:max-x($item), this:max-y($item), $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties := (
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
            )
            return (
              this:draw-polyline(this:min-px($item), this:min-py($item), this:max-px($item), this:max-py($item), $style-properties)
            )
          }
        )
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>

<!--rectangle.xsl-->
