<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Solids: regular tetrahedrons, cubes, octahedrons, dodecahedrons, icosahedrons,
 : ellipsoids (and spheres)
 : 
 : Copyright© Mary Holstege 2022-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/solid"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/box"
             xmlns:config="http://mathling.com/core/config"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:ellipse="http://mathling.com/geometric/ellipse"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:path="http://mathling.com/geometric/path"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:roots="http://mathling.com/core/roots"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric/solid"
             exclude-result-prefixes="this xs math map array config errors util roots v point box edge ellipse path">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/roots" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/box" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>

<!--
 :======================================================================
 : Constructors
 :======================================================================
 :-->

   <xsl:variable name="this:φ" as="xs:double">
      <xsl:sequence select='(1 + math:sqrt(5)) div 2'/>
   </xsl:variable>
   <xsl:variable name="this:ELLIPSOID-RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "center", "rx", "ry", "rx")'/>
   </xsl:variable>
   <xsl:variable name="this:FACE-RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "polygon")'/>
   </xsl:variable>
   <xsl:variable name="this:GENERIC-RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "vertices", "face-indexes")'/>
   </xsl:variable>
   <xsl:variable name="this:PLATONIC-RESERVED" as="xs:string*">
      <xsl:sequence select='($this:GENERIC-RESERVED, "center", "scale")'/>
   </xsl:variable>

   <!--
    : this:face()
    : Construct a 3D polygon, e.g. a face of a polyhedron.
    : Differs from a regular polygon wrt rendering: it will be projected.
    :-->
   <xsl:function name="this:face" as="map(xs:string, item()*)">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      map {
        "kind": "face",
        "polygon": path:polygon($edges)
      }
    '/>
   </xsl:function>

   <!--
    : this:sphere()
    : Construct a sphere with the given center and radius 
    :
    : $center: center of sphere (3D point)
    : $radius: radius of sphere
    :-->
   <xsl:function name="this:sphere" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="radius" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "sphere",
        "center": $center,
        "rx": $radius
      }
    '/>
   </xsl:function>

   <!--
    : this:ellipsoid()
    : Construct an ellipsoid with the given center and radii.
    : Only axis aligneds ellipsoids right now. If the radii are
    : all the same, will construct a sphere.
    : XYZZY: need axis angles for more general case
    :
    : $center: center of sphere (3D point)
    : $rx: radius of x-axis ellipse
    : $ry: radius of y-axis ellipse
    : $rz: radius of z-axis ellipse
    :-->
   <xsl:function name="this:ellipsoid" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="rx" as="xs:double"/>
      <xsl:param name="ry" as="xs:double"/>
      <xsl:param name="rz" as="xs:double"/>
      <xsl:sequence select='
      if ($rx = $ry and $ry = $rz) then this:sphere($center, $rx)
      else (
        map {
          "kind": "ellipsoid",
          "center": $center,
          "rx": $rx,
          "ry": $ry,
          "rz": $rz
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:polyhedron()
    : Construct a generic polyhedron with the given vertices.
    :
    : $vertices: the vertices of the polyhedron
    : $face-indexes: the definitions of the faces, as indexes into the vertex sequence
    :-->
   <xsl:function name="this:polyhedron" as="map(xs:string, item()*)">
      <xsl:param name="vertices" as="map(xs:string, item()*)*"/>
      <xsl:param name="face-indexes" as="array(xs:integer)*"/>
      <xsl:sequence select='
      map {
        "kind": "polyhedron",
        "center": point:point(avg($vertices!point:px(.)), avg($vertices!point:py(.)), avg($vertices!point:pz(.))),
        "vertices": $vertices,
        "face-indexes": $face-indexes
      }
    '/>
   </xsl:function>

   <!--
    : this:polyhedron()
    :-->
   <xsl:function name="this:polyhedron" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="vertices" as="map(xs:string, item()*)*"/>
      <xsl:param name="face-indexes" as="array(xs:integer)*"/>
      <xsl:sequence select='
      map {
        "kind": "polyhedron",
        "center": $center,
        "vertices": $vertices,
        "face-indexes": $face-indexes
      }
    '/>
   </xsl:function>

   <!--
    : this:polyhedron()
    : Convert one of the other solids into a polyhedron
    :-->
   <xsl:function name="this:polyhedron" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($solid) = ("tetrahedron", "cube", "octahedron", "icosahedron", "dodecahedron", "polyhedron", "block"))
      then (
        util:merge-into(
          this:property-map($solid),
          this:polyhedron(
            this:center($solid),
            this:vertices($solid),
            this:face-indexes($solid)
          )
        )
      ) else (
        errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:polyhedron"))
      )
    '/>
   </xsl:function>

   <!--
    : this:tetrahedron()
    : Construct a regular tetrahedron with the given edge length
    :
    : $location: placement of center of solid
    : $scale: length of sides 
    :-->
   <xsl:function name="this:tetrahedron" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "tetrahedron",
        "scale": $scale,
        "center": $location,
        "vertices": (
          point:point(1, 1, 1),
          point:point(1, -1, -1),
          point:point(-1, 1, -1),
          point:point(-1, -1, 1)
        )!point:times(., $scale div (2 * math:sqrt(2)))!point:add(., $location),
        "face-indexes": (
          [1, 3, 2], [1, 4, 2], [1, 3, 4], [2, 3, 4]
        )
      }
    '/>
   </xsl:function>

   <!--
    : this:cube()
    : Construct a regular cube with the given edge length
    :
    : $location: placement of center of solid
    : $scale: length of sides
    :-->
   <xsl:function name="this:cube" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "cube",
        "scale": $scale,
        "center": $location,
        "vertices": (
          point:point(-1,-1,-1), point:point(-1,-1,1),
          point:point(-1,1,-1), point:point(-1,1,1),
          point:point(1,-1,-1), point:point(1,-1,1),
          point:point(1,1,-1), point:point(1,1,1)
        )!point:times(., $scale div 2)!point:add(., $location),
        "face-indexes": (
          [1, 3, 4, 2], [1, 2, 6, 5], [1, 5, 7, 3],
          [8, 4, 3, 7], [8, 6, 5, 7], [8, 4, 2, 6]
        )
      }
    '/>
   </xsl:function>

   <!--
    : this:octahedron()
    : Construct a regular octahedron with the given edge length
    :
    : $location: placement of center of solid
    : $scale: length of sides
    :-->
   <xsl:function name="this:octahedron" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "octahedron",
        "scale": $scale,
        "center": $location,
        "vertices": (
          point:point(-1,0,0), point:point(1,0,0),
          point:point(0,-1,0), point:point(0,1,0),
          point:point(0,0,-1), point:point(0,0,1)
        )!point:times(., $scale div math:sqrt(2))!point:add(., $location),
        "face-indexes": (
          [4, 1, 6], [4, 1, 5], [4, 6, 2], [4, 2, 5],
          [3, 6, 1], [3, 2, 6], [3, 5, 1], [3, 2, 5]
        )
      }
    '/>
   </xsl:function>

   <!--
    : this:icosahedron()
    : Construct a regular icosahedron with the given edge length
    :
    : $location: placement of center of solid
    : $scale: length of sides
    :-->
   <xsl:function name="this:icosahedron" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "icosahedron",
        "scale": $scale,
        "center": $location,
        "vertices": (
          point:point(0, -1, -$this:φ), point:point(0, -1, $this:φ),
          point:point(0, 1, -$this:φ), point:point(0, 1, $this:φ),
          point:point(-1, -$this:φ, 0), point:point(-1, $this:φ, 0), 
          point:point(1, -$this:φ, 0), point:point(1, $this:φ, 0), 
          point:point(-$this:φ, 0, -1), point:point($this:φ, 0, -1), 
          point:point(-$this:φ, 0, 1), point:point($this:φ, 0, 1)
        )!point:times(., $scale div 2)!point:add(., $location),
        "face-indexes": (
          [1,3,9], [1,3,10], [1,5,7], [1,5,9], [1,7,10],
          [2,4,11], [2,4,12], [2,5,7], [2,5,11], [2,7,12],
          [3,6,8], [3,6,9], [3,8,10], [4,6,8], [4,6,11], [4,8,12],
          [5,9,11], [6,9,11], [7,10,12], [8,10,12]
        )
      }
    '/>
   </xsl:function>

   <!--
    : this:dodecahedron()
    : Construct a regular dodecahedron with the given edge length
    :
    : $location: placement of center of solid
    : $scale: length of sides
    :-->
   <xsl:function name="this:dodecahedron" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "dodecahedron",
        "scale": $scale,
        "center": $location,
        (: https://math.stackexchange.com/questions/2538897/equations-of-facelets-of-dodecahedron :)
        "vertices": (
           point:point(1 div $this:φ, 0, $this:φ),
           point:point($this:φ, 1 div $this:φ, 0),
           point:point(0, $this:φ, 1 div $this:φ),
           point:point(-1 div $this:φ, 0, -$this:φ),
           point:point(-$this:φ, 1 div $this:φ, 0),
           point:point(0, $this:φ, -1 div $this:φ),
           point:point(1 div $this:φ, 0, -$this:φ), 
           point:point($this:φ, -1 div $this:φ, 0), 
           point:point(0, -$this:φ, -1 div $this:φ),
           point:point(-1 div $this:φ, 0, $this:φ), 
           point:point(-$this:φ, -1 div $this:φ, 0),
           point:point(0, -$this:φ, 1 div $this:φ),
           point:point(1, 1, 1), point:point(1, 1, -1), 
           point:point(-1, 1, -1), point:point(-1, 1, 1),
           point:point(1, -1, -1), point:point(1, -1, 1),
           point:point(-1, -1, 1), point:point(-1, -1, -1)
        )!point:times(., $scale * $this:φ div 2)!point:add(., $location),
        (: distance btw points is 2/φ for these coordinates; so rescale :)
        "face-indexes": (
          [1, 10, 16, 3, 13], 
          [18, 12, 19, 10, 1],
          [1, 13, 2, 8, 18],
          [2, 8, 17, 7, 14],
          [2, 13, 3, 6, 14],
          [16, 5, 15, 6, 3],
          [15, 6, 14, 7, 4],
          [4, 7, 17, 9, 20],
          [20, 11, 5, 15, 4],
          [16, 10, 19, 11, 5],
          [18, 12, 9, 17, 8],
          [9, 12, 19, 11, 20]
        )
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $solid("kind")
    '/>
   </xsl:function>

   <!--
    : this:center()
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = ("sphere","ellipsoid","ellipsoid","tetrahedron","cube","octahedron","icosahedron","dodecahedron","polyhedron"))
      then $solid("center")
      else if ($solid("kind") = "face")
      then this:face-centroid($solid)
      else if ($solid("kind") = "block")
      then box:center($solid)
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:radius()
    :-->
   <xsl:function name="this:radius" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = "sphere")
      then $solid("rx")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:rx()
    :-->
   <xsl:function name="this:rx" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = "sphere") then $solid("rx")
      else if ($solid("kind") = "ellipsoid") then $solid("rx")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:ry()
    :-->
   <xsl:function name="this:ry" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = "sphere") then $solid("ry")
      else if ($solid("kind") = "ellipsoid") then $solid("ry")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:rz()
    :-->
   <xsl:function name="this:rz" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = "sphere") then $solid("rz")
      else if ($solid("kind") = "ellipsoid") then $solid("rz")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    :-->
   <xsl:function name="this:vertices" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = ("ellipsoid","tetrahedron","cube","octahedron","icosahedron","dodecahedron","polyhedron"))
      then $solid("vertices")
      else if ($solid("kind")="face")
      then this:polygon($solid)=>path:vertices()
      else if ($solid("kind")="block")
      then box:vertices($solid)
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:points()
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = ("ellipsoid","tetrahedron","cube","octahedron","icosahedron","dodecahedron","polyhedron"))
      then $solid("vertices")
      else if ($solid("kind")="face")
      then this:polygon($solid)=>path:points()
      else if ($solid("kind")="block")
      then box:points($solid)
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:face-indexes()
    :-->
   <xsl:function name="this:face-indexes" as="array(xs:integer)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = ("ellipsoid","tetrahedron","cube","octahedron","icosahedron","dodecahedron","polyhedron"))
      then $solid("face-indexes")
      else if ($solid("kind")="block")
      then ( (: like a cube :)
        [1, 3, 4, 2], [1, 2, 6, 5], [1, 5, 7, 3],
        [8, 4, 3, 7], [8, 6, 5, 7], [8, 4, 2, 6]
      ) else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:scale()
    :-->
   <xsl:function name="this:scale" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = ("tetrahedron","cube","octahedron","icosahedron","dodecahedron"))
      then $solid("scale")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:polygon()
    :-->
   <xsl:function name="this:polygon" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if ($solid("kind") = "face")
      then $solid("polygon")
      else errors:error("BAD-SOLID", util:quote($solid))
    '/>
   </xsl:function>

   <!--
    : this:n-faces()
    :-->
   <xsl:function name="this:n-faces" as="xs:integer">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      sum (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind = "sphere") then 1
        else if ($kind = "ellipsoid") then 1
        else if ($kind = "face") then 1
        else if ($kind = "tetrahedron") then 4
        else if ($kind = "cube") then 6
        else if ($kind = "octahedron") then 8
        else if ($kind = "icosahedron") then 20
        else if ($kind = "dodecahedron") then 12
        else if ($kind = "polyhedron") then count(this:faces($region))
        else if ($kind = "block") then 6
        else 1
      ) cast as xs:integer
    '/>
   </xsl:function>

   <!--
    : this:faces()
    :-->
   <xsl:function name="this:faces" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($solid) = "face") then $solid
      else if (this:kind($solid) = ("tetrahedron", "cube", "octahedron", "icosahedron", "dodecahedron", "polyhedron", "block"))
      then (
        let $face-properties := $solid("face-properties") return
        let $faces := 
          let $vertices := this:vertices($solid) return
          for $faceix in this:face-indexes($solid) return
          let $face-vertices :=
            for-each(array:flatten($faceix),
              function($ix as xs:integer) as map(xs:string,item()*) { $vertices[$ix] }
            )
          return (
            this:face(
              for $i in 2 to count($face-vertices)
              return edge:edge($face-vertices[$i - 1], $face-vertices[$i])
              (:
              for $v at $i in tail($face-vertices)
              return edge:edge($face-vertices[$i], $v)
              :)
            )
          )
       return
        for $ix in 1 to count($faces) return (
          $faces[$ix]=>this:with-properties(
            head(($face-properties[$ix], map {}))
          )
        )
      ) else (
        errors:error("GEOM-BADSOLID", (util:quote($solid), "solid:faces"))
      )
    '/>
   </xsl:function>

   <!--
    : this:ordered-edges()
    : All the edges, including duplicated, in face order
    :-->
   <xsl:function name="this:ordered-edges" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($solid) = "face") then (
        path:edges(this:polygon($solid))
      ) else if (this:kind($solid) = ("tetrahedron", "cube", "octahedron", "icosahedron", "dodecahedron", "polyhedron", "block")) then (
        let $vertices := this:vertices($solid) return
        for $faceindex in this:face-indexes($solid) return
        let $vixes := array:flatten($faceindex)
        return (
          for $i in 2 to count($vixes) return (
            edge:edge($vertices[$vixes[$i - 1]], $vertices[$vixes[$i]])
          ),
          edge:edge($vertices[$vixes[last()]], $vertices[$vixes[1]])
          (:
          for $vix at $i in (tail($vixes), head($vixes))
          return edge:edge($vertices[$vix], $vertices[$vixes[$i]])
          :)
        )
      ) else (
        errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:ordered-edges"))
      )
    '/>
   </xsl:function>

   <!--
    : this:edges()
    : Non-duplicated edges 
    :-->
   <xsl:function name="this:edges" as="map(xs:string, item()*)*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($solid) = "face") then (
        path:edges(this:polygon($solid))
      ) else if (this:kind($solid) = ("tetrahedron", "cube", "octahedron", "icosahedron", "dodecahedron", "polyhedron", "block")) then (
        let $vertices := this:vertices($solid) return
        let $pairs := 
          fold-left(this:face-indexes($solid), (),
            function($pairs as array(xs:integer)*, $faceindex as array(xs:integer)) {
              $pairs,
              let $vixes := array:flatten($faceindex)
              return (
                for $i in 2 to count($vixes) return (
                  if (
                    every $pair in $pairs satisfies
                      (not($pair(1)=$vixes[$i] and $pair(2)=$vixes[$i - 1]) and
                       not($pair(1)=$vixes[$i - 1] and $pair(2)=$vixes[$i])
                      )
                  ) then array { $vixes[$i - 1], $vixes[$i] } else ()
                ),
                if (
                  every $pair in $pairs satisfies
                    (not($pair(1)=$vixes[1] and $pair(2)=$vixes[last()]) and
                     not($pair(1)=$vixes[last()] and $pair(2)=$vixes[1])
                    )
                ) then array {$vixes[1], $vixes[last()]} else ()
                (:
                for $vix at $i in (tail($vixes), head($vixes))
                where every $pair in $pairs
                satisfies (not($pair(1)=$vix and $pair(2)=$vixes[$i]) and
                           not($pair(1)=$vixes[$i] and $pair(2)=$vix))
                return array{ $vix, $vixes[$i] }
                :)
              )
            }
          )
        return
        for $pair in $pairs
        return edge:edge($vertices[$pair(1)], $vertices[$pair(2)])
      ) else (
        errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:edges"))
      )
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Basic qualities
 :======================================================================
 :-->

   <!--
    : this:surface-area()
    : Total surface area of the solid. We're using an approximation for the
    : ellipsoid.
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:surface-area" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "face") then this:face-area($solid)
      else if ($kind = "sphere") then 4 * math:pi() * this:radius($solid) * this:radius($solid)
      else if ($kind = "ellipsoid") then (: approximation :)
        let $p := 1.6075E0 return
        let $ap := math:pow(this:rx($solid), $p) return
        let $bp := math:pow(this:ry($solid), $p) return
        let $cp := math:pow(this:rz($solid), $p) 
        return 4 * math:pi() * math:pow(($ap * $bp + $ap * $cp + $bp * $cp) div 3, 1 div $p)
      else if ($kind = "tetrahedron") then math:sqrt(3) * this:scale($solid) * this:scale($solid)
      else if ($kind = "cube") then 6 * this:scale($solid) * this:scale($solid)
      else if ($kind = "octahedron") then 2 * math:sqrt(3) * this:scale($solid) * this:scale($solid)
      else if ($kind = "icosahedron") then 5 * math:sqrt(3) * this:scale($solid) * this:scale($solid)
      else if ($kind = "dodecahedron") then 3 * math:sqrt(25 + 10*math:sqrt(5)) * this:scale($solid) * this:scale($solid)
      else if ($kind = "polyhedron") then sum(this:faces($solid)!this:surface-area(.))
      else if ($kind = "block") then box:surface-area($solid)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:surface-area"))
    '/>
   </xsl:function>

   <!--
    : this:volume()
    : Total volume of the solid.
    :
    : $solid: the solid
    :-->
   <xsl:function name="this:volume" as="xs:double">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "face") then 0
      else if ($kind = "sphere") then 4 * math:pi() * math:pow(this:radius($solid), 3) div 3
      else if ($kind = "ellipsoid") then 4 * math:pi() * this:rx($solid) * this:ry($solid) * this:rz($solid) div 3
      else if ($kind = "tetrahedron") then math:pow(this:scale($solid), 3) div (6 * math:sqrt(2))
      else if ($kind = "cube") then math:pow(this:scale($solid), 3)
      else if ($kind = "octahedron") then math:sqrt(2) * math:pow(this:scale($solid), 3) div 3
      else if ($kind = "icosahedron") then 5 * (3 + math:sqrt(5)) * math:pow(this:scale($solid), 3) div 12
      else if ($kind = "dodecahedron") then  (15 + 7*math:sqrt(5)) * math:pow(this:scale($solid), 3) div 4
      else if ($kind = "block") then box:width($solid)*box:height($solid)*box:depth($solid)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:volume"))
    '/>
   </xsl:function>

   <!--
    : this:circumsphere()
    : Sphere containing solid that touches the vertices (boundary for ellipsoid/sphee)
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:circumsphere" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "sphere") then $solid
      else if ($kind = "ellipsoid") then this:sphere($solid("center"), max((this:rx($solid), this:ry($solid), this:rz($solid))))
      else if ($kind = "tetrahedron") then this:sphere($solid("center"), math:sqrt(3 div 8)*this:scale($solid))
      else if ($kind = "cube") then this:sphere($solid("center"), math:sqrt(3)*this:scale($solid) div 2)
      else if ($kind = "octahedron") then this:sphere($solid("center"), this:scale($solid) div math:sqrt(2))
      else if ($kind = "icosahedron") then this:sphere($solid("center"), math:sqrt(10 + 2*math:sqrt(5)) * this:scale($solid) div 4)
      else if ($kind = "dodecahedron") then this:sphere($solid("center"), (1 + math:sqrt(5)) * math:sqrt(3) * this:scale($solid) div 4)
      else if ($kind = "block") then
        let $c := box:center($solid)
        return this:sphere($c, max(for $v in box:vertices($solid) return point:distance($c, $v)))
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:circumsphere"))
    '/>
   </xsl:function>

   <!--
    : this:insphere()
    : Sphere inside solid that is tangent to the faces (touches boundary for ellipsoid, sphere)
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:insphere" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "sphere") then $solid
      else if ($kind = "ellipsoid") then this:sphere($solid("center"), min((this:rx($solid), this:ry($solid), this:rz($solid))))
      else if ($kind = "tetrahedron") then this:sphere($solid("center"), this:scale($solid) div math:sqrt(24))
      else if ($kind = "cube") then this:sphere($solid("center"), this:scale($solid) div 2)
      else if ($kind = "octahedron") then this:sphere($solid("center"), this:scale($solid) div math:sqrt(6))
      else if ($kind = "icosahedron") then this:sphere($solid("center"), math:sqrt(3)*(3 + math:sqrt(5))*this:scale($solid) div 12)
      else if ($kind = "dodecahedron") then this:sphere($solid("center"), math:sqrt((5 div 2) + (11 div 10)*math:sqrt(5)) * this:scale($solid) div 2)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:insphere"))
    '/>
   </xsl:function>

   <!--
    : this:midsphere()
    : Sphere that is tangent to the edges. (Doesn't apply to ellispoid or sphere.)
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:midsphere" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "tetrahedron") then this:sphere($solid("center"), this:scale($solid) div math:sqrt(8))
      else if ($kind = "cube") then this:sphere($solid("center"), this:scale($solid) div math:sqrt(2))
      else if ($kind = "octahedron") then this:sphere($solid("center"), this:scale($solid) div 2)
      else if ($kind = "icosahedron") then this:sphere($solid("center"), (1 + math:sqrt(5)) * this:scale($solid) div 4)
      else if ($kind = "dodecahedron") then this:sphere($solid("center"), (3 + math:sqrt(5)) * this:scale($solid) div 4)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:midsphere"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property management
 :======================================================================
 :-->

   <!--
    : this:property-map()
    : Return the annotation properties of the solid as a map. Check whether this
    : is actually a known solid.
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "face") then util:exclude($solid, $this:FACE-RESERVED)
      else if ($kind = "sphere") then util:exclude($solid, $this:ELLIPSOID-RESERVED)
      else if ($kind = "ellipsoid") then util:exclude($solid, $this:ELLIPSOID-RESERVED)
      else if ($kind = "tetrahedron") then util:exclude($solid, $this:PLATONIC-RESERVED)
      else if ($kind = "cube") then util:exclude($solid, $this:PLATONIC-RESERVED)
      else if ($kind = "octahedron") then util:exclude($solid, $this:PLATONIC-RESERVED)
      else if ($kind = "icosahedron") then util:exclude($solid, $this:PLATONIC-RESERVED)
      else if ($kind = "dodecahedron") then util:exclude($solid, $this:PLATONIC-RESERVED)
      else if ($kind = "polyhedron") then util:exclude($solid, $this:GENERIC-RESERVED)
      else if ($kind = "block") then box:property-map($solid)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the solid
    : Check whether this is actually an solid
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "face") then ($solid=>map:keys())[not(. = $this:FACE-RESERVED)]
      else if ($kind = "sphere") then ($solid=>map:keys())[not(. = $this:ELLIPSOID-RESERVED)]
      else if ($kind = "ellipsoid") then ($solid=>map:keys())[not(. = $this:ELLIPSOID-RESERVED)]
      else if ($kind = "tetrahedron") then ($solid=>map:keys())[not(. = $this:PLATONIC-RESERVED)]
      else if ($kind = "cube") then ($solid=>map:keys())[not(. = $this:PLATONIC-RESERVED)]
      else if ($kind = "octahedron") then ($solid=>map:keys())[not(. = $this:PLATONIC-RESERVED)]
      else if ($kind = "icosahedron") then ($solid=>map:keys())[not(. = $this:PLATONIC-RESERVED)]
      else if ($kind = "dodecahedron") then ($solid=>map:keys())[not(. = $this:PLATONIC-RESERVED)]
      else if ($kind = "block") then box:properties($solid)
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the solid with some new properties and return the new solid.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually a known solid.
    : 
    : $solid: the solid
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "face") then util:merge-into($solid, util:exclude($properties, $this:FACE-RESERVED))
      else if ($kind = "sphere") then util:merge-into($solid, util:exclude($properties, $this:ELLIPSOID-RESERVED))
      else if ($kind = "ellipsoid") then util:merge-into($solid, util:exclude($properties, $this:ELLIPSOID-RESERVED))
      else if ($kind = "tetrahedron") then util:merge-into($solid, util:exclude($properties, $this:PLATONIC-RESERVED))
      else if ($kind = "cube") then util:merge-into($solid, util:exclude($properties, $this:PLATONIC-RESERVED))
      else if ($kind = "octahedron") then util:merge-into($solid, util:exclude($properties, $this:PLATONIC-RESERVED))
      else if ($kind = "icosahedron") then util:merge-into($solid, util:exclude($properties, $this:PLATONIC-RESERVED))
      else if ($kind = "dodecahedron") then util:merge-into($solid, util:exclude($properties, $this:PLATONIC-RESERVED))
      else if ($kind = "block") then box:with-properties($solid, $properties)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:with-properties"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Operations
 :======================================================================
 :-->

   <!--
    : this:snap()
    : Snap the coordinates of the points, returning the solids with snapped
    : (i.e. integer) points coordinates.
    :
    : $solids: the solids
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="solids" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $solid in $solids return let $kind := this:kind($solid) return
      if ($kind = "sphere") then
        util:merge-into($solid, 
          this:sphere(
            point:snap(this:center($solid)),
            round-half-to-even(this:radius($solid))
          )
        )
      else if ($kind = "ellipsoid") then
        util:merge-into($solid, 
          this:ellipsoid(
            point:snap(this:center($solid)),
            round-half-to-even(this:rx($solid)),
            round-half-to-even(this:ry($solid)),
            round-half-to-even(this:rx($solid))
          )
        )
      else if ($kind = "face") then
        util:merge-into(
          $solid,
          map {
            "polygon": path:snap(this:polygon($solid))
          }
        )
      (: Regular solids are a bit tricky: they may cease to be
       : regular after we snap them. Don&#x27;t snap them unless
       : precision within 1 unit is good enough. The scale
       : will not be snapped, note.
       :)
      else if ($kind = "tetrahedron") then
        let $vertices := point:snap($solid("vertices"))
        return
        util:merge-into(
          $solid,
          map {
            "center": point:snap(this:center($solid)),
            "vertices": $vertices,
            "scale": point:distance($vertices[1], $vertices[3])
          }
        )
      else if ($kind = "cube") then 
        let $vertices := point:snap($solid("vertices"))
        return
        util:merge-into(
          $solid,
          map {
            "center": point:snap(this:center($solid)),
            "vertices": $vertices,
            "scale": point:distance($vertices[1], $vertices[3])
          }
        )
      else if ($kind = "octahedron") then
        let $vertices := point:snap($solid("vertices"))
        return
        util:merge-into(
          $solid,
          map {
            "center": point:snap(this:center($solid)),
            "vertices": $vertices,
            "scale": point:distance($vertices[1], $vertices[4])
          }
        )
      else if ($kind = "icosahedron") then
        let $vertices := point:snap($solid("vertices"))
        return
        util:merge-into(
          $solid,
          map {
            "center": point:snap(this:center($solid)),
            "vertices": $vertices,
            "scale": point:distance($vertices[1], $vertices[2])
          }
        )
      else if ($kind = "dodecahedron") then
        let $vertices := point:snap($solid("vertices"))
        return
        util:merge-into(
          $solid,
          map {
            "center": point:snap(this:center($solid)),
            "vertices": $vertices,
            "scale": point:distance($vertices[1], $vertices[10])
          }
        )
      else if ($kind = "polyhedron") then
        util:merge-into(
          $solid,
          this:polyhedron(
            point:snap($solid("vertices")),
            this:face-indexes($solid)
          )
        )
      else if ($kind = "block") then box:snap($solid)
      else $solid
    '/>
   </xsl:function>

   <!--
    : this:decimal()
    : Perform decimal rounding on all the point coordinates (see util:decimal).
    :
    : $solids: the solid shapes to round
    : $digits: how many digits after the decimal point to keep
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="solids" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $solid in $solids return let $kind := this:kind($solid) return
      if ($kind = "sphere") then
        util:merge-into($solid, 
          this:sphere(
            point:decimal(this:center($solid), $digits),
            util:decimal(this:radius($solid), $digits)
          )
        )
      else if ($kind = "ellipsoid") then
        util:merge-into($solid, 
          this:ellipsoid(
            point:decimal(this:center($solid), $digits),
            util:decimal(this:rx($solid), $digits),
            util:decimal(this:ry($solid), $digits),
            util:decimal(this:rx($solid), $digits)
          )
        )
      else if ($kind = "face") then
        util:merge-into(
          $solid,
          map {
            "polygon": path:decimal(this:polygon($solid), $digits)
          }
        )
      (: Regular solids are a bit tricky: they may cease to be
       : regular after we snap them. Don&#x27;t snap them unless
       : precision within the number of digits is good enough. The scale
       : will not be snapped, note.
       :)
      else if ($kind = "tetrahedron") then
        let $vertices := point:decimal($solid("vertices"), $digits)
        return
        util:merge-into(
          $solid,
          map {
            "center": point:decimal(this:center($solid), $digits),
            "vertices": $vertices,
            "scale": util:decimal(point:distance($vertices[1], $vertices[3]), $digits)
          }
        )
      else if ($kind = "cube") then 
        let $vertices := point:decimal($solid("vertices"), $digits)
        return
        util:merge-into(
          $solid,
          map {
            "center": point:decimal(this:center($solid), $digits),
            "vertices": $vertices,
            "scale": util:decimal(point:distance($vertices[1], $vertices[3]), $digits)
          }
        )
      else if ($kind = "octahedron") then
        let $vertices := point:decimal($solid("vertices"), $digits)
        return
        util:merge-into(
          $solid,
          map {
            "center": point:decimal(this:center($solid), $digits),
            "vertices": $vertices,
            "scale": util:decimal(point:distance($vertices[1], $vertices[4]), $digits)
          }
        )
      else if ($kind = "icosahedron") then
        let $vertices := point:decimal($solid("vertices"), $digits)
        return
        util:merge-into(
          $solid,
          map {
            "center": point:decimal(this:center($solid), $digits),
            "vertices": $vertices,
            "scale": util:decimal(point:distance($vertices[1], $vertices[2]), $digits)
          }
        )
      else if ($kind = "dodecahedron") then
        let $vertices := point:decimal($solid("vertices"), $digits)
        return
        util:merge-into(
          $solid,
          map {
            "center": point:decimal(this:center($solid), $digits),
            "vertices": $vertices,
            "scale": util:decimal(point:distance($vertices[1], $vertices[10]), $digits)
          }
        )
      else if ($kind = "polyhedron") then
        util:merge-into(
          $solid,
          this:polyhedron(
            point:decimal($solid("vertices"), $digits),
            this:face-indexes($solid)
          )
        )
      else if ($kind = "block") then box:decimal($solid, $digits)
      else $solid
    '/>
   </xsl:function>

   <!--
    : this:quote()
    : Return a string value for the solids, suitable for debugging.
    : 
    : $solids: the sequence of solids to quote
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="solids" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      string-join(
        for $solid in $solids return let $kind := this:kind($solid) return
        if ($kind = "sphere") then "o"||this:radius($solid)||" "||point:quote(this:center($solid))
        else if ($kind = "ellipsoid") then "oo"||this:rx($solid)||"x"||this:ry($solid)||"x"||this:rz($solid)||" "||point:quote(this:center($solid))
        else if ($kind = "face") then "["||path:quote(this:polygon($solid))||"]"
        else if ($kind = "tetrahedron") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "cube") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "octahedron") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "icosahedron") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "dodecahedron") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "polyhedron") then "[["||point:quote(this:vertices($solid))||"]]"
        else if ($kind = "block") then box:quote($solid)
        else errors:quote($solid)
        ,
        " "
      )
    '/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for edges, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one edge
    : $other: the edge to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $this-kind := this:kind($this) return
      let $other-kind := this:kind($other)
      return
      (
        ($this-kind=$other-kind) and (
        if ($this-kind = "sphere") then
          point:same(this:center($this), this:center($other)) and
          this:radius($this) = this:radius($other)
        else if ($this-kind = "ellipsoid") then
          point:same(this:center($this), this:center($other)) and
          this:rx($this) = this:rx($other) and
          this:ry($this) = this:ry($other) and
          this:rz($this) = this:rz($other)
        else if ($this-kind = "face") then
          path:same(this:polygon($this), this:polygon($other))
        else if ($this-kind = "tetrahedron") then
          point:same(this:center($this), this:center($other)) and
          this:scale($this) = this:scale($other)
        else if ($this-kind = "cube") then
          point:same(this:center($this), this:center($other)) and
          this:scale($this) = this:scale($other)
        else if ($this-kind = "octahedron") then
          point:same(this:center($this), this:center($other)) and
          this:scale($this) = this:scale($other)
        else if ($this-kind = "icosahedron") then
          point:same(this:center($this), this:center($other)) and
          this:scale($this) = this:scale($other)
        else if  ($this-kind = "dodecahedron") then
          point:same(this:center($this), this:center($other)) and
          this:scale($this) = this:scale($other)
        else if ($this-kind = "polyhedron") then
          every $i in 1 to count(this:vertices($this))
          satisfies point:same(this:vertices($this)[$i], this:vertices($other)[$i])
        else if ($this-kind = "block") then box:same($this, $other)
        else deep-equal($this,$other)
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:mutate()
    : Run a function over a sequence of edges to produce a new sequence of
    : edges. The function maps points to points. Warning: When we mutate
    : the regular polyhedra we get a generic solid.
    : 
    : $solids: input sequence of slids
    : $mutate: function that takes a point as an argument and returns a new point
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="solids" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $solid in $solids 
      return (
        let $kind := this:kind($solid) return
        if ($kind = "sphere") then
          util:merge-into(
            $solid,
            this:sphere(
              point:mutate(this:center($solid), $mutate),
              this:radius($solid)
            )
          )
        else if ($kind = "ellipsoid") then
          util:merge-into(
            $solid,
            this:ellipsoid(
              point:mutate(this:center($solid), $mutate),
              this:rx($solid),
              this:ry($solid),
              this:rz($solid)
            )
          )
        else if ($kind = "face") then
          util:merge-into(
            $solid,
            this:face(
              path:mutate(this:polygon($solid), $mutate)=>path:edges()
            )
          )
        else if ($kind = "tetrahedron") then
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "cube") then 
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "octahedron") then
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "icosahedron") then
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "dodecahedron") then 
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "polyhedron") then
          util:merge-into(
            this:property-map($solid),
            this:polyhedron(
              point:mutate(this:vertices($solid), $mutate),
              this:face-indexes($solid)
            )
          )
        else if ($kind = "block") then box:mutate($solid, $mutate)
        else $solid
      )
    '/>
   </xsl:function>

   <!--
    : this:normal()
    : Compute the face normal vector. If you want the vertex normals, use one of the
    : point/vector perpendicular functions.
    :
    : $face: the face
    :-->
   <xsl:function name="this:normal" as="map(xs:string, item()*)">
      <xsl:param name="face" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($face)!="face") then errors:error("BAD-SOLID", (this:quote($face), "solid:normal")) else (
        let $vertices := this:vertices($face) return
        let $v1 := $vertices[1]=>point:pcoordinates() return
        let $v2 := $vertices[2]=>point:pcoordinates() return
        let $v3 := $vertices[3]=>point:pcoordinates() return
        let $e1 := $v1=>v:sub($v2) return
        let $e2 := $v2=>v:sub($v3)
        return point:vector( $e1=>v:cross($e2)=>v:normalize() )
      )
    '/>
   </xsl:function>

   <!--
    : this:face-area()
    :-->
<!-- https://math.stackexchange.com/questions/3207981/how-do-you-calculate-the-area-of-a-2d-polygon-in-3d -->
   <xsl:function name="this:face-area" as="xs:double">
      <xsl:param name="face" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($face)!="face") then errors:error("BAD-SOLID", (this:quote($face), "solid:normal")) else (
        sum(
          let $vertices := this:vertices($face)[position() &lt; last()] return
          let $p1 := ($vertices[1])=>point:pcoordinates() return
          for $i in 2 to count($vertices) - 1 return
          let $p2 := ($vertices[$i])=>point:pcoordinates() return
          let $p3 := ($vertices[$i+1])=>point:pcoordinates() return
          let $edge1 := $p2=>v:sub($p1) return
          let $edge2 := $p3=>v:sub($p1) return
          let $cp := $edge1=>v:cross($edge2)
          return ($cp=>v:magnitude() div 2)
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:face-centroid()
    :-->
<!-- https://stackoverflow.com/questions/22465832/center-point-calculation-for-3d-polygo -->
   <xsl:function name="this:face-centroid" as="map(xs:string, item()*)">
      <xsl:param name="face" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($face)!="face") then errors:error("BAD-SOLID", (this:quote($face), "solid:normal")) else (
        let $vertices := this:vertices($face)[position() &lt; last()] return
        let $p1 := ($vertices[1])=>point:pcoordinates() return
        let $data := 
          fold-left(2 to count($vertices) - 1, ($point:ORIGIN3D, 0),
            function($data as item()*, $i as xs:integer) as item()* {
              let $s := head($data) return
              let $total-area := tail($data) return
              let $p2 := point:pcoordinates($vertices[$i]) return
              let $p3 := point:pcoordinates($vertices[$i + 1]) return
              let $edge1 := $p3=>v:sub($p1) return
              let $edge2 := $p3=>v:sub($p2) return
              let $cp := $edge1=>v:cross($edge2) return
              let $area := $cp=>v:magnitude() div 2
              return (
                $s=>point:add(
                  point:vector(($p1=>v:add($p2)=>v:add($p3))=>v:times($area div 3))
                ),
                $total-area + $area
              )
            }
          )
        return (
          head($data)=>point:times(1 div tail($data))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:contains-point()
    : XYZZY INCOMPLETE 
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="solid" as="map(xs:string, item()*)"/>
      <xsl:param name="pt" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($solid) return
      if ($kind = "sphere") then point:distance(this:center($solid), $pt) &lt;= this:radius($solid)
      else if ($kind = "cube") then
         let $xs := this:vertices($solid)!point:px(.) return
         let $ys := this:vertices($solid)!point:py(.) return
         let $zs := this:vertices($solid)!point:pz(.)
         return (
           util:twixt(point:px($pt), min($xs), max($xs)) and
           util:twixt(point:py($pt), min($ys), max($ys)) and
           util:twixt(point:pz($pt), min($zs), max($zs))
         )
      else if ($kind = "block") then box:contains-point($solid, $pt)
      else errors:error("GEOM-BADSOLID", (util:quote($solid),"solid:contains-point"))
    '/>
   </xsl:function>

   <xsl:variable name="this:draw-solid-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($properties, this:property-map($item))) return
          let $faces := this:faces($item)
          return (
            $drawing("draw:draw")($faces, $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $faces := this:faces($item)
          return (
            $drawing("draw:draw")($faces, $style-properties, $drawing) 
          )
        }
      )
      '/>
   </xsl:variable>
     
   <xsl:variable name="this:draw-sphere-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $center := $drawing("draw:project")(this:center($item))
          return (
            $drawing("draw:draw")(ellipse:circle($center, this:radius($item)), $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $center := $drawing("draw:project")(this:center($item))
          return (
            $drawing("draw:draw")(ellipse:circle($center, this:radius($item)), $style-properties, $drawing) 
          )
        }
      )
      '/>
   </xsl:variable>
     
   <xsl:variable name="this:draw-ellipsoid-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
    (: Not a great drawing function: just an flattening of the ellipse, really:)
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $center := $drawing("draw:project")(this:center($item))
          return (
            $drawing("draw:draw")(ellipse:ellipse($center, this:rx($item), this:ry($item)), $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $center := $drawing("draw:project")(this:center($item))
          return (
            $drawing("draw:draw")(ellipse:ellipse($center, this:rx($item), this:ry($item)), $style-properties, $drawing)
          )
        }
      )
      '/>
   </xsl:variable>
     
   <xsl:variable name="this:draw-face-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $polygon := $drawing("draw:project")(this:polygon($item))
          return (
            $drawing("draw:draw")($polygon, $style-properties, $drawing) 
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
          let $center := $drawing("draw:project")(this:center($item)) return
          let $polygon := $drawing("draw:project")(this:polygon($item))
          return (
            $drawing("draw:draw")($polygon, $style-properties, $drawing) 
          )
        }
      )
      '/>
   </xsl:variable>

   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='
  let $kind := this:kind($item) return (
    if ($kind = "sphere") then $this:draw-sphere-impl($item, $properties, $drawing)
    else if ($kind = "ellipsoid") then $this:draw-ellipsoid-impl($item, $properties, $drawing)
    else if ($kind = "face") then $this:draw-face-impl($item, $properties, $drawing)
    else $this:draw-solid-impl($item, $properties, $drawing)
  )
     '/>
   </xsl:function>
</xsl:package>
<!--solid.xsl-->
