<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Spline calculation
 : Inspired by David J. Birnbaum's https://github.com/djbpitt/plot
 : This differs from that fairly substantially:
 : * XQuery, not XSLT
 : * Create path object, not SVG rendering of path object
 : * Leveraging existing geospatial types and operations
 : * Points do not need to be in sorted x-order (for art this is bad assumption)
 : * Some smoothing of z; not technically a spline, but it works for my purposes
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/spline"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:path="http://mathling.com/geometric/path"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric/spline"
             exclude-result-prefixes="this xs math map array errors point path edge">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>

   <!--
    : this:spline()
    : Compute a spline over the given points and return it as a path object
    : consisting of quadratic and cubic Bézier curves. The smoothing looks 
    : ahead two points. If there are only 2 points, return a simple path
    : connecting them; if only 1 return that point; if 0 return nothing.
    : Points are assumed to be in path order.
    :
    : $points: The points to compute the spline over
    : $scaling: A smoothing factor between 0 and 1; default is 0.4
    :-->
   <xsl:function name="this:spline" as="map(xs:string, item()*)?">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:param name="scaling" as="xs:double"/>
      <xsl:sequence select='
      if ($scaling &lt; 0 or $scaling > 1)
      then errors:error("ML-BADARGS", ("scaling",$scaling))
      else ()
      ,
      (:
       : If we have adjacent points that are the same, we have to drop them
       : or we&#x27;ll fail edge angle calculations and get NaN for anchor point
       : coordinates; also skip NaN coordinate points
       :)
      let $points := (
        let $n := count($points) return
        for $i in 1 to count($points) return (
          if (
            (($i = 1) or not(point:same($points[$i],  $points[$i - 1]))) and
            point:same($points[$i], $points[$i]) (: Exclude NaN :)
          ) then $points[$i] else ()
        )
      ) return
      let $n-points := count($points) (: May be less than original count($points) :)
      return (
        if ($n-points = 0) then ()
        else if ($n-points = 1) then $points
        else if ($n-points = 2) then path:path(edge:to-edges($points))
        else (
          let $spacings := (
            for $i in 1 to $n-points - 1
            return point:distance($points[$i], $points[$i + 1])
          ) return
          (: 
           : Compute the set of anchors: this is a sequence of pairs of anchor points.
           : First of pair is backwards looking, second is forwards looking.
           :)
          let $anchors := (
            for $i in 1 to $n-points - 2 return
            let $length := point:distance($points[$i], $points[$i + 2]) return
            let $dir := point:sub($points[$i + 2], $points[$i]) return
            let $normals := point:perpendiculars-with-z($dir) return
            (: Flipped because perpendiculars returned in 90°,-90° order
             : but we want the backwards one to come first here :)
            let $angles := (
              edge:angle($point:ORIGIN, $normals[2]) + 90,
              edge:angle($point:ORIGIN, $normals[1]) + 90
            ) return
            let $inclinations := (
              edge:inclination($point:ORIGIN3D, $normals[2]),
              edge:inclination($point:ORIGIN3D, $normals[1])
            ) return
            let $distances := (
              $scaling * $length * $spacings[$i] div ($spacings[$i] + $spacings[$i+1]),
              $scaling * $length * $spacings[$i+1] div ($spacings[$i] + $spacings[$i+1])
            )
            return (
              point:destination($points[$i+1], $angles[1], $inclinations[1], $distances[1]),
              point:destination($points[$i+1], $angles[2], $inclinations[2], $distances[2])
            )
          )
          return (
            (: 
             : Stitch together the spline. The ends are quadratic because we have
             : only one side to provide a control point. Everything else is a cubic
             : Bézier.
             :)
            path:path((
              edge:quad($points[1], $points[2], $anchors[1])
              ,
              for $i in 2 to $n-points - 2 return
              edge:cubic($points[$i], $points[$i+1], $anchors[2*($i - 1)], $anchors[2*$i - 1])
              ,
              edge:quad($points[last()-1], $points[last()], $anchors[last()])
            ))
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:spline()
    :-->
   <xsl:function name="this:spline" as="map(xs:string, item()*)*">
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:spline($points, 0.4)
    '/>
   </xsl:function>

</xsl:package>
<!--spline.xsl-->
