<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Transformation (2D) accumulator/wrapper
 : Builds up the appropriate transformation matrix
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/transform"
             xmlns:affine="http://mathling.com/geometric/affine"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric/transform"
             exclude-result-prefixes="this xs math map array config errors util point affine edge">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/affine" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>

   <!--
    : this:transform()
    : Construct a bare transformation matrix
    :-->
   <xsl:function name="this:transform" as="map(xs:string,item()*)">
      <xsl:sequence select='
      map {
        "kind": "transform",
        "dim": 2,
        "matrix": affine:identity2(),
        "operations": ()
      }
    '/>
   </xsl:function>

   <!--
    : this:matrix()
    : Accessor for transformation matrix
    : 
    : $transform: the transform
    :-->
   <xsl:function name="this:matrix" as="xs:double*">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $transform("matrix")
    '/>
   </xsl:function>

   <!--
    : this:matrix()
    : Compose another matrix onto the matrix
    :
    : $transform: the transform
    : $matrix: matrix to compose
    :-->
   <xsl:function name="this:matrix" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="matrix" as="xs:double*"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          $matrix,
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "matrix("||util:quote($matrix)||")")
    '/>
   </xsl:function>

   <!--
    : this:translate()
    : Compose a translation onto the matrix
    :
    : $transform: the transform
    : $tx: translation in x
    : $ty: translation in y
    :-->
   <xsl:function name="this:translate" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="tx" as="xs:double"/>
      <xsl:param name="ty" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:translation2($tx, $ty),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "translate("||$tx||" "||$ty||")")
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Compose a scaling onto the matrix
    :
    : $transform: the transform
    : $sx: scaling in x
    : $sy: scaling in y
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:scaling2($sx, $sy),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "scale("||$sx||" "||$sy||")")
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Compose a scaling about a center onto the matrix. 
    :
    : $transform: the transform
    : $center: center of scaling
    : $sx: scaling in x
    : $sy: scaling in y
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="center" as="item()"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:scaling2($center, $sx, $sy),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "translate("||point:quote($center)||") scale("||$sx||" "||$sy||") translate("||point:quote(point:minus($center))||")")
    '/>
   </xsl:function>

   <!--
    : this:reflect()
    : Compose a reflection across a center onto the matrix. 
    :
    : $transform: the transform
    : $center: center of reflection; if if is an edge of some kind, reflect 
    :   across the straight line from start to end of edge
    :-->
   <xsl:function name="this:reflect" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="center" as="item()"/>
      <xsl:sequence select='
      let $kind := point:kind($center)
      return (
        if ($kind="point") then (
          $transform=>map:put("matrix",
            affine:compose2(
              affine:reflection2($center),
              $transform("matrix")
            )
          )=>
          util:map-append("operations", "reflect("||point:quote($center)||")")
        ) else if ($kind=("edge","quad","cubic","arc")) then (
          $transform=>map:put("matrix",
            affine:compose2(
              affine:reflection2(edge:start($center), edge:end($center)),
              $transform("matrix")
            )
          )=>
          util:map-append("operations", "reflect("||point:quote(edge:start($center))||" "||point:quote(edge:end($center))||")")
        ) else (
          errors:error("GEOM-BADREGION", ($center, "reflect"))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:reflect()
    : Compose a reflection across an edge onto the matrix. 
    :
    : $transform: the transform
    : $start: start of line of reflection
    : $end: end of line of reflection
    :-->
   <xsl:function name="this:reflect" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="start" as="item()"/>
      <xsl:param name="end" as="item()"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:reflection2($start, $end),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "reflect("||point:quote($start)||" "||point:quote($end)||")")
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Compose a rotation onto the matrix.
    :
    : $transform: the transform
    : $degrees: number of degrees to rotate
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:rotation2($degrees),
          $transform("matrix")
        )
      )=>
      util:map-append("operations",
        if ($config:CANVAS-GEOMETRY)
        then "rotate("||-$degrees||")"
        else "rotate("||$degrees||")"
      )
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Compose a rotation about a center onto the matrix.
    :
    : $transform: the transform
    : $center: center of rotation
    : $degrees: number of degrees to rotate
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="center" as="item()"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:rotation2($center, $degrees),
          $transform("matrix")
        )
      )=>
      util:map-append("operations",
        if ($config:CANVAS-GEOMETRY)
        then "rotate("||-$degrees||" "||point:quote($center)||")"
        else "rotate("||$degrees||" "||point:quote($center)||")"
      )
    '/>
   </xsl:function>

   <!--
    : this:shear()
    : Compose a shearing onto the matrix.
    :
    : $transform: the transform
    : $sxy: shearing of x with respect to y
    : $syx: shearing of y with respect to x
    :-->
   <xsl:function name="this:shear" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="sxy" as="xs:double"/>
      <xsl:param name="syx" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:shearing2($sxy, $syx),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", "shear("||$sxy||" "||$syx||")")
    '/>
   </xsl:function>

   <!--
    : this:skew()
    : Compose a shearing onto the matrix, expressed as degrees of skew.
    : Shearings approach +/- infinity around 90° and 270°; use caution.
    : At 0 degrees the shearing is 0.
    :
    : $transform: the transform
    : $sx-degrees: degrees of skew of x
    : $sy-degrees: degrees of skew of y
    :-->
   <xsl:function name="this:skew" as="map(xs:string, item()*)">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:param name="sx-degrees" as="xs:double"/>
      <xsl:param name="sy-degrees" as="xs:double"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        affine:compose2(
          affine:skewing2($sx-degrees, $sy-degrees),
          $transform("matrix")
        )
      )=>
      util:map-append("operations", ("skewX("||$sx-degrees||")", "skewY("||$sy-degrees||")"))
    '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="transform" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      $transform=>map:put("matrix",
        this:matrix($transform)!util:decimal(.,$digits)
      )
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Give a string representation of the transform; suitable for using in 
    : a transform attribute in SVG. Note that SVG uses a column-major ordering
    : instead of a row-major ordering, hence the rejiggering.
    :
    : $transform: the transform
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $m := this:matrix($transform) return
      let $rowm := (
        $m[1], $m[4],
        $m[2], $m[5],
        $m[3], $m[6]
      )
      return 
        "matrix("||string-join($rowm!string(.), " ")||")"
    '/>
   </xsl:function>

   <!--
    : this:parse-matrix()
    :-->
   <xsl:function name="this:parse-matrix" as="xs:double*">
      <xsl:param name="matrix" as="xs:string"/>
      <xsl:sequence select='
      let $m := tokenize(substring-before(substring-after($matrix, "matrix("),")"), " ")!xs:double(.)
      return
      let $colm := (
        $m[1], $m[3], $m[5],
        $m[2], $m[4], $m[6]
      )
      return $colm
    '/>
   </xsl:function>

   <!--
    : this:describe()
    : Give a detailed debugging string representation of the transform.
    : Will give the full list of transforms, in reverse order (i.e. SVG style).
    : Note: this isn't SVG, because SVG doesn't have reflection as an operation, or
    : and SVG expresses skewing, not shearing. OTOH, if you avoid those operations
    : it will be valid SVG and more readable than the matrix form.
    :
    : $transform: the transform
    :-->
   <xsl:function name="this:describe" as="xs:string">
      <xsl:param name="transform" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      string-join(reverse($transform("operations")), " ")
    '/>
   </xsl:function>

</xsl:package>
<!--transform.xsl-->
