<?xml version="1.0" encoding="UTF-8"?>
<!--
 : 2D SDF functions
 : After https://www.iquilezles.org/www/articles/distfunctions2d/distfunctions2d.htm (MIT License)
 :
 : These functions all return a function that takes a point (as xs:double*)
 : and return the signed distance. Signature is xs:double* because they
 : may also return a material. The base functions don't, in general. See 
 : the operations, however.
 :
 : Copyright© Mary Holstege 2021-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/sdf/2d"
             xmlns:ann="http://mathling.com/sdf/annotations"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:f="http://mathling.com/core/callable"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/sdf/2d"
             exclude-result-prefixes="this xs math map array util f v point edge ann">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf/annotations" package-version="*"/>

   <!--
    : this:sdCircle()
    : Circle centered at origin of the given radius
    :-->
   <xsl:function name="this:sdCircle" as="map(*)">
      <xsl:param name="radius" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd2:sdCircle", function($point as xs:double*) as xs:double* {
        v:magnitude($point) - $radius
      })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedBox()
    : Box centered at origin with corners rounded with four corner sizes
    :-->
   <xsl:function name="this:sdRoundedBox" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="r" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2)
      return
        ann:f("sd2:sdRoundedBox", function($point as xs:double*) as xs:double* {
          let $rx := 
            if (v:px($point) > 0) then (
              if (v:py($point) > 0) then point:px($r) else point:py($r)
            ) else (
              if (v:pw($point) > 0) then point:pz($r) else point:pw($r)
            )
          return
          let $q := v:abs($point)=>v:sub($bv=>v:add(($rx, $rx)))
          return
            min(( max($q), 0.0 )) + v:magnitude(v:at-max($q, 0)) - $rx
        })
    '/>
   </xsl:function>

   <!--
    : this:sdBox()
    : Box centered at origin with given width and height
    :-->
   <xsl:function name="this:sdBox" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2)
      return
        ann:f("sd2:sdBox",
          function($point as xs:double*) as xs:double* {
            let $d := v:abs($point)=>v:sub($bv)
            return
              v:magnitude(v:at-max($d, 0)) + min(( max($d), 0))
          }
        )
    '/>
   </xsl:function>

   <!--
    : this:sdOrientedBox()
    : Line from a to b of thickness th
    :-->
   <xsl:function name="this:sdOrientedBox" as="map(*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="th" as="xs:double"/>
      <xsl:sequence select='
      let $av := point:pcoordinates($a, 2) return
      let $bv := point:pcoordinates($b, 2) return
      let $b-a := $bv=>v:sub($av) return
      let $l := v:magnitude($b-a) return
      let $d := v:normalize($b-a) return
      let $half-ab := $av=>v:add($bv)=>v:times(0.5) return
      let $d-matrix := (v:px($d), -v:py($d), v:py($d), v:px($d))
      return 
        ann:f("sd2:sdOrientedBox", function($point as xs:double*) as xs:double* {
          let $q := $point=>v:sub($half-ab) return
          let $q := ( (: Matrix multiplication :)
            $d-matrix[1] * v:px($q) + $d-matrix[2] * v:py($q),
            $d-matrix[3] * v:px($q) + $d-matrix[4] * v:py($q)
          ) return
          let $q := (: abs(q) - (l,th)*0.5 :)
            v:abs($q)=>v:sub(($l, $th)=>v:times(0.5))
          return
            v:magnitude(v:at-max($q, 0)) +
            min(( max((v:px($q), v:py($q))), 0))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdSegment()
    : Line segment from a to b
    :-->
   <xsl:function name="this:sdSegment" as="map(*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $av := point:pcoordinates($a, 2) return
      let $bv := point:pcoordinates($b, 2) return
      let $b-a := $bv=>v:sub($av)
      return
        ann:f("sd2:sdSegment", function($point as xs:double*) as xs:double* {
          let $p-a := $point=>v:sub($av) return
          let $h :=
            util:clamp(
              v:dot($p-a, $b-a) div v:dot2($b-a),
              0.0, 1.0
            )
          return (: |pa - ba*h| :)
            v:magnitude(
              v:sub($p-a, $b-a=>v:times($h))
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRhombus()
    : Rhombus centered at origin with width and height
    :-->
   <xsl:function name="this:sdRhombus" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2)
      return
        ann:f("sd2:sdRhombus", function($point as xs:double*) as xs:double* {
          let $q := $point=>v:abs() return
          let $h :=
            util:clamp(
              (-2.0 * v:ndot($q, $bv) + v:ndot($bv,$bv)) div v:dot2($bv),
              -1.0, 1.0
            )  return
          let $d :=
            v:magnitude(
              $q=>v:sub($bv=>v:times(0.5)=>v:multiply((1.0 - $h, 1.0 + $h)))
            )
          return
            util:sign(v:px($q)*v:py($bv) + v:py($q)*v:px($bv) - v:px($bv)*v:py($bv)) * $d
        })
    '/>
   </xsl:function>

   <!--
    : this:sdTrapezoid()
    : Trapezoid centered at the origin with lower and upper widths and given
    : height; axis oriented
    :-->
   <xsl:function name="this:sdTrapezoid" as="map(*)">
      <xsl:param name="width1" as="xs:double"/>
      <xsl:param name="width2" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $width1 := $width1 div 2 return
      let $width2 := $width2 div 2 return
      let $height := $height div 2 return
      let $k1 := ($width2, $height) return
      let $k2 := ($width2 - $width1, 2.0*$height)
      return
        ann:f("sd2:sdTrapezoid", function($point as xs:double*) as xs:double* {
          let $p := (abs(head($point)), tail($point)) return
          let $ca := (
            max((0.0, v:px($p) - (if (v:py($p) &lt; 0) then $width1 else $width2)))
            ,
            abs(v:py($p)) - $height
          ) return
          let $cb :=
            $p=>v:sub($k1)=>
              v:add(
                $k2=>v:times(
                  util:clamp(v:dot(v:sub($k1,$p),$k2) div v:dot2($k2), 0.0, 1.0)
                )
              ) return
          let $s := if (v:px($cb) &lt; 0 and v:py($ca) &lt; 0) then -1.0 else 1.0
          return $s * math:sqrt( min((v:dot2($ca), v:dot2($cb))) )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdParallelogram()
    : Parallelogram centered at the origin with given width and height
    : given amount of skew (degrees)
    :-->
   <xsl:function name="this:sdParallelogram" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="skew" as="xs:double"/>
      <xsl:sequence select='
      let $width := $width div 2 return
      let $height := $height div 2 return
      let $e := (util:radians($skew), $height) return
      let $e2 := v:dot2($e)
      return
        ann:f("sd2:sdParallelogram", function($point as xs:double*) as xs:double* {
          let $p := if (v:py($point) &lt; 0) then v:minus($point) else $point  return
          (: Horizontal edge :)
          let $w := $p=>v:sub($e) return
          let $w := (v:px($w) - util:clamp(v:px($w), -$width, $width), tail($w)) return
          let $d := (v:dot2($w), -v:py($w)) return
          (: Vertical edge :)
          let $s := v:determinant($p, $e) return
          let $p := if ($s &lt; 0.0) then v:minus($p) else $p return
          let $v := $p=>v:sub(($width, 0)) return
          let $sub :=
            $e=>v:times(util:clamp(v:dot($v, $e) div $e2, -1.0, 1.0)) return
          let $v := $v=>v:sub($sub) return
          let $dv := v:min((v:dot2($v), $width*$height - abs($s)), $d)
          return util:sign(-v:py($dv)) * math:sqrt(v:px($dv))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdEquilateralTriangle()
    : Equilateral triangle centered on the origin with sides of the given size.
    :-->
   <xsl:function name="this:sdEquilateralTriangle" as="map(*)">
      <xsl:param name="s" as="xs:double"/>
      <xsl:sequence select='
      let $k := math:sqrt(3.0) return
      let $r := $s div 2
      return
        ann:f("sd2:sdEquilateralTriangle", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)) - $r, v:py($point) + ($r div $k)) return
          let $p :=
            if (v:px($p) + $k*v:py($p) > 0) then 
              ((v:px($p) - $k*v:py($p)) div 2.0, (-$k*v:px($p) - v:py($p)) div 2.0)
            else $p
          return
          let $p := (v:px($p) - util:clamp(v:px($p), -2.0*$r, 0.0), v:py($p))
          return -util:sign(v:py($p)) * v:magnitude($p)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdIsoscelesTriangle()
    : Isosceles triangle with vertex as origin and base at height from origin with
    : base width of width. Negative height puts base below vertex.
    :-->
   <xsl:function name="this:sdIsoscelesTriangle" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $qv := ($width div 2, $height)
      return
        ann:f("sd2:sdIsoscelesTriangle", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point)) return
          let $a := $p=>v:sub($qv=>v:times(util:clamp(v:dot($p, $qv) div v:dot2($qv), 0.0, 1.0))) return
          let $b := $p=>v:sub($qv=>v:multiply((util:clamp(v:px($p) div v:px($qv), 0.0, 1.0), 1.0)) ) return
          let $k := util:sign(v:py($qv)) return
          let $d := min((v:dot2($a), v:dot2($b))) return
          let $s := max(($k*v:determinant($p,$qv), $k*(v:py($p) - v:py($qv))))
          return util:sign($s)*math:sqrt($d)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdTriangle()
    : A triangle with the given vertices
    :-->
   <xsl:function name="this:sdTriangle" as="map(*)">
      <xsl:param name="p0" as="map(xs:string, item()*)"/>
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $pv0 := point:pcoordinates($p0, 2) return
      let $pv1 := point:pcoordinates($p1, 2) return
      let $pv2 := point:pcoordinates($p2, 2) return
      let $e0 := $pv1=>v:sub($pv0) return
      let $e1 := $pv2=>v:sub($pv1) return
      let $e2 := $pv0=>v:sub($pv2)
      return
        ann:f("sd2:sdTriangle", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $v0 := $p=>v:sub($pv0) return
          let $v1 := $p=>v:sub($pv1) return
          let $v2 := $p=>v:sub($pv2) return
          let $pq0 :=
            $v0=>v:sub(
              $e0=>v:times(util:clamp(v:dot($v0,$e0) div v:dot2($e0), 0.0, 1.0))
            ) return
          let $pq1 :=
            $v1=>v:sub(
              $e1=>v:times(util:clamp(v:dot($v1,$e1) div v:dot2($e1), 0.0, 1.0))
            ) return
          let $pq2 :=
            $v2=>v:sub(
              $e2=>v:times(util:clamp(v:dot($v2,$e2) div v:dot2($e2), 0.0, 1.0))
            ) return
          let $s := util:sign( v:determinant($e0, $e2) ) return
          let $d :=
            v:min(
              v:min(
                (v:dot2($pq0), $s*v:determinant($v0, $e0)),
                (v:dot2($pq1), $s*v:determinant($v1, $e1))
              ),
              (v:dot2($pq2), $s*v:determinant($v2, $e2))
            )
          return -util:sign(v:py($d))*math:sqrt(v:px($d))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdUnevenCapsule()
    : Two circles smoothly joined: center of one is at origin, the other is
    : at (0,height).
    :-->
   <xsl:function name="this:sdUnevenCapsule" as="map(*)">
      <xsl:param name="radius1" as="xs:double"/>
      <xsl:param name="radius2" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $b := ($radius1 - $radius2) div $height return
      let $a := math:sqrt(1.0 - $b*$b)
      return
        ann:f("sd2:sdUnevenCapsule", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point)) return
          let $k := v:dot($p, (-$b, $a))
          return
            if ($k &lt; 0.0) then (
              v:magnitude($p) - $radius1
            ) else if ($k > $a*$height) then (
              v:magnitude($p=>v:sub((0.0,$height))) - $radius2
            ) else (
              v:dot($p, ($a, $b)) - $radius1
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdPentagon()
    : Regular pentagon centered at origin with given radius (distance to vertex)
    :
    : Original gave "radius" as distance to midpoint, which is bonkers, so 
    : adjusting:
    : Interior angles: 
    : sum=(n-2)*180 = 540 divided by 10 to get right triangle as half
    : slice of pie segments from middle
    : sin(54) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(54)
    :   => d(midpoint) = d(vertex) sin(54)
    : cos(54) = ½ side / d(vertex) => side = 2 d(vertex) cos(54)
    :-->
   <xsl:function name="this:sdPentagon" as="map(*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $r := $r * math:sin(util:radians(54)) return (: d(midpoint) :)
      let $angle := math:pi() div 5 return
      let $k := (math:cos($angle), math:sin($angle), math:tan($angle)) return
      let $rkz := v:pz($k)*$r return
      let $kxy := (v:px($k), v:py($k)) return
      let $nkxy := (-v:px($k), v:py($k))
      return
        ann:f("sd2:sdPentagon", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point)) return
          let $p := $p=>v:sub(
            $nkxy=>v:times(2.0*min((v:dot($nkxy, $p), 0)))
          ) return
          let $p := $p=>v:sub(
            $kxy=>v:times(2.0*min((v:dot($kxy,$p), 0)))
          ) return
          let $p := $p=>v:sub(
            (util:clamp(v:px($p), -$rkz, $rkz), $r)
          )
          return util:sign(v:py($p))*v:magnitude($p)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdHexagon()
    : Regular hexagon centered at origin with given radius (distance to vertex)
    :
    : Original gave "radius" as distance to midpoint, which is bonkers, so 
    : adjusting:
    : Interior angles: 
    : sum=(n-2)*180 = 720 divided by 12 to get right triangle as half = 60
    : slice of pie segments from middle
    : sin(60) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(60)
    :   => d(midpoint) = d(vertex) sin(60)
    : cos(60) = ½ side / d(vertex) => side = 2 d(vertex) cos(60)
    : cos(60) = 1/2 sin(60)=½√3
    :-->
   <xsl:function name="this:sdHexagon" as="map(*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $r := $r * math:sin(util:radians(60)) return (: d(midpoint) :)
      let $angle := math:pi() div 6 return
      let $k := (-math:cos($angle), math:sin($angle), math:tan($angle)) return
      let $rkz := v:pz($k)*$r return
      let $kxy := (v:px($k), v:py($k))
      return
        ann:f("sd2:sdHexagon", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs() return
          let $p := $p=>v:sub(
            $kxy=>v:times(2.0*min((v:dot($kxy,$p), 0)))
          ) return
          let $p := $p=>v:sub(
            (util:clamp(v:px($p), -$rkz, $rkz), $r)
          )
          return util:sign(v:py($p))*v:magnitude($p)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdOctogon()
    : Regular octogon centered at origin with given radius (distance to vertex)
    :
    : Original gave "radius" as distance to midpoint, which is bonkers, so 
    : adjusting:
    : Interior angles: 
    : sum=(n-2)*180 = 1080 divided by 16 to get right triangle as half = 67.5
    : slice of pie segments from middle
    : sin(67.5) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(67.5)
    :   => d(midpoint) = d(vertex) sin(67.5)
    :-->
   <xsl:function name="this:sdOctogon" as="map(*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $r := $r * math:sin(util:radians(67.5)) return (: d(midpoint) :)
      let $angle := math:pi() div 8 return
      (: let $k := (-0.9238795325, 0.3826834323, 0.4142135623) :)
      let $k := (-math:cos($angle), math:sin($angle), math:tan($angle)) return
      let $rkz := v:pz($k)*$r return
      let $kxy := (v:px($k), v:py($k)) return
      let $nkxy := (-v:px($k), v:py($k))
      return
        ann:f("sd2:sdOctogon", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs() return
          let $p := $p=>v:sub(
            $kxy=>v:times(2.0*min((v:dot($kxy, $p),0)))
          ) return
          let $p := $p=>v:sub(
            $nkxy=>v:times(2.0*min((v:dot($nkxy, $p),0)))
          ) return
          let $p := $p=>v:sub(
            (util:clamp(v:px($p), -$rkz, $rkz), $r)
          )
          return util:sign(v:py($p))*v:magnitude($p)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdStar()
    : n-sided star with given radius and interior angle, centered at origin
    :
    : $n: number of sides
    : $r: radius (distance(origin, outer vertex))
    : $m: interior angle factor [2:n]; 2 => regular ngon, n => n-pointed asterisk
    :-->
   <xsl:function name="this:sdStar" as="map(*)">
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="m" as="xs:double"/>
      <xsl:sequence select='
      (: mod function for OpenGL is not defined like XQuery mod :)
      (: Produces different values for mixed sign operands :)
      let $mod := 
        function ($x as xs:double, $y as xs:double) as xs:double
        {
           $x - $y * floor($x div $y)
        }
      return
      let $an := math:pi() div $n return
      let $en := math:pi() div $m  return (: m is between 2 and n :)
      let $racs := ($r*math:cos($an), $r*math:sin($an)) return
      let $ecs := (math:cos($en), math:sin($en)) return (: ecs=vec2(0,1) for regular polygon :)
      let $y-ratio := v:py($racs) div v:py($ecs)
      return
        ann:f("sd2:sdStar", function($point as xs:double*) as xs:double* {
          let $p := $point return
          (: Reduce to first sector: point to top :)
          let $bn := $mod(math:atan2(v:px($p), v:py($p)), 2.0*$an) - $an return
          let $p := (math:cos($bn), abs(math:sin($bn)))=>v:times(v:magnitude($p)) return
    
          (: Line SDF :)
          let $p := $p=>v:sub($racs) return
          let $p := $p=>v:add(
            $ecs=>v:times(util:clamp(-v:dot($p,$ecs), 0.0, $y-ratio))
          )
          return util:sign(v:px($p))*v:magnitude($p)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdPie()
    : Pie segment with given arc size and radius centered at the origin.
    : Arc is oriented to 0°
    :
    : $size: size of arc (degrees)
    : $radius: radius of arc
    :-->
   <xsl:function name="this:sdPie" as="map(*)">
      <xsl:param name="size" as="xs:double"/>
      <xsl:param name="radius" as="xs:double"/>
      <xsl:sequence select='
      let $aperture := 360 - $size return
      let $c := (math:sin(util:radians($aperture div 2)), math:cos(util:radians($aperture div 2)))
      return
        ann:f("sd2:sdPie", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), v:py($point)) return
          let $l := v:magnitude($p) - $radius return
          let $m :=
            v:magnitude($p=>v:sub($c=>v:times(util:clamp(v:dot($p,$c),0.0,$radius))))
          return
            max(($l, util:sign(v:determinant($p, $c))*$m))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdArc()
    : Arc segment of the given radius and thickness with opening oriented 
    : as per orientation, and the aperture with as given.
    :
    : This isn't entirely reliable and I can't figure out why: for very small
    : or 0 rb, you get positive values even along the arc towards the ends. The
    : rb padding extents outside the ends as well, you with large rb the negative
    : values will occur well outside the arc.
    : I have found that (in canvas coordinates) 0.7 works OK as a thickness
    : (i.e. half diagonal of a pixel) for a pure arc.
    :
    : $orientation: orientation of opening, degrees
    : $aperture: width of opening, degrees 
    : $ra: center radius (+/- rb) 
    : $rb: thickness (padding on each side of ra)
    :-->
   <xsl:function name="this:sdArc" as="map(*)">
      <xsl:param name="orientation" as="xs:double"/>
      <xsl:param name="aperture" as="xs:double"/>
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      let $sca := (math:sin(util:radians($orientation)), math:cos(util:radians($orientation))) return
      let $scb := (math:sin(util:radians($aperture div 2)), math:cos(util:radians($aperture div 2)))
      return
        ann:f("sd2:sdArc", function($point as xs:double*) as xs:double* {
          let $p := ( 
            abs(v:px($sca)* v:px($point) + v:py($sca) * v:py($point)),
            -v:py($sca) * v:px($point) + v:px($sca) * v:py($point)
          ) return
          let $k :=
            if (v:determinant($p, $scb) > 0.0)
            then v:dot($p, $scb)
            else v:magnitude($p)
          return
            math:sqrt( max((0.0, v:dot2($p) + $ra*$ra - 2.0*$ra*$k)) ) - $rb
        })
    '/>
   </xsl:function>

   <!--
    : this:sdHorseshoe()
    : Arced base with straight arms.
    : Angle is angle of arms (degrees)
    : Center of arcs is origin.
    : Length of arms is widths.x
    : Radius of center of circle = r
    : Radius of outer circle = r + width/2
    : Radius of inner circle = r - width/2
    : Distance from center to end of arms = length
    :-->
   <xsl:function name="this:sdHorseshoe" as="map(*)">
      <xsl:param name="angle" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="length" as="xs:double"/>
      <xsl:sequence select='
      let $angle := $angle div 2.0 return
      let $c := (math:sin(util:radians($angle)), math:cos(util:radians($angle))) return
      let $matrix := (-v:px($c), v:py($c), v:py($c), v:px($c)) return
      let $w := ($length, $width div 2.0)
      return
        ann:f("sd2:sdHorseshoe", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point)) return
          let $l := v:magnitude($p) return
          let $p := ( (: Matrix multiplication :)
            $matrix[1] * v:px($p) + $matrix[2] * v:py($p),
            $matrix[3] * v:px($p) + $matrix[4] * v:py($p)
          ) return
          let $p := (
            if (v:py($p) > 0 or v:px($p) > 0)
            then v:px($p)
            else util:sign(-v:px($c))*$l,
            if (v:px($p) > 0) then v:py($p) else $l
          ) return
          let $p := (v:px($p), abs(v:py($p) - $r))=>v:sub($w)
          return v:magnitude($p=>v:at-max(0)) + min((0, max($p)))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdLens()
    : Vesica = intersection of two disks centered at origin
    : Distance from origin at 0 will be +/-(r - d)
    :
    : $r: radius of disks
    : $d: amount of overlap
    :-->
   <xsl:function name="this:sdLens" as="map(*)">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="d" as="xs:double"/>
      <xsl:sequence select='
      let $b := math:sqrt($r*$r - $d*$d)
      return
        ann:f("sd2:sdLens", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs()
          return
            if ($d*(v:py($p) - $b) > v:px($p)*$b)
            then util:sign($d)*v:magnitude($p=>v:sub((0,$b)))
            else v:magnitude($p=>v:sub((-$d,0))) - $r
        })
    '/>
   </xsl:function>

   <!--
    : this:sdCrescent()
    : Moon shape centered at origin. Outer edge has radius ra, inner edge has
    : radius rb, overlap determined by d (center of inner edge offset by d)
    :-->
   <xsl:function name="this:sdCrescent" as="map(*)">
      <xsl:param name="d" as="xs:double"/>
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      let $a := ($ra*$ra - $rb*$rb + $d*$d) div (2.0*$d) return
      let $b := math:sqrt(max(($ra*$ra - $a*$a, 0)))
      return
        ann:f("sd2:sdCresent", function($point as xs:double*) as xs:double* {
          let $p := (v:px($point), abs(v:py($point)))
          return
            if ($d*(v:px($p)*$b - v:py($p)*$a) > $d*$d*max(($b - v:py($p),0)))
            then v:magnitude($p=>v:sub(($a,$b)))
            else max((v:magnitude($p) - $ra, -v:magnitude($p=>v:sub(($d,0))) - $rb))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedCross()
    : A cross formed of arcs with the height/width ratio as given, centered on
    : the origin.
    :
    : $radius: distance from center to point
    : $height: ratio of height to width (0,1]; use rotation to get higher
    :   that wide
    :-->
   <xsl:function name="this:sdRoundedCross" as="map(*)">
      <xsl:param name="radius" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $k := 0.5*($height+1.0 div $height)
      return
        ann:f("sd2:sdRoundedCross", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs()=>v:div($radius)
          return
            if (v:px($p) &lt; 1 and v:py($p) &lt; v:px($p)*($k - $height)+$height)
            then
              $k - math:sqrt(v:dot2($p=>v:sub((1,$k))))
            else
              math:sqrt( min((v:dot2($p=>v:sub((0,$height))), v:dot2($p=>v:sub((1,0))))) )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdEgg()
    : An egg shape: Vesica stuck on top of a circle. Center is (0,-rb).
    : 
    : $ra: radius of circle
    : $rb: rb=ra=> circle, rb>ra => squished circle, rb<ra => pointed circle
    :   smaller rb => more pointed
    :-->
   <xsl:function name="this:sdEgg" as="map(*)">
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      let $k := math:sqrt(3.0) return
      let $r := $ra - $rb
      return
        ann:f("sd2:sdEgg", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point))
          return
            (
              if (v:py($p) &lt; 0) then (
                v:magnitude($p) - $r
              ) else if ($k * (v:px($p) + $r) &lt; v:py($p)) then (
                v:magnitude($p=>v:sub((0, -$k*$r)))
              ) else (
                v:magnitude($p=>v:add(($r,0))) - 2.0*$r
              )
            ) - $rb
    })
    '/>
   </xsl:function>

   <!--
    : this:sdHeart()
    : Heart shape, with point at origin
    : Centers of circular parts at +/- (0.25s,-0.75s), radius = √2/4
    : Point between circles is (0,-s)
    :-->
   <xsl:function name="this:sdHeart" as="map(*)">
      <xsl:param name="size" as="xs:double"/>
      <xsl:sequence select='
      let $k := math:sqrt(2.0) div 4.0
      return
        ann:f("sd2:sdHeart", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), -v:py($point))=>v:div($size)
          return
            if (v:px($p)+v:py($p) > 1.0) then (
              let $v := $p=>v:sub((0.25,0.75))
              return math:sqrt(v:dot2($v)) - $k
            ) else (
              let $v1 := $p=>v:sub((0.0,1.0)) return
              let $v2 := $p=>v:sub((max($p),0)=>v:times(0.5))
              return
                util:sign(v:px($p) - v:py($p)) *
                math:sqrt(min((v:dot2($v1),v:dot2($v2))))
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdCross()
    : Cross with equal arms, centered on origin.
    : Arms joined with arc curve of given radius.
    : Exact exterior, bound interior
    : 
    : $arm: length of each of the four arms (distance from origin)
    : $thickness: thickness of each arm
    : $r: radius of joining arc
    :-->
   <xsl:function name="this:sdCross" as="map(*)">
      <xsl:param name="arm" as="xs:double"/>
      <xsl:param name="thickness" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($arm, $thickness)
      return
        ann:f("sd2:sdCross", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs() return
          let $p := if (v:py($p) > v:px($p)) then reverse($p=>v:as-dimension(2)) else $p return
          let $q := $p=>v:sub($bv) return
          let $k := max($q) return
          let $w := if ($k > 0) then $q else (v:py($bv) - v:px($p), -$k) return
          let $d := v:magnitude($w=>v:at-max(0))
          return (if ($k > 0) then $d else -$d) + $r
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedX()
    : X with rounded ends (45 degree angle). Bounding box is w X w, so 
    : each arm is √(½w²) = w/√2
    : 
    : $w: width/height of bounding box
    : $r: radius of rounding
    :-->
   <xsl:function name="this:sdRoundedX" as="map(*)">
      <xsl:param name="w" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd2:sdRoundedX", function($point as xs:double*) as xs:double* {
        let $p := $point=>v:abs() return
        let $k := 0.5 * min((v:px($p) + v:py($p), $w))
        return
          v:magnitude(v:map(function ($c) {$c - $k}, $p)) - $r
      })
    '/>
   </xsl:function>

   <!--
    : this:sdPolygon()
    : Polygon with straight edges.
    :-->
   <xsl:function name="this:sdPolygon" as="map(*)">
      <xsl:param name="v" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      (: Ensure loop is closed :)
      let $v := if (point:same($v[1], $v[last()])) then $v else ($v, $v[1]) return
      let $n := count($v)
      return
        ann:f("sd2:sdPolygon", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $sd :=
            fold-left(
              2 to $n,
              (1.0, v:dot2($p=>v:sub(point:pcoordinates($v[1], 2)))),
              function($sd, $i as xs:integer) {
                let $s := head($sd) return
                let $d := tail($sd) return
                let $vj := point:pcoordinates($v[$i - 1], 2) return
                let $vi := point:pcoordinates($v[$i], 2) return
                let $e := $vj=>v:sub($vi) return
                let $w := $p=>v:sub($vi) return
                let $b :=
                  $w=>v:sub(
                    $e=>v:times(
                      util:clamp(v:dot($w,$e) div v:dot2($e), 0.0, 1.0)
                    )
                  )
                return
                let $c := (
                  v:py($p) >= v:py($vi),
                  v:py($p) &lt; v:py($vj),
                  v:determinant($e, $w) > 0
                )
                return (
                  if ((every $b in $c satisfies $b) or
                      (every $b in $c satisfies not($b)))
                  then (
                    -$s, min(($d, v:dot2($b)))
                  ) else (
                    $s, min(($d, v:dot2($b)))
                  )
                )
              }
            )
          return
          let $s := head($sd) return
          let $d := tail($sd)
          return $s*math:sqrt($d)
        })
    '/>
   </xsl:function>

   <!--
    : this:msign()
    :-->
   <xsl:function name="this:msign" as="xs:double">
      <xsl:param name="x" as="xs:double"/>
      <xsl:sequence select='
      if ($x &lt; 0.0) then -1.0 else 1.0
    '/>
   </xsl:function>

   <!--
    : this:sdEllipse()
    : Ellipse centered at origin with given x and y axes
    :-->
   <xsl:function name="this:sdEllipse" as="map(*)">
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      if ($ra=$rb) then this:sdCircle($ra)
      else
      ann:f("sd2:sdEllipse", function($point as xs:double*) as xs:double* {
        let $p := $point=>v:abs() return
        let $flip := v:px($p) > v:py($p) return
        let $p := if ($flip) then reverse($p=>v:as-dimension(2)) else $p return
        let $a := if ($flip) then $rb else $ra return
        let $b := if ($flip) then $ra else $rb return
        
        let $l := $b*$b - $a*$a return
        
        let $m := $a*v:px($p) div $l return
        let $n := $b*v:py($p) div $l return
        let $m2 := $m*$m return
        let $n2 := $n*$n return
    
        let $c := ($m2 + $n2 - 1.0) div 3.0 return
        let $c3 := $c*$c*$c return
    
        let $d := $c3 + $m2*$n2 return
        let $q := $d + $m2*$n2 return
        let $g := $m + $m*$n2 return
    
        let $co :=
          if ($d &lt; 0.0) then (
            let $h := math:acos($q div $c3) div 3.0 return
            let $s := math:cos($h) + 2.0 return
            let $t := math:sin($h) * math:sqrt(3.0) return
            let $rx := math:sqrt( $m2 - $c*($s + $t) ) return
            let $ry := math:sqrt( $m2 - $c*($s - $t) )
            return $ry + util:sign($l)*$rx + (abs($g) div ($rx*$ry))
          ) else (
            let $h := 2.0*$m*$n*math:sqrt($d) return
            let $s := this:msign($q + $h)*math:pow(abs($q + $h), 1.0 div 3.0) return
            let $t := this:msign($q - $h)*math:pow(abs($q - $h), 1.0 div 3.0) return
            let $rx := -($s + $t) - $c*4.0 + 2.0*$m2 return
            let $ry := ($s - $t)*math:sqrt(3.0) return
            let $rm := math:sqrt( $rx*$rx + $ry*$ry )
            return ($ry div math:sqrt($rm - $rx)) + 2.0*$g div $rm
          )
        return
        let $co := ($co - $m) div 2.0 return
        let $si := math:sqrt( max((1.0 - $co*$co, 0.0)) ) return
        let $r := ($a,$b)=>v:multiply(($co, $si))
        return util:sign(v:py($p) - v:py($r))*v:magnitude($r=>v:sub($p))
      })
    '/>
   </xsl:function>

   <!--
    : this:sdParabola()
    : Parabola with minimal point at origin
    : f(x) = ax²  k = a
    : Focus = (0, 1/(4a)), semi-latus rectum p = 1/(2a)
    :-->
   <xsl:function name="this:sdParabola" as="map(*)">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      let $ik := 1.0 div $k
      return
        ann:f("sd2:sdParabola", function($point as xs:double*) as xs:double* {
          let $pos := (abs(v:px($point)), tail($point)) return
          let $p := $ik*(v:py($pos) - 0.5*$ik) div 3.0 return
          let $q := 0.25*$ik*$ik*v:px($pos) return
          let $h := $q*$q - $p*$p*$p return
          let $r := math:sqrt(abs($h)) return
          let $x :=
            if ($h > 0.0)
            then math:pow($q+$r,1.0 div 3.0) - math:pow(abs($q - $r),1.0 div 3.0)*util:sign($r - $q)
            else 2.0*math:cos(math:atan2($r,$q) div 3.0)*math:sqrt($p)
          return
            if (v:px($pos)=0)
            then v:py($pos)
            else util:sign(v:px($pos) - $x) * v:magnitude($pos=>v:sub(($x,$k*$x*$x)))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdParabolaSegment()
    : Parabola segment, running from (-w,0) to (w,0) with apex in middle at 
    : (0,h)
    :
    : f(x) = ax² + c
    : c = h, a = -h/w²
    : Focus = (0, (4ac + 1)/(4a)), semi-latus rectum p = 1/(2a)
    :-->
   <xsl:function name="this:sdParabolaSegment" as="map(*)">
      <xsl:param name="wi" as="xs:double"/>
      <xsl:param name="he" as="xs:double"/>
      <xsl:sequence select='
      let $ik := $wi*$wi div $he
      return
        ann:f("sd2:sdParabolaSegment", function($point as xs:double*) as xs:double* {
          let $pos := (abs(v:px($point)), tail($point)) return
          let $p := $ik*($he - v:py($pos) - 0.5*$ik) div 3.0 return
          let $q := v:px($pos)*$ik*$ik*0.25 return
          let $h := $q*$q - $p*$p*$p return
          let $r := math:sqrt(abs($h)) return
          let $x := min((
            if ($h > 0.0)
            then math:pow($q + $r, 1.0 div 3.0) - math:pow(abs($q - $r), 1.0 div 3.0)*util:sign($r - $q)
            else 2.0*math:cos(math:atan2($r,$q) div 3.0)*math:sqrt($p),
            $wi
          ))
          return
            (: use util:sign(pos.x - x) and negative is only within segment
             : swept to infinity vs whole parabola including x outside width :)
            if (v:px($pos)=0)
            then v:py($pos) - $he
            else util:sign($ik*(v:py($pos) - $he)+v:px($pos)*v:px($pos)) *
                 v:magnitude($pos=>v:sub(($x,$he - $x*$x div $ik)))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdQuad()
    : Quadratic Bezier signed distance. Negative is on the same side as control
    :-->
   <xsl:function name="this:sdQuad" as="map(*)">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $start := point:pcoordinates(edge:start($quad), 2) return
      let $end := point:pcoordinates(edge:end($quad), 2) return
      let $control := point:pcoordinates(edge:controls($quad)[1], 2) return
      let $A := $start (: v0 :) return
      let $B := $control (: v1 :) return
      let $C := $end  (: v2 :) return
      let $a := $B=>v:sub($A) return
      let $b := $A=>v:sub($B=>v:times(2.0))=>v:add($C) return
      let $c := $a=>v:times(2.0) return
      let $kk := 1.0 div v:dot2($b) return
      let $kx := $kk * v:dot($a,$b)
      return
        ann:f("sd2:sdQuad", function($point as xs:double*) as xs:double* {
          let $pos := $point return
          let $d := $A=>v:sub($pos) return
          let $ky := $kk * (2.0*v:dot2($a)+v:dot($d,$b)) div 3.0 return
          let $kz := $kk * v:dot($d,$a) return
    
          let $p := $ky - $kx*$kx return
          let $p3 := $p*$p*$p return
          let $q := $kx*(2.0*$kx*$kx - 3.0*$ky) + $kz return
          let $h := $q*$q + 4.0*$p3 return
    
          let $res-and-sign :=
            if ($h >= 0.0) then (
              let $h := math:sqrt($h) return
              let $x := ($h,-$h)=>v:sub(($q,$q))=>v:div(2.0) return
              let $uv := v:map( function($c) { util:sign($c) * math:pow(abs($c), 1.0 div 3.0) }, $x) return
              let $t := util:clamp(v:px($uv) + v:py($uv) - $kx, 0.0, 1.0) return
              let $qv := $d=>v:add($c=>v:add($b=>v:times($t))=>v:times($t)) return
              let $cb := $c=>v:add($b=>v:times(2.0*$t))
              return (
                v:dot2($qv), (: res :)
                v:determinant($cb, $qv) (: sign :)
              )
            ) else (
              let $z := math:sqrt(-$p) return
              let $v := math:acos( $q div ($p*$z*2.0) ) div 3.0 return
              let $m := math:cos($v) return
              let $n := math:sin($v)*1.732050808 return
              let $tx := util:clamp(($m + $m)*$z - $kx, 0.0, 1.0) return
              let $ty := util:clamp((-$n - $m)*$z - $kx, 0.0, 1.0) return
              (: let $tz := util:clamp(($n - $m)*$z - $kx, 0.0, 1.0)  return :)
              let $qx := $d=>v:add($c=>v:add($b=>v:times($tx))=>v:times($tx)) return
              let $qy := $d=>v:add($c=>v:add($b=>v:times($ty))=>v:times($ty)) return
              (: let $qz := $d=>v:add($c=>v:add($b=>v:times($tz))=>v:times($tz))  return:)
              let $dx := v:dot2($qx) return
              let $dy := v:dot2($qy) return
              (: let $dz := v:dot2($qz)  return:)
              let $sx := v:determinant($c=>v:add($b=>v:times(2.0*$tx)), $qx) return
              let $sy := v:determinant($c=>v:add($b=>v:times(2.0*$ty)), $qy)
              (: let $sz := v:determinant($c=>v:add($b=>v:times(2.0*$tz)), $qz) :)
              return (
                (: the third root cannot be the closest :)
                if ($dx &lt; $dy) then ( $dx, $sx ) else ( $dy, $sy )
              )
            )
          return
          let $res := head($res-and-sign) return
          let $sign := head(tail($res-and-sign))
          return (
            util:sign($sign)*math:sqrt($res)
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:udQuad()
    : Quadratic Bezier unsigned distance. 
    :-->
   <xsl:function name="this:udQuad" as="map(*)">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $start := point:pcoordinates(edge:start($quad), 2) return
      let $end := point:pcoordinates(edge:end($quad), 2) return
      let $control := point:pcoordinates(edge:controls($quad)[1], 2) return
      let $A := $start (: v0 :) return
      let $B := $control (: v1 :) return
      let $C := $end  (: v2 :) return
      let $a := $B=>v:sub($A) return
      let $b := $A=>v:sub($B=>v:times(2.0))=>v:add($C) return
      let $c := $a=>v:times(2.0) return
      let $kk := 1.0 div v:dot2($b) return
      
      let $kx := $kk * v:dot($a,$b)
      return
        ann:f("sd2:udQuad", function($point as xs:double*) as xs:double* {
          let $pos := $point return
          let $d := $A=>v:sub($pos) return
          let $ky := $kk * (2.0*v:dot2($a)+v:dot($d,$b)) div 3.0 return
          let $kz := $kk * v:dot($d,$a)  return
    
          let $p := $ky - $kx*$kx return
          let $p3 := $p*$p*$p return
          let $q := $kx*(2.0*$kx*$kx - 3.0*$ky) + $kz return
          let $h := $q*$q + 4.0*$p3 return
    
          let $res :=
            if ($h >= 0.0) then (
              let $h := math:sqrt($h) return
              let $x := ($h,-$h)=>v:sub(($q,$q))=>v:div(2.0) return
              let $uv := v:map( function($c) { util:sign($c) * math:pow(abs($c), 1.0 div 3.0) }, $x) return
              let $t := util:clamp(v:px($uv) + v:py($uv) - $kx, 0.0, 1.0) return
              let $qv := $d=>v:add($c=>v:add($b=>v:times($t))=>v:times($t))
              return (
                v:dot2($qv)
              )
            ) else (
              let $z := math:sqrt(-$p) return
              let $v := math:acos( $q div ($p*$z*2.0) ) div 3.0 return
              let $m := math:cos($v) return
              let $n := math:sin($v)*1.732050808 return
              let $tx := util:clamp(($m + $m)*$z - $kx, 0.0, 1.0) return
              let $ty := util:clamp((-$n - $m)*$z - $kx, 0.0, 1.0) return
              (: let $tz := util:clamp(($n - $m)*$z - $kx, 0.0, 1.0) return :)
              let $qx := $d=>v:add($c=>v:add($b=>v:times($tx))=>v:times($tx)) return
              let $qy := $d=>v:add($c=>v:add($b=>v:times($ty))=>v:times($ty)) return
              (: let $qz := $d=>v:add($c=>v:add($b=>v:times($tz))=>v:times($tz)) return :)
              let $dx := v:dot2($qx) return
              let $dy := v:dot2($qy)
              (: let $dz := v:dot2($qz) :)
              return (
                (: the third root cannot be the closest :)
                if ($dx &lt; $dy) then $dx else $dy
              )
            )
          return (
            math:sqrt($res)
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdBlobbyCross()
    : A cross where the arms are made of parabolic segments.
    : Looks better if you opRound it.
    :
    : $radius: size of arms
    : $he:
    :-->
   <xsl:function name="this:sdBlobbyCross" as="map(*)">
      <xsl:param name="radius" as="xs:double"/>
      <xsl:param name="he" as="xs:double"/>
      <xsl:sequence select='
      let $he6 := $he * 6.0 return
      let $he216 := $he * $he * 16.0 return
      let $root2over2 := math:sqrt(2.0) div 2.0
      return
        ann:f("sd2:sdBlobbyCross", function($point as xs:double*) as xs:double* {
          let $pos := $point=>v:abs()=>v:div($radius) return
          let $pos :=
            (abs(v:px($pos) - v:py($pos)), 1.0 - v:px($pos) - v:py($pos))=>
              v:div(math:sqrt(2.0))
          return
          let $p := ($he - v:py($pos) - 0.25 div $he) div $he6 return
          let $q := v:px($pos) div $he216 return
          let $h := $q*$q - $p*$p*$p return
        
          let $x :=
            min((
              if ($h > 0.0) then (
                let $r := math:sqrt($h)
                return math:pow($q + $r, 1.0 div 3.0) - math:pow(abs($q - $r), 1.0 div 3.0)*util:sign($r - $q)
              ) else (
                let $r := math:sqrt($p)
                return 2.0 * $r* math:cos(math:acos($q div ($p*$r)) div 3.0)
              ),
              $root2over2
            ))
          return
          let $z := ($x, $he*(1.0 - 2.0*$x*$x))=>v:sub($pos)
          return  util:sign(v:py($z)) * v:magnitude($z)
      })
    '/>
   </xsl:function>

   <!--
    : this:sdWobble()
    : Numbers that wobble around 
    :-->
   <xsl:function name="this:sdWobble" as="map(*)">
      <xsl:param name="ax" as="xs:double"/>
      <xsl:param name="by" as="xs:double"/>
      <xsl:param name="d" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd2:sdWobble", function($point as xs:double*) as xs:double* {
         (math:sin($ax * v:px($point)) *
          math:sin($by * v:py($point)) *
          $d)
      })
    '/>
   </xsl:function>

   <!--
    : this:sdNoise()
    : Calculate the given noise function over the point
    :-->
   <xsl:function name="this:sdNoise" as="map(*)">
      <xsl:param name="noise" as="item()"/>
      <xsl:sequence select='
      let $noisef := f:function($noise)
      return (
        ann:f("sd2:sdNoise", 
          function($point as xs:double*) as xs:double* {
            $noisef(point:vector((v:px($point),v:py($point))))
          },
          $noise
        )
      )
    '/>
   </xsl:function>

</xsl:package>
<!--sd2.xsl-->
