<?xml version="1.0" encoding="UTF-8"?>
<!--
 : 3D SDF functions
 : After https://www.iquilezles.org/www/articles/distfunctions/distfunctions.htm
 :
 : These functions all return a function that takes a point (as xs:double*)
 : and return the signed distance. Signature is xs:double* because they
 : may also return a material. The base functions don't, in general. See 
 : the operations, however.
 :
 : Copyright© Mary Holstege 2021-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/sdf/3d"
             xmlns:ann="http://mathling.com/sdf/annotations"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:f="http://mathling.com/core/callable"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/sdf/3d"
             exclude-result-prefixes="this xs math map array util f v point ann">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/random" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/distribution" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf/annotations" package-version="*"/>

   <!--
    : this:sdSphere()
    : Sphere centered at origin of the given radius
    :-->
   <xsl:function name="this:sdSphere" as="map(*)">
      <xsl:param name="radius" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd3:sdSphere", function($point as xs:double*) as xs:double* {
        v:magnitude($point) - $radius
      })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedBox()
    : Box centered at origin with corners rounded with radius
    :-->
   <xsl:function name="this:sdRoundedBox" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="depth" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2, $depth div 2)
      return
        ann:f("sd3:sdRoundedBox", function($point as xs:double*) as xs:double* {
          let $q := $point=>v:abs()=>v:sub($bv)
          return
            v:magnitude($q=>v:at-max(0)) + min((max($q), 0.0)) - $r
        })
    '/>
   </xsl:function>

   <!--
    : this:sdBox()
    : Box centered at origin with given width, height, and depth
    :-->
   <xsl:function name="this:sdBox" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="depth" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2, $depth div 2)
      return
        ann:f("sd3:sdBox", function($point as xs:double*) as xs:double* {
          let $q := $point=>v:abs()=>v:sub($bv)
          return
            v:magnitude($q=>v:at-max(0)) + min((max($q), 0.0))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdBoxFrame()
    : Frame (edges) of a box centered at origin with given width, height, and depth
    : and the given edge thickness
    :-->
   <xsl:function name="this:sdBoxFrame" as="map(*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="depth" as="xs:double"/>
      <xsl:param name="e" as="xs:double"/>
      <xsl:sequence select='
      let $bv := ($width div 2, $height div 2, $depth div 2) return
      let $ev := ($e, $e, $e)
      return
        ann:f("sd3:sdBoxFrame", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs()=>v:sub($bv) return
          let $q := v:abs($p=>v:add($ev))=>v:sub($ev)
          return
            min((
              v:magnitude(v:at-max((v:px($p),v:py($q),v:pz($q)),0)) +
                min((max((v:px($p),v:py($q),v:pz($q))),0)),
              v:magnitude(v:at-max((v:px($q),v:py($p),v:pz($q)),0)) +
                min((max((v:px($q),v:py($p),v:pz($q))),0)),
              v:magnitude(v:at-max((v:px($q),v:py($q),v:pz($p)),0)) +
                min((max((v:px($q),v:py($q),v:pz($p))),0))
            ))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdTorus()
    : Torus with large radius $ra and small cross radius $rb
    :-->
   <xsl:function name="this:sdTorus" as="map(*)">
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd3:sdTorus", function($point as xs:double*) as xs:double* {
        let $p := $point return
        let $q := (v:magnitude((v:px($p),v:pz($p))) - $ra, v:py($p))
        return v:magnitude($q) - $rb
      })
    '/>
   </xsl:function>

   <!--
    : this:sdCappedTorus()
    : A wedge of a torus with angle being the size of the torus wedge
    : ends are rounded by $rb as well
    :-->
   <xsl:function name="this:sdCappedTorus" as="map(*)">
      <xsl:param name="angle" as="xs:double"/>
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      let $sc := (math:sin(util:radians($angle div 2)), math:cos(util:radians($angle div 2))) return
      let $ra2 := $ra*$ra
      return 
        ann:f("sd3:sdCappedTorus", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), tail($point))  return
          let $pxy := (v:px($p), v:pz($p)) return
          let $k :=
            if (v:determinant($pxy,$sc) > 0)
            then v:dot($pxy, $sc)
            else v:magnitude($pxy)
          return math:sqrt( v:dot($p,$p) + $ra2 - 2.0*$ra*$k ) - $rb
        })
    '/>
   </xsl:function>

   <!--
    : this:sdLink()
    : A link centered on the origin
    : $le gives the extension (+/-y) from a torus
    : $r1 gives the radius of the central oval
    : $r2 gives radius of link loop
    : Top of link is at le + r1 + r2 (at z=0)
    : Side of link is at r1 + r2 (at z = 0)
    :-->
   <xsl:function name="this:sdLink" as="map(*)">
      <xsl:param name="le" as="xs:double"/>
      <xsl:param name="r1" as="xs:double"/>
      <xsl:param name="r2" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd3:sdLink", function($point as xs:double*) as xs:double* {
        let $q := (v:px($point), max((abs(v:py($point)) - $le, 0)), v:pz($point))
        return
          v:magnitude((v:magnitude($q=>v:as-dimension(2)) - $r1, v:pz($q))) - $r2
      })
    '/>
   </xsl:function>

   <!--
    : this:sdCylinder()
    : Infinite cylinder with given offset from x-z origin and given radius
    :-->
   <xsl:function name="this:sdCylinder" as="map(*)">
      <xsl:param name="offset" as="map(xs:string, item()*)"/>
      <xsl:param name="radius" as="xs:double"/>
      <xsl:sequence select='
      let $cv := point:pcoordinates($offset, 2)
      return
        ann:f("sd3:sdCylinder", function($point as xs:double*) as xs:double* {
          v:magnitude(
            (v:px($point), v:pz($point))=>v:sub($cv)
          ) - $radius
        })
    '/>
   </xsl:function>

   <!--
    : this:sdCone()
    : Cone of height $h with angle of extent $angle, base on x-z plane
    : Apex is at h above origin; cone opens down
    :-->
   <xsl:function name="this:sdCone" as="map(*)">
      <xsl:param name="angle" as="xs:double"/>
      <xsl:param name="h" as="xs:double"/>
      <xsl:sequence select='
      let $q := (math:tan(util:radians($angle)), -1.0)=>v:times($h)
      return
        ann:f("sd3:sdCone", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $w := ( v:magnitude((v:px($p), v:pz($p))), v:py($p) - $h ) return
          let $a := $w=>v:sub($q=>v:times(util:clamp(v:dot($w,$q) div v:dot($q,$q), 0.0, 1.0))) return
          let $b := $w=>v:sub($q=>v:multiply((util:clamp(v:px($w) div v:px($q), 0.0, 1.0), 1.0))) return
          let $k := util:sign( v:py($q) ) return
          let $d := min(( v:dot($a, $a), v:dot($b, $b) )) return
          let $s := max(( $k*v:determinant($w, $q), $k*(v:py($w) - v:py($q)) ))
          return util:sign($s)*math:sqrt($d)
        })
    '/>
   </xsl:function>

   <!--
    : this:sdPlane()
    : A flat plane with the given normal vector and offset 
    :-->
   <xsl:function name="this:sdPlane" as="map(*)">
      <xsl:param name="n" as="map(xs:string, item()*)"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      let $nn := point:pcoordinates($n, 3)=>v:normalize()
      return
        ann:f("sd3:sdPlane", function($point as xs:double*) as xs:double* {
          v:dot($point, $nn) + $height
        })
    '/>
   </xsl:function>

   <!--
    : this:sdHexPrism()
    : Hexagonal prism centered at the origin with given height and thickness
    : Height is flat side to flat side
    :-->
   <xsl:function name="this:sdHexPrism" as="map(*)">
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="thickness" as="xs:double"/>
      <xsl:sequence select='
      (: h1 and h2 are 1/2 extents :)
      let $h1 := $height div 2 return
      let $h2 := $thickness div 2 return
      (: let $k := (-0.8660254, 0.5, 0.57735) return :)
      let $k := (-math:sin(util:radians(60)), math:cos(util:radians(60)), util:cot(util:radians(60))) return
      let $kxy := (v:px($k), v:py($k)) return
      let $kzh1 := v:pz($k)*$h1
      return
        ann:f("sd3:sdHexPrism", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs() return
          let $pxy := (v:px($p), v:py($p)) return
          let $pxy :=
            $pxy=>v:sub($kxy=>v:times(2.0*min((v:dot($kxy, $pxy), 0)))) return
          let $d := (
             util:sign(v:py($pxy) - $h1) * 
             v:magnitude(
               $pxy=>v:sub(
                 (util:clamp(v:px($pxy), -$kzh1, $kzh1), $h1)
               )
             ),
             v:pz($p) - $h2
          )
          return v:magnitude(v:at-max($d, 0)) + min((max($d), 0))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdTriPrism()
    : Triangular prism centered on origin
    : Height gives height above center, so 1/2 height below for total height
    : of 1.5height
    :-->
   <xsl:function name="this:sdTriPrism" as="map(*)">
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="thickness" as="xs:double"/>
      <xsl:sequence select='
      (: h1 measures height above 0; height is split at center 2/3 above 1/3 below
       : h2 is 1/2 extent :)
      let $h1 := $height return
      let $h2 := $thickness div 2 return
      let $sin := math:sin(util:radians(60)) return
      let $cos := 0.5 (: math:sin(util:radians(60)) :)
      return
      ann:f("sd3:sdTriPrism", function($point as xs:double*) as xs:double* {
        let $p := $point return
        let $q := $p=>v:abs()
        return (
          max((v:pz($q) - $h2, max((v:px($q)*$sin+v:py($p)*$cos, -v:py($p))) - $h1*$cos))
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:sdRod()
    : A 3D line from a to b, with r as the radius of the rod
    :-->
   <xsl:function name="this:sdRod" as="map(*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $av := point:pcoordinates($a, 3) return
      let $bv := point:pcoordinates($b, 3) return
      let $b-a := $bv=>v:sub($av) return
      let $ba-dot := v:dot($b-a, $b-a)
      return
        ann:f("sd3:sdRod", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $p-a := $p=>v:sub($av) return
          let $h := util:clamp( v:dot($p-a, $b-a) div $ba-dot, 0.0, 1.0 )
          return (
            v:magnitude($p-a=>v:sub($b-a=>v:times($h))) - $r
          )
      })
    '/>
   </xsl:function>

   <!--
    : this:sdVerticalRod()
    : A 3D line of the given height, starting at the origin, 
    : with r as the radius of the rod
    :-->
   <xsl:function name="this:sdVerticalRod" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd3:sdVerticalRod", function($point as xs:double*) as xs:double* {
        let $p := (v:px($point), v:py($point) - util:clamp(v:py($point), 0.0, $h), v:pz($point))
        return v:magnitude($p) - $r
      })
    '/>
   </xsl:function>

   <!--
    : this:sdCappedCylinder()
    : Cylinder centered at x-z origin with given radius and height, sitting on
    : x-z plane
    :-->
   <xsl:function name="this:sdCappedCylinder" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      let $h := $h div 2.0 return
      let $rh := ($r, $h)
      return
        ann:f("sd3:sdCappedCylinder", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $d :=
            (v:magnitude((v:px($p), v:pz($p))), v:py($p) - $h)=>v:abs()=>v:sub($rh)
          return
            v:magnitude($d=>v:at-max(0)) + min((max($d), 0))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedCylinder()
    : Rounded cylinder centered at x-z origin with given height and radias
    : $rb is radius of rounding
    :-->
   <xsl:function name="this:sdRoundedCylinder" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:sequence select='
      let $r := $r div 2 return
      let $h := $h div 2 return
      let $k := -2.0 * $r + $rb
      return
        ann:f("sd3:sdRoundedCylinder", function($point as xs:double*) as xs:double* {
          let $p := (v:px($point), v:py($point) - $h, v:pz($point)) return
          let $d := (v:magnitude((v:px($p), v:pz($p))) + $k, abs(v:py($p)) - $h + $rb)
          return
            v:magnitude($d=>v:at-max(0)) + min((max($d), 0)) - $rb
        })
    '/>
   </xsl:function>

   <!--
    : this:sdCappedCone()
    : Fustrum of a cone: top chopped off, centered on the origin in the x-z plane
    : height is $h, $r1 is base radius, $r2 is top radius
    :-->
   <xsl:function name="this:sdCappedCone" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="r1" as="xs:double"/>
      <xsl:param name="r2" as="xs:double"/>
      <xsl:sequence select='
      let $h := $h div 2 return
      let $k1 := ($r2, $h) return
      let $k2 := ($r2 - $r1, 2.0*$h) return
      let $k2dot := v:dot($k2, $k2)
      return
        ann:f("sd3:sdCappedCone", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $q := ( v:magnitude((v:px($p), v:pz($p))), v:py($p) - $h ) return
          let $ca := (
            v:px($q) - min((v:px($q), if (v:py($q) &lt; 0) then $r1 else $r2)),
            abs(v:py($q)) - $h
          ) return
          let $cb := (
            $q=>v:sub($k1)=>
                v:add( $k2=>
                  v:times(util:clamp(v:dot($k1=>v:sub($q), $k2) div $k2dot, 0.0, 1.0)))
          ) return
          let $s := if (v:px($cb) &lt; 0 and v:py($ca) &lt; 0) then -1.0 else 1.0
          return
            $s * math:sqrt( min((v:dot($ca,$ca),v:dot($cb,$cb))) )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdSolidAngle()
    : A chunk of a sphere from origin to positive y with the given angle and radius
    : of intersection with the sphere
    :-->
   <xsl:function name="this:sdSolidAngle" as="map(*)">
      <xsl:param name="angle" as="xs:double"/>
      <xsl:param name="ra" as="xs:double"/>
      <xsl:sequence select='
      let $c := (math:sin(util:radians($angle)), math:cos(util:radians($angle))) return
      let $ra := 2*$ra
      return
        ann:f("sd3:sdSolidAngle", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $q := ( v:magnitude((v:px($p), v:pz($p))), v:py($p) ) return
          let $l := v:magnitude($q) - $ra return
          let $m := v:magnitude($q=>v:sub($c=>v:times(util:clamp(v:dot($q,$c),0.0,$ra))))
          return max(($l, util:sign(v:determinant($q,$c))*$m))
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRoundedCone()
    : Rounded fustrum of cone, with spherical endcaps of the given radii; tangent
    : point at origin and at origin of x-z plane at y=height
    :-->
   <xsl:function name="this:sdRoundedCone" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="r1" as="xs:double"/>
      <xsl:param name="r2" as="xs:double"/>
      <xsl:sequence select='
      let $h :=  $h div 2.0 return
      let $b := ($r1 - $r2) div $h return
      let $a := math:sqrt(1.0 - $b*$b) return
      let $ah := $a*$h
      return
        ann:f("sd3:sdRoundedCone", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $q := ( v:magnitude((v:px($p), v:pz($p))), v:py($p) - $h + $r2 ) return
          let $k := v:dot($q, (-$b, $a))
          return
            if ($k &lt; 0.0) then v:magnitude($q) - $r1
            else if ($k > $ah) then v:magnitude($q=>v:sub((0,$h))) - $r2
            else v:dot($q, ($a, $b)) - $r1
        })
    '/>
   </xsl:function>

   <!--
    : this:sdEllipsoid()
    : (Not exact)
    :-->
   <xsl:function name="this:sdEllipsoid" as="map(*)">
      <xsl:param name="ra" as="xs:double"/>
      <xsl:param name="rb" as="xs:double"/>
      <xsl:param name="rc" as="xs:double"/>
      <xsl:sequence select='
      if ($ra=$rb and $ra=$rc) then this:sdSphere($ra)
      else
      let $rv := ($ra, $rb, $rc)
      return
        ann:f("sd3:sdEllipsoid", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $k0 := v:magnitude($p=>v:divide($rv)) return
          let $k1 := v:magnitude($p=>v:divide($rv=>v:multiply($rv)))
          return
            if ($k1=0) (: p=origin :)
            then -max($rv)
            else $k0*($k0 - 1.0) div $k1
        })
    '/>
   </xsl:function>

   <!--
    : this:sdRhombus()
    : Rounded rhombus centered on origin with total width (vertex to vertex)
    : in one direction width1 and the other width2 and total height height
    : $ra is rounding radius on sides 0.0=no rounding
    :-->
   <xsl:function name="this:sdRhombus" as="map(*)">
      <xsl:param name="width1" as="xs:double"/>
      <xsl:param name="width2" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="ra" as="xs:double"/>
      <xsl:sequence select='
      let $la := $width1 div 2 return
      let $lb := $width2 div 2 return
      let $h := $height div 2 return
      let $b := ($la, $lb) return
      let $bdot := v:dot2($b)
      return
        ann:f("sd3:sdRhombus", function($point as xs:double*) as xs:double* {
          let $p := $point=>v:abs() return
          let $pxz := (v:px($p), v:pz($p)) return
          let $f :=
            util:clamp(
              v:ndot($b, $b=>v:sub($pxz=>v:times(2.0))) div $bdot, -1.0, 1.0
            ) return
          let $q := (
            (: v:sign(v:determinant($pxz, $b)) * :)
            v:sign(v:px($pxz)*v:py($b) + v:py($pxz)*v:px($b) - v:px($b)*v:py($b))*
            v:magnitude(
              $pxz=>v:sub(
                $b=>v:multiply((1.0 - $f, 1.0 + $f))=>v:times(0.5)
              )
            ) - $ra,
            v:py($p) - $h
          )
          return
            v:magnitude($q=>v:at-max(0)) + min((max($q), 0)) 
        })
    '/>
   </xsl:function>

   <!--
    : this:sdOctahedron()
    : Regular octahedron centered on origin with sides of length $s
    :-->
   <xsl:function name="this:sdOctahedron" as="map(*)">
      <xsl:param name="s" as="xs:double"/>
      <xsl:sequence select='
      let $tan := -math:tan(30)
      return
      ann:f("sd3:sdOctahedron", function($point as xs:double*) as xs:double* {
        let $p := $point=>v:abs() return
        let $m := sum($p) - $s return
        let $q :=
          if (3.0 * v:px($p) &lt; $m) then $p
          else if (3.0 * v:py($p)&lt; $m) then (v:py($p), v:pz($p), v:px($p))
          else if (3.0 * v:pz($p)&lt; $m) then (v:pz($p), v:px($p), v:py($p))
          else ()
        return (
          if (empty($q)) then $m*$tan
          else (
            let $k := util:clamp( 0.5*(v:pz($q) - v:py($q) + $s), 0.0, $s)
            return v:magnitude((v:px($q), v:py($q) - $s + $k, v:pz($q) - $k))
          )
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:sdApproxOctahedron()
    : Regular octahedron centered on origin with sides of length $s
    : Faster but not anywhere close to exact, just bounded
    :-->
   <xsl:function name="this:sdApproxOctahedron" as="map(*)">
      <xsl:param name="s" as="xs:double"/>
      <xsl:sequence select='
      let $tan := -math:tan(30)
      return
      ann:f("sd3:sdApproxOctahedron", function($point as xs:double*) as xs:double* {
        let $p := $point=>v:abs()
        return (sum($p) - $s)*$tan 
      })
    '/>
   </xsl:function>

   <!--
    : this:sdPyramid()
    : Pyramid with height $h and base with sides $b planted on the xz plane
    :-->
   <xsl:function name="this:sdPyramid" as="map(*)">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $m2 := $h*$h + 0.25*$b*$b
      return
        ann:f("sd3:sdPyramid", function($point as xs:double*) as xs:double* {
          let $p := (abs(v:px($point)), v:py($point), abs(v:pz($point))) return
          let $p :=
            if (v:pz($p) > v:px($p)) then (
              (v:pz($p), v:py($p), v:px($p))
            ) else (
              $p
            ) return
          let $p := $p=>v:sub((0.5*$b, 0.0, 0.5*$b)) return
          let $q := (v:pz($p), $h*v:py($p) - 0.5*$b*v:px($p), $h*v:px($p) + 0.5*$b*v:py($p)) return
          let $d2 :=
            if (min((v:py($q), -v:px($q)*$m2 - v:py($q)*0.5)) > 0.0) then (
              0.0
            ) else (
              let $s := max((-v:px($q), 0.0)) return
              let $t := util:clamp( (v:py($q) - 0.5*v:pz($p)) div ($m2 + 0.25*$b*$b), 0.0, 1.0) return
    
              let $a := $m2 * (v:px($q) + $s)*(v:px($q) + $s) + v:py($q)*v:py($q) return
              let $b := $m2 * (v:px($q) + 0.5*$t)*(v:px($q) + 0.5*$t) + (v:py($q) - $m2*$t)*(v:py($q) - $m2*$t)
              return min(($a, $b))
            )
          return
            util:sign( max((v:pz($q), -v:py($p))) ) *
              math:sqrt( ($d2 + v:pz($q)*v:pz($q)) div $m2 )
        })
    '/>
   </xsl:function>

   <!--
    : this:udTriangle()
    : A flat triangle defined by three points. (Unsigned distance: on triangle =0)
    :-->
   <xsl:function name="this:udTriangle" as="map(*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $av :=  point:pcoordinates($a, 3) return
      let $bv :=  point:pcoordinates($b, 3) return
      let $cv :=  point:pcoordinates($c, 3) return
      let $b-a := $bv=>v:sub($av) return
      let $c-b := $cv=>v:sub($bv) return
      let $a-c := $av=>v:sub($cv) return
      let $nor := $b-a=>v:cross($a-c) return
      let $baXnor := $b-a=>v:cross($nor) return
      let $cbXnor := $c-b=>v:cross($nor) return
      let $acXnor := $a-c=>v:cross($nor)
      return
        ann:f("sd3:udTriangle", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $p-a := $p=>v:sub($av) return
          let $p-b := $p=>v:sub($bv) return
          let $p-c := $p=>v:sub($cv)
          return
            if (
              util:sign(v:dot($baXnor, $p-a)) +
              util:sign(v:dot($cbXnor, $p-b)) +
              util:sign(v:dot($acXnor, $p-c)) &lt; 2
            ) then (
              min((
                v:dot2($b-a=>v:times(util:clamp(v:dot($b-a,$p-a) div v:dot2($b-a), 0.0, 1.0))=>v:sub($p-a)),
                v:dot2($c-b=>v:times(util:clamp(v:dot($c-b,$p-b) div v:dot2($c-b), 0.0, 1.0))=>v:sub($p-b)),
                v:dot2($a-c=>v:times(util:clamp(v:dot($a-c,$p-c) div v:dot2($a-c), 0.0, 1.0))=>v:sub($p-c))
              ))
            ) else (
              v:dot($nor, $p-a)*v:dot($nor, $p-a) div v:dot2($nor)
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:udQuad()
    : A flat quadrangle defined by 4 points. (Unsigned distance: on quad =0)
    :-->
   <xsl:function name="this:udQuad" as="map(*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:param name="d" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $av :=  point:pcoordinates($a, 3) return
      let $bv :=  point:pcoordinates($b, 3) return
      let $cv :=  point:pcoordinates($c, 3) return
      let $dv :=  point:pcoordinates($d, 3) return
      let $b-a := $bv=>v:sub($av) return
      let $c-b := $cv=>v:sub($bv) return
      let $d-c := $dv=>v:sub($cv) return
      let $a-d := $av=>v:sub($dv) return
      let $nor := $b-a=>v:cross($a-d) return
      let $baXnor := $b-a=>v:cross($nor) return
      let $cbXnor := $c-b=>v:cross($nor) return
      let $dcXnor := $d-c=>v:cross($nor) return
      let $adXnor := $a-d=>v:cross($nor)
      return
        ann:f("sd3:udQuad", function($point as xs:double*) as xs:double* {
          let $p := $point return
          let $p-a := $p=>v:sub($av) return
          let $p-b := $p=>v:sub($bv) return
          let $p-c := $p=>v:sub($cv) return
          let $p-d := $p=>v:sub($dv)
          return
            if (
              util:sign(v:dot($baXnor, $p-a)) +
              util:sign(v:dot($cbXnor, $p-b)) +
              util:sign(v:dot($dcXnor, $p-c)) +
              util:sign(v:dot($adXnor, $p-d)) &lt; 3
            ) then (
              min((
                v:dot2($b-a=>v:times(util:clamp(v:dot($b-a,$p-a) div v:dot2($b-a), 0.0, 1.0))=>v:sub($p-a)),
                v:dot2($c-b=>v:times(util:clamp(v:dot($c-b,$p-b) div v:dot2($c-b), 0.0, 1.0))=>v:sub($p-b)),
                v:dot2($d-c=>v:times(util:clamp(v:dot($d-c,$p-c) div v:dot2($d-c), 0.0, 1.0))=>v:sub($p-c)),
                v:dot2($a-d=>v:times(util:clamp(v:dot($a-d,$p-c) div v:dot2($a-d), 0.0, 1.0))=>v:sub($p-d))
              ))
            ) else (
              v:dot($nor, $p-a)*v:dot($nor, $p-a) div v:dot2($nor)
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:sdWobble()
    :-->
   <xsl:function name="this:sdWobble" as="map(*)">
      <xsl:param name="ax" as="xs:double"/>
      <xsl:param name="by" as="xs:double"/>
      <xsl:param name="cz" as="xs:double"/>
      <xsl:param name="d" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sd3:sdWobble", function($point as xs:double*) as xs:double* {
         (math:sin($ax * v:px($point)) *
          math:sin($by * v:py($point)) *
          math:sin($cz * v:pz($point)) * $d)
      })
    '/>
   </xsl:function>

   <!--
    : this:sdNoise()
    :-->
   <xsl:function name="this:sdNoise" as="map(*)">
      <xsl:param name="noise" as="item()"/>
      <xsl:sequence select='
      let $noisef := f:function($noise)
      return (
        ann:f("sd3:sdNoise",
          function($point as xs:double*) as xs:double* {
            $noisef(point:vector((v:px($point),v:py($point),v:pz($point))))
          },
          $noise
        )
      )
    '/>
   </xsl:function>

</xsl:package>
<!--sd3.xsl-->
