<?xml version="1.0" encoding="UTF-8"?>
<!--
 : SDF support functions
 :
 : SDF modifiers 
 : After https://www.iquilezles.org/www/articles/distfunctions/distfunctions.htm
 : and https://iquilezles.org/www/articles/functions/functions.htm
 : Also some aspects from https://github.com/electricsquare/raymarching-workshop
 : See also https://www.alanzucconi.com/2016/07/01/surface-shading/ for more
 : information on surface lighting models
 :
 : Modifier functions, like SDF functions, return a function that 
 : takes a point and returns the SDF+material.
 :
 : Rendering functions also return functions which take the vector returned
 : by the ray caster (distance, material, iterations), the ray origin and 
 : direction, and the overall SDF function (needed for normal calculations etc.)
 : Rendering functions return colours in linear (XYZ) colour space even
 : though for the most part they take RGB colours as constant parameters.
 : The main render function will need to map back to RGB at the end.
 :
 : As of 202302 these functions do not return annotated functions, but wrapped
 : callables. You'll need to use callable:function() to get the actual function to use.
 : sdf:make-point-render() and sdf:make-vector-render() take care of this for you, 
 : but when using SDFs outside of ray casting contexts, you'll have to manage this.
 :
 : Copyright© Mary Holstege 2021-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
  :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/sdf"
             xmlns:affine="http://mathling.com/geometric/affine"
             xmlns:ann="http://mathling.com/sdf/annotations"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/rectangle"
             xmlns:config="http://mathling.com/core/config"
             xmlns:cs="http://mathling.com/colour/space"
             xmlns:dist="http://mathling.com/type/distribution"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:ellipse="http://mathling.com/geometric/ellipse"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:f="http://mathling.com/core/callable"
             xmlns:gradient="http://mathling.com/svg/gradients"
             xmlns:light="http://mathling.com/type/light"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:path="http://mathling.com/geometric/path"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:rand="http://mathling.com/core/random"
             xmlns:rgb="http://mathling.com/colour/rgb"
             xmlns:sd2="http://mathling.com/sdf/2d"
             xmlns:sd3="http://mathling.com/sdf/3d"
             xmlns:slot="http://mathling.com/type/slot"
             xmlns:solid="http://mathling.com/geometric/solid"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:wrapper="http://mathling.com/type/wrapper"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             
             xmlns:xyz="http://mathling.com/colour/xyz"
             version="3.0"
             name="http://mathling.com/sdf"
             exclude-result-prefixes="this xs math map array errors config rand dist wrapper slot light util f v affine point box edge path ellipse solid gradient rgb xyz cs sd2 sd3 ann">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/random" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/distribution" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/wrapper" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/slot" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/light" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/affine" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/rectangle" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/solid" package-version="*"/>
   <xsl:use-package name="http://mathling.com/svg/gradients" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/rgb" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/xyz" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/space" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf/2d" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf/3d" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf/annotations" package-version="*"/>

<!--
 :======================================================================
 : SDF functions
 :======================================================================
 :-->

   <!--
    : this:trace()
    : Pass through to function, with tracing
    :-->
   <xsl:function name="this:trace" as="map(*)">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $fn := f:function($f)
      return
      ann:f(util:function-name($f), 
        function($point as xs:double*) as xs:double* {
          $fn($point)=>trace(util:quote($f)||"("||util:quote($point)||")")
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:as-point()
    : Pass through to point-as-point function, converting from vector
    : $f is function(map(xs:string,item()*)) as map(xs:string,item()*) or a callable
    : wrapper of same
    : Result is callable function(xs:double*) as xs:double*
    :-->
   <xsl:function name="this:as-point" as="map(*)">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      let $fn := f:function($f)
      return
      ann:f(util:function-name($f), 
        function($point as xs:double*) as xs:double* {
          $fn(point:vector($point))=>point:pcoordinates()
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:as-vector()
    : Pass through to make a point-as-point function, converting from vector
    : Can be used to convert these functions to mutation functions, for example
    : $f is function(xs:double*) as xs:double* or a f: wrapper of same
    : Result is callable function(map(xs:string,item()*)) as map(xs:string,item()*)
    :-->
   <xsl:function name="this:as-vector" as="map(*)">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      let $fn := f:function($f)
      return
      ann:f(util:function-name($f),
        function($point as map(xs:string,item()*)) as map(xs:string,item()*) {
          $fn(point:pcoordinates($point))=>point:vector()
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:opConstant()
    : Return a constant value, for any point.
    : $k: the constant value
    :-->
   <xsl:function name="this:opConstant" as="map(*)">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:opConstant", function($point as xs:double*) as xs:double* {
        $k
      })
    '/>
   </xsl:function>

   <!--
    : this:opMaterial()
    : Append a constant value as an additional coordinate.
    : $material: the constant value
    :-->
   <xsl:function name="this:opMaterial" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="material" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opMaterial",
        function($point as xs:double*) as xs:double* {
          ($basef($point), $material)
        },
        ($base, $material)
      )
    '/>
   </xsl:function>

   <!--
    : this:opTranslate()
    : Translate the point before applying the base function.
    : This means the sign is the opposite of what you expect: to move a box,
    : say, to center at (1,1) do opTranslate(-1,-1)
    :-->
   <xsl:function name="this:opTranslate" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="loc" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:translation3(point:px($loc), point:py($loc), point:pz($loc)) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opTranslate", function($point as xs:double*) as xs:double* {
            $basef(affine:affine3($point, $matrix))
          },
          ($base, $loc)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRotate()
    : Rotate (2D) the point before applying the base function.
    :-->
   <xsl:function name="this:opRotate" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="rotation" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:rotation2($rotation) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opRotate",
          function($point as xs:double*) as xs:double* {
            $basef(affine:affine2($point, $matrix))
          },
          ($base, $rotation)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRotate()
    : Rotate (3D) the point before applying the base function.
    :-->
   <xsl:function name="this:opRotate" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="roll" as="xs:double"/>
      <xsl:param name="pitch" as="xs:double"/>
      <xsl:param name="yaw" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:rotation3($roll, $pitch, $yaw) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opRotate",
          function($point as xs:double*) as xs:double* {
            $basef(affine:affine3($point=>v:as-dimension(3), $matrix))
          },
          ($base, $roll, $pitch, $yaw)
        )
    '/>
   </xsl:function>

   <!--
    : this:opScale()
    : Scale the point before applying the base function.
    :-->
   <xsl:function name="this:opScale" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opScale",
        function($point as xs:double*) as xs:double* {
          $basef($point=>v:div($scale))=>v:times($scale)
        },
        ($base, $scale)
      )
    '/>
   </xsl:function>

   <!--
    : this:opReflect()
    : Reflect the point across a center before applying the base function.
    :-->
   <xsl:function name="this:opReflect" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:reflection3($center) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opReflect",
          function($point as xs:double*) as xs:double* {
            $basef(affine:affine3($point, $matrix))
          },
          ($base, $center)
        )
    '/>
   </xsl:function>

   <!--
    : this:opReflectXZ()
    : Reflect the point X and Z coordinates across a line before applying the base function.
    :-->
   <xsl:function name="this:opReflectXZ" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:reflection2($start, $end) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opReflectXZ",
          function($point as xs:double*) as xs:double* {
            let $qxz := affine:affine2((v:px($point), v:pz($point)), $matrix)
            return $basef((v:px($qxz), v:py($point), v:pz($qxz)))
          },
          ($base, $start, $end)
        )
    '/>
   </xsl:function>

   <!--
    : this:opReflectXY()
    : Reflect the point X and Y coordinates across a line before applying the base function.
    :-->
   <xsl:function name="this:opReflectXY" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:reflection2($start, $end) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opReflectXY",
          function($point as xs:double*) as xs:double* { 
            let $qxy := affine:affine2((v:px($point), v:py($point)), $matrix)
            return $basef((v:px($qxy), v:py($qxy), v:pz($point)))
          },
          ($base, $start, $end)
        )
    '/>
   </xsl:function>

   <!--
    : this:opReflectYZ()
    : Reflect the point Y and Z coordinates across a line before applying the base function.
    :-->
   <xsl:function name="this:opReflectYZ" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $matrix := affine:reflection2($start, $end) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opReflectYZ",
          function($point as xs:double*) as xs:double* {
            let $qyz := affine:affine2((v:py($point), v:pz($point)), $matrix)
            return $basef((v:px($point), v:py($qyz), v:pz($qyz)))
          },
          ($base, $start, $end)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRepeat()
    : Repeat the base function with the given periods (corresponding coordinates of
    : periods point).
    :-->
   <xsl:function name="this:opRepeat" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="periods" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $c := point:pcoordinates($periods) return
      let $half-c := $c=>v:times(0.5) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opRepeat",
          function($point as xs:double*) as xs:double* {
            let $q := $point=>v:add($half-c)=>v:modulo($c)=>v:sub($half-c)
            return $basef($q)
          },
          ($base, $periods)
        )
    '/>
   </xsl:function>

   <!--
    : this:opLimitedRepeat()
    : Repeat the base function with the given periods (corresponding coordinates of
    : periods point), with repetitions limited by by lower-limits and upper-limits.
    : we get lower-limits.x to upper-limits.x by lower-limits.y to upper-limits.y
    :-->
   <xsl:function name="this:opLimitedRepeat" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="periods" as="map(xs:string, item()*)"/>
      <xsl:param name="lower-limits" as="map(xs:string, item()*)"/>
      <xsl:param name="upper-limits" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $c := point:pcoordinates($periods) return
      let $lim-a := point:pcoordinates($lower-limits) return
      let $lim-b := point:pcoordinates($upper-limits) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opLimitedRepeat",
          function($point as xs:double*) as xs:double* {
            $basef(
              $point=>v:sub(
                v:clampv(v:round($point=>v:divide($c)), $lim-a, $lim-b)=>v:multiply($c)
              )
            )
          },
          ($base, $periods, $lower-limits, $upper-limits)
        )
    '/>
   </xsl:function>

   <!--
    : this:opElongate()
    : Extends a base shape by h in the given dimensions
    : e.g. h=(2,0) will extend all x dimensions outward by 2 (+/-)
    : $h gives the amount of elongation in each dimension
    :-->
   <xsl:function name="this:opElongate" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="h" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $hv := point:pcoordinates($h) return
      let $basef := f:function($base)
      return
        ann:f("sdf:opElongate",
          function($point as xs:double*) as xs:double* {
            let $q := $point=>v:abs()=>v:sub($hv)
            return $basef($q=>v:at-max(0)) + min((max($q),0))
          },
          $base
        )
    '/>
   </xsl:function>

   <!--
    : this:opRound()
    : Round the shape, with radius of rounding r
    :-->
   <xsl:function name="this:opRound" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opRound",
        function($point as xs:double*) as xs:double* {
          $basef($point) - $r
        },
        ($base, $r)
      )
    '/>
   </xsl:function>

   <!--
    : this:opOnion()
    : Create onion ring of the basic shape; kind of a scooping out of a smaller
    : verstion of the outer shape.
    : To create multiple layers, iterate onion
    : $thickness is width of layer between inner and outer shape; core shape is
    : extended +/- by this
    :-->
   <xsl:function name="this:opOnion" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="thickness" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $r := $thickness div 2 return
      let $basef := f:function($base)
      return
        ann:f("sdf:opOnion",
          function($point as xs:double*) as xs:double* {
            abs($basef($point)) - $r
          },
          ($base, $thickness)
        )
    '/>
   </xsl:function>

   <!--
    : this:opManyOnion()
    :-->
   <xsl:function name="this:opManyOnion" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="thickness" as="xs:double"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      fold-left(
        1 to $n,
        $base,
        function($f as item(), $i as xs:integer) as map(*) {
          this:opOnion($f, $thickness*math:pow(2, -($i - 1)))
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:opExtrude()
    : Convert a 2D shape to 3D by extruding into the third dimension by h
    :-->
   <xsl:function name="this:opExtrude" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="h" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opExtrude",
        function($point as xs:double*) as xs:double* {
          let $d := $basef($point=>v:as-dimension(2)) return
          let $w := ( $d, abs(v:pz($point)) - $h )
          return min((max($w), 0)) + v:magnitude($w=>v:at-max(0))
        },
        ($base, $h)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRevolve()
    : Convert a 2D shape into a 3D one by revolving around y axis at $o distance
    :-->
   <xsl:function name="this:opRevolve" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="o" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opRevolve",
        function($point as xs:double*) as xs:double* {
          $basef( ( v:magnitude((v:px($point), v:pz($point))) - $o, v:py($point) ) )
        },
        ($base, $o)
      )
    '/>
   </xsl:function>

   <!--
    : this:opUnion()
    : Union of two shapes.
    :-->
   <xsl:function name="this:opUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opUnion",
        function($point as xs:double*) as xs:double* {
          let $v1 := $af($point) return
          let $v2 := $bf($point)
          return
            if (head($v1) &lt; head($v2)) then $v1 else $v2
        },
        ($a, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:opUnion()
    : Union of multiple shapes.
    :-->
   <xsl:function name="this:opUnion" as="map(*)">
      <xsl:param name="fs" as="item()*"/>
      <xsl:sequence select='
      for $f in $fs return if (not(ann:is-sdf($f))) then errors:error("ML-BADARGS", ("fs", $f)) else (),
      let $ffs := for $f in $fs return f:function($f)
      return
      ann:f("sdf:opUnion",
        function($point as xs:double*) as xs:double* {
          let $ix :=
            util:min-index($ffs,
              function($f as function(xs:double*) as xs:double*) as xs:double {
                head($f($point))
              }
            )
          return (
            $ffs[$ix]($point)
          )
        },
        $fs
      )
    '/>
   </xsl:function>

   <!--
    : this:min()
    :-->
   <xsl:function name="this:min" as="xs:double">
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      min(($a,$b))
    '/>
   </xsl:function>

   <!--
    : this:sminQuad()
    :-->
   <xsl:function name="this:sminQuad" as="xs:double">
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      let $h := max(($k - abs($a - $b), 0)) div $k
      return min(($a,$b)) - $h*$h*$k*(1.0 div 4.0)
    '/>
   </xsl:function>

   <!--
    : this:sminCubic()
    :-->
   <xsl:function name="this:sminCubic" as="xs:double">
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      let $h := max(($k - abs($a - $b), 0)) div $k
      return min(($a,$b)) - $h*$h*$h*$k*(1.0 div 6.0)
    '/>
   </xsl:function>

   <!--
    : this:opUnion()
    :-->
   <xsl:function name="this:opUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="min" as="(function(xs:double, xs:double) as xs:double)"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opUnion",
        function($point as xs:double*) as xs:double* {
          let $v1 := $af($point) return
          let $v2 := $bf($point)
          return (
            $min(head($v1), head($v2)),
            if (head($v1) &lt; head($v2)) then tail($v1) else tail($v2)
          )
        },
        ($a, $b, $min)
      )
    '/>
   </xsl:function>

   <!--
    : this:opSubtract()
    : Difference of two shapes. a - b
    :-->
   <xsl:function name="this:opSubtract" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSubtract",
        function($point as xs:double*) as xs:double* {
          let $v1 := $af($point) return
          let $v2 := $bf($point)
          return (
            max((head($v1), -head($v2))),
            if (head($v1) > -head($v2)) then tail($v1) else tail($v2)
          )
        },
        ($a, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:opIntersect()
    : Intersection of two shapes.
    :-->
   <xsl:function name="this:opIntersect" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opIntersect",
        function($point as xs:double*) as xs:double* {
          let $v1 := $af($point) return
          let $v2 := $bf($point)
          return (
            if (head($v1) > head($v2)) then $v1 else $v2
          )
        },
        ($a, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:opIntersect()
    : Intersection of multiple shapes.
    :-->
   <xsl:function name="this:opIntersect" as="map(*)">
      <xsl:param name="fs" as="item()*"/>
      <xsl:sequence select='
      for $f in $fs return if (not(ann:is-sdf($f))) then errors:error("ML-BADARGS", ("fs", $f)) else (),
      let $ffs := for $f in $fs return f:function($f)
      return
      ann:f("sdf:opIntersect",
        function($point as xs:double*) as xs:double* {
          let $ix :=
            util:max-index($ffs,
              function($f as function(xs:double*) as xs:double*) as xs:double {
                head($f($point))
              }
            )
          return (
            $ffs[$ix]($point)
          )
        },
        ($fs)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRoundUnion()
    : Rounded union of two shapes
    :-->
   <xsl:function name="this:opRoundUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opRoundUnion",
        function($point as xs:double*) as xs:double* {
          let $v2 := $af($point) return
          let $v1 := $bf($point) return
          let $d1 := head($v1) return
          let $d2 := head($v2) return
          let $h := max(($k - abs($d1 - $d2), 0))
          return (
            min(($d1, $d2)) - $h*$h*0.25 div $k,
            if (head($v2) &lt; head($v1)) then tail($v2) else tail($v1)
          )
        },
        ($a, $b, $k)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRoundSubtract()
    : Rounded subtraction of two shapes a - b
    :-->
   <xsl:function name="this:opRoundSubtract" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opRoundSubtract",
        function($point as xs:double*) as xs:double* {
          let $l := v:magnitude($point) return
          let $v2 := $af($point) return
          let $v1 := $bf($point) return
          let $d1 := head($v1) return
          let $d2 := head($v2) return
          let $h :=  max(($k - abs(-$d1 - $d2),0))
          return (
            max((-$d1, $d2)) + $h*$h*0.25 div $k,
            if ($d2 > -$d1) then tail($v2) else tail($v1)
          )
        },
        ($a, $b, $k)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRoundIntersect()
    : Rounded intersection of two shapes
    :-->
   <xsl:function name="this:opRoundIntersect" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opRoundIntersect",
        function($point as xs:double*) as xs:double* {
          let $v2 := $af($point) return
          let $v1 := $bf($point) return
          let $d1 := head($v1) return
          let $d2 := head($v2) return
          let $h := max(($k - abs($d1 - $d2), 0))
          return (
            max(($d1, $d2)) + $h*$h*0.25 div $k,
            if (head($v2) > head($v1)) then tail($v2) else tail($v1)
          )
        },
        ($a, $b, $k)
      )
    '/>
   </xsl:function>

  <!--
   : opSmoothUnion()
   : Smoothed union of two shapes
   :-->
   <xsl:function name="this:opSmoothUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSmoothUnion",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv) return
          let $h := util:clamp(0.5 + (0.5 * ($bd - $ad)) div $k, 0, 1)
          return (
            util:mix($bd, $ad, $h) - $k * $h * (1 - $h),
            if (head($av) &lt; head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
   </xsl:function>

   <!--
    : opSmoothSubtract()
    : Smoothed difference of two shapes
    :-->
    <xsl:function name="this:opSmoothSubtract" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSmoothSubtract",
        function($point as xs:double*) as xs:double* {
          let $l := v:magnitude($point) return
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv) return
          let $h :=  util:clamp(0.5 - (0.5 * ($ad + $bd)) div $k, 0, 1)
          return (
            util:mix($ad, -$bd, $h) + $k * $h * (1 - $h),
            if ($ad > -$bd) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
    </xsl:function>

    <!--
     : opSmoothIntersect()
     : Smoothed intersection of two shapes
     :-->
    <xsl:function name="this:opSmoothIntersect" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSmoothIntersect",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv) return
          let $h :=  util:clamp(0.5 - (0.5 * ($bd - $ad)) div $k, 0, 1)
          return (
            util:mix($bd, $ad, $h) + $k * $h * (1 - $h),
            if ($ad > -$bd) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
    </xsl:function>

    <!--
     : opChamferUnion()
     : Chamfered union of two shapes
     :-->
    <xsl:function name="this:opChamferUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b) return
      let $sqrt_2_2 := math:sqrt(2) div 2
      return
      ann:f("sdf:opChamferUnion",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            min(($ad, $bd, ($ad - $k + $bd) * $sqrt_2_2)),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
    </xsl:function>

    <!--
     : opChamferIntersect()
     : Chamfered intersection of two shapes
     :-->
    <xsl:function name="this:opChamferIntersect" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b) return
      let $sqrt_2_2 := math:sqrt(2) div 2
      return
      ann:f("sdf:opChamferIntersect",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            max(($ad, $bd, ($ad + $k + $bd) * $sqrt_2_2)),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
    </xsl:function>

    <!--
     : opChamferSubtract()
     : Chamfered difference of two shapes
     :-->
    <xsl:function name="this:opChamferSubtract" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b) return
      let $sqrt_2_2 := math:sqrt(2) div 2
      return
      ann:f("sdf:opChamferSubtract",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            max(($ad, -$bd, ($ad + $k - $bd) * $sqrt_2_2)),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $k)
      )
      '/>
    </xsl:function>

    <xsl:function name="this:steps" as="xs:double">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      let $s := $r div $n return
      let $u := $b - $r
      return min(($a, $b, 0.5 * ($u + $a + abs(($u - $a + $s) mod (2 * $s)))))
      '/>
    </xsl:function>

    <!--
     : opSteppedUnion()
     : Stepped union of two shapes
     :-->
    <xsl:function name="this:opSteppedUnion" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSteppedUnion",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            this:steps($ad, $bd, $r, $n),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $r, $n)
      )
      '/>
    </xsl:function>

    <!--
     : opSteppedIntersect()
     : Stepped intersection of two shapes
     :-->
    <xsl:function name="this:opSteppedIntersect" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSteppedIntersection",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            -this:steps(-$ad, -$bd, $r, $n),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $r, $n)
      )
      '/>
    </xsl:function>

    <!--
     : opSteppedSubtract()
     : Stepped difference of two shapes
     :-->
    <xsl:function name="this:opSteppedSubtract" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opSteppedSubtract",
        function($point as xs:double*) as xs:double* {
          let $av := $af($point) return
          let $bv := $bf($point) return
          let $ad := head($av) return
          let $bd := head($bv)
          return (
            -this:steps(-$ad, $bd, $r, $n),
            if (head($av) > head($bv)) then tail($av) else tail($bv)
          )
        },
        ($a, $b, $r, $n)
      )
      '/>
    </xsl:function>
    
   <!--
    : this:opTwist()
    : 3D twisting
    :-->
   <xsl:function name="this:opTwist" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opTwist",
        function($point as xs:double*) as xs:double* {
          let $c := math:cos($k * v:py($point)) return
          let $s := math:sin($k * v:py($point)) return
          let $matrix := ($c, -$s, $s, $c) return
          let $q := (
            $matrix[1] * v:px($point) + $matrix[2] * v:pz($point),
            $matrix[3] * v:px($point) + $matrix[4] * v:pz($point),
            v:py($point)
          )
          return $basef($q)
        },
        ($base, $k)
      )
    '/>
   </xsl:function>

   <!--
    : this:opBend()
    : Simple bend of the base
    :-->
   <xsl:function name="this:opBend" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opBend",
        function($point as xs:double*) as xs:double* {
          let $c := math:cos($k * v:px($point)) return
          let $s := math:sin($k * v:px($point)) return
          let $matrix := ($c, -$s, $s, $c) return
          let $q := (
            $matrix[1] * v:px($point) + $matrix[2] * v:py($point),
            $matrix[3] * v:px($point) + $matrix[4] * v:py($point),
            tail(tail($point))
          )
          return $basef($q)
        },
        ($base, $k)
      )
    '/>
   </xsl:function>

   <!--
    : this:opAdd()
    : Add the two distance functions.
    : e.g. to add in a wobble
    :-->
   <xsl:function name="this:opAdd" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opAdd",
        function($point as xs:double*) as xs:double* {
          $af($point)=>v:add($bf($point))
        },
        ($a, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:opAddInto()
    : Add the second distance function to the first, but keep any materials or
    : other secondary values from the first.
    : e.g. to add in a wobble but keep the same material
    :-->
   <xsl:function name="this:opAddInto" as="map(*)">
      <xsl:param name="a" as="item()"/>
      <xsl:param name="b" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($a))) then errors:error("ML-BADARGS", ("a", $a)) else (),
      if (not(ann:is-sdf($b))) then errors:error("ML-BADARGS", ("b", $b)) else (),
      let $af := f:function($a) return
      let $bf := f:function($b)
      return
      ann:f("sdf:opAddInto",
        function($point as xs:double*) as xs:double* {
          let $v1 := $af($point) return
          let $v2 := $bf($point)
          return (head($v1) + head($v2), tail($v1))
        },
        ($a, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:opInvert()
    : Flip insides with outsides
    :-->
   <xsl:function name="this:opInvert" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opInvert",
        function($point as xs:double*) as xs:double* {
          let $res := $basef($point)
          return (-head($res), tail($res))
        },
        $base
      )
    '/>
   </xsl:function>

   <!--
    : this:opRescale()
    : Rescale the output distance.
    :-->
   <xsl:function name="this:opRescale" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="amplitude" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      let $basef := f:function($base)
      return
      ann:f("sdf:opRescale",
        function($point as xs:double*) as xs:double* {
          let $res := $basef($point)
          return ($amplitude * head($res), tail($res))
        },
        ($base, $amplitude)
      )
    '/>
   </xsl:function>

   <!--
    : this:opModifyDistance()
    : Apply a gain function to the raw distance value.
    :-->
   <xsl:function name="this:opModifyDistance" as="map(*)">
      <xsl:param name="base" as="item()"/>
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-sdf($base))) then errors:error("ML-BADARGS", ("base", $base)) else (),
      if (not(ann:is-distance-function($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $basef := f:function($base) return
      let $ff := f:function($f)
      return
      ann:f("sdf:opModifyDistance",
        function($point as xs:double*) as xs:double* {
          let $res := $basef($point)
          return ($ff(head($res)), tail($res))
        },
        ($base, $f)
      )
    '/>
   </xsl:function>

<!--
 : Distance modifiers: 
 : NOTE: Some of these operating unit interval so be sure distance is scaled first 
 :-->

   <!--
    : this:modAlmostIdentity()
    : Modify value by returning the result except close to 0.
    :
    : $m: lower bound below which we modify value
    : $n: lower bound of output value
    :-->
   <xsl:function name="this:modAlmostIdentity" as="item()">
      <xsl:param name="m" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      let $a := 2.0*$n - $m return
      let $b := 2.0*$m - 3.0*$n
      return  
        ann:f("sdf:modAlmostIdentity", function($x as xs:double) as xs:double {
          if (abs($x) > $m) then $x
          else (
            let $t := $x div $m
            return (($a * $t + $b)*$t*$t + $n)
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:modExpImpulse()
    : Modify value fast growth plus slow decay.
    : Maximum value becomes 1 at d=1/k
    :
    : $k: stretching
    :-->
   <xsl:function name="this:modExpImpulse" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modExpImpulse", function($x as xs:double) as xs:double {
        let $d := abs($x) return
        let $sign := util:sign($x)
        return (
          let $h := $k * $d
          return ($sign * $h * math:exp(1.0 - $h), tail($d))
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modQuadImpulse()
    : Modify value fast growth plus slow decay.
    : Maximum value becomes 1 at d=sqrt(1/k)
    :
    : $k: stretching
    :-->
   <xsl:function name="this:modQuadImpulse" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      let $two-root-k := 2.0 * math:sqrt($k)
      return
        ann:f("sdf:modQuadImpulse", function($x as xs:double) as xs:double {
          let $d := abs($x) return
          let $sign := util:sign($x)
          return (
            $sign * $two-root-k * $d div (1.0 + $k*$d*$d)
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:modExpSustainedImpulse()
    : Modify value fast growth plus slow decay, with control over width of
    : both attack and release.
    :
    : $k: attack
    : $f: release
    :-->
   <xsl:function name="this:modExpSustainedImpulse" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:param name="f" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modExpSustainedImpulse", function($x as xs:double) as xs:double {
        let $d := abs($x) return
        let $sign := util:sign($x) return
        let $s := max(($d - $f, 0.0))
        return (
          $sign * min(( ($d*$d) div ($f*$f), 1 + (2.0 div $f)*$s*math:exp(-$k*$s)))
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modCubicPulse()
    : Modify value zero with a spread curve in the middle; pseudo-Gaussian
    :
    : $c: center value
    : $w: width
    :-->
   <xsl:function name="this:modCubicPulse" as="item()">
      <xsl:param name="c" as="xs:double"/>
      <xsl:param name="w" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modCubicPulse", function($x as xs:double) as xs:double {
        let $x := abs($x - $c)
        return (
          if ($x > $w) then 0.0
          else (
            let $x := $x div $w
            return (1.0 - $x*$x*(3.0 - 2.0*$x))
          )
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modExpStep()
    : Modify value exponential decay
    :
    : $k: decay rate
    : $n: sharpness of decay; higher n => more like a step 
    :-->
   <xsl:function name="this:modExpStep" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:param name="n" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modExpStep", function($x as xs:double) as xs:double {
        let $d := abs($x) return
        let $sign := if ($x=0) then 1 else util:sign($x)
        return (
          $sign * math:exp( -$k * math:pow($d, $n) )
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modGain()
    : Modify value expanding sides and compressing center
    : k=1 is identity, k>1 is ess curve, k<1 is symmetric inverse (k=a vs k=1/a)
    :
    : $k: gain 
    :-->
   <xsl:function name="this:modGain" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modGain", function($x as xs:double) as xs:double {
        let $d := abs($x) return
        let $sign := util:sign($x) return
        let $a := 0.5 * math:pow(2.0*(if ($d &lt; 0.5) then $d else 1.0 - $d), $k)
        return (
          if ($d &lt; 0.5) then $sign*$a else $sign*(1.0 - $a)
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modPowerCurve()
    : Modify value map [0,1] to [0,1] with corners mapped to 0 and a curve
    : in between.
    :
    : $a, $b: curve shape constants
    :-->
   <xsl:function name="this:modPowerCurve" as="item()">
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $k := math:pow($a + $b, $a + $b) div (math:pow($a, $a) * math:pow($b, $b))
      return
        ann:f("sdf:modPowerCurve", function($x as xs:double) as xs:double {
          let $d := abs($x) return
          let $sign := util:sign($x)
          return (
            $sign * $k * math:pow($d, $a) * math:pow(1.0 - $d, $b)
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:modSinc()
    : Modify value map with a phase shifted sinc curve (sine cardinal) 
    : sinc(x) = 1 if x=0, sin(x)/x otherwise
    :
    : $k: controls degree of phase shift
    : $scale: rescales result
    :-->
   <xsl:function name="this:modSinc" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:param name="scale" as="xs:double"/>
      <xsl:sequence select='
      ann:f("sdf:modSinc", function($x as xs:double) as xs:double {
        let $d := abs($x) return
        let $sign := util:sign($x) return
        let $a := math:pi()*($k * $d - 1.0)
        return (
          if ($a = 0) then $sign * $scale
          else $sign * $scale * math:sin($a) div $a
        )
      })
    '/>
   </xsl:function>

   <!--
    : this:modSinc()
    :-->
   <xsl:function name="this:modSinc" as="item()">
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:modSinc($k, 1.0)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Rendering functions
 :======================================================================
 :-->

   <!--
    : this:renderTrace()
    : Trace passthrough for rendering functions
    : f and result are both function(xs:double*, xs:double*, xs:double*, item()) as xs:double* or a callable of same
    : final argument is function(xs:double*) as xs:double* or callable of same
    :-->
   <xsl:function name="this:renderTrace" as="map(*)">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $fn := f:function($f)
      return
      ann:f(util:function-name($f), 
        function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          $fn($tm, $origin, $direction, $sdf)=>trace(util:quote($f)||"("||util:quote($tm)||")")
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:renderPerformance()
    : Render the scene by mapping iterations into the given gradient.
    :-->
   <xsl:function name="this:renderPerformance" as="map(*)">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      let $colours := gradient:gradient-points($gradient)
      return 
        ann:f("sdf:renderPerformance", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          (: 64 = max iterations :)
          point:pcoordinates(
            cs:rgb-to-xyz(cs:interpolate-colour($tm[3] div 64, $colours))
          )
        })
    '/>
   </xsl:function>

   <!--
    : this:renderDepth()
    : Render the scene by mapping depth to the given gradient.
    : $field defines the range for scaling
    :-->
   <xsl:function name="this:renderDepth" as="map(*)">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:param name="field" as="xs:double*"/>
      <xsl:sequence select='
      let $colours := gradient:gradient-points($gradient) return
      let $n := count($colours) return
      let $scale := $n div ($field[2] - $field[1])
      return 
        ann:f("sdf:renderDepth", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $adjusted-t := util:round(($tm[1] - $field[1]) * $scale) div $n
          return
            point:pcoordinates(
              cs:rgb-to-xyz(cs:interpolate-colour($adjusted-t, $colours))
            )
        })
    '/>
   </xsl:function>

   <!--
    : this:renderColourPick()
    : Map from a material index to a colour. Colours should be RGB; function
    : maps to XYZ.
    :-->
   <xsl:function name="this:renderColourPick" as="map(*)">
      <xsl:param name="pick-table" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $n := count($pick-table) return
      let $xyz-pick-table := $pick-table!cs:rgb-to-xyz(.)
      return
        ann:f("sdf:renderColourPick", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          point:pcoordinates($xyz-pick-table[util:modix(util:round($tm[2]), $n)])
        })
    '/>
   </xsl:function>

   <!--
    : this:renderMaterial()
    : Map from a material index to a colouring function.
    :-->
   <xsl:function name="this:renderMaterial" as="map(*)">
      <xsl:param name="pick-table" as="item()*"/>
      <xsl:sequence select='
      for $pick in $pick-table return if (not(ann:is-renderer($pick))) then errors:error("ML-BADARGS", ("pick-table", $pick)) else (),
      let $pick-tablef := for $pick in $pick-table return f:function($pick) return
      let $n := count($pick-table)
      return
        ann:f("sdf:renderMaterial",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $f := $pick-tablef[util:modix(util:round($tm[2]), $n)]
            return $f($tm, $origin, $direction, $sdf)
          },
          $pick-table
        )
    '/>
   </xsl:function>

   <!--
    : this:renderChoice()
    : Render everything deeper than (or equal to) a cutoff depth using one
    : renderer; everything nearer with a different one.
    :-->
   <xsl:function name="this:renderChoice" as="map(*)">
      <xsl:param name="t-pivot" as="xs:double"/>
      <xsl:param name="background" as="item()"/>
      <xsl:param name="foreground" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($background))) then errors:error("ML-BADARGS", ("background", $background)) else (),
      if (not(ann:is-renderer($foreground))) then errors:error("ML-BADARGS", ("foreground", $foreground)) else (),
      let $foregroundf := f:function($foreground) return
      let $backgroundf := f:function($background)
      return
      ann:f("sdf:renderChoice",
        function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          if (head($tm) &lt;= $t-pivot)
          then $backgroundf($tm, $origin, $direction, $sdf)
          else $foregroundf($tm, $origin, $direction, $sdf)
        },
        ($t-pivot, $background, $foreground)
      )
    '/>
   </xsl:function>

   <!--
    : this:renderConstant()
    : Return a constant colour. Colour should be given in RGB; function maps  
    : to XYZ.
    :-->
   <xsl:function name="this:renderConstant" as="map(*)">
      <xsl:param name="colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $colourv := point:pcoordinates(cs:rgb-to-xyz($colour))
      return
        ann:f("sdf:renderConstant", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          $colourv
        })
    '/>
   </xsl:function>

   <!--
    : this:renderRandom()
    : Return a constant colour. Colour should be given in RGB; function maps  
    : to XYZ.
    :-->
   <xsl:function name="this:renderRandom" as="map(*)" new-each-time="yes">
      <xsl:param name="colours" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $xyz-colours := $colours!cs:rgb-to-xyz(.)
      return
        ann:f("sdf:renderRandom", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          rand:select-random($xyz-colours)!point:pcoordinates(.)
        })
    '/>
   </xsl:function>

   <!--
    : this:renderGradient()
    : Apply a shading gradient based on t
    : Base colour should be given in RGB, function maps to XYZ.
    :-->
   <xsl:function name="this:renderGradient" as="map(*)">
      <xsl:param name="base-colour" as="map(xs:string, item()*)"/>
      <xsl:param name="strength" as="xs:double"/>
      <xsl:sequence select='
      let $colourv := point:pcoordinates(cs:rgb-to-xyz($base-colour))
      return
        ann:f("sdf:renderGradient", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $shade := head($tm) * $strength
          return $colourv=>v:sub(($shade, $shade, $shade))
        })
    '/>
   </xsl:function>

   <!--
    : this:renderXGradient()
    : Apply a shading gradient based on x coordinate
    : Base colour should be given in RGB, function maps to XYZ.
    :-->
   <xsl:function name="this:renderXGradient" as="map(*)">
      <xsl:param name="base-colour" as="map(xs:string, item()*)"/>
      <xsl:param name="strength" as="xs:double"/>
      <xsl:sequence select='
      let $colourv := point:pcoordinates(cs:rgb-to-xyz($base-colour))
      return
        ann:f("sdf:renderXGradient", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $shade := v:px($direction) * $strength
          return $colourv=>v:sub(($shade, $shade, $shade))
        })
    '/>
   </xsl:function>

   <!--
    : this:renderYGradient()
    : Apply a shading gradient based on y coordinate
    : Base colour should be given in RGB, function maps to XYZ.
    :-->
   <xsl:function name="this:renderYGradient" as="map(*)">
      <xsl:param name="base-colour" as="map(xs:string, item()*)"/>
      <xsl:param name="strength" as="xs:double"/>
      <xsl:sequence select='
      let $colourv := point:pcoordinates(cs:rgb-to-xyz($base-colour))
      return
        ann:f("sdf:renderYGradient", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $shade := v:py($direction) * $strength
          return $colourv=>v:sub(($shade, $shade, $shade))
        })
    '/>
   </xsl:function>

   <!--
    : this:renderZGradient()
    : Apply a shading gradient based on y coordinate
    : Base colour should be given in RGB, function maps to XYZ.
    :-->
   <xsl:function name="this:renderZGradient" as="map(*)">
      <xsl:param name="base-colour" as="map(xs:string, item()*)"/>
      <xsl:param name="strength" as="xs:double"/>
      <xsl:sequence select='
      let $colourv := point:pcoordinates(cs:rgb-to-xyz($base-colour))
      return
        ann:f("sdf:renderZGradient", function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $shade := v:pz($direction) * $strength
          return $colourv=>v:sub(($shade, $shade, $shade))
        })
    '/>
   </xsl:function>

   <!--
    : this:opRenderFuzz()
    : Fuzz out the depth of field
    :-->
   <xsl:function name="this:opRenderFuzz" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="field" as="xs:double*"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour)
      return
      ann:f("sdf:opRenderFuzz",
        function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $col := $base-colourf($tm, $origin, $direction, $sdf) return
          let $smooth := util:smoothstep($field[1], $field[2], head($tm))
          return $col=>v:times(1.0 - $smooth)  
        },
        ($base-colour, $field)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRenderDiffuseLight()
    : Diffuse lighting.
    :
    : Lambert reflection: Id = L·N C Il 
    : Id = surface brightness, L = direction of light, N = surface normal
    : C = colour, Il = incoming light intensity
    :-->
   <xsl:function name="this:opRenderDiffuseLight" as="map(*)">
      <xsl:param name="surface-colour" as="item()"/>
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($surface-colour))) then errors:error("ML-BADARGS", ("surface-colour", $surface-colour)) else (),
      let $surface-colourf := f:function($surface-colour) return
      let $light-dir := light:direction($light) return
      let $light-colour := light:directional($light) return
      let $intensity := light:intensity($light) return
      let $ambient-intensity := light:ambient-intensity($light) return
      let $ambient := light:ambient($light)=>v:times($ambient-intensity)
      return
        ann:f("sdf:opRenderDiffuseLight",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $surface := $surface-colourf($tm, $origin, $direction, $sdf) return
            let $pos := this:linear($origin, $direction, head($tm)) return
            let $n := this:calc-normal($pos, $sdf) return
            let $surface-brightness := max((v:dot($n, $light-dir), 0.0)) * $intensity return
            let $directional := $light-colour=>v:times($surface-brightness)
            return
              $surface=>v:multiply( $directional=>v:add($ambient) )
          },
          ($surface-colour, $light)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRenderHardShadow()
    : Modify colours to provide hard shadows.
    :-->
   <xsl:function name="this:opRenderHardShadow" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:param name="field" as="xs:double*"/>
      <xsl:param name="shadow-scale" as="xs:double"/>
      <xsl:param name="shadow-strength" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour) return
      let $light-dir := light:direction($light)
      return
        ann:f("sdf:opRenderHardShadow",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $col := $base-colourf($tm, $origin, $direction, $sdf) return
            let $pos := this:linear($origin, $direction, head($tm)) return
            let $normal := this:calc-normal($pos, $sdf) return
            let $shadow-origin := $pos=>v:add($normal=>v:times($shadow-scale)) return
            let $st := this:cast-ray($shadow-origin, $light-dir, $field, $sdf)[2] return
            let $shadow := if ($st != -1.0) then 1.0 else 0.0 
            return v:mix($col, $col=>v:times($shadow-strength), $shadow)
          },
          ($base-colour, $light, $field, $shadow-scale, $shadow-strength)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRenderSoftShadow()
    : Modify colours to provide soft shadows.
    :-->
   <xsl:function name="this:opRenderSoftShadow" as="map(*)" new-each-time="yes">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:param name="field" as="xs:double*"/>
      <xsl:param name="shadow-scale" as="xs:double"/>
      <xsl:param name="shadow-strength" as="xs:double"/>
      <xsl:param name="n-shadows" as="xs:integer"/>
      <xsl:param name="shadow-falloff" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour) return
      let $light-dir := light:direction($light) return
      let $rand := dist:uniform(-1.0, 1.0 - $config:ε)
      return
        ann:f("sdf:opRenderSoftShadow",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $col := $base-colourf($tm, $origin, $direction, $sdf) return
            let $pos := this:linear($origin, $direction, head($tm)) return
            let $normal := this:calc-normal($pos, $sdf) return
            let $shadow-origin := $pos=>v:add($normal=>v:times($shadow-scale)) return
            let $shadow :=
              fold-left(
                1 to $n-shadows,
                0.0,
                function($shadow, $s) {
                  let $fall := $shadow-falloff * rand:randomize($rand) return
                  let $shadow-dir := $light-dir=>v:add(($fall, $fall, $fall)) return
                  let $st := this:cast-ray($shadow-origin, $shadow-dir, $field, $sdf)[2]
                  return (
                    if ($st != -1.0) then $shadow + 1.0 else $shadow
                  )
                }
              )
            return v:mix($col, $col=>v:times($shadow-strength), $shadow div $n-shadows)
          },
          ($base-colour, $light, $field, $shadow-scale, $shadow-strength, $n-shadows, $shadow-falloff)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRenderComplexLighting()
    : Modify colours to provide soft shadows in diffuse light.
    :-->
   <xsl:function name="this:opRenderComplexLighting" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:param name="shadow-scale" as="xs:double"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour) return
      let $light-direction := light:direction($light) return
      let $light-colour :=
        light:directional($light)=>
          v:times( light:intensity($light) )=>
          v:add( light:ambient($light)=>v:times( light:ambient-intensity($light) ) )
      return
        ann:f("sdf:opRenderComplexLighting",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $pos := this:linear($origin, $direction, head($tm)) return
            let $col := $base-colourf($tm, $origin, $direction, $sdf) return
            let $normal := this:calc-normal($pos, $sdf) return
            let $shade-origin := $direction=>v:add($normal=>v:times($shadow-scale)) return
            let $shade := head(
              util:while(
                function($res, $t, $i, $working) { $i &lt;= 16 and $working },
                function($res, $t, $i, $working) {
                  let $h := head($sdf( this:linear($shade-origin, $light-direction, $t))) return
                  let $res := min(($res, $k*$h div $t)) return
                  let $t := $t + util:clamp($h, 0.01, 1.0) (: fineness of calc :)
                  return (
                    $res, $t, $i + 1, not($h &lt; 0.0001)
                  )
                },
                1.0, 0.0, 1, true()
              )
            ) return
            let $shade :=
              util:clamp($shade, 0.0, 1.0) *
              util:clamp(v:dot($normal, $light-direction), 0.0, 1.0)
            return
            let $occlusion := head(
              util:while(
                function ($res as xs:double, $sca as xs:double, $i as xs:integer) as xs:boolean { $i &lt; 5 },
                function ($res as xs:double, $sca as xs:double, $i as xs:integer) as item()* {
                  let $hr := 0.02 + 0.025 * xs:double($i) * xs:double($i) return
                  let $aopos := $pos=>v:add($normal=>v:times($hr)) return
                  let $dd := head($sdf($aopos))
                  return ($res + -($dd - $hr)*$sca, $sca * 0.995, $i + 1)
                },
                0.0, 1.0, 0
              )
            ) return
            let $occlusion :=
              (1.0 - util:clamp($occlusion, 0.0, 1.0))*util:clamp(v:py($normal), 0.0, 1.0)
            return
              ($col=>v:times($occlusion))=>             (: base + ambient shadows :)
                v:add($light-colour=>v:times($shade))=> (: lighting :)
                v:times(math:exp(-0.2*head($tm)))       (: softening :)
          },
          ($base-colour, $light, $shadow-scale, $k)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRenderColourWaves()
    : Waves of colour bands
    :-->
   <xsl:function name="this:opRenderColourWaves" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="darkness" as="xs:double"/>
      <xsl:param name="frequency" as="xs:double"/>
      <xsl:param name="strength" as="xs:double"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour)
      return
      ann:f("sdf:opRenderColourWaves",
        function (
          $tm as xs:double*,
          $origin as xs:double*,
          $direction as xs:double*,
          $sdf as item()
        ) as xs:double*
        {
          let $d := head($tm)
          return
            $base-colourf($tm, $origin, $direction, $sdf)=>
              v:times(1.0 - math:exp($darkness*abs($d)))=>
              v:times((1 - $strength) + $strength*math:cos($frequency*$d))
        },
        ($base-colour, $darkness, $frequency, $strength)
      )
    '/>
   </xsl:function>

   <!--
    : this:opRenderColourBlend()
    : Blend a colour into the source colouring based on the distance
    :-->
   <xsl:function name="this:opRenderColourBlend" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="blend-colour" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      let $base-colourf := f:function($base-colour) return
      let $blend := point:pcoordinates($blend-colour)
      return
        ann:f("sdf:opRenderColourBlend",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $col := $base-colourf($tm, $origin, $direction, $sdf)
            return v:mix($col, $blend, $tm[1])
          },
          ($base-colour, $blend-colour)
        )
    '/>
   </xsl:function>

   <!--
    : this:opRenderModifyColour()
    : Blend a colour into the source colouring based on a function of the distance
    :-->
   <xsl:function name="this:opRenderModifyColour" as="map(*)">
      <xsl:param name="base-colour" as="item()"/>
      <xsl:param name="blend-colour" as="map(xs:string, item()*)"/>
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-renderer($base-colour))) then errors:error("ML-BADARGS", ("base-colour", $base-colour)) else (),
      if (not(ann:is-distance-function($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $base-colourf := f:function($base-colour) return
      let $ff := f:function($f) return
      let $blend := point:pcoordinates($blend-colour)
      return
        ann:f("sdf:opRenderModifyColour",
          function (
            $tm as xs:double*,
            $origin as xs:double*,
            $direction as xs:double*,
            $sdf as item()
          ) as xs:double*
          {
            let $col := $base-colourf($tm, $origin, $direction, $sdf) return
            let $mod := $ff($tm[1])
            return v:mix($col, $blend, $mod)
          },
          ($base-colour, $blend-colour, $f)
        )
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Ray casting main drivers
 :======================================================================
 :-->

   <!--
    : this:camera()
    : Define the camera view: its position and target
    :-->
   <xsl:function name="this:camera" as="map(xs:string, item()*)">
      <xsl:param name="position" as="map(xs:string, item()*)"/>
      <xsl:param name="target" as="map(xs:string, item()*)"/>
      <xsl:param name="perspective" as="xs:double"/>
      <xsl:sequence select='
      let $forward := point:normalize(point:sub($target, $position)) return
      let $right := point:normalize(point:cross(point:point(0,1,0), $forward)) return
      let $up := point:normalize(point:cross($forward, $right))
      return
        map {
          "kind": "camera",
          "position": point:pcoordinates($position),
          "view": point:pcoordinates($position),
          "target": point:pcoordinates($target),
          "perspective": $perspective,
          "forward": point:pcoordinates($forward)=>v:times($perspective),
          "right": point:pcoordinates($right),
          "up": point:pcoordinates($up)
        }
    '/>
   </xsl:function>

   <!--
    : this:camera()
    : Define camera POV based on view angles
    :-->
   <xsl:function name="this:camera" as="map(xs:string, item()*)">
      <xsl:param name="angle" as="xs:double"/>
      <xsl:param name="view" as="map(xs:string, item()*)"/>
      <xsl:param name="target" as="map(xs:string, item()*)"/>
      <xsl:param name="perspective" as="xs:double"/>
      <xsl:sequence select='
      let $radians := util:radians($angle) return
      let $origin :=
        point:pcoordinates($view)=>
          v:multiply((math:cos($radians), 1.0, math:sin($radians)))
      return
      let $forward := v:normalize(point:pcoordinates($target)=>v:sub($origin)) return
      let $right := v:normalize(v:cross($forward,(0,1,0))) return
      let $up := v:normalize(v:cross($forward, $right))
      return
        map {
          "kind": "camera",
          "position": $origin,
          "angle": $angle,
          "view": point:pcoordinates($view),
          "target": point:pcoordinates($target),
          "perspective": $perspective,
          "forward": $forward=>v:times($perspective),
          "right": $right,
          "up": $up
        }
    '/>
   </xsl:function>

   <!--
    : this:castDepth()
    : A caster that calculates the depth relative to a particular ray
    :-->
   <xsl:function name="this:castDepth" as="map(*)">
      <xsl:param name="field" as="xs:double*"/>
      <xsl:sequence select='
      ann:f("sdf:castDepth", function($origin as xs:double*, $direction as xs:double*, $sdf as item()) as xs:double* {
        this:cast-ray($origin, $direction, $field, $sdf)
      })
    '/>
   </xsl:function>

   <!--
    : this:castDirect()
    : A caster that just returns the SDF unaltered
    :-->
   <xsl:function name="this:castDirect" as="map(*)">
      <xsl:sequence select='
      ann:f("sdf:castDirect", function($origin as xs:double*, $direction as xs:double*, $sdf as item()) as xs:double* {
        $sdf($direction)
      })
    '/>
   </xsl:function>

   <!--
    : this:render()
    : Compute rendering for a normalized point (scaled to canvas)
    :-->
   <xsl:function name="this:render" as="map(xs:string, item()*)">
      <xsl:param name="uv" as="map(xs:string, item()*)"/>
      <xsl:param name="camera" as="map(xs:string, item()*)?"/>
      <xsl:param name="caster" as="item()"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:param name="render" as="item()"/>
      <xsl:sequence select='
      (:
       : This is called a lot: you are better off with make-point-render() 
       :)
      if (not(ann:is-caster($caster))) then errors:error("ML-BADARGS", ("caster", $caster)) else (),
      if (not(ann:is-sdf($sdf))) then errors:error("ML-BADARGS", ("sdf", $sdf)) else (),
      if (not(ann:is-renderer($render))) then errors:error("ML-BADARGS", ("render", $render)) else (),
      let $casterf := f:function($caster) return
      let $sdff := f:function($sdf) return
      let $renderf := f:function($render)
      return (
        if (empty($camera)) then (
          (: Take each point verbatim; really only useful for 2D :)
          let $ray-origin := point:pcoordinates($point:ORIGIN3D) return
          let $ray-direction := point:pcoordinates($uv) return
          let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
          let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
          return cs:xyz-to-rgb(point:vector($col))
        ) else (
          let $ray-origin := $camera("position") return
          let $ray-direction := this:camera-ray-direction($uv, $camera) return
          let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
          let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff) 
          return cs:xyz-to-rgb(point:vector($col))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:make-point-render()
    :-->
   <xsl:function name="this:make-point-render"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"
                >
      <xsl:param name="camera" as="map(xs:string, item()*)?"/>
      <xsl:param name="caster" as="item()"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:param name="render" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-caster($caster))) then errors:error("ML-BADARGS", ("caster", $caster)) else (),
      if (not(ann:is-sdf($sdf))) then errors:error("ML-BADARGS", ("sdf", $sdf)) else (),
      if (not(ann:is-renderer($render))) then errors:error("ML-BADARGS", ("render", $render)) else (),
      let $casterf := f:function($caster) return
      let $sdff := f:function($sdf) return
      let $renderf := f:function($render)
      return (
        if (empty($camera)) then (
          let $ray-origin := point:pcoordinates($point:ORIGIN3D) return
          function($uv as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $ray-direction := point:pcoordinates($uv) return
            let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
            let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
            return cs:xyz-to-rgb(point:vector($col))
          }
        ) else (
          let $ray-origin := $camera("position") return
          function($uv as map(xs:string,item()*)) as map(xs:string,item()*) {
            let $ray-direction := this:camera-ray-direction($uv, $camera) return
            let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
            let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
            return cs:xyz-to-rgb(point:vector($col))
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:renderv()
    :-->
   <xsl:function name="this:renderv" as="map(xs:string, item()*)">
      <xsl:param name="uv" as="xs:double*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)?"/>
      <xsl:param name="caster" as="item()"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:param name="render" as="item()"/>
      <xsl:sequence select='
      (:
       : This is called a lot: you are better off with make-vector-render() 
       :)
      if (not(ann:is-caster($caster))) then errors:error("ML-BADARGS", ("caster", $caster)) else (),
      if (not(ann:is-sdf($sdf))) then errors:error("ML-BADARGS", ("sdf", $sdf)) else (),
      if (not(ann:is-renderer($render))) then errors:error("ML-BADARGS", ("render", $render)) else (),
      let $casterf := f:function($caster) return
      let $sdff := f:function($sdf) return
      let $renderf := f:function($render)
      return (
        if (empty($camera)) then (
          (: Take each point verbatim; really only useful for 2D :)
          let $ray-origin := point:pcoordinates($point:ORIGIN3D) return
          let $ray-direction := $uv return
          let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
          let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff) 
          return cs:xyz-to-rgb(point:vector($col))
        ) else (
          let $ray-origin := $camera("position") return
          let $ray-direction := this:camera-ray-directionv($uv, $camera) return
          let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
          let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
          return cs:xyz-to-rgb(point:vector($col))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:make-vector-render()
    :-->
   <xsl:function name="this:make-vector-render"
                 as="(function(xs:double*) as map(xs:string, item()*))"
                >
      <xsl:param name="camera" as="map(xs:string, item()*)?"/>
      <xsl:param name="caster" as="item()"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:param name="render" as="item()"/>
      <xsl:sequence select='
      if (not(ann:is-caster($caster))) then errors:error("ML-BADARGS", ("caster", $caster)) else (),
      if (not(ann:is-sdf($sdf))) then errors:error("ML-BADARGS", ("sdf", $sdf)) else (),
      if (not(ann:is-renderer($render))) then errors:error("ML-BADARGS", ("render", $render)) else (),
      let $casterf := f:function($caster) return
      let $sdff := f:function($sdf) return
      let $renderf := f:function($render)
      return (
        if (empty($camera)) then (
          let $ray-origin := point:pcoordinates($point:ORIGIN3D) return
          function($uv as xs:double*) as map(xs:string,item()*) {
            let $ray-direction := $uv return
            let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
            let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
            return cs:xyz-to-rgb(point:vector($col))
          }
        ) else (
          let $ray-origin := $camera("position") return
          function($uv as xs:double*) as map(xs:string,item()*) {
            let $ray-direction := this:camera-ray-directionv($uv, $camera) return
            let $tm := $casterf($ray-origin, $ray-direction, $sdff) return
            let $col := $renderf($tm, $ray-origin, $ray-direction, $sdff)
            return cs:xyz-to-rgb(point:vector($col))
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:camera-ray-direction()
    :-->
   <xsl:function name="this:camera-ray-direction" as="xs:double*">
      <xsl:param name="uv" as="map(xs:string, item()*)"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      v:normalize(
        (
          ($camera("right"))=>v:times(point:px($uv))
        )=>v:add(
          ($camera("up"))=>v:times(point:py($uv))
        )=>v:add(
          $camera("forward")
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:camera-ray-directionv()
    :-->
   <xsl:function name="this:camera-ray-directionv" as="xs:double*">
      <xsl:param name="uv" as="xs:double*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      v:normalize(
        (
          ($camera("right"))=>v:times(v:px($uv))
        )=>v:add(
          ($camera("up"))=>v:times(v:py($uv))
        )=>v:add(
          $camera("forward")
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:cast-ray()
    : Main ray-casting function
    :-->
   <xsl:function name="this:cast-ray" as="xs:double*">
      <xsl:param name="ray-origin" as="xs:double*"/>
      <xsl:param name="ray-direction" as="xs:double*"/>
      <xsl:param name="field" as="xs:double*"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:sequence select='
      util:assert(count($ray-origin)=3, "bad ray origin"),
      (: util:assert(count($ray-direction)=3, "bad ray direction"), :)
      let $ray-direction := if (count($ray-direction)=2) then ($ray-direction,0) else $ray-direction return
      let $ε := 0.0001 return
      let $start := $field[1] return
      let $end := $field[2] return
      let $max-iterations := 64 return
      let $data :=
        util:while(
          function($working as xs:boolean, $t as xs:double, $m as xs:double, $i as xs:integer) as xs:boolean {
            $working and $i &lt;= $max-iterations
          },
          function($working as xs:boolean, $t as xs:double, $m as xs:double, $i as xs:integer) {
            let $pos := this:linear($ray-origin, $ray-direction, $t) return
            let $result := $sdf($pos) return
            let $distance := $result[1] return
            let $material := ($result[2],0)[1] 
            return (
              if ($distance &lt; $ε) then (
                false(), $t, $m, $i + 1
              ) else if ($t + $distance > $end) then (
                false(), -1.0, -1.0, $i + 1
              ) else (
                true(), $t + $distance, $material, $i + 1
              )
            )
          },
          true(), $start, -1.0, 1
        )
      return
      let $working := head($data)
      return
        if ($working) then (-1.0, -1.0, $data[4]) else tail($data)
    '/>
   </xsl:function>

   <!--
    : this:calc-normal()
    :-->
   <xsl:function name="this:calc-normal" as="xs:double*">
      <xsl:param name="point" as="xs:double*"/>
      <xsl:param name="sdf" as="item()"/>
      <xsl:sequence select='
      let $c := head($sdf($point)) return
      let $epsilon := 0.001 
      (: Compare sdf in offsets from point :)
      return
        v:sub(
          v:normalize((
            head($sdf($point=>v:add(($epsilon, 0, 0)))),
            head($sdf($point=>v:add((0, $epsilon, 0)))),
            head($sdf($point=>v:add((0, 0, $epsilon))))
          )),
          ($c, $c, $c)
        )
    '/>
   </xsl:function>

   <!--
    : this:linear()
    :-->
   <xsl:function name="this:linear" as="xs:double*">
      <xsl:param name="origin" as="xs:double*"/>
      <xsl:param name="direction" as="xs:double*"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      util:zip(
        function($a, $b) {$a + $t * $b},
          $origin,
          $direction
      )
    '/>
   </xsl:function>

   <!--
    : this:normalize-to-canvas()
    : Map point from canvas space to unit box. Need to shift and scale.
    : Canvas: [0:w, 0:h] Unit box: [-1:1, -1:1]
    :-->
   <xsl:function name="this:normalize-to-canvas"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="pt" as="map(xs:string, item()*)"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := box:width($canvas) return
      let $height := box:height($canvas)
      return (
        point:vector((
          2.0 * (point:px($pt) - ($width div 2)) div $width,
          2.0 * (point:py($pt) - ($height div 2)) div $height,
          point:pcoordinates($pt)[position() > 2]
        ))
      )
    '/>
   </xsl:function>

   <!--
    : this:normalize-to-canvasv()
    :-->
   <xsl:function name="this:normalize-to-canvasv" as="xs:double*">
      <xsl:param name="pt" as="xs:double*"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := box:width($canvas) return
      let $height := box:height($canvas)
      return (
        (
          2.0 * (v:px($pt) - ($width div 2)) div $width,
          2.0 * (v:py($pt) - ($height div 2)) div $height,
          ($pt)[position() > 2]
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:normalize-to-canvas()
    : Map point from canvas space to unit box. Need to shift and scale.
    : Canvas: [0:w, 0:h] Unit box: [-1:1, -1:1]
    : A version defined for efficiency when we are scanning through a whole
    : matrix of x/y values.
    :-->
   <xsl:function name="this:normalize-to-canvas"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="aspect" as="xs:double"/>
      <xsl:sequence select='
      point:point(
        2.0 * ($x - ($width div 2)) div $width,
        2.0 * ($y - ($height div 2)) div ($aspect * $height)
      )
    '/>
   </xsl:function>

   <!--
    : this:normalize-to-canvasv()
    :-->
   <xsl:function name="this:normalize-to-canvasv" as="xs:double*">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="aspect" as="xs:double"/>
      <xsl:sequence select='
      (
        2.0 * ($x - ($width div 2)) div $width,
        2.0 * ($y - ($height div 2)) div ($aspect * $height)
      )
    '/>
   </xsl:function>

   <!--
    : this:scale-from-canvas()
    : Scale a distance metric from canvas space to distance space. Scaling only.
    : (Scales by width.)
    : Canvas distance: [-w/2:w/2] Unit distance: [-1:1] 
    :-->
   <xsl:function name="this:scale-from-canvas" as="xs:double">
      <xsl:param name="d" as="xs:double"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := $canvas("width")
      return (
        2.0 * $d div $width
      )
    '/>
   </xsl:function>

   <!--
    : this:scale-from-unit()
    : Scale a distance metric from distance space to canvas space. Scaling only.
    : (Scales by width.)
    : Unit distance: [-1:1] Canvas distance: [-w/2:w/2] 
    :-->
   <xsl:function name="this:scale-from-unit" as="xs:double">
      <xsl:param name="d" as="xs:double"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := $canvas("width")
      return (
        0.5 * $d * $width
      )
    '/>
   </xsl:function>

   <!--
    : this:regionToSDF()
    :-->
   <xsl:function name="this:regionToSDF" as="map(*)" visibility="private">
      <xsl:param name="region" as="map(xs:string,item()*)"/>
      <xsl:param name="op" as="xs:string"/>
      <xsl:param name="options" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      let $kind := util:kind($region) return
      if ($kind = "point") then (
        let $size := $options("radius") return
        let $pt := $region=>point:minus()
        return this:opTranslate(sd2:sdCircle($size), $pt)
      )
      else if ($kind = "box") then (
        if (point:same(box:min-point($region), box:max-point($region)))
        then this:regionToSDF(box:min-point($region), $op, $options)
        else (
          let $min := box:min-point($region) return
          let $max := box:max-point($region) return
          let $sides := point:sub($max, $min) return
          let $offset := $sides=>point:times(0.5)=>point:add($min)=>point:minus()
          return
            this:opTranslate(
              sd2:sdBox(point:px($sides), point:py($sides)),
              $offset
            )
        )
      )
      else if ($kind = "block") then (
        let $center := box:center($region) return
        let $offset := $center=>point:minus()
        return (
          this:opTranslate(
            sd3:sdBox(box:width($region), box:height($region), box:depth($region)),
            $offset
          )
        )
      )
      else if ($kind = "space") then (
        if (point:same(box:min-point($region), box:max-point($region)))
        then this:regionToSDF(box:min-point($region), $op, $options)
        else (
          let $min := box:min-point($region) return
          let $max := box:max-point($region) return
          let $sides := point:sub($max, $min) return
          let $offset := $sides=>point:times(0.5)=>point:add($min)=>point:minus()
          return
            this:opTranslate(
              sd2:sdBox(point:px($sides), point:py($sides)),
              $offset
            )
        )
      )
      else if ($kind = "ellipse") then (
        if (ellipse:is-circle($region)) then (
          let $center := ellipse:center($region) return
          let $r := ellipse:radius($region)
          return
            this:opTranslate(
              sd2:sdCircle($r),
              $center=>point:minus()
            )
        ) else (
          let $center := ellipse:center($region) return
          let $a := ellipse:rx($region) return
          let $b := ellipse:ry($region) return
          let $rotate := ellipse:rotation($region)
          return
            this:opTranslate(
              if ($rotate = 0) 
              then sd2:sdEllipse($a, $b)
              else this:opRotate(sd2:sdEllipse($a, $b), $rotate),
              $center=>point:minus()
            ) 
        )
      )
      else if ($kind = "edge") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $options)
        else (
          let $start := edge:start($region) return
          let $end := edge:end($region)
          return sd2:sdSegment($start, $end)
        )
      )
      else if ($kind = "quad") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $options)
        else (
          let $start := edge:start($region) return
          let $end := edge:end($region) return
          let $control := edge:controls($region)[1]
          return
            if ($options("ud"))
            then sd2:udQuad(edge:quad($start, $end, $control))
            else sd2:sdQuad(edge:quad($start, $end, $control))
        )
      )
      else if ($kind = "cubic") then ( (: Approximate with interpolation :)
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $options)
        else (
          let $interpolate := $options("interpolate") return
          let $n := if ($interpolate > 0) then $interpolate else 4 return
          let $edges := edge:to-edges(edge:interpolate($n, $region))
          return
            if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
            then this:toSDF($edges, $options)
            else (
              this:toSDF($edges,
                util:merge-into($options, map {"op":"union"})
            )
          )
        )
      )
      else if ($kind = "arc") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $options)
        else (
          let $arc-width := $options("width") return
          let $arc := edge:as-angle-arc($region) return
          let $angles := ($arc=>edge:arc-angles())!util:remap-degrees(.) return
          let $large := edge:arc-large($arc) return
          let $flipped := edge:arc-flipped($arc) return
          let $delta := $angles[2] - $angles[1] return
          let $swap :=
            ($large and
              ($delta &lt; -180 or ($delta > 0 and $delta &lt; 180))
            ) or
            (not($large) and
              (($delta &lt; 0 and abs($delta) &lt; 180) or ($delta > 180))
            )
          return
          let $adjust-from :=
            ($large and $delta > 0 and $delta &lt; 180) or
            (not($large) and $delta > 180)
         return
         let $adjust-to :=
            ($large and $delta &lt; 0 and abs($delta) &lt;= 180) or
            (not($large) and $delta &lt;= -180)
          return
          let $adjusted-from :=
            if ($adjust-from) then $angles[1] + 360 else $angles[1]
          return
          let $adjusted-to :=
            if ($adjust-to) then $angles[2] + 360 else $angles[2]
          return
          let $from-angle :=
            if ($swap) then $adjusted-to else $adjusted-from
          return
          let $to-angle := 
            if ($swap) then $adjusted-from else $adjusted-to
          return
          let $center := $arc=>edge:arc-center() return
          let $radius := $arc=>edge:arc-radius() return
          (: 
           : If we run from A to B, orientation is 180 from angle in 
           : the middle of the range
           : orientation = 180 + (A + B)/2 
           : aperture = (B - A)
           : However, orientation of degrees is flipped, so we need to negate orientation
           :)
          let $orientation := util:remap-degrees(-(180 + ($to-angle + $from-angle) div 2.0)) return
          let $aperture := util:remap-degrees(($to-angle - $from-angle)) return
          (: center and radius already normalized :)
          let $from := point:destination($center, $from-angle, $radius) return
          let $to := point:destination($center, $to-angle, $radius)
          return (
            this:opTranslate(
              if ( ($swap and not($flipped)) or (not($swap) and $flipped) ) then (
                this:opReflectXY(
                  sd2:sdArc($orientation, $aperture, $radius, $arc-width),
                  $from,
                  $to
                )
              ) else (
                sd2:sdArc($orientation, $aperture, $radius, $arc-width)
              ),
              $center=>point:minus()
            )
          )
        )
      )
      else if ($kind = "path") then (
        if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
        then this:toSDF(path:edges($region), $options)
        else (
          this:toSDF(path:edges($region), 
            util:merge-into($options, map {"op":"union"})
          )
        )
      )
      else if ($kind = "polygon") then (
        let $interpolate := $options("interpolate") return
        if ($interpolate=0) then (
          sd2:sdPolygon(path:vertices($region))
        ) else if ($interpolate > 0) then (
          sd2:sdPolygon(
            edge:interpolate($interpolate, path:edges($region))
          )
        ) else (
          if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
          then this:toSDF(path:edges($region), $options)
          else (
            this:toSDF(path:edges($region),
              util:merge-into($options, map{"op":"union"})
            )
          )
        )
      )
      else if ($kind = "sphere") then (
        let $center := solid:center($region) return
        let $r := solid:radius($region) return
        let $offset := $center=>point:minus()
        return
          this:opTranslate(
            sd3:sdSphere($r),
            $offset
          )
      )
      else if ($kind = "ellipsoid") then (
        let $center := solid:center($region) return
        let $ra := solid:rx($region) return
        let $rb := solid:ry($region) return
        let $rc := solid:rz($region) return
        let $offset := $center=>point:minus()
        return
          this:opTranslate(
            sd3:sdEllipsoid($ra, $rb, $rc),
            $offset
          )
      )
      else if ($kind = "tetrahedron") then (
        let $center := solid:center($region) return
        let $size := solid:scale($region) return
        let $offset := $center=>point:minus()
        return             
          this:opTranslate(
            sd3:sdPyramid($size * math:sqrt(3) div 2, $size),
            $offset
          )
      )
      else if ($kind = "cube") then (
        let $center := solid:center($region) return
        let $size := solid:scale($region) return
        let $offset := $center=>point:minus()
        return
          this:opTranslate(
            sd3:sdBox($size, $size, $size),
            $offset
          )
      )
      else if ($kind = "octahedron") then (
        let $center := solid:center($region) return
        let $size := solid:scale($region) return
        let $offset := $center=>point:minus()
        return
          this:opTranslate(
            sd3:sdOctahedron($size),
            $offset
          )
      )
      else if ($kind = "face") then (
        let $vertices := solid:vertices($region)[position() &lt; last()]
        return (
          if (count($vertices) = 3) then (
            sd3:udTriangle($vertices[1], $vertices[2], $vertices[3])
          ) else (
            let $k := $options("smoothing") return
            let $n := $options("steps") return
            let $triangles :=
              for $i in 2 to count($vertices) - 1
              return sd3:udTriangle($vertices[1], $vertices[$i], $vertices[$i + 1])
            return (
              if ($op="smooth-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opSmoothUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="round-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opRoundUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="chamfer-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opChamferUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="stepped-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opSteppedUnion($sdf, $triangle, $k, $n)
                  }
                )
              ) else (
                this:opUnion($triangles)
              )
            )
          )
        )
      )
      else if ($kind = "polyhedron") then (
        if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
        then this:toSDF(solid:faces($region), $options)
        else (
          this:toSDF(solid:faces($region), 
            util:merge-into($options, map{"op":"union"})
          )
        )
      )
      else if ($kind = "wrapper") then (
        this:toSDF(wrapper:body($region), $options)
      )
      else if ($kind = "slot") then (
        (: XYZZY: NOT accounting for transforms :)
        this:toSDF(slot:body($region), $options)
      )
      else errors:error("GEOM-BADREGION", ($region, "regionToSDF"))
    '/>
   </xsl:function>

   <xsl:function name="this:regionToSDF" as="map(*)" visibility="private">
      <xsl:param name="region" as="map(xs:string,item()*)"/>
      <xsl:param name="op" as="xs:string"/>
      <xsl:param name="canvas" as="map(xs:string,item()*)"/>
      <xsl:param name="options" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      let $kind := util:kind($region) return
      if ($kind = "point") then (
        let $size := this:scale-from-canvas($options("radius"),$canvas) return
        let $pt := this:normalize-to-canvas($region, $canvas)=>point:minus()
        return this:opTranslate(sd2:sdCircle($size), $pt)
      )
      else if ($kind = "box") then (
        if (point:same(box:min-point($region), box:max-point($region)))
        then this:regionToSDF(box:min-point($region), $op, $canvas, $options)
        else (
          let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
          let $min := this:normalize-to-canvas(box:min-point($region), $canvas) return
          let $max := this:normalize-to-canvas(box:max-point($region), $canvas) return
          let $sides := point:sub($max, $min) return
          let $offset := $sides=>point:times(0.5)=>point:add($min)=>point:minus()=>point:add($origin)
          return
            this:opTranslate(
              sd2:sdBox(point:px($sides), point:py($sides)),
              $offset
            )
        )
      )
      else if ($kind = "block") then (
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $center := this:normalize-to-canvas(box:center($region), $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return (
          this:opTranslate(
            sd3:sdBox(box:width($region), box:height($region), box:depth($region)),
            $offset
          )
        )
      )
      else if ($kind = "space") then (
        if (point:same(box:min-point($region), box:max-point($region)))
        then this:regionToSDF(box:min-point($region), $op, $canvas, $options)
        else (
          let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
          let $min := this:normalize-to-canvas(box:min-point($region), $canvas) return
          let $max := this:normalize-to-canvas(box:max-point($region), $canvas) return
          let $sides := point:sub($max, $min) return
          let $offset := $sides=>point:times(0.5)=>point:add($min)=>point:minus()=>point:add($origin)
          return
            this:opTranslate(
              sd2:sdBox(point:px($sides), point:py($sides)),
              $offset
            )
        )
      )
      else if ($kind = "ellipse") then (
        if (ellipse:is-circle($region)) then (
          let $center := this:normalize-to-canvas(ellipse:center($region), $canvas) return
          let $r := this:scale-from-canvas(ellipse:radius($region), $canvas)
          return
            this:opTranslate(
              sd2:sdCircle($r),
              $center=>point:minus()
            )
        ) else (
          let $center := this:normalize-to-canvas(ellipse:center($region), $canvas) return
          let $a := this:scale-from-canvas(ellipse:rx($region), $canvas) return
          let $b := this:scale-from-canvas(ellipse:ry($region), $canvas) return
          let $rotate := ellipse:rotation($region)
          return
            this:opTranslate(
              if ($rotate = 0) 
              then sd2:sdEllipse($a, $b)
              else this:opRotate(sd2:sdEllipse($a, $b), $rotate),
              $center=>point:minus()
            ) 
        )
      )
      else if ($kind = "edge") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $canvas, $options)
        else (
          let $start := this:normalize-to-canvas(edge:start($region), $canvas) return
          let $end := this:normalize-to-canvas(edge:end($region), $canvas)
          return sd2:sdSegment($start, $end)
        )
      )
      else if ($kind = "quad") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $canvas, $options)
        else (
          let $start := this:normalize-to-canvas(edge:start($region), $canvas) return
          let $end := this:normalize-to-canvas(edge:end($region), $canvas) return
          let $control := this:normalize-to-canvas(edge:controls($region)[1], $canvas)
          return
            if ($options("ud"))
            then sd2:udQuad(edge:quad($start, $end, $control))
            else sd2:sdQuad(edge:quad($start, $end, $control))
        )
      )
      else if ($kind = "cubic") then ( (: Approximate with interpolation :)
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $canvas, $options)
        else (
          let $interpolate := $options("interpolate") return
          let $n := if ($interpolate > 0) then $interpolate else 4 return
          let $edges := edge:to-edges(edge:interpolate($n, $region))
          return
            if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
            then this:toSDF($edges, $canvas, $options)
            else (
              this:toSDF($edges, $canvas,
                util:merge-into($options, map {"op":"union"})
            )
          )
        )
      )
      else if ($kind = "arc") then (
        if (point:same(edge:start($region), edge:end($region)))
        then this:regionToSDF(edge:start($region), $op, $canvas, $options)
        else (
          let $arc-width := this:scale-from-canvas($options("width"), $canvas) return
          let $arc := edge:as-angle-arc($region) return
          let $angles := ($arc=>edge:arc-angles())!util:remap-degrees(.) return
          let $large := edge:arc-large($arc) return
          let $flipped := edge:arc-flipped($arc) return
          let $delta := $angles[2] - $angles[1] return
          let $swap :=
            ($large and
              ($delta &lt; -180 or ($delta > 0 and $delta &lt; 180))
            ) or
            (not($large) and
              (($delta &lt; 0 and abs($delta) &lt; 180) or ($delta > 180))
            )
          return
          let $adjust-from :=
            ($large and $delta > 0 and $delta &lt; 180) or
            (not($large) and $delta > 180)
         return
         let $adjust-to :=
            ($large and $delta &lt; 0 and abs($delta) &lt;= 180) or
            (not($large) and $delta &lt;= -180)
          return
          let $adjusted-from :=
            if ($adjust-from) then $angles[1] + 360 else $angles[1]
          return
          let $adjusted-to :=
            if ($adjust-to) then $angles[2] + 360 else $angles[2]
          return
          let $from-angle :=
            if ($swap) then $adjusted-to else $adjusted-from
          return
          let $to-angle := 
            if ($swap) then $adjusted-from else $adjusted-to
          return
          let $center := this:normalize-to-canvas($arc=>edge:arc-center(), $canvas) return
          let $radius := this:scale-from-canvas($arc=>edge:arc-radius(), $canvas) return
          (: 
           : If we run from A to B, orientation is 180 from angle in 
           : the middle of the range
           : orientation = 180 + (A + B)/2 
           : aperture = (B - A)
           : However, orientation of degrees is flipped, so we need to negate orientation
           :)
          let $orientation := util:remap-degrees(-(180 + ($to-angle + $from-angle) div 2.0)) return
          let $aperture := util:remap-degrees(($to-angle - $from-angle)) return
          (: center and radius already normalized :)
          let $from := point:destination($center, $from-angle, $radius) return
          let $to := point:destination($center, $to-angle, $radius)
          return (
            this:opTranslate(
              if ( ($swap and not($flipped)) or (not($swap) and $flipped) ) then (
                this:opReflectXY(
                  sd2:sdArc($orientation, $aperture, $radius, $arc-width),
                  $from,
                  $to
                )
              ) else (
                sd2:sdArc($orientation, $aperture, $radius, $arc-width)
              ),
              $center=>point:minus()
            )
          )
        )
      )
      else if ($kind = "path") then (
        if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
        then this:toSDF(path:edges($region), $canvas, $options)
        else (
          this:toSDF(path:edges($region), $canvas,
            util:merge-into($options, map {"op":"union"})
          )
        )
      )
      else if ($kind = "polygon") then (
        let $interpolate := $options("interpolate") return
        if ($interpolate=0) then (
          sd2:sdPolygon(
            for $pt in path:vertices($region)
            return this:normalize-to-canvas($pt, $canvas)
          )
        ) else if ($interpolate > 0) then (
          sd2:sdPolygon(
            for $pt in edge:interpolate($interpolate, path:edges($region))
            return this:normalize-to-canvas($pt, $canvas)
          )
        ) else (
          if ($op=("union","smooth-union","round-union","stepped-union","chamfer-union"))
          then this:toSDF(path:edges($region), $canvas, $options)
          else (
            this:toSDF(path:edges($region), $canvas,
              util:merge-into($options, map{"op":"union"})
            )
          )
        )
      )
      else if ($kind = "sphere") then (
        let $center := this:normalize-to-canvas(solid:center($region), $canvas) return
        let $r := this:scale-from-canvas(solid:radius($region), $canvas) return
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return
          this:opTranslate(
            sd3:sdSphere($r),
            $offset
          )
      )
      else if ($kind = "ellipsoid") then (
        let $center := this:normalize-to-canvas(solid:center($region), $canvas) return
        let $ra := this:scale-from-canvas(solid:rx($region), $canvas) return
        let $rb := this:scale-from-canvas(solid:ry($region), $canvas) return
        let $rc := this:scale-from-canvas(solid:rz($region), $canvas) return
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return
          this:opTranslate(
            sd3:sdEllipsoid($ra, $rb, $rc),
            $offset
          )
      )
      else if ($kind = "tetrahedron") then (
        let $center := this:normalize-to-canvas(solid:center($region), $canvas) return
        let $size := this:scale-from-canvas(solid:scale($region), $canvas) return
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return             
          this:opTranslate(
            sd3:sdPyramid($size * math:sqrt(3) div 2, $size),
            $offset
          )
      )
      else if ($kind = "cube") then (
        let $center := this:normalize-to-canvas(solid:center($region), $canvas) return
        let $size := this:scale-from-canvas(solid:scale($region), $canvas) return
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return
          this:opTranslate(
            sd3:sdBox($size, $size, $size),
            $offset
          )
      )
      else if ($kind = "octahedron") then (
        let $center := this:normalize-to-canvas(solid:center($region), $canvas) return
        let $size := this:scale-from-canvas(solid:scale($region), $canvas) return
        let $origin := this:normalize-to-canvas($point:ORIGIN, $canvas) return
        let $offset := $center=>point:minus()=>point:add($origin)
        return
          this:opTranslate(
            sd3:sdOctahedron($size),
            $offset
          )
      )
      else if ($kind = "face") then (
        let $vertices :=
          for $v in solid:vertices($region)[position() &lt; last()]
          return this:normalize-to-canvas($v, $canvas)
        return (
          if (count($vertices) = 3) then (
            sd3:udTriangle($vertices[1], $vertices[2], $vertices[3])
          ) else (
            let $k := $options("smoothing") return
            let $n := $options("steps") return
            let $triangles :=
              for $i in 2 to count($vertices) - 1
              return sd3:udTriangle($vertices[1], $vertices[$i], $vertices[$i + 1])
            return (
              if ($op="smooth-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opSmoothUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="round-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opRoundUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="chamfer-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opChamferUnion($sdf, $triangle, $k)
                  }
                )
              ) else if ($op="stepped-union") then (
                fold-left(tail($triangles), head($triangles),
                  function($sdf, $triangle) {
                    this:opSteppedUnion($sdf, $triangle, $k, $n)
                  }
                )
              ) else (
                this:opUnion($triangles)
              )
            )
          )
        )
      )
      else if ($kind = "polyhedron") then (
        if ($op=("smooth-union","round-union","stepped-union","chamfer-union"))
        then this:toSDF(solid:faces($region), $canvas, $options)
        else (
          this:toSDF(solid:faces($region), $canvas,
            util:merge-into($options, map{"op":"union"})
          )
        )
      )
      else if ($kind = "wrapper") then (
        this:toSDF(wrapper:body($region), $canvas, $options)
      )
      else if ($kind = "slot") then (
        (: XYZZY: NOT accounting for transforms :)
        this:toSDF(slot:body($region), $canvas, $options)
      )
      else errors:error("GEOM-BADREGION", ($region, "regionToSDF"))
    '/>
   </xsl:function>

   <!--
    : this:toSDF()
    :-->
   <xsl:function name="this:toSDF" as="map(*)">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="options" as="map(xs:string, item()*)"/>
      <xsl:sequence select="this:toSDF($regions, (), $options)"/>
   </xsl:function>
   
   <xsl:function name="this:toSDF" as="map(*)">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)?"/>
      <xsl:param name="options" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $options :=
        util:merge-into((
          map {
            "op": "union",
            "smoothing": 0.05,
            "interpolate": 0,
            "steps": 3,
            "radius": 1,
            "width": 1,
            "ud": false()
          },
          $options
        ))
      return
      let $op := $options("op") return
      let $k := $options("smoothing") return
      let $n := $options("steps") return 
      let $translated-regions :=
        if (exists($canvas)) then (
          for $region in $regions return (
            this:regionToSDF($region, $op, $canvas, $options)
          )
        ) else (
          for $region in $regions return (
            this:regionToSDF($region, $op, $options)
          )
        )
      return (
        if ($op="union") then (
          this:opUnion($translated-regions)
        ) else if ($op="intersect") then (
          this:opIntersect($translated-regions)
        ) else (
          fold-left(
            tail($translated-regions), head($translated-regions),
            function($sdf, $current) {
              if (empty($sdf)) then $current
              else if ($op = "intersect") then this:opIntersect($sdf, $current)
              else if ($op = "subtract") then this:opSubtract($sdf, $current)
              else if ($op = "round-union") then this:opRoundUnion($sdf, $current, $k)
              else if ($op = "round-intersect") then this:opRoundIntersect($sdf, $current, $k)
              else if ($op = "round-subtract") then this:opRoundSubtract($sdf, $current, $k)
              else if ($op = "smooth-union") then this:opSmoothUnion($sdf, $current, $k)
              else if ($op = "smooth-intersect") then this:opSmoothIntersect($sdf, $current, $k)
              else if ($op = "smooth-subtract") then this:opSmoothSubtract($sdf, $current, $k)
              else if ($op = "chamfer-union") then this:opChamferUnion($sdf, $current, $k)
              else if ($op = "chamfer-intersect") then this:opChamferIntersect($sdf, $current, $k)
              else if ($op = "chamfer-subtract") then this:opChamferSubtract($sdf, $current, $k)
              else if ($op = "stepped-union") then this:opSteppedUnion($sdf, $current, $k, $n)
              else if ($op = "stepped-intersect") then this:opSteppedIntersect($sdf, $current, $k, $n)
              else if ($op = "stepped-subtract") then this:opSteppedSubtract($sdf, $current, $k, $n)
              else (: union :) this:opUnion($sdf, $current)
            }
          )
        )
      )
    '/>
   </xsl:function>

</xsl:package>
<!--sdf.xsl-->
