<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Constructors and accessors for random distributions.
 : Parameters are usually numeric but can be functions to create dynamically
 : mutating distributions. See core/random.xqy for details. Port of 
 : distributions.xqy.
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:this="http://mathling.com/type/distribution"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:art="http://mathling.com/art"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"   
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             exclude-result-prefixes="this util xs math map array"
             version="3.0"
             name="http://mathling.com/type/distribution">

  <xsl:expose component="variable" names="this:*" visibility="public"/>
  <xsl:expose component="function" names="this:*" visibility="public"/>

  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

<!--
 :======================================================================
 : Basic distribution constructors
 :======================================================================
 :-->

 <!--
  : this:constant()
  : Construct a constant distribution with the given value
  :
  : $value: the value of the constant distribution
  :-->
  <xsl:function name="this:constant" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="value" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "constant",
      "min" : $value
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:uniform()
   : Construct a uniform distribution over the given value range
   :
   : $min: minimum value of the distribution
   : $max: maximum value of the distribution
   :-->
  <xsl:function name="this:uniform" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="min" as="item()"/>
    <xsl:param name="max" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "uniform",
      "min" : $min,
      "max" : $max
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:uniform-index()
   : Construct a uniform distribution that functions as an index over
   : integers starting
   :
   : $start: starting index
   : $count: how many index values
   :-->
  <xsl:function name="this:uniform-index" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="start" as="xs:integer"/>
    <xsl:param name="count" as="xs:integer"/>
    <xsl:sequence select='
    map {
      "distribution" : "uniform",
      "min" : $start,
      "max" : $start + $count,
      "cast": "integer"
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:uniform-index-of()
   : Construct a uniform distribution that functions as an index over
   : the values
   :
   : $values: the values
   :-->
  <xsl:function name="this:uniform-index-of" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="values" as="item()*"/>
    <xsl:sequence select='
     map {
       "distribution" : "uniform",
       "min" : 1,
       "max" : count($values),
       "cast": "integer",
       "keys": $values
     }=>this:set-is-complex()
     '/>
  </xsl:function>

  <!--
   : this:normal()
   : Construct a normal distribution 
   :
   : $mean: mean value of the distribution
   : $std: standard deviation of the distribution
   :-->
  <xsl:function name="this:normal" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="mean" as="item()"/>
    <xsl:param name="std" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "normal",
      "mean" : $mean,
      "std" : $std
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:normal()
   : Construct a normal distribution with default mean (0) and std (1)
   :-->
  <xsl:function name="this:normal" as="map(xs:string, item()*)" art:distribution="1">
     <xsl:sequence select='
     this:normal(0, 1)
     '/>
  </xsl:function>

  <!--
   : this:skewed()
   : Construct a skewed normal distribution 
   :
   : $mean: mean value of the distribution
   : $std: standard deviation of the distribution
   : $skew: skew of the distribution
   :
   : Skewed distribution is defined by parameters α, ξ, and ω
   : α is skew 
   :   α = 0 => normal, α > 0 right skewed, α < 0 left skewed
   :      right skewed => long tail on right
   : ξ = location = shift along x
   : ω = scaling along y
   :   ω is positive 
   :
   : Mean is mean value; if have long tail on right, means will have more 
   : instances below the mean, but larger variance to higher values
   :
   : mean = ξ + ω*δ*sqrt(2/π) δ=α/sqrt(1+α*α) 
   : std = ω*ω*(1 - 2*δ*δ/π)
   :
   : So:
   : ω = sqrt(std / (1 - 2*δ*δ/π))
   : ξ = mean - ω*δ*sqrt(2/π)
   :-->
  <xsl:function name="this:skewed" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="mean" as="item()"/>
    <xsl:param name="std" as="item()"/>
    <xsl:param name="skew" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "skewed",
      "mean" : $mean,
      "std" : $std,
      "skew" : $skew
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:skewed()
   : Construct a skewed normal distribution with default mean, 
   : standard deviation, and skew
   :-->
  <xsl:function name="this:skewed" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:sequence select='
    this:skewed(0, 1, 0)
    '/>
  </xsl:function>

  <!--
   : this:bernoulli()
   : Construct a Bernoulli distribution
   :
   : $p: Probability of returning 1
   :-->
  <xsl:function name="this:bernoulli" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="p" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "bernoulli",
      "p" : $p
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:bernoulli()
   : Construct a Bernoulli distribution with default probability
   :-->
  <xsl:function name="this:bernoulli" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:sequence select='
    this:bernoulli(50)
    '/>
  </xsl:function>

  <!--
   : this:flip()
   : Construct a Bernoulli distribution rendered as boolean values
   :
   : $p: Probability of returning true()
   :-->
  <xsl:function name="this:flip" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="p" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "flip",
      "p" : $p
    }=>this:set-is-complex()
   '/>
  </xsl:function>

  <!--
   : this:flip()
   : Construct a Bernoulli distribution rendered as boolean values with
   : default probability
   :-->
  <xsl:function name="this:flip" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:sequence select='
    this:flip(50)
    '/>
 </xsl:function>

  <!--
   : this:zipf-sums()
   : Returns a sequence of cumulative Zipf probabilities, to be used
   : by select-index() to perform selections. This is to ensure we don't have 
   : to keep recomputing the cumulative sums for each selection.
   : 
   : $alpha: the alpha parameter, should be >=1
   :   That said, numbers < 1 work:
   :   0 = α => uniform
   :   0 < α < 1 => get more representation of higher ranks
   :   α < 0 => invert the range, so that higher ranks are more likely than 
   :      lower ones
   :   (Rounded) cumulative percents for various α values:
   :                   1    2    3    4    5    6    7    8    9   10
   :     z(-3,10)=    0.0  0.3  1.2  3.3  7.4 14.6 25.9 42.8 66.9 100.0
   :     z(-2,10)=    0.3  1.3  3.6  7.8 14.3 23.6 36.4 53.0 74.0 100.0
   :     z(-1.1,10)=  1.5  4.7  9.8 16.7 25.6 36.4 49.2 64.1 81.0 100.0
   :     z(-1,10)=    1.8  5.5 10.9 18.2 27.3 38.2 50.9 65.5 81.8 100.0
   :     z(-0.9,10)=  2.2  6.3 12.1 19.8 29.1 40.0 52.6 66.8 82.6 100.0
   :     z(-0.5,10)=  4.5 10.7 18.5 27.4 37.3 48.2 60.0 72.6 85.9 100.0
   :     z(-0.3,10)=  6.2 13.9 22.6 32.0 42.1 52.7 63.9 75.5 87.6 100.0
   :     z(0,10)=    10.0 20.0 30.0 40.0 50.0 60.0 70.0 80.0 90.0 100.0
   :     z(0.3,10)=  15.4 27.9 38.9 49.1 58.5 67.5 76.1 84.3 92.3 100.0
   :     z(0.5,10)=  19.9 34.0 45.5 55.5 64.4 72.5 80.0 87.1 93.7 100.0
   :     z(0.9,10)=  31.0 47.7 59.2 68.1 75.4 81.6 87.0 91.8 96.1 100.0
   :     z(1,10)=    34.1 51.2 62.6 71.1 78.0 83.6 88.5 92.8 96.6 100.0
   :     z(1.1,10)=  37.3 54.7 65.9 74.0 80.3 85.5 89.9 93.7 97.0 100.0
   :     z(2,10)=    64.5 80.7 87.8 91.9 94.4 96.2 97.5 98.6 99.4 100.0
   :     z(3,10)=    83.5 93.9 97.0 98.3 99.0 99.4 99.6 99.8 99.9 100.0
   : $n: number to generate
   :-->
  <xsl:function name="this:zipf-sums" as="xs:double*" new-each-time="no">
    <xsl:param name="alpha" as="xs:double"/>
    <xsl:param name="n" as="xs:integer"/>
    <xsl:variable name="c" as="xs:double">
      <xsl:sequence select='
      1 div
        fold-left(
          1 to $n, 0,
          function($z as xs:double, $a as xs:integer) as xs:double {
            $z + 1 div (math:pow(xs:double($a), $alpha))
          }
        )
      '/>
    </xsl:variable>
    <xsl:sequence select='
    fold-left(
      1 to $n, 0,
      function($z as xs:double*, $a as xs:integer) as xs:double* {
        $z,
        $z[last()] + $c div math:pow(xs:double($a), $alpha)
      }
    )=>tail()
    '/>
  </xsl:function>

  <!--
   : this:zipf()
   : Construct a Zipf distribution as an integer index from 1 to N
   :
   : p(n)=(1/k^α)/sum(i=1 to n)(1/i^α)
   :    n = number of elements
   :    k = rank
   :    α = exponent
   :
   : $alpha: the α parameter, should be >=1
   :   For English words this by some estimates is around 1.6
   :   For city sizes 1.07
   : $n: number of Zipf values in range
   :-->
  <xsl:function name="this:zipf" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="alpha" as="xs:double"/>
    <xsl:param name="n" as="xs:integer"/>
    <xsl:sequence select='
    map {
      "distribution" : "zipf",
      "cast" : "integer",
      "alpha" : $alpha,
      "sums" : this:zipf-sums($alpha,$n+1),
      "max" : $n,
      "truncation" : "resample"
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:zipf()
   : Construct a Zipf distribution as an integer index from 1 to N with
   : default alpha and limit
   :-->
  <xsl:function name="this:zipf" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:sequence select='
    this:zipf(1,1000)
    '/>
  </xsl:function>

  <!--
   : this:zipf-index-of()
   : Construct a Zipf index over the set of values, taking them in the order
   : given.
   :-->
  <xsl:function name="this:zipf-index-of" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="alpha" as="xs:double"/>
    <xsl:param name="values" as="item()*"/>
    <xsl:variable name="n" as="xs:integer" select='count($values)'/>
    <xsl:sequence select='
    map {
      "distribution" : "zipf",
      "cast" : "integer",
      "alpha" : $alpha,
      "sums" : this:zipf-sums($alpha,$n+1),
      "max" : $n,
      "truncation" : "resample",
      "keys": $values
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:markov-sums()
   : Return a Markov probability matrix as a matrix where each row contains
   : the cumulative probability sums for that row. This allows the Markov
   : distribution selection to run more efficiently.
   :
   : Example:
   :   markov-sums(3, (
   :     0.2, 0.4, 0.4,
   :     0.5, 0.5, 0,
   :     0.4, 0.2, 0.4
   :   )) => (
   :     0.2, 0.6, 1,
   :     0.5, 1, 1,
   :     0.4, 0.6, 1
   :   )
   :
   :  To account for rounding, the last item in the row is always forced to 1
   :  and any sums greater than 1 are also rounded down to 1.
   :
   : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
   : $matrix: the input non-cumulative Markov matrix
   :-->
  <xsl:function name="this:markov-sums" as="xs:double*" new-each-time="no">
    <xsl:param name="dim" as="xs:integer"/>
    <xsl:param name="matrix" as="xs:double*"/>
    <xsl:for-each select="1 to $dim">
      <xsl:variable name="i" as="xs:integer" select="."/>
      <xsl:variable name="row" as="xs:double*">
        <xsl:sequence select='
        $matrix[position()=(($i - 1) * $dim + 1 to $i * $dim)]
        '/>
      </xsl:variable>
      <xsl:variable name="new-row" as="xs:double*">
        <xsl:sequence select='
        fold-left(
          $row, 0,
          function($z as xs:double*, $a as xs:double) as xs:double* {
            $z,
            min(($z[last()] + $a, 1))
          }
        )=>tail()
        '/>
      </xsl:variable>
      <xsl:sequence select='
      if ($new-row[last()] lt 1) 
      then ($new-row[position()=(1 to last()-1)],1) 
      else $new-row
      '/>
    </xsl:for-each>
  </xsl:function>

  <!--
   : this:markov-percent-sums()
   : Return a Markov probability matrix as a matrix where each row contains
   : the cumulative probability sums for that row. This allows the Markov
   : distribution selection to run more efficiently.
   :
   : Example:
   :   markov-percent-sums(3, (
   :     20, 40, 40,
   :     50, 50, 0,
   :     40, 20, 40
   :   )) => (
   :     0.2, 0.6, 1,
   :     0.5, 1, 1,
   :     0.4, 0.6, 1
   :   )
   :
   :  To account for rounding, the last item in the row is always forced to 1
   :  and any sums greater than 1 are also rounded down to 1.
   :
   : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
   : $matrix: the input non-cumulative Markov matrix
   :-->
  <xsl:function name="this:markov-percent-sums" as="xs:double*" new-each-time="no">
    <xsl:param name="dim" as="xs:integer"/>
    <xsl:param name="matrix" as="xs:integer*"/>
    <xsl:for-each select="1 to $dim">
      <xsl:variable name="i" as="xs:integer" select="."/>
      <xsl:variable name="row" as="xs:integer*">
        <xsl:sequence select='
        $matrix[position()=(($i - 1) * $dim + 1 to $i * $dim)]
        '/>
      </xsl:variable>
      <xsl:variable name="new-row" as="xs:double*">
        <xsl:sequence select='
        fold-left(
          $row, 0,
          function($z as xs:double*, $a as xs:integer) as xs:double* {
            $z,
            min(($z[last()] + ($a div 100), 1))
          }
        )=>tail()
        '/>
      </xsl:variable>
      <xsl:sequence select='
      if ($new-row[last()] lt 1) 
      then ($new-row[position()=(1 to last()-1)],1) 
      else $new-row
      '/>
    </xsl:for-each>
  </xsl:function>

  <!--
   : this:markov()
   : Construct a Markov distribution given a square matrix of probabilities
   : expressed as percentages (0 to 100).
   :
   : Example:
   : (0, 10, 90
   :  10, 0, 90,
   :  33, 33, 33)
   : Says state 1 transitions to state 2 10% of the time and state 3 90% of the
   : time; state 2 transitions to state 1 10% of the time and state 3 90% of the
   : time; state3 transitions to each state 1/3 of the time (actually, due to
   : rounding, state 3 transitions to state 3 34% of the time)
   :
   : $start: Starting state (an index in the range 1 to sqrt(count($matrix))
   : $matrix: Square matrix of percentages
   :-->
  <xsl:function name="this:markov" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="start" as="xs:integer"/>
    <xsl:param name="percent-matrix" as="xs:integer*"/>
    <xsl:variable name="dim" as="xs:integer">
      <xsl:sequence select='
      math:sqrt(count($percent-matrix)) cast as xs:integer
      '/>
    </xsl:variable>
    <xsl:sequence select='
    map {
      "distribution" : "markov",
      "dim" : $dim,
      "sums" : this:markov-percent-sums($dim, $percent-matrix),
      "start" : $start
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:markov()
   : Construct a Markov distribution given a square matrix of probabilities
   : expressed as percentages (0 to 100). 1 as start
   :
   : $percent-matrix: Square matrix of percentages
   :-->
  <xsl:function name="this:markov" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="percent-matrix" as="xs:integer*"/>
    <xsl:sequence select='
    this:markov(1, $percent-matrix)
    '/>
  </xsl:function>

  <!--
   : this:percent-sums()
   : Returns a sequence of cumulative probabilities, to be used
   : by select-index() to perform selections. This is to ensure we don't have to
   : keep recomputing the cumulative sums for each selection.
   :
   : To account for rounding, the last item is always forced to 1 and any sums
   : greater than 1 are also rounded down to 1. To get expected distributions, 
   : ensure that the input percentages sum to 100.
   :
   : Sums are returned in map:keys() order, so this assumes stability of that
   : function between calls to percent-sums() and calls to selection(). 
   : 
   : $weights a map from key to percent (expressed as number from 0 to 100)
   :-->
  <xsl:function name="this:percent-sums" as="xs:double*" new-each-time="no">
    <xsl:param name="weights" as="map(xs:anyAtomicType, xs:numeric)"/>
    <xsl:variable name='percentages' as="xs:numeric*">
      <xsl:sequence select='
      for $k in $weights=>map:keys() return $weights=>map:get($k)
      '/>
    </xsl:variable>
    <xsl:variable name="sums" as="xs:double*">
      <xsl:sequence select='
      fold-left(
        $percentages, 0, 
        function($z as xs:double*, $a as xs:numeric) as xs:double* {
          $z,
          min(($z[last()] + ($a div 100), 1))
        }
      )=>tail()
      '/>
    </xsl:variable>
    <xsl:sequence select='
    if ($sums[last()] lt 1) then ($sums[position() lt last()],1) else $sums
    '/>
  </xsl:function>

  <!--
   : this:simple-sums()
   : Construct a cumulative weight table from an ordered set of weights. [0,1]
   :-->
  <xsl:function name="this:simple-sums" as="xs:double*" new-each-time="no">
    <xsl:param name="weights" as="xs:double*"/>
    <xsl:sequence select='
    let $sums :=
      fold-left(
        $weights, 0, 
        function($z as xs:double*, $a as xs:double) as xs:double* {
          $z,
          min(($z[last()] + $a, 1))
        }
      )=>tail()
    return (
      if ($sums[last()] lt 1) then ($sums[position() lt last()],1) else $sums
    )
    '/>
  </xsl:function>

  <!--
   : this:sums()
   : Construct an ad hoc index distribution based on a weight map.
   : Assumes that keys are returned in a consistent order from the table
   : 
   : Example:
   :    {"a": 10, "b": 30, "c": 60}
   : If keys are returned in that order, will return 1 10% of the time, 2
   : 30% of the time, and 3 60% of the time.
   :
   : $weight-table: map of keys to percentages
   :-->
  <xsl:function name="this:sums" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="weight-table" as="map(xs:anyAtomicType, xs:numeric)"/>
    <xsl:sequence select='
    map {
      "distribution" : "sums",
      "cast" : "integer",
      "sums" : this:percent-sums($weight-table),
      "keys": $weight-table=>map:keys()
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:weighted-distribution-of()
   : Construct a sums() distribution based on a set of *relative* weights:
   : will rescale to percentages.
   :-->
  <xsl:function name="this:weighted-distribution-of" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="weights" as="xs:double*"/>
    <xsl:param name="values" as="item()*"/>
    <xsl:variable name="multiplier" as="xs:double">
      <xsl:sequence select='1.0 div sum($weights)'/>
    </xsl:variable>
    <xsl:sequence select='
    util:assert(count($weights) = count($values), "Different numbers of weights and values"),
    map {
      "distribution" : "sums",
      "cast" : "integer",
      "sums" : this:simple-sums($weights!(. * $multiplier)),
      "keys": $values
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <xsl:function name="this:weighted-distribution-of" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="weight-table" as="map(xs:anyAtomicType, xs:numeric)"/>
    <xsl:sequence select='
    let $weights := for $k in $weight-table=>map:keys() return $weight-table=>map:get($k) return
    let $values := $weight-table=>map:keys() return
    let $multiplier := 1.0 div sum($weights) 
    return
      map {
        "distribution" : "sums",
        "cast" : "integer",
        "sums" : this:simple-sums($weights!(. * $multiplier)),
        "keys": $values
      }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:multimodal()
   : Construct a multimodal distribution with a specific selection distribution
   :
   : $distributions: the component distributions
   : $selector: selection distribution, must return integers in range [1,|dist|]
   :-->
  <xsl:function name="this:multimodal" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="distributions" as="map(xs:string, item()*)*"/>
    <xsl:param name="selector" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    map {
      "distribution" : "multimodal",
      "distributions" : $distributions,
      "selector": $selector
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:multimodal()
   :-->
  <xsl:function name="this:multimodal" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="distributions" as="map(xs:string, item()*)*"/>
    <xsl:sequence select='
    map {
      "distribution" : "multimodal",
      "distributions" : $distributions,
      "selector": this:uniform(1, count($distributions))=>this:cast("integer")
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:binomial-poisson()
   : Construct a Poisson binomial distribution
   :
   : $probabilities: probabilities of the component Bernoulli distributions 
   :   as percentages
   :-->
  <xsl:function name="this:binomial-poisson" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="probabilities" as="item()*"/>
    <xsl:sequence select='
    map {
      "distribution" : "binomial-poisson",
      "probabilities" : $probabilities
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:poisson()
   : Construct a Poisson distribution
   :
   : $λ: the λ parameter of the distribution
   :-->
  <xsl:function name="this:poisson" as="map(xs:string, item()*)">
    <xsl:param name="λ" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "poisson",
      "mean": $λ
    }=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:exponential()
   :-->
  <xsl:function name="this:exponential" as="map(xs:string, item()*)" art:distribution="1">
    <xsl:param name="λ" as="item()"/>
    <xsl:sequence select='
    map {
      "distribution" : "exponential",
      "mean": 1 div $λ
    }=>this:set-is-complex()
    '/>
  </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

  <!--
   : this:distribution()
   :-->
  <xsl:function name="this:distribution" as="xs:string">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("distribution")
    '/>
  </xsl:function>

  <!--
   : this:cast()
   :-->
  <xsl:function name="this:cast" as="xs:string">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("cast")
    '/>
  </xsl:function>

  <!--
   : this:min()
   :-->
  <xsl:function name="this:min" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("min")
    '/>
  </xsl:function>

  <!--
   : this:max()
   :-->
  <xsl:function name="this:max" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("max")
    '/>
  </xsl:function>

  <!--
   : this:pre-multiplier()
   :-->
  <xsl:function name="this:pre-multiplier" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("pre-multiplier")
    '/>
  </xsl:function>

  <!--
   : this:post-multiplier()
   :-->
  <xsl:function name="this:post-multiplier" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("post-multiplier")
    '/>
  </xsl:function>

  <!--
   : this:post-shift()
   :-->
  <xsl:function name="this:post-shift" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("post-shift")
    '/>
  </xsl:function>

  <!--
   : this:truncation()
   :-->
  <xsl:function name="this:truncation" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("truncation")
    '/>
  </xsl:function>

  <!--
   : this:mean()
   :-->
  <xsl:function name="this:mean" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("mean")
    '/>
  </xsl:function>

  <!--
   : this:std()
   :-->
  <xsl:function name="this:std" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("std")
    '/>
  </xsl:function>

  <!--
   : this:skew()
   :-->
  <xsl:function name="this:skew" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("skew")
    '/>
  </xsl:function>

  <!--
   : this:p()
   :-->
  <xsl:function name="this:p" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("p")
    '/>
  </xsl:function>

  <!--
   : this:probabilities()
   :-->
  <xsl:function name="this:probabilities" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("probabilities")
    '/>
  </xsl:function>

  <!--
   : this:start()
   :-->
  <xsl:function name="this:start" as="item()">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("start")
    '/>
  </xsl:function>

  <!--
   : this:keys()
   :-->
  <xsl:function name="this:keys" as="item()*">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("keys")
    '/>
  </xsl:function>

  <!--
   : this:distributions()
   :-->
  <xsl:function name="this:distributions" as="map(xs:string, item()*)*">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("distributions")
    '/>
  </xsl:function>

  <!--
   : this:selector()
   :-->
  <xsl:function name="this:selector" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("selector")
    '/>
  </xsl:function>

  <!--
   : this:is-complex()
   :-->
  <xsl:function name="this:is-complex" as="xs:boolean">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution("is_complex")
    '/>
  </xsl:function>

<!--
 :======================================================================
 : Modifier functions
 :======================================================================
 :-->

  <!--
   : this:cast()
   : Modify the distribution to cast values. Applies after other modifiers.
   :
   : $distribution: the distribution to modify
   : $cast: one of "integer", "decimal", "boolean", "string", or "double"
   :   decimal will cast to a double value with $rand:DECIMAL-DIGITS following
   :   the decimal point
   :   Some of the distributions (e.g. zipf() automatically add an integer cast;
   :   the "double" cast will remove that
   :-->
  <xsl:function name="this:cast" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="cast" as="xs:string"/>
    <xsl:sequence select='
    if ($cast="double") then (
      $distribution=>map:remove("cast")=>this:set-is-complex()
    ) else (
      $distribution=>map:put("cast", $cast)=>this:set-is-complex()
    )
    '/>
  </xsl:function>

  <!--
   : this:min()
   : Modify the distribution to set a minimum value (after pre-multiplier)
   :
   : $distribution: the distribution to modify
   : $min: the minimum value
   :   for distributions (like normal) that may generate results less than
   :   this, the truncation setting will determine how to handle the case
   :-->
  <xsl:function name="this:min" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="min" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("min", $min)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:max()
   : Modify the distribution to set a maximum value (after pre-multiplier)
   :
   : $distribution: the distribution to modify
   : $max: the maximum value
   :   for distributions (like normal) that may generate results greater than
   :   this, the truncation setting will determine how to handle the case
   :-->
  <xsl:function name="this:max" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="max" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("max", $max)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:pre-multiplier()
   : Modify the distribution to set a multiplier to apply to the value before
   : applying min/max and other modifications.
   :
   : $distribution: the distribution to modify
   : $pre-multiplier: the multiplier
   :-->
  <xsl:function name="this:pre-multiplier" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="pre-multiplier" as="item()"/>
    <xsl:sequence select='
    $distribution=>
      map:put("pre-multiplier", $pre-multiplier)=>
      this:set-is-complex()
    '/>
  </xsl:function>
  
  <!--
   : this:post-multiplier()
   : Modify the distribution to set a multiplier to apply to the value after
   : applying min/max.
   :
   : $distribution: the distribution to modify
   : $post-multiplier: the multiplier
   :-->
  <xsl:function name="this:post-multiplier" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="post-multiplier" as="item()"/>
    <xsl:sequence select='
    $distribution=>
      map:put("post-multiplier", $post-multiplier)=>
      this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:post-shift()
   : Modify the distribution to set add a shift to the value after
   : applying min/max.
   :
   : $distribution: the distribution to modify
   : $post-shift: the shift
   :-->
  <xsl:function name="this:post-shift" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="post-shift" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("post-shift", $post-shift)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:truncation()
   : Modify the distribution to set a policy for handling values out of the
   : min/max range. 
   :
   : $distribution: the distribution to modify
   : $truncation: the truncation policy
   :    resample: (the default) retry up to $rand:RESAMPLE-LIMIT times 
   :      and then cap the value to minimum or maximum
   :    ceiling: cap the value to minimum or maximum
   :      Faster, but skews the distribution, perhaps badly
   :    drop: drop the value entirely (i.e. return empty sequence)
   :      Fairly useless for most applications that want a definitive number
   :      of random values
   :-->
  <xsl:function name="this:truncation" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="truncation" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("truncation", $truncation)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:mean()
   : Modify the distribution to set a mean value. (normal, skewed)
   :
   : $distribution: the distribution to modify
   : $mean: mean value
   :-->
  <xsl:function name="this:mean" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="mean" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("mean", $mean)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:std()
   : Modify the distribution to set a standard deviation value (normal, skewed)
   :
   : $distribution: the distribution to modify
   : $std: standard deviation value
   :-->
  <xsl:function name="this:std" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="std" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("std", $std)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:skew()
   : Modify the distribution to set a skew value (skewed)
   :
   : $distribution: the distribution to modify
   : $skew: skew value
   :-->
  <xsl:function name="this:skew" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="skew" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("skew", $skew)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:p()
   : Modify the distribution to set a probability value (bernoulli, flip)
   :
   : $distribution: the distribution to modify
   : $p: the probability as a percent in [0,100]
   :-->
  <xsl:function name="this:p" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="p" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("p", $p)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:start()
   : Modify the distribution to set a start value (markov)
   :
   : $distribution: the distribution to modify
   : $start: the starting value of the chain
   :-->
  <xsl:function name="this:start" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="start" as="item()"/>
    <xsl:sequence select='
    $distribution=>map:put("start", $start)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:keys()
   : Modify the distribution to set a set of keys (for distributions that
   : define an index). Number of keys should match index range.
   :
   : $distribution: the distribution to modify
   : $keys: the keys
   :-->
  <xsl:function name="this:keys" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="keys" as="item()*"/>
    <xsl:sequence select='
    $distribution=>map:put("keys", $keys)=>this:set-is-complex()
    '/>
  </xsl:function>

  <!--
   : this:is-complex()
   : Performance hint. If you use the constructors and modifiers in this
   : API you don't need this, but if you are constructing from a vanilla map,
   : and you know the distribution is or is not complex, set it so the 
   : randomizer functions don't have to keep recalculating it. 
   :
   : Don't be wrong here: thing will break. 
   :-->
  <xsl:function name="this:is-complex" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="is-complex" as="xs:boolean"/>
    <xsl:sequence select='
    $distribution=>map:put("is_complex", $is-complex)
    '/>
  </xsl:function>

  <!--
   : this:set-is-complex()
   : Another performance hint, but does the looking for you.
   : Look in algorithm descriptor to see if it needs special handling.
   : The constructors and setters in this API do this automatically; you
   : don't need to.
   :-->
  <xsl:function name="this:set-is-complex" as="map(xs:string, item()*)">
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    $distribution=>map:put("is_complex",
      some $k in map:keys($distribution)[not(. = "keys")]
      satisfies (
        ($distribution($k) instance of xs:QName) or
        ($distribution($k) instance of function(*))
      )
    )
    '/>
  </xsl:function>

  <!--
   : this:describe()
   : Dump the algorithm map in form suitable for debugging.
   : Function values will have their QNames dumped (if possible)
   :   distribution: Distribution to use, one of "constant", "uniform", "normal",
   :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal"
   :   min: mimumum value (optional)
   :   max: maximum value (optional)
   :   pre-multiplier: multiplier before min/max (optional)
   :   post-multiplier: multiplier after min/max (optional)
   :   post-shift: shift to add after min/max (optional)
   :   cast: cast type (optional)
   :   mean: mean of distribution (normal, skewed) (default=0)
   :   std: standard deviation of distribution (normal, skewed) (default=mean)
   :   skew: skew of distribution (skewed) (default=0)
   :   p: probability (bernoulli, flip) (default=50)
   :   sums: cumulative probability sums (zipf, markov)
   :   alpha: alpha parameter (zipf) (needed if no sums) (default=0)
   :   limit: number of sums (zipf) (needed if no sums) (default=1000)
   :   start: index of starting symbol (markov) (default=uniform[1,dim])
   :   dim: size of each dimension of Markov matrix (markov)
   :   matrix: raw Markov matrix (used if sums not provided, not recommended)
   :   sums: cumulative probability sums (zipf, markov)
   :     Single element, sequence of doubles
   :   matrix: raw Markov matrix (used if sums not provided, not recommended)
   :     Single element, sequence of doubles
   :   distributions: component distributions (multimodal)
   :-->
  <xsl:function name="this:describe" as="xs:string">
    <xsl:param name="dist" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
     if (exists($dist("describe"))) then $dist("describe")($dist) else
    "["||$dist("distribution")||" "||
    string-join((
      this:describe("min",$dist),
      this:describe("max",$dist),
      this:describe("pre-multiplier",$dist),
      this:describe("post-multiplier",$dist),
      this:describe("post-shift",$dist),
      this:describe("cast",$dist),
      this:describe("mean",$dist),
      this:describe("std",$dist),
      this:describe("skew",$dist),
      this:describe("p",$dist),
      this:describe("alpha",$dist),
      this:describe("limit",$dist),
      this:describe("dim",$dist),
      this:describe("start",$dist),
      this:describe("sums",$dist,true()),
      this:describe("matrix",$dist,true()),
      this:describe("keys",$dist,true())
      ), " "
    )||
    (if ($dist=>map:contains("distributions")) then "[" else "")||
    string-join(
      for $distribution in $dist("distributions")
      return this:describe($distribution),
      " "
    )||
    (if ($dist=>map:contains("distributions")) then "]" else "")||
    (if ($dist=>map:contains("selector")) then "@"||this:describe($dist("selector")) else "")
    '/>
  </xsl:function>

  <xsl:variable name="this:CRLF" as="xs:string" visibility="private">
    <xsl:text>
</xsl:text>
  </xsl:variable>
  
  <xsl:function name="this:describe" as="xs:string" visibility="private">
    <xsl:param name="k" as="xs:string"/>
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:param name="break" as="xs:boolean"/>
    <xsl:sequence select='
    if ($distribution=>map:contains($k)) then (
      (if ($break) then $this:CRLF else "")||
      $k||":"||util:quote($distribution($k))||
      (if ($break) then $this:CRLF else "")
    ) else ()
    '/> 
  </xsl:function>

  <xsl:function name="this:describe" as="xs:string" visibility="private">
    <xsl:param name="k" as="xs:string"/>
    <xsl:param name="distribution" as="map(xs:string, item()*)"/>
    <xsl:sequence select='
    this:describe($k, $distribution, false())
    '/>
  </xsl:function>
</xsl:package>
<!--distributions.xsl-->
