<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Light definitions
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/light"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:cs="http://mathling.com/colour/space"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:rgb="http://mathling.com/colour/rgb"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xyz="http://mathling.com/colour/xyz"
             version="3.0"
             name="http://mathling.com/type/light"
             exclude-result-prefixes="this xs math map array errors v point cs rgb xyz">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/space" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/rgb" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/xyz" package-version="*"/>

   <!--
    : this:diffuse-light()
    : Constructor for a diffuse light source
    :
    : $direction: light direction (normalized vector)
    : $directional: light colour, RGB
    : $ambient: ambient light colour, RGB
    : $intensity: incoming light intensity [0,1]
    : $ambient-intensity: ambient light intensity [0,1]
    :-->
   <xsl:function name="this:diffuse-light" as="map(xs:string, item()*)">
      <xsl:param name="direction" as="map(xs:string, item()*)"/>
      <xsl:param name="directional" as="map(xs:string, item()*)"/>
      <xsl:param name="ambient" as="map(xs:string, item()*)"/>
      <xsl:param name="intensity" as="xs:double"/>
      <xsl:param name="ambient-intensity" as="xs:double"/>
      <xsl:sequence select='
      if (every $c in point:pcoordinates($direction) satisfies $c=0)
      then errors:error("ML-BADARGS", ("direction", $direction))
      else ()
      ,
      map {
        "colours": (rgb:to-string($directional), rgb:to-string($ambient)),
        "ambient": xyz:raw-coordinates(cs:rgb-to-xyz($ambient))[position()&lt;=3],
        "directional": xyz:raw-coordinates(cs:rgb-to-xyz($directional))[position()&lt;=3],
        "direction": v:normalize(point:pcoordinates($direction, 3)),
        "raw-direction": point:pcoordinates($direction, 3),
        "intensity": $intensity,
        "ambient-intensity": $ambient-intensity
      }
    '/>
   </xsl:function>

   <!--
    : this:diffuse-light()
    : Constructor for a diffuse light source with full intensity light
    :
    : $direction: light direction (normalized vector)
    : $directional: light colour, RGB
    : $ambient: ambient light colour, RGB
    :-->
   <xsl:function name="this:diffuse-light" as="map(xs:string, item()*)">
      <xsl:param name="direction" as="map(xs:string, item()*)"/>
      <xsl:param name="directional" as="map(xs:string, item()*)"/>
      <xsl:param name="ambient" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:diffuse-light($direction, $directional, $ambient, 1.0, 1.0)
    '/>
   </xsl:function>

   <!--
    : this:diffuse-light()
    : Constructor for a diffuse light source with full intensity light, white
    : ambient light
    :
    : $direction: light direction (normalized vector)
    : $directional: light colour, RGB
    :-->
   <xsl:function name="this:diffuse-light" as="map(xs:string, item()*)">
      <xsl:param name="direction" as="map(xs:string, item()*)"/>
      <xsl:param name="directional" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:diffuse-light($direction, $directional, rgb:rgb("white"), 1.0, 1.0)
    '/>
   </xsl:function>

   <!--
    : this:colours()
    :-->
   <xsl:function name="this:colours" as="map(xs:string, item()*)*">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("colours")
    '/>
   </xsl:function>

   <!--
    : this:ambient()
    :-->
   <xsl:function name="this:ambient" as="xs:double*">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("ambient")
    '/>
   </xsl:function>

   <!--
    : this:directional()
    :-->
   <xsl:function name="this:directional" as="xs:double*">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("directional")
    '/>
   </xsl:function>

   <!--
    : this:direction()
    :-->
   <xsl:function name="this:direction" as="xs:double*">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("direction")
    '/>
   </xsl:function>

   <!--
    : this:raw-direction()
    :-->
   <xsl:function name="this:raw-direction" as="xs:double*">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("raw-direction")
    '/>
   </xsl:function>

   <!--
    : this:intensity()
    :-->
   <xsl:function name="this:intensity" as="xs:double">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("intensity")
    '/>
   </xsl:function>

   <!--
    : this:ambient-intensity()
    :-->
   <xsl:function name="this:ambient-intensity" as="xs:double">
      <xsl:param name="light" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $light("ambient-intensity")
    '/>
   </xsl:function>

</xsl:package>
<!--light.xsl-->
