<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Constructors and accessors for masks: masking objects
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/mask"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/type/mask"
             exclude-result-prefixes="this xs math map array config errors point util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

<!--
 :======================================================================
 : Basic constructors
 :======================================================================
 :-->

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "body", "id")'/>
   </xsl:variable>

   <!--
    : this:mask()
    : Make a mask 
    :   $id: the mask ID
    :   $body: contents of the mask
    :   $properties: additional properties
    :-->
   <xsl:function name="this:mask" as="map(xs:string, item()*)">
      <xsl:param name="id" as="xs:string"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "mask",
          "id": $id,
          "body": $body
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:mask()
    : Make a mask
    :   $body: contents of the mask
    :-->
   <xsl:function name="this:mask" as="map(xs:string, item()*)">
      <xsl:param name="id" as="xs:string"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      map {
        "kind": "mask",
        "id": $id,
        "body": $body
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $mask: the mask
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $mask("kind")
    '/>
   </xsl:function>

   <!--
    : this:body()
    : Get the contents of the mask
    : 
    : $mask: the mask
    :-->
   <xsl:function name="this:body" as="item()*">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $mask("body")
    '/>
   </xsl:function>

   <!--
    : this:id()
    : Get the unique id of the mask
    : 
    : $mask: the mask
    :-->
   <xsl:function name="this:id" as="xs:string">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $mask("id")
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Setters
 :======================================================================
 :-->

   <!--
    : this:body()
    : Set the contents of the mask, returning the new mask
    : 
    : $mask: the mask
    : $body: the contents
    :-->
   <xsl:function name="this:body" as="map(xs:string, item()*)">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      $mask=>map:put("body", $body)
    '/>
   </xsl:function>

   <!--
    : this:id()
    : Set the id of the mask, returning the new mask
    : 
    : $mask: the mask
    : $id: the id
    :-->
   <xsl:function name="this:id" as="map(xs:string, item()*)">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:param name="id" as="xs:string"/>
      <xsl:sequence select='
      $mask=>map:put("id", $id)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property manipulation and access
 :======================================================================
 :-->

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="mask" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($mask) = "mask") then util:exclude($mask,$this:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="mask" as="item()"/>
      <xsl:sequence select='
      if (this:kind($mask) = "mask") then ($mask=>map:keys())[not(. = $this:RESERVED)]
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="mask" as="item()"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($mask,
        if (this:kind($mask) = "mask") then util:exclude($properties,$this:RESERVED)
        else map {}
      )
    '/>
   </xsl:function>

   <xsl:function name="this:draw-mask" as="element()" visibility="private">
     <xsl:param name="id" as="xs:string"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:mask id="{$id}">{
        <xsl:sequence select='$drawing("draw:draw")($content, $style-properties, $drawing)'/>
      </art:mask>
   </xsl:function>

   <xsl:function name="this:draw-svg-mask" as="element()" visibility="private">
     <xsl:param name="id" as="xs:string"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <svg:defs>
       <svg:mask id="{$id}">
          <xsl:sequence select='$drawing("draw:draw")($content, $style-properties, $drawing)'/>
       </svg:mask>
      </svg:defs>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties := util:merge-into(($properties, this:property-map($item)))
          return (
            this:draw-mask(this:id($item), this:body($item), $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
          return (
            this:draw-svg-mask(this:id($item), this:body($item), $style-properties, $drawing)
          )
        }
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>
<!--mask.xsl-->
