<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Polynomials; represented as vector of coefficients
 : Most operations require us to process them in reverse order, so we 
 : store them that way. The accessor flips the order back, so internally we 
 : generally do not use it.
 :
 : Copyright© Mary Holstege 2022-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/polynomial"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             name="http://mathling.com/type/polynomial"
             version="3.0"
             exclude-result-prefixes="this xs math map array errors">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>

   <!--
    : this:polynomial()
    : Construct a representation of a polynomial
    : $coefficients: coefficients of polynomial in descreasing order, e.g. 3x³ - 2 = (3, 0, 0, -1)
    :-->
   <xsl:function name="this:polynomial" as="map(xs:string, item()*)">
      <xsl:param name="coefficients" as="xs:double*"/>
      <xsl:sequence select='
      (: Prune leading zeros :)
      let $useful :=
        let $first-nonzero := head(for $i in 1 to count($coefficients) return if ($coefficients[$i]!=0) then $i else ())
        return $coefficients[position() &gt;= $first-nonzero]
      return
      (: Make sure we have something :)
      let $useful :=
        if (empty($useful)) then (0) else $useful
      return
        map {
          "kind": "polynomial",
          "coefficients": reverse($useful) (: Most operations need to use reverse :)
        }
      '/>
   </xsl:function>

   <!--
    : this:coefficients()
    : Accessor for coefficients of polynomial
    :
    : $poly: the polynomial
    :-->
   <xsl:function name="this:coefficients" as="xs:double*">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      reverse($poly("coefficients"))
     '/>
   </xsl:function>

   <!--
    : this:order()
    : Largest exponent. e.g. order(3x³ - 2) = 3
    :
    : $poly: the polynomial
    :-->
   <xsl:function name="this:order" as="xs:integer">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      count($poly("coefficients")) - 1
     '/>
   </xsl:function>

   <!--
    : this:value()
    : Compute f(x) where f is a polynomial in x
    :
    : $poly: the polynomial
    : $x: the value to compute the function of
    :-->
   <xsl:function name="this:value" as="xs:double">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:sequence select='
      let $cs := $poly("coefficients")
      return sum(for $i in 1 to count($cs) return math:pow($x, $i - 1) * $cs[$i])
     '/>
   </xsl:function>

   <!--
    : this:antiderivative()
    : Anti-derivative of a polynomial, which is another polynomial. We use 0 as the
    : integration constant.
    : 
    : $poly: the polynomial
    :-->
   <xsl:function name="this:antiderivative" as="map(xs:string, item()*)">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $cs := $poly("coefficients")
      return
        this:polynomial(
          (reverse(for $i in 1 to count($cs) return $cs[$i] div $i), 0)
      )
     '/>
   </xsl:function>

   <!--
    : this:derivative()
    : Derivative of a polynomial, which is another polynomial.
    : 
    : $poly: the polynomial
    :-->
   <xsl:function name="this:derivative" as="map(xs:string, item()*)">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $cs := $poly("coefficients")
      return (
        this:polynomial(
          if (count($cs) = 1) then 0
          else reverse(for $i in 2 to count($cs) return $cs[$i] * ($i - 1))
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:definite-integral()
    : Definite integral of a polynomial (exact)
    : 
    : $poly: the polynomial
    : $a: start of range of integration
    : $b: end of range of integration
    :-->
   <xsl:function name="this:definite-integral" as="xs:double">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $anti := this:antiderivative($poly)
      return $anti=&gt;this:value($b) - $anti=&gt;this:value($a)
      '/>
   </xsl:function>

   <!--
    : this:describe()
    : Describe the polynomial
    :
    : $poly: the polynomial
    :-->
   <xsl:function name="this:describe" as="xs:string">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (count($poly("coefficients"))=1) then xs:string($poly("coefficients"))
      else
      string-join(
        reverse(
          let $cs := $poly("coefficients") return
          for $i in 1 to count($cs) return
          let $c := $cs[$i]
          return (
            if ($c = 0) then ()
            else (
              let $xton :=
                if ($i - 1 = 0) then ""
                else if ($i - 1 = 1) then "x"
                else "x^"||($i - 1)
              return
              let $k :=
                if ($c = 1 and $i &gt; 1) then ""
                else if ($c = -1 and $i &gt; 1) then "-"
                else xs:string($c)
              return $k||$xton (: c*x^(i-1) :)
            )
          )
        ),
        " + "
      )
     '/>
   </xsl:function>

   <!--
    : this:function()
    : Return a function that takes values of the polynomial, e.g. for curve plotting
    :
    : $poly: the polynomial
    :-->
   <xsl:function name="this:function" as="(function(xs:double) as xs:double)">
      <xsl:param name="poly" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      function($v as xs:double) as xs:double {
        $poly=&gt;this:value($v)
      }
     '/>
   </xsl:function>

<!--
  :======================================================================
  : Operations between two polynomials
  :======================================================================
  :-->

 <!--
    : this:add()
    : Add two polynomials
    :
    : $p1, $p2: the polynomials
    :-->
   <xsl:function name="this:add" as="map(xs:string, item()*)">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $c1 :=
        if (this:order($p1) &gt; this:order($p2))
        then $p1("coefficients")
        else $p2("coefficients")
      return
      let $c2 := 
        if (this:order($p1) &gt; this:order($p2))
        then $p2("coefficients")
        else $p1("coefficients")
      return (
        this:polynomial(
          reverse(
            for $i in 1 to count($c1) return $c1[$i] + ($c2[$i],0)[1]
          )
        )
      )
     '/>
   </xsl:function>

   <!--
    : this:sub()
    : Subtract two polynomials
    :
    : $p1, $p2: the polynomials
    :-->
   <xsl:function name="this:sub" as="map(xs:string, item()*)">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $c1 :=
        if (this:order($p1) &gt; this:order($p2))
        then $p1("coefficients")
        else $p2("coefficients")!(-.)
      return
      let $c2 := 
        if (this:order($p1) &gt; this:order($p2))
        then $p2("coefficients")!(-.)
        else $p1("coefficients")
      return (
        this:polynomial(
          reverse(
            for $i in 1 to count($c1) return $c1[$i] + ($c2[$i],0)[1]
          )
        )
      )
     '/>
   </xsl:function>
</xsl:package>

<!--polynomial.xsl-->
