<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Constructors and accessors for slots (containers for other objects)
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/slot"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/type/slot"
             exclude-result-prefixes="this xs math map array errors config point util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "body", "location", "skew", "scale", "rotation", "center")'/>
   </xsl:variable>

<!--
 :======================================================================
 : Basic constructors
 :======================================================================
 :-->

   <!--
    : this:slot()
    : slot()
    : Make a slot at the given location
    :   $location: location for the slot (x,y); also creates rotation center
    :   $body: contents of the slot
    :   $properties: additional properties
    :-->
   <xsl:function name="this:slot" as="map(xs:string, item()*)">
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "slot",
          "body": $body,
          "location": $location,
          "center": $location
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:slot()
    : Make a slot
    :   $body: contents of the slot
    :   $properties: additional properties
    :-->
   <xsl:function name="this:slot" as="map(xs:string, item()*)">
      <xsl:param name="body" as="item()*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "slot",
          "body": $body
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:slot()
    : Make a slot at the origin
    :   $body: contents of the slot
    :-->
   <xsl:function name="this:slot" as="map(xs:string, item()*)">
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      map {
        "kind": "slot",
        "body": $body
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $slot: the slot
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='$slot("kind")'/>
   </xsl:function>

   <!--
    : this:body()
    : Get the contents of the slot
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:body" as="item()*">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("body")
    '/>
   </xsl:function>

   <!--
    : this:location()
    : Get the location of the slot (x,y)
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:location" as="map(xs:string, item()*)?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("location")
    '/>
   </xsl:function>

   <!--
    : this:center()
    : Get the rotation center of the slot (cx,cy)
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("center")
    '/>
   </xsl:function>

   <!--
    : this:rotation()
    : Get the rotation angle (degrees) of the slot (rotate)
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:rotation" as="xs:double?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("rotation")
    '/>
   </xsl:function>

   <!--
    : this:skew()
    : Get the X and Y skew of the slot (skewX, skewY)
    : Returned as a point; the values are angles in degrees
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:skew" as="map(xs:string, item()*)?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("skew")
    '/>
   </xsl:function>

   <!--
    : this:skew-x()
    : Get the X skew of the slot (degrees)
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:skew-x" as="xs:double?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("skew")=>point:x()
    '/>
   </xsl:function>

   <!--
    : this:skew-y()
    : skew-y()
    : Get the Y skew of the slot (degrees)
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:skew-y" as="xs:double?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("skew")=>point:y()
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Get the X and Y scale of the slot (sx, sy)
    : Returned as a point
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("scale")
    '/>
   </xsl:function>

   <!--
    : this:scale-x()
    : Get the X scale of the slot 
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:scale-x" as="xs:double?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("scale")=>point:x()
    '/>
   </xsl:function>

   <!--
    : this:scale-y()
    : Get the Y scale of the slot 
    : 
    : $slot: the slot
    :-->
   <xsl:function name="this:scale-y" as="xs:double?">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot("scale")=>point:y()
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Setters
 :======================================================================
 :-->

   <!--
    : this:body()
    : Set the contents of the slot, returning the new slot
    : 
    : $slot: the slot
    : $body: the contents
    :-->
   <xsl:function name="this:body" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      $slot=>map:put("body", $body)
    '/>
   </xsl:function>

   <!--
    : this:location()
    : Set the location of the slot, returning the new slot
    : Note: you will need to reset the center separately
    : 
    : $slot: the slot
    : $location: the location
    :-->
   <xsl:function name="this:location" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="location" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot=>map:put("location", $location)
    '/>
   </xsl:function>

   <!--
    : this:center()
    : Set the rotation center of the slot (cx,cy), returning the new slot
    : 
    : $slot: the slot
    : $center: the center of rotation
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot=>map:put("center", $center)
    '/>
   </xsl:function>

   <!--
    : this:rotation()
    : Set the rotation angle (degrees) of the slot (rotate), returning the new slot
    : 
    : $slot: the slot
    : $rotation: degrees of rotation
    :-->
   <xsl:function name="this:rotation" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="rotation" as="xs:double"/>
      <xsl:sequence select='
      $slot=>map:put("rotation", $rotation)
    '/>
   </xsl:function>

   <!--
    : this:skew()
    : Set the X and Y skew of the slot (skewX, skewY), returning the new slot
    : 
    : $slot: the slot
    : $skew: the X and Y skew as a point
    :-->
   <xsl:function name="this:skew" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="skew" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $slot=>map:put("skew", $skew)
    '/>
   </xsl:function>

   <!--
    : this:skew()
    : Set the X and Y skew of the slot (skewX, skewY), returning the new slot
    : 
    : $slot: the slot
    : $skewX: the X skew (degrees)
    : $skewY: the Y skew (degrees)
    :-->
   <xsl:function name="this:skew" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="skewX" as="xs:double"/>
      <xsl:param name="skewY" as="xs:double"/>
      <xsl:sequence select='
      $slot=>map:put("skew", point:point($skewX, $skewY))
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Set the X and Y scale of the slot (sx, sy), returning the new slot
    : 
    : $slot: the slot
    : $scale: the X and Y scaling as a point or double (if both the same)
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="scale" as="item()"/>
      <xsl:sequence select='
      if ($scale instance of map(xs:string,item()*)) then (
        if (point:kind($scale) = "point") then $slot=>map:put("scale", $scale=>point:as-dimension(2))
        else errors:error("ML-BADARGS",("scale",$scale))
      ) else if ($scale instance of xs:numeric) then (
        $slot=>map:put("scale", point:point($scale, $scale))
      ) else (
        errors:error("ML-BADARGS",("scale",$scale))
      )
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Set the X and Y scale of the slot (sx, sy), returning the new slot
    : 
    : $slot: the slot
    : $scaleX: the X scaling
    : $scaleY: the Y scaling
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:param name="scaleX" as="xs:double"/>
      <xsl:param name="scaleY" as="xs:double"/>
      <xsl:sequence select='
      $slot=>map:put("scale", point:point($scaleX, $scaleY))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property manipulation and access
 :======================================================================
 :-->

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($slot) = "slot") then util:exclude($slot,$this:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="slot" as="item()"/>
      <xsl:sequence select='
      if (this:kind($slot) = "slot") then ($slot=>map:keys())[not(. = $this:RESERVED)]
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="slot" as="item()"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($slot,
        if (this:kind($slot) = "slot") then util:exclude($properties,$this:RESERVED)
        else map {}
      )
    '/>
   </xsl:function>

   <xsl:function name="this:draw-slot" as="element()" visibility="private">
     <xsl:param name="location" as="map(xs:string,item()*)?"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="rotation" as="xs:double?"/>
     <xsl:param name="center" as="map(xs:string,item()*)?"/>
     <xsl:param name="scale" as="map(xs:string,item()*)?"/>
     <xsl:param name="skew" as="map(xs:string,item()*)?"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:slot>
       <xsl:if test='exists($location)'>
         <xsl:attribute name="x" select='point:px($location)'/>
         <xsl:attribute name="y" select='point:py($location)'/>
       </xsl:if>
       <xsl:if test='exists($rotation)'>
          <xsl:attribute name="rotate" select='$rotation'/>
          <xsl:if test='exists($center)'>
            <xsl:attribute name="cx" select='point:px($center)'/>
            <xsl:attribute name="cy" select='point:py($center)'/>
          </xsl:if>
       </xsl:if>
        <xsl:if test='exists($scale)'>
          <xsl:if test='exists($center) and empty($rotation)'>
            <xsl:attribute name="cx" select='point:px($center)'/>
            <xsl:attribute name="cy" select='point:py($center)'/>
          </xsl:if>
          <xsl:attribute name="sx" select='point:px($scale)'/>
          <xsl:attribute name="sy" select='point:py($scale)'/>
        </xsl:if>
        <xsl:if test='exists($skew)'>
          <xsl:attribute name="skewX" select='point:px($skew)'/>
          <xsl:attribute name="skewY" select='point:py($skew)'/>
        </xsl:if>
        <xsl:sequence select='util:as-attributes($style-properties)'/>
        <xsl:sequence select='$drawing("draw:draw")($content, map{}, $drawing)'/>
     </art:slot>
   </xsl:function>

   <xsl:function name="this:draw-group" as="element()" visibility="private">
     <xsl:param name="location" as="map(xs:string,item()*)?"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="rotation" as="xs:double?"/>
     <xsl:param name="center" as="map(xs:string,item()*)?"/>
     <xsl:param name="scale" as="map(xs:string,item()*)?"/>
     <xsl:param name="skew" as="map(xs:string,item()*)?"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:variable name="translating" select='
      if (empty($location)) then () else (
        "translate("||point:px($location)||","||point:py($location)||")"
      )
      '/>
      <xsl:variable name="scaling" select='
      if (empty($scale)) then () else (
        (if (empty($center)) then () else (
          "translate("||point:px($center)||","||point:py($center)||") "
        ))||
        ("scale("||point:px($scale)||","||point:py($scale)||")")||
        (if (empty($center)) then () else (
          "translate("||-point:px($center)||","||-point:py($center)||")"
        ))
      )
      '/>
      <xsl:variable name="rotating" select='
      if (empty($rotation) or $rotation eq 0) then () else (
        "rotate("||$rotation||(
          if (empty($center)) then () else (
            ","||point:px($center)||","||point:py($center)
          )
        )
      )
      '/>
      <xsl:variable name="skewing" select='
       if (empty($skew)) then () else (
         "skewX("||point:px($skew)||") "||"skewY("||point:py($skew)||")"
       )
       '/>
       <xsl:variable name="transform" select='
       string-join(($translating, $scaling, $rotating, $skewing)," ")
       '/>
       <svg:g>
         <xsl:if test='exists($transform) and $transform ne ""'>
           <xsl:attribute name="transform" select="$transform"/>
         </xsl:if>
         <xsl:sequence select='util:as-attributes($style-properties)'/>
         <xsl:sequence select='$drawing("draw:draw")($content, map{}, $drawing)'/>
       </svg:g>
   </xsl:function>
   
   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item))) return
            let $location := this:location($item)!point:snap(.) return
            let $body := this:body($item) return
            let $rotation := this:rotation($item) return
            let $center := this:center($item)!point:snap(.) return
            let $scale := this:scale($item)!point:snap(.) return
            let $skew := this:skew($item)!point:snap(.)
            return (
              this:draw-slot($location, $body, $rotation, $center, $scale, $skew, $style-properties, $drawing)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties := util:merge-into(($properties, this:property-map($item))) return
            let $location := this:location($item) return
            let $body := this:body($item) return
            let $rotation := this:rotation($item) return
            let $center := this:center($item) return
            let $scale := this:scale($item) return
            let $skew := this:skew($item)
            return (
              this:draw-slot($location, $body, $rotation, $center, $scale, $skew, $style-properties, $drawing)
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
            let $location := this:location($item)!point:snap(.) return
            let $body := this:body($item) return
            let $rotation := this:rotation($item) return
            let $center := this:center($item)!point:snap(.) return
            let $scale := this:scale($item)!point:snap(.) return
            let $skew := this:skew($item)!point:snap(.)
            return (
              this:draw-group($location, $body, $rotation, $center, $scale, $skew, $style-properties, $drawing)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item)))) return
            let $location := this:location($item) return
            let $body := this:body($item) return
            let $rotation := this:rotation($item) return
            let $center := this:center($item) return
            let $scale := this:scale($item) return
            let $skew := this:skew($item)
            return (
              this:draw-group($location, $body, $rotation, $center, $scale, $skew, $style-properties, $drawing)
            )
          }
        )
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>

<!--slot.xsl-->
