<?xml version="1.0" encoding="UTF-8"?>
<!-- 
 : Constructors and accessors for wrappers: objects that wrap content with
 : dynamic parameters.
 :
 : Copyright© Mary Holstege 2022-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/wrapper"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:mask="http://mathling.com/type/mask"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:slot="http://mathling.com/type/slot"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:art="http://mathling.com/art"
             xmlns:svg="http://www.w3.org/2000/svg"
             version="3.0"
             name="http://mathling.com/type/wrapper"
             exclude-result-prefixes="this xs math map array errors config util slot mask">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/slot" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/mask" package-version="*"/>

<!--
 :======================================================================
 : Basic constructors
 :======================================================================
 :-->

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "body", "dynamics")'/>
   </xsl:variable>

   <!--
    : this:wrapper()
    : Make a wrapper for the content
    :   $body: contents of the wrapper
    :   $dynamics: map of dynamic parameter values
    :-->
   <xsl:function name="this:wrapper" as="map(xs:string, item()*)">
      <xsl:param name="body" as="item()*"/>
      <xsl:param name="dynamics" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      map {
        "kind": "wrapper",
        "body": $body,
        "dynamics": $dynamics
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $wrapper: the wrapper
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="wrapper" as="map(xs:string, item()*)"/>
      <xsl:sequence select='$wrapper("kind")'/>
   </xsl:function>

   <!--
    : this:body()
    : Get the contents of the wrapper
    : 
    : $wrappers: sequence of wrappers
    :-->
   <xsl:function name="this:body" as="item()*">
      <xsl:param name="wrappers" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $wrapper in $wrappers return (
        if ($wrapper("kind") = "wrapper")  
        then $wrapper("body")
        else $wrapper
      )
    '/>
   </xsl:function>

   <!--
    : this:dynamics()
    : Get the dynamic parameter values of the wrapper
    : Returns a single merged map if given a sequence of wrappers
    : 
    : $wrappers: the wrapper
    :-->
   <xsl:function name="this:dynamics" as="map(xs:string, item()*)">
      <xsl:param name="wrappers" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      util:merge-into((
        map {},
        for $wrapper in $wrappers return $wrapper("dynamics")
      ))
    '/>
   </xsl:function>

   <!--
    : this:get-shapes()
    :-->
   <xsl:function name="this:get-shapes" as="item()*" visibility="private">
      <xsl:param name="stuff" as="item()*"/>
      <xsl:sequence select='
      for $item in $stuff return
      if ($item instance of map(xs:string,item()*)) then (
        let $kind := $item("kind") return
        if ($kind = "wrapper") then this:shapes($item)
        else if ($kind = "def") then ()
        else if ($kind = "ref") then ()
        else if ($kind = "slot") then this:get-shapes(slot:body($item))
        else if ($kind = "mask") then this:get-shapes(mask:body($item))
        else $item
      )
      else ()
    '/>
   </xsl:function>

   <!--
    : this:shapes()
    : Pull actual shapes from wrappers, tossing out things like defs
    : Doesn't handle semantics of slot modifiers, so you get the raw shapes
    :-->
   <xsl:function name="this:shapes" as="map(xs:string, item()*)*">
      <xsl:param name="wrappers" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:get-shapes(this:body($wrappers))[. instance of map(xs:string,item()*)]
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Setters
 :======================================================================
 :-->

   <!--
    : this:body()
    : Set the contents of the wrapper, returning the new wrapper
    : 
    : $wrapper: the wrapper
    : $body: the contents
    :-->
   <xsl:function name="this:body" as="map(xs:string, item()*)">
      <xsl:param name="wrapper" as="map(xs:string, item()*)"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      $wrapper=>map:put("body", $body)
    '/>
   </xsl:function>

   <!--
    : this:dynamics()
    : Set the dynamic parameters of the wrapper, returning the new wrapper
    : 
    : $wrapper: the wrapper
    : $dynamics: the dynamic parameters
    :-->
   <xsl:function name="this:dynamics" as="map(xs:string, item()*)">
      <xsl:param name="wrapper" as="map(xs:string, item()*)"/>
      <xsl:param name="dynamics" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $wrapper=>map:put("dynamics", $dynamics)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property manipulation and access
 :======================================================================
 :-->

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="wrapper" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($wrapper) = "wrapper") then util:exclude($wrapper,$this:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="wrapper" as="item()"/>
      <xsl:sequence select='
      if (this:kind($wrapper) = "wrapper") then ($wrapper=>map:keys())[not(. = $this:RESERVED)]
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="wrapper" as="item()"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($wrapper,
        if (this:kind($wrapper) = "wrapper") then util:exclude($properties,$this:RESERVED)
        else map {}
      )
    '/>
   </xsl:function>

   <xsl:function name="this:draw-metadata" as="element()" visibility="private">
     <xsl:param name="item" as="item()*"/>
     <svg:metadata>
       <xsl:sequence select='$item'/>
     </svg:metadata>
   </xsl:function>
   
   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($properties, this:property-map($item)))
          return (
            $drawing("draw:dump-dynamic-parameters")(this:dynamics($item)),
            $drawing("draw:draw")(this:body($item), $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
          return (
            this:draw-metadata(
              $drawing("draw:dump-dynamic-parameters")(this:dynamics($item))
            ),
            $drawing("draw:draw")(this:body($item), $style-properties, $drawing)
          )
        }
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>
<!--wrapper.xsl-->
